/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.bhie.dao.patient;

import gov.va.med.bhie.beans.NoteImage;
import gov.va.med.bhie.common.AppConfig;
import gov.va.med.bhie.dao.DaoBase;
import gov.va.med.bhie.dao.DaoException;
import gov.va.med.bhie.dataservice.BHIERestServiceHandler;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;

/**
 * BHIE Note data access object.
 */
public class NoteDao extends DaoBase {

    /**
     * Retrieves a complex note image (as binary data) from the BHIE interface.
     * @param eventId Note event identifier.
     * @return NoteImage bean.
     * @throws DaoException If there is a failure.
     */
    public NoteImage getNoteImage(String eventId) throws DaoException {
        if (eventId == null || eventId.isEmpty()) {
            throw new IllegalArgumentException("Event id cannot be null or empty");
        }

        try {
            return BHIERestServiceHandler.getInstance().queryNoteImage(eventId);
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    /**
     * Retrieves a complex note image (as binary data) from the BHIE interface.
     * @param urlString Complex note URL string.
     * @return NoteImage bean.
     * @throws DaoException If there is a failure.
     */
    public NoteImage getNoteImageFromUrl(String urlString) throws DaoException {
        if (urlString == null || urlString.isEmpty()) {
            throw new IllegalArgumentException("Note image url cannot be null or empty");
        }

        URL url = null;
        URL bhieEndpoint = null;

        if(urlString.contains("^")){
            urlString = urlString.replace("^", "%5E");
        }

        try {
            url = new URL(urlString);
            bhieEndpoint = new URL(AppConfig.getInstance().getBhieEndpoint());
        } catch (MalformedURLException e) {
            throw new IllegalArgumentException(e);
        }

        //ensure path matches complex note path
        if (!url.getPath().contains(AppConfig.getInstance().getComplexNotePath())) {
            throw new IllegalArgumentException("Invalid URL path: " + url.getPath());
        }

        try {
            //if host starts with "BHIEDomain"
            if (url.getHost().startsWith(AppConfig.getInstance().getComplexNoteHostPlaceholder())) {
                StringBuilder urlSB = new StringBuilder("");
                urlSB.append(bhieEndpoint.getProtocol());
                urlSB.append("://");
                urlSB.append(bhieEndpoint.getHost());
                if (bhieEndpoint.getPort() > -1) {
                    urlSB.append(":");
                    urlSB.append(bhieEndpoint.getPort());
                }
                urlSB.append(url.getPath());
                url = new URL(urlSB.toString());
            }
            return BHIERestServiceHandler.getInstance().queryNoteImage(new URI(url.toString()));
        } catch (Exception e) {
            throw new DaoException(e);
        }

    }

}
