package gov.va.med.bhie.util;

import gov.va.med.bhie.common.AppConfig;

import java.io.*;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.Locale;
import java.util.TimeZone;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.log4j.Logger;


/**
 *  Database / file logging class.
 *
 * @author KWeller
 * @version 1.0
 */
public class DBLogger {

    /**
     * Exception string.
     */
    protected static final String EXCEPTION = "Exception: ";

    /**
     * Logger.
     */
    private static final Logger LOGGER =
        Logger.getLogger(DBLogger.class.getName());

    /**
     * Lock for synchronizing file writes.
     */
    private static final Object lock = new Object();
    
    /**
     * Table created flag.
     */
    private static boolean createtable = false;
    
    /**
     * Connection to the database.
     */
    private Connection conn;
    
    /**
     *  Connection pool.
     */
    private DataSource pool;
    
    /**
     * Database type.
     */
    private String dbtype = "";
    
    /**
     * Filename.
     */
    private String filename = "";
    
    /**
     * Filename.
     */
    private String filenamedate = "";
    
    public DBLogger() {
        init(AppConfig.getInstance().getDataSource());
    }
    
    public DBLogger(String dsource) {
        init(dsource);
    }
    
    /**
     * Constructor.
     */
    private void init(String dsource) {
        Context env;
        
        try {
//            String dsource = AppConfig.getInstance().getDataSource();
            if (dsource == null || dsource.length() == 0) {
                dsource = "";
                conn = null;
                pool = null;
            } else {
                LOGGER.debug("DSOURCE = " + dsource);
                // jdbc/jXPDS
                // xml/d:/oracle/middleware/bhielog.xml
                // text/d:/oracle/middleware/bhielog.txt
                if (dsource.startsWith("text")) {
                    dbtype = "text";
                    filename = dsource.substring(5);
                    int dir = filename.lastIndexOf('/');
                    if(dir > 2){ //ensure we don't count escape slashes (> 2)
                        String path = filename.substring(0,dir+1);
                        String file = filename.substring(dir+1);
                        filename = path + getHostName() + "_" + file;
                    }
                } else if (dsource.startsWith("xml")) {
                    dbtype = "xml";
                    filename = dsource.substring(4);
                } else {
                    env = new InitialContext();
                    this.pool = (DataSource) env.lookup(dsource);
                    this.conn = this.pool.getConnection();
                    if (this.conn == null) {
                        LOGGER.error("CONN is null.");
                    } else {
                        DatabaseMetaData meta = conn.getMetaData();
                        if (meta != null) {
                            dbtype = meta.getDatabaseProductName();
                            LOGGER.debug("DBTYPE = " + dbtype);
                        }
                    }
                }
            }
        } catch (NamingException ex) {
            LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
        } catch (SQLException ex) {
            LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
        }
    }

    /**
     * Store log event.
     *
     * @param edipn    patient id
     * @param type     type of request
     * @param errorstr error string
     * @param msec     length of call
     * @param count    number of items returned
     * @param cached   cached flag
     */
    public void logEvent(String edipn, String type, String errorstr, long msec, int count, String cached) {
        logEvent(edipn, type, errorstr, msec, count, cached, "");
    }

    /**
     * Store log event.
     * 
     * @param edipn patient id
     * @param type type of request
     * @param errorstr error string
     * @param msec length of call
     * @param count number of items returned
     * @param cached cached flag
     */
    public void logEvent(String edipn, String type, String errorstr, long msec, int count, String cached, String queryinfo) {
        //clean and limit error string
        errorstr = cleanString(errorstr);
        if(errorstr.length() > 500){
            errorstr = errorstr.substring(0,500);
        }

        if ("text".equals(dbtype)) {
            storeText(edipn, type, errorstr, msec, count, cached, queryinfo);
        } else if ("xml".equals(dbtype)) {
            storeXML(edipn, type, errorstr, msec, count, cached, queryinfo);
        } else {
            if (conn != null) {
                if ("oracle".equalsIgnoreCase(dbtype)) {
                    storeOracle(edipn, type, errorstr, msec, count, cached, queryinfo);
                } else if ("Microsoft SQL Server".equalsIgnoreCase(dbtype)) {
                    storeMSSQL(edipn, type, errorstr, msec, count, cached, queryinfo);
                } else if (dbtype != null && dbtype.length() > 0) {
                    storeGeneric(edipn, type, errorstr, msec, count, cached, queryinfo);
                }
            }
        }
    }
    
    /**
     * Store event to oracle.
     * 
     * @param edipn patient id
     * @param type type of request
     * @param errorstr error string
     * @param msec length of call
     * @param count number of items returned
     * @param cached cached flag
     */
    private void storeOracle(String edipn, String type, String errorstr, long msec, int count, String cached, String queryinfo) {
        synchronized (lock) {
            String message = "Message: Relay service / edi: " + edipn + " / type: " 
                + type + " / msec: " + Long.toString(msec) + " / count: " + count + " / cached: " 
                + cached + " / error: " + errorstr + " / query: " + queryinfo;
    
            int nval = getLogNextval();
            
            String sql = "INSERT INTO audit_log(log_id, log_date_time, user_id, "
                     + "item_id_system, item_id_subsystem, message_id, message) "
                     + "VALUES (?, current_timestamp, 'SYSTEM', 2030, 2031, 95, ?)";
            
            PreparedStatement statement = null;
            
            try {
                statement = conn.prepareStatement(sql);
                statement.setInt(1, nval);
                statement.setString(2, message);
                statement.execute();
            } catch (SQLException ex) {
                LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
                try {
                    if (!createtable) {
                        createtable = true;
                        try {
                            if (statement != null) {
                                statement.close();
                            }
                        } catch (SQLException ex2) {
                            LOGGER.error(EXCEPTION + ex2.getLocalizedMessage(), ex2);
                        }
                        LOGGER.debug("storeOracle - create table");
                        statement = conn.prepareStatement(" CREATE TABLE AUDIT_LOG ( LOG_ID NUMBER(18,0), LOG_DATE_TIME DATE, ITEM_ID_SYSTEM NUMBER(10,0), ITEM_ID_SUBSYSTEM NUMBER(10,0), MESSAGE_ID NUMBER(10,0), MESSAGE VARCHAR2(2000),  REASON VARCHAR2(1000),  USER_ID VARCHAR2(40), CONSTRAINT AUDIT_LOG_PK PRIMARY KEY ( LOG_ID ) )");
                        statement.execute();
                    }
                } catch (SQLException ex2) {
                    LOGGER.error(EXCEPTION + ex2.getLocalizedMessage(), ex2);
                } 
            } finally {
                try {
                    if (statement != null) {
                        statement.close();
                    }
                } catch (SQLException ex) {
                    LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
                }
            }
        }
    }

    /**
     * Store event to SQL Server.
     *
     * @param edipn    patient id
     * @param type     type of request
     * @param errorstr error string
     * @param msec     length of call
     * @param count    number of items returned
     * @param cached   cached flag
     */
    private void storeMSSQL(String edipn, String type, String errorstr, long msec, int count, String cached, String queryinfo) {
        synchronized (lock) {
            String sql = "INSERT INTO audit_log (edipn, calltype, errormsg, msec, count, cached, queryinfo) VALUES (?, ?, ?, ?, ?, ?, ?)";
    
            PreparedStatement statement = null;
    
            String ledipn = "";
            if (edipn != null) {
                ledipn = edipn;
                if (ledipn.length() > 50) {
                    ledipn = edipn.substring(0, 50);
                }
            }
            String lcalltype = "";
            if (type != null) {
                lcalltype = type;
                if (lcalltype.length() > 255) {
                    lcalltype = type.substring(0, 255);
                }
            }
            String lerrorstr = "";
            if (errorstr != null) {
                lerrorstr = errorstr;
                if (lerrorstr.length() > 255) {
                    lerrorstr = errorstr.substring(0, 255);
                }
            }
            String lcached = "";
            if (cached != null) {
                lcached = cached;
                if (lcached.length() > 50) {
                    lcached = cached.substring(0, 50);
                }
            }
    
            String lqinfo = "";
            if (queryinfo != null) {
                lqinfo = queryinfo;
                if (lqinfo.length() > 255) {
                    lqinfo = queryinfo.substring(0, 255);
                }
            }
            try {
                statement = conn.prepareStatement(sql);
                statement.setString(1, ledipn);
                statement.setString(2, lcalltype);
                statement.setString(3, lerrorstr);
                statement.setLong(4, msec);
                statement.setInt(5, count);
                statement.setString(6, lcached);
                statement.setString(7, lqinfo);
                statement.execute();
            } catch (SQLException ex) {
                if (statement != null) {
                    try {
                        statement.close();
                    } catch (SQLException e) {
                        LOGGER.error(e.getLocalizedMessage());
                    }
                }
                LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
                try {
                    LOGGER.error("storeMSSQL - create table");
                    statement = conn.prepareStatement("CREATE TABLE [dbo].[AUDIT_LOG]("
                            + " [AUDIT_ID] [uniqueidentifier] ROWGUIDCOL NOT NULL,"
                            + " [LOG_DATE_TIME] [datetime] NOT NULL,"
                            + " [EDIPN] [varchar](50) NULL,"
                            + " [CALLTYPE] [varchar](255) NULL,"
                            + " [ERRORMSG] [varchar](255) NULL,"
                            + " [MSEC] [numeric](18, 0) NULL,"
                            + " [COUNT] [numeric](18, 0) NULL,"
                            + " [CACHED] [varchar](50) NULL,"
                            + " CONSTRAINT [PK_AUDIT_LOG2] PRIMARY KEY CLUSTERED "
                            + "("
                            + " [AUDIT_ID] ASC"
                            + ")WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, IGNORE_DUP_KEY = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) ON [PRIMARY]"
                            + ") ON [PRIMARY]");
                    statement.execute();
                    statement.close();
                    statement = conn.prepareStatement("ALTER TABLE [dbo].[AUDIT_LOG] ADD  CONSTRAINT [DF_AUDIT_LOG2_AUDIT_ID]  DEFAULT (newid()) FOR [AUDIT_ID]");
                    statement.execute();
                    statement.close();
                    statement = conn.prepareStatement("ALTER TABLE [dbo].[AUDIT_LOG] ADD  CONSTRAINT [DF_AUDIT_LOG2_LOG_DATE_TIME]  DEFAULT (getdate()) FOR [LOG_DATE_TIME]");
                    statement.execute();
                    statement.close();
                    statement = conn.prepareStatement("ALTER TABLE [dbo].[AUDIT_LOG] ADD ( [QUERYINFO] [varchar](255) NULL )");
                    statement.execute();
                } catch (SQLException ex2) {
                    LOGGER.error(EXCEPTION + ex2.getLocalizedMessage(), ex2);
                }
            } finally {
                try {
                    if (statement != null) {
                        statement.close();
                    }
                } catch (SQLException ex) {
                    LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
                }
            }
        }
    }

    /**
     * Get sequence number.
     *
     * @return number
     */
    private int getLogNextval() {
        int rtc = 0;
        String sql = "SELECT log_id_seq.nextval from dual";
        PreparedStatement statement = null;
        ResultSet resultset = null;

        try {
            statement = conn.prepareStatement(sql);
            resultset = statement.executeQuery();

            if (resultset != null && resultset.next()) {
                rtc = resultset.getInt(1);
            }
        } catch (SQLException ex) {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    LOGGER.error(e.getLocalizedMessage());
                }
            }
            LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
            try {
                LOGGER.debug("storeOracle - create sequence");
                statement = conn.prepareStatement("CREATE SEQUENCE LOG_ID_SEQ  MINVALUE 1 MAXVALUE 999999999999999999999999999 INCREMENT BY 1 START WITH 2 CACHE 20 ORDER  NOCYCLE");
                statement.execute();
                rtc = 1;
            } catch (SQLException ex2) {
                LOGGER.error(EXCEPTION + ex2.getLocalizedMessage(), ex2);
            } 
        } finally {
            try {
                if (resultset != null) {
                    resultset.close();
                }
            } catch (SQLException ex) {
                LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
            }

            try {
                if (statement != null) {
                    statement.close();
                }
            } catch (SQLException ex) {
                LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
            }
        }
        return rtc;
    }
    
    /**
     * Store event to generic schema.
     * 
     * @param edipn patient id
     * @param type type of request
     * @param errorstr error string
     * @param msec length of call
     * @param count number of items returned
     * @param cached cached flag
     */
    private void storeGeneric(String edipn, String type, String errorstr,
                              long msec, int count, String cached, String queryinfo) {
        String sql = "INSERT INTO audit_log (log_date_time, call_type, "
                + "error_msg, resp_time, doc_count, resp_cached, queryinfo, systemid)"
                + "VALUES (current_timestamp, ?, ?, ?, ?, ?, ?, ?)";
   
        PreparedStatement statement = null;
        
        try {
            statement = conn.prepareStatement(sql);
            statement.setString(1, type);
            statement.setString(2, errorstr);
            statement.setLong(3, msec);
            statement.setInt(4, count);
            statement.setString(5, cached);
            statement.setString(6, queryinfo);
            statement.setString(7, getId());
            statement.execute();
        } catch (SQLException ex) {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    LOGGER.error(e.getLocalizedMessage());
                }
            }
            LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
            try {
                LOGGER.debug("storeGeneric - create table");
                statement = conn.prepareStatement(" CREATE TABLE AUDIT_LOG ( LOG_DATE_TIME DATE, CALL_TYPE VARCHAR(255), ERROR_MSG VARCHAR(255), RESP_TIME NUMBER(10,0), DOC_COUNT NUMBER(10,0),  RESP_CACHED VARCHAR(80), QUERYINFO VARCHAR(255), SYSTEMID VARCHAR(255) )");
                statement.execute();
            } catch (SQLException ex2) {
                LOGGER.error(EXCEPTION + ex2.getLocalizedMessage(), ex2);
            }
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
            } catch (SQLException ex) {
                LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
            }
        }
    }

    /**
     * Close connection to database.
     */
    public final void close() {
        boolean commitfailed = false;
        try {
            if (this.conn != null) {
                this.conn.commit();
            }
        } catch (SQLException ex) {
            if (ex != null && ex.getLocalizedMessage() != null && ex.getLocalizedMessage().indexOf("autocommit") == -1) {
                commitfailed = true;
            }
        }
        try {
            if (this.conn != null) {
                this.conn.close();
            }
        } catch (SQLException ex) {
            if (commitfailed) {
                LOGGER.error(EXCEPTION + ex.getLocalizedMessage() + " (commit also failed)", ex);
            } else {
                LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
            }
        }
        this.conn = null;
        this.pool = null;
    }

    /**
     * Store event to text file.
     * 
     * @param edipn patient id
     * @param type type of request
     * @param errorstr error string
     * @param msec length of call
     * @param count number of items returned
     * @param cached cached flag
     */
    private void storeText(String edipn, String type, String errorstr,
                           long msec, int count, String cached, String queryinfo) {
        StringBuffer sql = new StringBuffer();
        
        sql.append("INSERT INTO audit_log (log_date_time, call_type, "
                + "error_msg, resp_time, doc_count, resp_cached, queryinfo, systemid)"
                + " VALUES (");
        sql.append('\'' + getCurrentDateTime() + '\'');
        sql.append(", ");
        sql.append('\'' + fixSQLString(type) + '\'');
        sql.append(", ");
        sql.append('\'' + fixSQLString(errorstr) + '\'');
        sql.append(", ");
        sql.append(msec);
        sql.append(", ");
        sql.append(count);
        sql.append(", ");
        sql.append('\'' + fixSQLString(cached) + '\'');
        sql.append(", ");
        sql.append('\'' + fixSQLString(queryinfo) + '\'');
        sql.append(", ");
        sql.append('\'' + fixSQLString(getId()) + '\'');
        sql.append(")");
        sql.append(System.getProperty("line.separator"));
        sql.append("GO");
        sql.append(System.getProperty("line.separator"));

        writeLine(sql.toString());
        
    }
    
    /**
     * Store event to XML file.
     * 
     * @param edipn patient id
     * @param type type of request
     * @param errorstr error string
     * @param msec length of call
     * @param count number of items returned
     * @param cached cached flag
     */
    private void storeXML(String edipn, String type, String errorstr,
                          long msec, int count, String cached, String queryinfo) {
        StringBuffer xml = new StringBuffer();
        
        xml.append("<audit_log log_date_time=");
        xml.append('\"' + getCurrentDateTime() + '\"');
        xml.append(" call_type=");
        xml.append('\"' + fixSQLString(type) + '\"');
        xml.append(" error_msg=");
        xml.append('\"' + fixSQLString(errorstr) + '\"');
        xml.append(" resp_time=");
        xml.append(msec);
        xml.append(" doc_count=");
        xml.append(count);
        xml.append(" resp_cached=");
        xml.append('\"' + fixSQLString(cached) + '\"');
        xml.append(" queryinfo=");
        xml.append('\"' + fixSQLString(queryinfo) + '\"');
        xml.append(" />");
        xml.append(System.getProperty("line.separator"));

        writeLine(xml.toString());
        
    }

    private void writeLine(String line) {
        checkRolling();
        synchronized (lock) {
            PrintWriter out = null;
            BufferedWriter bwrite = null;
            OutputStreamWriter fwrite = null;
            try {
                fwrite = new OutputStreamWriter(new FileOutputStream(filenamedate, true), "UTF-8");
                bwrite = new BufferedWriter(fwrite);
                out = new PrintWriter(bwrite);
                out.print(line);
            } catch (IOException ex) {
                LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
            } finally {
                if (out != null) {
                    out.close();
                }
                if (bwrite != null) {
                    try {
                        bwrite.close();
                    } catch (IOException ex) {
                        LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
                    }
                }
                if (fwrite != null) {
                    try {
                        fwrite.close();
                    } catch (IOException ex) {
                        LOGGER.error(EXCEPTION + ex.getLocalizedMessage(), ex);
                    }
                }
            }
        }
    }
    
    private String getCurrentDateTime() {
        Calendar cal = Calendar.getInstance(TimeZone.getDefault());
        java.text.SimpleDateFormat sdf =
            new java.text.SimpleDateFormat("yyyyMMddHHmmss.SSSZ");
        sdf.setTimeZone(TimeZone.getDefault());

        return sdf.format(cal.getTime());
    }

    /**
     * Check to see if file should be rolled.
     */
    protected void checkRolling() {
        synchronized (lock) {
            Date cdate = new Date();
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd", Locale.getDefault());
            String dstr = sdf.format(cdate);
            int dot = filename.lastIndexOf('.');
            if (dot >= 0) {
                String name = filename.substring(0, dot);
                String ext = filename.substring(dot + 1);
                filenamedate = name + dstr + "." + ext;
            } else {
                filenamedate = filename + dstr;
            }
        }
    }
    
    /**
     * Get computer id.
     * 
     * @return id
     */
    private static String getId() {
        StringBuffer idstring = new StringBuffer();

        String hostname = System.getProperty("COMPUTERNAME", System.getenv().get("COMPUTERNAME"));
        if (hostname == null || hostname.length() == 0) {
            hostname = System.getProperty("HOSTNAME", System.getenv().get("HOSTNAME"));
        }
        if (hostname != null && hostname.length() > 0) {
            idstring.append(hostname);
            idstring.append('|');
        }

        try {
            Enumeration<NetworkInterface> nets = NetworkInterface.getNetworkInterfaces();
            for (NetworkInterface netint : Collections.list(nets)) {
                if (netint.isUp() && (!netint.isLoopback()) && (!netint.isPointToPoint()) && (!netint.isVirtual())) {
                    Enumeration<InetAddress> inetAddresses = netint.getInetAddresses();
                    for (InetAddress inetAddress : Collections.list(inetAddresses)) {
                        if (inetAddress instanceof Inet4Address) {
                            if (idstring.length() == 0) {
                                idstring.append(inetAddress.getHostName());
                                idstring.append('|');
                            }
                            idstring.append(inetAddress.getHostAddress());
                            idstring.append('|');
                        }
                    }
//                    byte[] mac = netint.getHardwareAddress(); 
//                    if (mac != null) { 
//                        for (int i = 0; i < mac.length; i++) { 
//                            if (i > 0) {
//                                idstring.append('-');    
//                            }
//                            idstring.append(String.format("%02X", mac[i]));
//                        } 
//                    }
//                    idstring.append('|');
                }
            }
        } catch (SocketException ex) {
            idstring.append('|');
        }
        return idstring.toString();
    }

    /**
     * Get computer hostName.
     *
     * @return hostName
     */
    private static String getHostName() {
        StringBuffer idstring = new StringBuffer();
        idstring.append("");
        String hostname = System.getProperty("COMPUTERNAME", System.getenv().get("COMPUTERNAME"));
        if (hostname == null || hostname.length() == 0) {
            hostname = System.getProperty("HOSTNAME", System.getenv().get("HOSTNAME"));
        }
        if (hostname != null && hostname.length() > 0) {
            idstring.append(hostname);
        } else {
            try {
                Enumeration<NetworkInterface> nets = NetworkInterface.getNetworkInterfaces();
                netLoop:
                for (NetworkInterface netint : Collections.list(nets)) {
                    if (netint.isUp() && (!netint.isLoopback()) && (!netint.isPointToPoint()) && (!netint.isVirtual())) {
                        Enumeration<InetAddress> inetAddresses = netint.getInetAddresses();
                        for (InetAddress inetAddress : Collections.list(inetAddresses)) {
                            if (inetAddress instanceof Inet4Address) {
                                if (idstring.length() == 0) {
                                    idstring.append(inetAddress.getHostName());
                                    break netLoop;
                                }
                                idstring.append(inetAddress.getHostAddress());
                                break netLoop;
                            }
                        }
                    }
                }
            } catch (SocketException ex) {
                idstring.append("noHostName");
            }
        }

        return idstring.toString();
    }

    /**
     * Format data for SQL statements.
     * 
     * @param inStr input
     * @return output
     */
    protected String fixSQLString(String inStr) {
        StringBuffer outStr = new StringBuffer();
        if (inStr != null && inStr.length() > 0) {
//            String chrstring = "\'||" + System.getProperty("line.separator") + "Chr";
            String chrstring = "\'+" + System.getProperty("line.separator") + "Char";

            for (int i = 0; i < inStr.length(); ++i) {
                char chr = inStr.charAt(i);
                if (chr == '\'') {
                    outStr.append(chrstring);
                    outStr.append("(39)||\'");
                } else if (chr == '&') {
                    outStr.append(chrstring);
                    outStr.append("(38)||\'");
                } else if (chr == '\"') {
                    outStr.append(chrstring);
                    outStr.append("(34)||\'");
                } else if (chr == '\\')  {
                    outStr.append(chrstring);
                    outStr.append("(92)||\'");
                } else if (chr == 10)  {
                    outStr.append(chrstring);
                    outStr.append("(10)||\'");
                } else if (chr == 13)  {
                    outStr.append(chrstring);
                    outStr.append("(13)||\'");
                } else {
                    outStr.append(chr);
                }
            }
        }
        return outStr.toString();
    }

    public static String cleanString(String input) {
        String message = "";
        if (input != null) {
            message = input;
        }
        // ensure no CRLF injection into logs for forging records
        String clean = escapeHTML(message.replace('\n', '_').replace('\r', '_'));
        return clean;
    }

    public static String escapeHTML(String s) {
        StringBuilder out = new StringBuilder(Math.max(16, s.length()));
        for (int i = 0; i < s.length(); i++) {
            char c = s.charAt(i);
            if (c > 127 || c == '"' || c == '<' || c == '>' || c == '&') {
                out.append("&#");
                out.append((int) c);
                out.append(';');
            } else {
                out.append(c);
            }
        }
        return out.toString();
    }

    /**
     * Make sure things get cleaned up.
     *
     * @throws Throwable throw
     */
    protected void finalize() throws Throwable{
        try {
            close();
        } finally {
            super.finalize();
        }
    }
    
 //   public static void main(String[] args) {
 //       ARMLog log = new ARMLog("text//tmp/test.log");
 //       log.logEvent("1000099999", "type", "error", 1500, 1, "true");
 //       log.logEvent("1000099999", "type", "error", 5000, 1, "true");
 //       log.logEvent("1000099999", "type", "error", 500, 1, "true");
 //   }
}
