/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
package gov.va.med.bhie.webservice;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import gov.va.med.bhie.beans.*;
import gov.va.med.bhie.common.AppConfig;
import gov.va.med.bhie.common.ClinicalDomainLoincCode;
import gov.va.med.bhie.dao.DaoException;
import gov.va.med.bhie.dao.patient.NoteDao;
import gov.va.med.bhie.dao.patient.V4Dao;
import gov.va.med.bhie.dao.patient.VLERDao;
import gov.va.med.bhie.dataservice.BHIECacheClient;
import gov.va.med.bhie.dataservice.SnareworksHandler;
import gov.va.med.bhie.resultcollection.ResultCollection;
import gov.va.med.bhie.resultcollection.V4ResultCollection;
import gov.va.med.bhie.util.DBLogger;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Resource;
import javax.jws.HandlerChain;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.mail.MessagingException;
import javax.mail.internet.MimeUtility;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import static gov.va.med.bhie.common.ClinicalDomainLoincCode.VLERCCDA;

//import gov.va.med.foundations.utilities.ARMLog;

/**
 * BHIE Relay web service.
 *
 * @author brunos
 * @author <a href="mailto:kweller@hawaiirg.com">kweller</a>
 * @version 1.0
 *
 */
@WebService
@HandlerChain(file = "SoapHandlerChain.xml")
public class BHIERelay {

    /**
     * Logger.
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(BHIERelay.class);

    /**
     * Web context.
     */
    @Resource
    private WebServiceContext wsContext;


    /**
     * Get deployment forms.
     *
     * @param queryBean query bean
     * @return deployment list collection
     */
    @WebMethod(operationName = "getDeploymentForms")
    public V4ResultCollection getDeploymentForms(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.DEPLOYMENT_FORMS, queryBean);
    }
    
    /**
     * Get patient encounter notes.
     *
     * @param queryBean query bean
     * @return note collection
     */
    @WebMethod(operationName = "getEncounterNotes")
    public V4ResultCollection getEncounterNotes(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.ENCOUNTER_NOTES, queryBean);
    }

    /**
     * Get HAIMS documents.
     *
     * @param queryBean query bean
     * @return note collection
     */
    @WebMethod(operationName = "getHAIMSDocuments")
    public V4ResultCollection getHAIMSDocuments(@WebParam(name = "queryBean") QueryBean queryBean) {
        String oid = AppConfig.getInstance().getProperties().getProperty(
                "includedsources." + ClinicalDomainLoincCode.DOCUMENTS.getLoincCode() + ".haims");
        List<String> includedSources = new ArrayList<String>();
        includedSources.add(oid);
        queryBean.setIncludedSources(includedSources);
        return getV4(ClinicalDomainLoincCode.DOCUMENTS, queryBean);
    }

    /**
     * Get MHS Genesis documents.
     *
     * @param queryBean query bean
     * @return note collection
     */
    @WebMethod(operationName = "getMhsGenesisDocuments")
    public V4ResultCollection getMhsGenesisDocuments(@WebParam(name = "queryBean") QueryBean queryBean) {
        String oid = AppConfig.getInstance().getProperties().getProperty(
                "includedsources." + ClinicalDomainLoincCode.DOCUMENTS.getLoincCode() + ".genesis");
        List<String> includedSources = new ArrayList<String>();
        includedSources.add(oid);
        queryBean.setIncludedSources(includedSources);
        return getV4(ClinicalDomainLoincCode.DOCUMENTS, queryBean);
    }

//    /**
//     * Get inpatient notes (Essentris notes).
//     *
//     * @param queryBean query bean
//     * @return note collection
//     */
//    @WebMethod(operationName = "getInpatientNotes")
//    public V4ResultCollection getInpatientNotes(@WebParam(name = "queryBean") QueryBean queryBean) {
//        return getV4(ClinicalDomainLoincCode.INPATIENT_NOTES, queryBean);
//    }

    /**
     * Get note image.
     *
     * @param eventId event ID
     * @return note image
     */
    @WebMethod(operationName = "getNoteImage")
    public NoteImage getNoteImage(@WebParam(name = "eventId") String eventId) {
        NoteDao noteDao = new NoteDao();
        try {
            NoteImage rtc;
            rtc = noteDao.getNoteImage(eventId);
            if (!"false".equalsIgnoreCase((String) AppConfig.getInstance().getProperties().get("fixrtf"))) {
                rtc = NoteFixer.fixNoteImage(rtc, "pdf".equalsIgnoreCase(AppConfig.getInstance().getNoteConvertFormat()));
            }

            if (rtc != null) {
                String contentType = rtc.getContentType();
                /* flatten pdf note types to remove fillable pdf */
                if ("application/octet-stream".equalsIgnoreCase(contentType)
                        || "application/pdf".equalsIgnoreCase(contentType)) {
                    rtc = NoteFixer.flattenNoteForm(rtc);
                }
            }


            return rtc;
        } catch (DaoException e) {
            logError(e);
            throw new RuntimeException(e);
        }
    }

    /**
     * Get note image encoded.
     *
     * @param url Note image url.
     * @return note image
     */
    @WebMethod(operationName = "getNoteImageEncodedFromUrl")
    public NoteImageEncoded getNoteImageEncodedFromUrl(@WebParam(name = "url") String url) {
        NoteDao noteDao = new NoteDao();
        try {
            NoteImage rtc;
            rtc = noteDao.getNoteImageFromUrl(url);
            if (!"false".equalsIgnoreCase((String) AppConfig.getInstance().getProperties().get("fixrtf"))) {
                rtc = NoteFixer.fixNoteImage(rtc, "pdf".equalsIgnoreCase(AppConfig.getInstance().getNoteConvertFormat()));
            }

            String contentType = rtc.getContentType();
            /* flatten pdf note types to remove fillable pdf */
            if (rtc != null && "application/octet-stream".equalsIgnoreCase(contentType)
                    || "application/pdf".equalsIgnoreCase(contentType)) {
                rtc = NoteFixer.flattenNoteForm(rtc);
            }

            NoteImageEncoded rtc2 = new NoteImageEncoded();
            rtc2.setContentDisposition(rtc.getContentDisposition());
            rtc2.setContentType(rtc.getContentType());

            ByteArrayOutputStream baos = null;
            OutputStream b64os = null;
            try {
                baos = new ByteArrayOutputStream();
                b64os = MimeUtility.encode(baos, "base64");
                b64os.write(rtc.getNoteBytes());
                b64os.close();
                rtc2.setNoteBytesEncoded(new String(baos.toByteArray(), "UTF-8"));
            } finally {
                if (b64os != null) {
                    b64os.close();
                }
                if (baos != null) {
                    baos.close();
                }
            }
            return rtc2;
        } catch (IOException e) {
            logError(e);
            throw new RuntimeException(e);
        } catch (MessagingException e) {
            logError(e);
            throw new RuntimeException(e);
        } catch (DaoException e) {
            logError(e);
            throw new RuntimeException(e);
        }
    }

    /**
     * Get note image encoded.
     *
     * @param eventId event ID
     * @return note image
     */
    @WebMethod(operationName = "getNoteImageEncoded")
    public NoteImageEncoded getNoteImageEncoded(@WebParam(name = "eventId") String eventId) {
        NoteDao noteDao = new NoteDao();
        try {
            NoteImage rtc;
            rtc = noteDao.getNoteImage(eventId);
            if (!"false".equalsIgnoreCase((String) AppConfig.getInstance().getProperties().get("fixrtf"))) {
                rtc = NoteFixer.fixNoteImage(rtc, "pdf".equalsIgnoreCase(AppConfig.getInstance().getNoteConvertFormat()));
            }

            if (rtc != null) {
                String contentType = rtc.getContentType();
                /* flatten pdf note types to remove fillable pdf */
                if ("application/octet-stream".equalsIgnoreCase(contentType)
                        || "application/pdf".equalsIgnoreCase(contentType)) {
                    rtc = NoteFixer.flattenNoteForm(rtc);
                }
            }

            NoteImageEncoded rtc2 = new NoteImageEncoded();
            rtc2.setContentDisposition(rtc.getContentDisposition());
            rtc2.setContentType(rtc.getContentType());

            ByteArrayOutputStream baos = null;
            OutputStream b64os = null;
            try {
                baos = new ByteArrayOutputStream();
                b64os = MimeUtility.encode(baos, "base64");
                b64os.write(rtc.getNoteBytes());
                b64os.close();
                rtc2.setNoteBytesEncoded(new String(baos.toByteArray(), "UTF-8"));
            } finally {
                if (b64os != null) {
                    b64os.close();
                }
                if (baos != null) {
                    baos.close();
                }
            }
            return rtc2;
        } catch (IOException e) {
            logError(e);
            throw new RuntimeException(e);
        } catch (MessagingException e) {
            logError(e);
            throw new RuntimeException(e);
        } catch (DaoException e) {
            logError(e);
            throw new RuntimeException(e);
        }
    }

    /**
     * Get note image.
     *
     * @param url Note image url.
     * @return note image
     */
    @WebMethod(operationName = "getNoteImageFromUrl")
    public NoteImage getNoteImageFromUrl(@WebParam(name = "url") String url) {
        NoteDao noteDao = new NoteDao();
        try {
            NoteImage rtc;
            rtc = noteDao.getNoteImageFromUrl(url);
            if (!"false".equalsIgnoreCase((String) AppConfig.getInstance().getProperties().get("fixrtf"))) {
                rtc = NoteFixer.fixNoteImage(rtc, "pdf".equalsIgnoreCase(AppConfig.getInstance().getNoteConvertFormat()));
            }

            if (rtc != null) {
                String contentType = rtc.getContentType();
                /* flatten pdf note types to remove fillable pdf */
                if ("application/octet-stream".equalsIgnoreCase(contentType)
                        || "application/pdf".equalsIgnoreCase(contentType)) {
                    rtc = NoteFixer.flattenNoteForm(rtc);
                }
            }



            return rtc;
        } catch (DaoException e) {
            logError(e);
            throw new RuntimeException(e);
        }
    }

    /**
     * Get outpatient procedures.
     * 
     * @param queryBean query bean
     * @return collection
     */
    @WebMethod(operationName = "getOutpatientProcedures")
    public V4ResultCollection getOutpatientProcedures(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.PROCEDURES, queryBean);
    }

    /**
     * Get patient admissions.
     *
     * @param queryBean query bean
     * @return admissions collection
     */
    @WebMethod(operationName = "getPatientAdmissions")
    public V4ResultCollection getPatientAdmissions(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.ADMISSIONSV4, queryBean);
    }
    
    /**
     * Get patient allergies.
     *
     * @param queryBean query bean
     * @return allergy collection
     */
    @WebMethod(operationName = "getPatientAllergies")
    public V4ResultCollection getPatientAllergies(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.ALLERGIESV4, queryBean);
    }

    /**
     * Get patient appointments.
     *
     * @param queryBean query bean
     * @return appointment collection
     */
    @WebMethod(operationName = "getPatientAppointments")
    public V4ResultCollection getPatientAppointments(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.APPOINTMENTS, queryBean);
    }
//
//    /**
//     * Get consult orders.
//     *
//     * @param queryBean query bean
//     * @return lab order collection
//     */
//    @WebMethod(operationName = "getPatientConsultOrders")
//    public V4ResultCollection getPatientConsultOrders(@WebParam(name = "queryBean") QueryBean queryBean) {
//        return getV4(ClinicalDomainLoincCode.CONSULT_ORDERS, queryBean);
//    }

    /**
     * Get patient consults.
     *
     * @param queryBean query bean
     * @return consult collection
     */
    @WebMethod(operationName = "getPatientConsults")
    public V4ResultCollection getPatientConsults(@WebParam(name = "queryBean") QueryBean queryBean) {
        //call notes, and filter out in jMeadows
        return getV4(ClinicalDomainLoincCode.NOTES, queryBean);
    }

    /**
     * Get patient demographics.
     *
     * @param queryBean query bean
     * @return demographic collection
     */
    @WebMethod(operationName = "getPatientDemographics")
    public V4ResultCollection getPatientDemographics(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.PATIENT_DEMOGRAPHICSV4, queryBean);
    }

    /**
     * Get discharge summaries.
     *
     * @param queryBean query bean
     * @return summary collection
     */
    @WebMethod(operationName = "getPatientDischargeSummaries")
    public V4ResultCollection getPatientDischargeSummaries(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.INPATIENT_NOTES, queryBean);
    }

    /**
     * Get patient encounters.
     *
     * @param queryBean query bean
     * @return encounter collection
     */
    @WebMethod(operationName = "getPatientEncounters")
    public V4ResultCollection getPatientEncounters(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.ENCOUNTERS, queryBean);
    }

    /**
     * Get patient family history list.
     *
     * @param queryBean query bean
     * @return family history list collection
     */
    @WebMethod(operationName = "getPatientFamilyHistory")
    public V4ResultCollection getPatientFamilyHistory(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.FAMILY_HISTORIES, queryBean);
    }
    
    /**
     * Get patient immunizations.
     *
     * @param queryBean query bean
     * @return immunization collection
     */
    @WebMethod(operationName = "getPatientImmunizations")
    public V4ResultCollection getPatientImmunizations(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.IMMUNIZATIONS, queryBean);
    }

    /**
     * Get patient insurance.
     * 
     * @param queryBean query bean
     * @return collection
     */
    @WebMethod(operationName = "getPatientInsurances")
    public V4ResultCollection getPatientInsurances(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.INSURANCES, queryBean);
    }

        /**
     * Get chemistry labs.
     *
     * @param queryBean query bean
     * @return chemistry collection
     */
    @WebMethod(operationName = "getPatientLabChemistries")
    public V4ResultCollection getPatientLabChemistries(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.LAB_CHEMISTRIES, queryBean);

    }


    /**
     * Get patient AP tests.
     *
     * @param queryBean query bean
     * @return AP test collection
     */
    @WebMethod(operationName = "getPatientLabAnatomicPathologies")
    public V4ResultCollection getPatientLabAnatomicPathologies(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.LAB_ANATOMIC_PATHOLOGIES, queryBean);

    }

    /**
     * Get patient micros.
     *
     * @param queryBean query bean
     * @return micro collection
     */
    @WebMethod(operationName = "getPatientLabMicrobiologies")
    public V4ResultCollection getPatientLabMicrobiologies(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.LAB_MICROBIOLOGIES, queryBean);

    }


//    /**
//     * Get lab orders.
//     *
//     * @param queryBean query bean
//     * @return lab order collection
//     */
//    @WebMethod(operationName = "getPatientLabOrders")
//    public V4ResultCollection getPatientLabOrders(@WebParam(name = "queryBean") QueryBean queryBean) {
//        return getV4(ClinicalDomainLoincCode.LAB_ORDERS, queryBean);
//    }

//    /**
//     * Get medication orders.
//     *
//     * @param queryBean query bean
//     * @return medication order collection
//     */
//    @WebMethod(operationName = "getPatientMedicationOrders")
//    public V4ResultCollection getPatientMedicationOrders(@WebParam(name = "queryBean") QueryBean queryBean) {
//        return getV4(ClinicalDomainLoincCode.MEDICATION_ORDERS, queryBean);
//    }
//

    /**
     * Get patient management. (patient flags)
     *
     * @param queryBean query bean
     * @return patientManagement list collection
     */
    @WebMethod(operationName = "getPatientManagement")
    public V4ResultCollection getPatientManagement(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.PATIENT_MANAGEMENT, queryBean);
    }

    /**
     * Get patient medications.
     *
     * @param queryBean query bean
     * @return medication collection
     */
    @WebMethod(operationName = "getPatientMedications")
    public V4ResultCollection getPatientMedications(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.MEDICATIONS, queryBean);
    }

    /**
     * Get patient notes.
     *
     * @param queryBean query bean
     * @return note collection
     */
    @WebMethod(operationName = "getPatientNotes")
    public V4ResultCollection getPatientNotes(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.NOTES, queryBean);
    }

    /**
     * Get patient orders.
     *
     * @param queryBean query bean
     * @return radiology order collection
     */
    @WebMethod(operationName = "getPatientOrders")
    public V4ResultCollection getPatientOrders(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.ALL_ORDERS, queryBean);
    }

    /**
     * Get patient other history list.
     *
     * @param queryBean query bean
     * @return other history list collection
     */
    @WebMethod(operationName = "getPatientOtherHistory")
    public V4ResultCollection getPatientOtherHistory(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.OTHER_PAST_MEDICAL_HISTORIES, queryBean);
    }

    /**
     * Get patient problem list.
     *
     * @param queryBean query bean
     * @return problem list collection
     */
    @WebMethod(operationName = "getPatientProblemList")
    public V4ResultCollection getPatientProblemList(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.PROBLEMS, queryBean);
    }

    /**
     * Get Questionnaire.
     *
     * @param queryBean query bean
     * @return questionnaire list collection
     */
    @WebMethod(operationName = "getPatientQuestionnaires")
    public V4ResultCollection getPatientQuestionnaires(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.QUESTIONNAIRES, queryBean);
    }

//    /**
//     * Get radiology orders.
//     *
//     * @param queryBean query bean
//     * @return radiology order collection
//     */
//    @WebMethod(operationName = "getPatientRadiologyOrders")
//    public V4ResultCollection getPatientRadiologyOrders(@WebParam(name = "queryBean") QueryBean queryBean) {
//        return getV4(ClinicalDomainLoincCode.RADIOLOGY_ORDERS, queryBean);
//    }

    /**
     * Get patient radiology reports.
     *
     * @param queryBean query bean
     * @return report collection
     */
    @WebMethod(operationName = "getPatientRadiologyReports")
    public V4ResultCollection getPatientRadiologyReports(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.RADIOLOGY_REPORTSV4, queryBean);
    }

    /**
     * Get patient social history list.
     *
     * @param queryBean query bean
     * @return social history list collection
     */
    @WebMethod(operationName = "getPatientSocialHistory")
    public V4ResultCollection getPatientSocialHistory(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.SOCIAL_HISTORIES, queryBean);
    }

    /**
     * Get patient vitals.
     *
     * @param queryBean query bean
     * @return vital collection
     */
    @WebMethod(operationName = "getPatientVitals")
    public V4ResultCollection getPatientVitals(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getV4(ClinicalDomainLoincCode.VITALS, queryBean);
    }

    /**
     * Get VLER patient admissions.
     * 
     * @param queryBean query bean
     * @return collection
     */
    @WebMethod(operationName = "getVLERPatientAdmissions")
    public V4ResultCollection getVLERPatientAdmissions(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERADMISSIONS, queryBean);
    }
    
    /**
     * Get VLER patient allergies.
     *
     * @param queryBean query bean
     * @return allergy collection
     */
    @WebMethod(operationName = "getVLERPatientAllergies")
    public V4ResultCollection getVLERPatientAllergies(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERALLERGIES, queryBean);
    }

    /**
     * Get VLER patient appointments.
     *
     * @param queryBean query bean
     * @return appointment collection
     */
    @WebMethod(operationName = "getVLERPatientAppointments")
    public V4ResultCollection getVLERPatientAppointments(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERAPPOINTMENTS, queryBean);
    }
    
    /**
     * Get VLER CCDA document.
     * 
     * @param queryBean query bean.
     * @return collection
     */
    @WebMethod(operationName = "getVLERPatientCCDA")
    public V4ResultCollection getVLERPatientCCDA(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERCCDA, queryBean);
    }

    /**
     * Get VLER CCDA document as JSON string.
     * 
     * @param queryBean query bean
     * @return JSON CCDA
     */
    @WebMethod(operationName = "getVLERPatientCCDARaw")
    public String getVLERPatientCCDARaw(@WebParam(name = "queryBean") QueryBean queryBean) {
        String rtc = "";
        V4ResultCollection rcoll = getVLER(VLERCCDA, queryBean);
        ObjectMapper objectMapper = new ObjectMapper();
        try {
            rtc = objectMapper.writeValueAsString(rcoll);
        } catch (JsonProcessingException e) {
            rtc = "";
        }
        return rtc;
    }

    /**
     * Get VLER patient demographics.
     * 
     * @param queryBean query bean.
     * @return collection
     */
    @WebMethod(operationName = "getVLERPatientDemographics")
    public V4ResultCollection getVLERPatientDemographics(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERPATIENT_DEMOGRAPHICS, queryBean);
    }
    
    /**
     * Get VLER patient Immunizations.
     *
     * @param queryBean query bean
     * @return immunization collection
     */
    @WebMethod(operationName = "getVLERPatientImmunizations")
    public V4ResultCollection getVLERPatientImmunizations(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERIMMUNIZATIONS, queryBean);
    }
    
    /**
     * Get VLER patient insurance.
     * 
     * @param queryBean query bean
     * @return collection
     */
    @WebMethod(operationName = "getVLERPatientInsurances")
    public V4ResultCollection getVLERPatientInsurances(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERINSURANCE, queryBean);
    }
    
    /**
     * Get VLER patient Medications.
     *
     * @param queryBean query bean
     * @return medication collection
     */
    @WebMethod(operationName = "getVLERPatientMedications")
    public V4ResultCollection getVLERPatientMedications(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERMEDICATIONS, queryBean);
    }
    
    /**
     * Get VLER patient problem list.
     *
     * @param queryBean query bean
     * @return problem list collection
     */
    @WebMethod(operationName = "getVLERPatientProblemList")
    public V4ResultCollection getVLERPatientProblemList(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERPROBLEMS, queryBean);
    }
    
    /**
     * Get VLER results.
     *
     * @param queryBean query bean
     * @return results collection
     */
    @WebMethod(operationName = "getVLERPatientResults")
    public V4ResultCollection getVLERPatientResults(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERRESULTS, queryBean);
    }
    
    /**
     * Get VLER patient vitals.
     *
     * @param queryBean query bean
     * @return vitals collection
     */
    @WebMethod(operationName = "getVLERPatientVitals")
    public V4ResultCollection getVLERPatientVitals(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERVITALS, queryBean);
    }
    
    /**
     * Get VLER patient procedures.
     *
     * @param queryBean query bean
     * @return procedures collection
     */
    @WebMethod(operationName = "getVLERPatientProcedures")
    public V4ResultCollection getVLERPatientProcedures(@WebParam(name = "queryBean") QueryBean queryBean) {
        return getVLER(ClinicalDomainLoincCode.VLERPROCEDURES, queryBean);
    }
    
    /**
     * Log request.
     *
     * @param queryBean query bean
     * @return start time
     */
    private long logRequest(QueryBean queryBean) {
        if (wsContext != null) {
            MessageContext msgctx = wsContext.getMessageContext();
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss", Locale.getDefault());

            if (queryBean == null) {
                LOGGER.error("begin {} (queryBean is null) ", msgctx.get(MessageContext.WSDL_OPERATION));
            } else {
                LOGGER.info("begin " + msgctx.get(MessageContext.WSDL_OPERATION) + " (" + DBLogger.cleanString(queryBean.getPatientId()) + ") ");
                if (LOGGER.isDebugEnabled()) {
                    if (queryBean.getStartDate() != null) {
                        LOGGER.debug("info " + msgctx.get(MessageContext.WSDL_OPERATION) + " (" + DBLogger.cleanString(queryBean.getPatientId()) + ") startDate = " + sdf.format(queryBean.getStartDate().getTime()));
                    }
                    if (queryBean.getEndDate() != null) {
                        LOGGER.debug("info " + msgctx.get(MessageContext.WSDL_OPERATION) + " (" + DBLogger.cleanString(queryBean.getPatientId()) + ") endDate = " + sdf.format(queryBean.getEndDate().getTime()));
                    }
                    if (queryBean.getIncludedSources() != null) {
                        List<String> src = queryBean.getIncludedSources();
                        if (CollectionUtils.isNotEmpty(src)) {
                            int srcSize = src.size();
                            for (int i = 0; i < srcSize; ++i) {
                                LOGGER.debug("info " + msgctx.get(MessageContext.WSDL_OPERATION) + " (" + DBLogger.cleanString(queryBean.getPatientId()) + ") include src = " + DBLogger.cleanString(src.get(i)));
                            }
                        }
                    }
                    if (queryBean.getExcludedSources() != null) {
                        List<String> src = queryBean.getExcludedSources();
                        if (CollectionUtils.isNotEmpty(src)) {
                            int srcSize = src.size();
                            for (int i = 0; i < srcSize; ++i) {
                                LOGGER.debug("info " + msgctx.get(MessageContext.WSDL_OPERATION) + " (" + DBLogger.cleanString(queryBean.getPatientId()) + ") exclude src = " + DBLogger.cleanString(src.get(i)));
                            }
                        }
                    }
                }
            }
        }
        return System.currentTimeMillis();
    }

    /**
     * Log response.
     *
     * @param queryBean query bean
     * @param exception exception
     * @param rtc collection
     * @param count count in return collection
     * @param stime start time
     * @param cached cached
     */
    private void logResponse(String operation, QueryBean queryBean, String exception, ResultCollection rtc, int count, long stime, String cached) {
        long etime = System.currentTimeMillis();
        String patientId = "Patient ID was not set";
        String queryInfo = "";
        int scount = 0;
        int ecount = 0;

        if (wsContext != null) {
            if (queryBean != null) {
                patientId = queryBean.getPatientId();
                BHIELimit limit = BHIELimit.getInstance();
                queryInfo = limit.getKey(queryBean, "");
            }
            
            if (rtc == null) {
                BHIEStatus.updateStatistics(operation, "false", true, etime - stime);
                LOGGER.info("end {} milliseconds {} ({}) rtc=null exception={}",  (etime - stime), operation,  DBLogger.cleanString(patientId), exception);
                String dsource = AppConfig.getInstance().getDataSource();
                DBLogger dbl = new DBLogger(dsource);
                dbl.logEvent(patientId, "BHIERelay." + operation, exception, etime - stime, 0, "false", queryInfo);
                dbl.close();
            } else {
                BHIEStatus.updateStatistics(operation, cached, false, etime - stime);
                String errormsg = ""; //rtc.getError();
                if (rtc.getStatusList() != null) {
                    scount = rtc.getStatusList().size();
                }
                if (rtc.getErrorList() != null) {
                    ecount = rtc.getErrorList().size();

                    List<gov.va.med.bhie.beans.Error> src = rtc.getErrorList();
                    if (CollectionUtils.isNotEmpty(src)) {
                        int srcSize = src.size();
                        for (int i = 0; i < srcSize; ++i) {
                            if (src.get(i).getErrorCode() != null && src.get(i).getErrorCode().length() > 0) {
                                errormsg = src.get(i).getErrorCode();
                                break;
                            }
                        }
                    }
                    if (ecount > 0 && (errormsg == null || errormsg.isEmpty())) {
                        errormsg = Integer.toString(ecount) + " errors returned";
                    }
                }
                if (errormsg == null) {
                    errormsg = "";
                }
                LOGGER.info("end {} milliseconds {} ({}) Complete rtc.size()={} errorsize={} statussize={}", (etime - stime), operation, DBLogger.cleanString(patientId), count, ecount, scount);
                String dsource = AppConfig.getInstance().getDataSource();
                DBLogger dbl = new DBLogger(dsource);
                dbl.logEvent(patientId, "BHIERelay." + operation, errormsg, etime - stime, count, cached);
                dbl.close();
                if (LOGGER.isDebugEnabled()) {
                    if (rtc.getQueryComplete() != null && rtc.getQueryComplete().length() > 0) {
                        LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") queryComplete = " + DBLogger.cleanString(rtc.getQueryComplete()));
                    }
                    if (rtc.getQueryComplete() != null && rtc.getErrorList().size() > 0) {
                        LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Error = " + DBLogger.cleanString(errormsg));
                    }
                    if (rtc.getStatusList() != null) {
                        LOGGER.debug("complete info " + operation + " (" + DBLogger.cleanString(patientId) + ") StatusList.size() = " + rtc.getStatusList().size());
                        List<SiteStatus> src = rtc.getStatusList();
                        if (CollectionUtils.isNotEmpty(src)) {
                            int srcSize = src.size();
                            for (int i = 0; i < srcSize; ++i) {
                                if (src.get(i).getExpectedCount() != null && src.get(i).getExpectedCount().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Status[" + i + "] expected count=" + DBLogger.cleanString(src.get(i).getExpectedCount()));
                                }
                                if (src.get(i).getRetrievedCount() != null && src.get(i).getRetrievedCount().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Status[" + i + "] retrieved count=" + DBLogger.cleanString(src.get(i).getRetrievedCount()));
                                }
                                if (src.get(i).getSiteId() != null && src.get(i).getSiteId().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Status[" + i + "] site id=" + DBLogger.cleanString(src.get(i).getSiteId()));
                                }
                                if (src.get(i).getSiteName() != null && src.get(i).getSiteName().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Status[" + i + "] site name=" + DBLogger.cleanString(src.get(i).getSiteName()));
                                }
                                if (src.get(i).getSiteStatus() != null && src.get(i).getSiteStatus().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Status[" + i + "] site status=" + DBLogger.cleanString(src.get(i).getSiteStatus()));
                                }
                            }
                        }
                    }
                    if (rtc.getErrorList() != null) {
                        LOGGER.debug("complete info " + operation + " (" + DBLogger.cleanString(patientId) + ") ErrorList.size() = " + rtc.getErrorList().size());
                        List<gov.va.med.bhie.beans.Error> src = rtc.getErrorList();
    
                        if (CollectionUtils.isNotEmpty(src)) {
                            int srcSize = src.size();
                            for (int i = 0; i < srcSize; ++i) {
                                if (src.get(i).getErrorCode() != null && src.get(i).getErrorCode().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Error[" + i + "] Error code=" + DBLogger.cleanString(src.get(i).getErrorCode()));
                                }
                                if (src.get(i).getErrorValue() != null && src.get(i).getErrorValue().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Error[" + i + "] Error value=" + DBLogger.cleanString(src.get(i).getErrorValue()));
                                }
                                if (src.get(i).getErrorLocation() != null && src.get(i).getErrorLocation().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Error[" + i + "] Error loc=" + DBLogger.cleanString(src.get(i).getErrorLocation()));
                                }
                                if (src.get(i).getErrorSeverity() != null && src.get(i).getErrorSeverity().length() > 0) {
                                    LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Error[" + i + "] Error sev=" + DBLogger.cleanString(src.get(i).getErrorSeverity()));
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    /**
     * Log response.
     *
     * @param queryBean query bean
     * @param exception exception
     * @param rtc collection
     * @param count count in return collection
     * @param stime start time
     * @param cached cached
     */
    private void logVLERResponse(String operation, QueryBean queryBean, String exception, V4ResultCollection rtc, int count, long stime, String cached) {
        //TODO: Refactor both logResponse methods into one.
        long etime = System.currentTimeMillis();
        String patientId = "Patient ID was not set";
        String queryInfo = "";
        int scount = 0;
        int ecount = 0;

        if (wsContext != null) {
            if (queryBean != null) {
                patientId = queryBean.getPatientId();
                BHIELimit limit = BHIELimit.getInstance();
                queryInfo = limit.getKey(queryBean, "");
            }
            
            if (rtc == null) {
                BHIEStatus.updateStatistics(operation, "false", true, etime - stime);
                LOGGER.info("end {} milliseconds {} ({}) rtc=null exception={} ", (etime - stime), operation, DBLogger.cleanString(patientId), DBLogger.cleanString(exception));
                String dsource = AppConfig.getInstance().getDataSource();
                DBLogger dbl = new DBLogger(dsource);
                dbl.logEvent(patientId, "BHIERelay." + operation, exception, etime - stime, 0, "false", queryInfo);
                dbl.close();
            } else {
                String errormsg = "";
                if (rtc.getErrorList() != null && !rtc.getErrorList().isEmpty()) {
                    errormsg = rtc.getErrorListv4().get(0);
                }
                BHIEStatus.updateStatistics(operation, cached, false, etime - stime);
                if (rtc.getStatusList() != null) {
                    scount = rtc.getStatusList().size();
                }
                LOGGER.info("end {} milliseconds {} ({}) Complete rtc.size()={} errorsize={} statussize={}", (etime - stime), operation, DBLogger.cleanString(patientId), count, ecount, scount);
                String dsource = AppConfig.getInstance().getDataSource();
                DBLogger dbl = new DBLogger(dsource);
                dbl.logEvent(patientId, "BHIERelay." + operation, errormsg, etime - stime, count, cached);
                dbl.close();
                if (LOGGER.isDebugEnabled()) {
                    if (rtc.getQueryComplete() != null && rtc.getQueryComplete().length() > 0) {
                        LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") queryComplete = " + DBLogger.cleanString(rtc.getQueryComplete()));
                    }
                    if (rtc.getQueryComplete() != null && errormsg.length() > 0) {
                        LOGGER.debug("rtc info " + operation + " (" + DBLogger.cleanString(patientId) + ") Error = " + DBLogger.cleanString(errormsg));
                    }
                }
            }
        }
    }
    
    /**
     * Get VLER data.
     * 
     * @param LOINC LOINC code
     * @param queryBean query bean
     * @return collection
     */
    private V4ResultCollection getVLER(ClinicalDomainLoincCode LOINC, QueryBean queryBean) {
        V4ResultCollection rtc = null;
        long stime = logRequest(queryBean);

        try {
            VLERDao dao = new VLERDao();
            int count = 0;
            if(queryBean.getExcludedSources() != null)
                queryBean.getExcludedSources().clear();
            if(queryBean.getIncludedSources() != null)
                queryBean.getIncludedSources().clear();
            if(queryBean.getStartDate() != null)
                queryBean.setStartDate(null);
            if(queryBean.getEndDate() != null)
                queryBean.setEndDate(null);

            rtc = V4Limit.getInstance().getV4Cache("VLER" + VLERCCDA.getLoincCode(), queryBean);
            if (rtc == null) {
                rtc = dao.getVLERData(queryBean);
                if (rtc != null) {
                    count = getSize(LOINC, rtc);
                    if (AppConfig.getInstance().getRelayEndpointCount() > 1) {
                        String key = V4Limit.getInstance().getKey("VLER" + VLERCCDA.getLoincCode(), queryBean);
                        String rcoll = "";
                        ObjectMapper objectMapper = new ObjectMapper();
                        try {
                            rcoll = objectMapper.writeValueAsString(rtc);
                        } catch (JsonProcessingException e) {
                            rcoll = "";
                        }                    
                        distributeCache("V4ResultCollection", key, rcoll);
                    } else {
                        V4Limit.getInstance().putV4Cache("VLER" + VLERCCDA.getLoincCode(), queryBean, rtc);
                    }
                }
                logVLERResponse(VLERCCDA.getLoincCode(), queryBean, "", rtc, count, stime, "false");
            } else {
                logVLERResponse(VLERCCDA.getLoincCode(), queryBean, "", rtc, getSize(LOINC, rtc), stime, "true");
            }
        } catch (DaoException e) {
            logResponse(LOINC.getLoincCode(), queryBean, e.getLocalizedMessage(), null, 0, stime, "false");
            logError(e);
            throw new RuntimeException(e);
        }
        if (rtc != null) {
            rtc.setVler(true);
        }
        return rtc;
    }
    
    /**
     * Get BDA V4 data.
     * 
     * @param loinc loinc
     * @param queryBean query bean
     * @return collection 
     */
    private V4ResultCollection getV4(ClinicalDomainLoincCode loinc, QueryBean queryBean) {
        V4ResultCollection rtc = null;
        long stime = logRequest(queryBean);

        try {
            V4Dao dao = new V4Dao();
            int count = 0;
            rtc = dao.getV4Data(loinc, queryBean);
            if (rtc != null) {
                count = getSize(loinc, rtc);
            }
            logVLERResponse(loinc.getLoincCode(), queryBean, "", rtc, count, stime, "false");
        } catch (DaoException e) {
            logResponse(loinc.getLoincCode(), queryBean, e.getLocalizedMessage(), null, 0, stime, "false");
            logError(e);
            throw new RuntimeException(e);
        }
        return rtc;
    }
    
    /**
     * Get size of data list.
     * 
     * @param LOINC LOINC
     * @param rtc data
     * @return size
     */
    private int getSize(ClinicalDomainLoincCode LOINC, V4ResultCollection rtc) {
        int count = 0;
        if (rtc != null && rtc.getDataRecordList() != null) {
            if (LOINC == ClinicalDomainLoincCode.ADMISSIONSV4 && rtc.getDataRecordList().getAdmissions() != null) {
                count = rtc.getDataRecordList().getAdmissions().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERADMISSIONS && rtc.getDataRecordList().getAdmissions() != null) {
                count = rtc.getDataRecordList().getAdmissions().size();
            } else if (LOINC == ClinicalDomainLoincCode.ALLERGIESV4 && rtc.getDataRecordList().getAllergy() != null) {
                count = rtc.getDataRecordList().getAllergy().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERALLERGIES && rtc.getDataRecordList().getAllergy() != null) {
                count = rtc.getDataRecordList().getAllergy().size();
            } else if (LOINC == ClinicalDomainLoincCode.APPOINTMENTS && rtc.getDataRecordList().getAppointments() != null) {
                count = rtc.getDataRecordList().getAppointments().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERAPPOINTMENTS && rtc.getDataRecordList().getAppointments() != null) {
                count = rtc.getDataRecordList().getAppointments().size();
//            } else if (LOINC == ClinicalDomainLoincCode.CONSULT_ORDERS && rtc.getDataRecordList().getConsultOrders() != null) {
//                count = rtc.getDataRecordList().getConsultOrders().size();
            } else if (LOINC == ClinicalDomainLoincCode.PATIENT_DEMOGRAPHICSV4 && rtc.getDataRecordList().getDemographics() != null) {
                count = rtc.getDataRecordList().getDemographics().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERPATIENT_DEMOGRAPHICS && rtc.getDataRecordList().getDemographics() != null) {
                count = rtc.getDataRecordList().getDemographics().size();
            } else if (LOINC == ClinicalDomainLoincCode.PATIENT_MANAGEMENT && rtc.getDataRecordList().getPatientManagement() != null) {
                count = rtc.getDataRecordList().getPatientManagement().size();
            } else if (LOINC == ClinicalDomainLoincCode.DEPLOYMENT_FORMS && rtc.getDataRecordList().getDeploymentForms() != null) {
                count = rtc.getDataRecordList().getDeploymentForms().size();
            } else if (LOINC == ClinicalDomainLoincCode.ENCOUNTER_NOTES && rtc.getDataRecordList().getEncounterNotes() != null) {
                count = rtc.getDataRecordList().getEncounterNotes().size();
            } else if (LOINC == ClinicalDomainLoincCode.ENCOUNTERS && rtc.getDataRecordList().getEncounters() != null) {
                count = rtc.getDataRecordList().getEncounters().size();
            } else if (LOINC == ClinicalDomainLoincCode.IMMUNIZATIONS && rtc.getDataRecordList().getImmunizations() != null) {
                count = rtc.getDataRecordList().getImmunizations().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERIMMUNIZATIONS && rtc.getDataRecordList().getImmunizations() != null) {
                count = rtc.getDataRecordList().getImmunizations().size();
            } else if (LOINC == ClinicalDomainLoincCode.LAB_ANATOMIC_PATHOLOGIES && rtc.getDataRecordList().getLabApResults() != null) {
                count = rtc.getDataRecordList().getLabApResults().size();
            } else if (LOINC == ClinicalDomainLoincCode.LAB_CHEMISTRIES && rtc.getDataRecordList().getLabChemResults() != null) {
                count = rtc.getDataRecordList().getLabChemResults().size();
            } else if (LOINC == ClinicalDomainLoincCode.LAB_MICROBIOLOGIES && rtc.getDataRecordList().getLabMicroResults() != null) {
                count = rtc.getDataRecordList().getLabMicroResults().size();
//            } else if (LOINC == ClinicalDomainLoincCode.LAB_ORDERS && rtc.getDataRecordList().getLabOrders() != null) {
//                count = rtc.getDataRecordList().getLabOrders().size();
//            } else if (LOINC == ClinicalDomainLoincCode.MEDICATION_ORDERS && rtc.getDataRecordList().getMedicationOrders() != null) {
//                count = rtc.getDataRecordList().getMedicationOrders().size();
            } else if (LOINC == ClinicalDomainLoincCode.MEDICATIONS && rtc.getDataRecordList().getMedications() != null) {
                count = rtc.getDataRecordList().getMedications().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERMEDICATIONS && rtc.getDataRecordList().getMedications() != null) {
                count = rtc.getDataRecordList().getMedications().size();
            } else if (LOINC == ClinicalDomainLoincCode.NOTES && rtc.getDataRecordList().getNotes() != null) {
                count = rtc.getDataRecordList().getNotes().size();
            } else if (LOINC == ClinicalDomainLoincCode.INSURANCES && rtc.getDataRecordList().getPayers() != null) {
                count = rtc.getDataRecordList().getPayers().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERINSURANCE && rtc.getDataRecordList().getPayers() != null) {
                count = rtc.getDataRecordList().getPayers().size();
            } else if (LOINC == ClinicalDomainLoincCode.PROBLEMS && rtc.getDataRecordList().getProblems() != null) {
                count = rtc.getDataRecordList().getProblems().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERPROBLEMS && rtc.getDataRecordList().getProblems() != null) {
                count = rtc.getDataRecordList().getProblems().size();
            } else if (LOINC == ClinicalDomainLoincCode.PROCEDURES && rtc.getDataRecordList().getProcedures() != null) {
                count = rtc.getDataRecordList().getProcedures().size();
            } else if (LOINC == ClinicalDomainLoincCode.VLERPROCEDURES && rtc.getDataRecordList().getProcedures() != null) {
                count = rtc.getDataRecordList().getProcedures().size();
            } else if (LOINC == ClinicalDomainLoincCode.QUESTIONNAIRES && rtc.getDataRecordList().getQuestionnaires() != null) {
                count = rtc.getDataRecordList().getQuestionnaires().size();
//            } else if (LOINC == ClinicalDomainLoincCode.RADIOLOGY_ORDERS && rtc.getDataRecordList().getRadiologyOrders() != null) {
//                count = rtc.getDataRecordList().getRadiologyOrders().size();
            } else if (LOINC == ClinicalDomainLoincCode.RADIOLOGY_REPORTSV4 && rtc.getDataRecordList().getRadiologyReports() != null) {
                count = rtc.getDataRecordList().getRadiologyReports().size();
//            } else if (LOINC == ClinicalDomainLoincCode.VLERRESULTS && rtc.getDataRecordList().getResults() != null) {
//                count = rtc.getDataRecordList().getResults().size();
            } else if (LOINC == ClinicalDomainLoincCode.VITALS && rtc.getDataRecordList().getVitals() != null) {
                count = rtc.getDataRecordList().getVitals().size();
            }
        }
        return count;
    }
    
    /**
     * Get user record from snareworks.
     * 
     * @param certificate users certificate
     * @param user user
     * @param credential credential
     * @return user
     */
    @WebMethod(operationName = "getSnareworksUser")
    public SnareUser getSnareworksUser(@WebParam(name = "certificate") String certificate, @WebParam(name = "user") String user, @WebParam(name = "credential") String credential) {
        SnareUser rtc = null;
        String host = AppConfig.getInstance().getProperties().getProperty("SNAREWORKS_HOST", "172.18.53.16");
        String port = AppConfig.getInstance().getProperties().getProperty("SNAREWORKS_PORT", "510");
        LOGGER.debug("getSnareworksUser Cert=" + DBLogger.cleanString(certificate));
        LOGGER.debug("getSnareworksUser User=" + DBLogger.cleanString(user));
        LOGGER.debug("getSnareworksUser Credential=" + DBLogger.cleanString(credential));
        int iport = REDACTED;
        try {
            iport = Integer.parseInt(port);
        } catch (NumberFormatException e) {
            iport = REDACTED;
        }
        SnareworksHandler snare = new SnareworksHandler();
        if (user != null && user.length() > 0) {
            rtc = snare.checkCredentials(user, credential, host, iport);
        }
        if (rtc == null && certificate != null && certificate.length() > 0) {
            rtc = snare.checkCert(certificate, host, iport);
        }
        return rtc;
    }

    /**
     * Reset status info.
     * 
     */
    @WebMethod(operationName = "resetStatus")
    public void resetStatus() {
        BHIEStatus.resetStatus();
    }
    
    /**
     * Get status info.
     * 
     * @return status
     */
    @WebMethod(operationName = "getStatus")
    public String getStatus() {
        return BHIEStatus.getStatus();
    }
    
    /**
     * Add cache record.
     * 
     * @param type document type
     * @param key key
     * @param value cached record
     * @param timestamp time stamp
     */
    @WebMethod(operationName = "addCache")
    public void addCache(@WebParam(name = "type") String type, @WebParam(name = "key") String key, @WebParam(name = "value") String value, @WebParam(name = "timestamp") String timestamp) {
        LOGGER.debug("addCache " + DBLogger.cleanString(type) + ", " + DBLogger.cleanString(key) + ", " + DBLogger.cleanString(value) + ", " + DBLogger.cleanString(timestamp));
        if ("V4ResultCollection".equals(type)) {
            V4Limit.getInstance().putV4Cache(key, value);
        } else {
            ObjectMapper objectMapper = new ObjectMapper();
            objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

            LOGGER.error("addCache - unknown type: {}", DBLogger.cleanString(type));
        }
    }
    
//    /**
//     * Call addCache on all relay services.
//     *
//     * @param type type of cache entry
//     * @param queryBean query bean
//     * @param resultCollection value to cache
//     */
//    private void distributeCache(String type, QueryBean queryBean, String resultCollection) {
//        String key = BHIELimit.getInstance().getKey(queryBean, type);
//        distributeCache(type, key, resultCollection);
//    }
    
    /**
     * Call addCache on all relay services.
     * 
     * @param type type of cache entry
     * @param key key of cache entry
     * @param resultCollection value to cache
     */
    private void distributeCache(String type, String key, String resultCollection) {
        String endpoints = AppConfig.getInstance().getRelayEndpoints();
        String auths = AppConfig.getInstance().getRelayEndpointAuths();
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss", Locale.getDefault());
        Date now = new Date();
        String tstamp = sdf.format(now);
        if (endpoints != null && endpoints.length() > 0) {
            String[] endpoint = endpoints.split("\\|");
            String[] auth = null;
            if (auths != null && auths.length() > 0) {
                auth = auths.split("\\|");
            }
            for (int i = 0; i < endpoint.length; ++i) {
                if (auth != null && auth.length > i && auth[i] != null && auth[i].length() > 0) {
                    (new BHIECacheClient(endpoint[i], auth[i], type, key, resultCollection, tstamp)).start();
                } else {
                    (new BHIECacheClient(endpoint[i], "", type, key, resultCollection, tstamp)).start();
                }
            }
        }
    }

    private void logError(Exception e){
        LOGGER.error("BHIERelayService ERROR: " + e.getMessage(), e);
    }
}
