/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
package gov.va.med.bhie.webservice;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.concurrent.ConcurrentHashMap;

/**
 * BHIE Relay Status.
 *
 * @author <a href="mailto:kweller@hawaiirg.com">kweller</a>
 * @version 1.0
 *
 */
public class BHIEStatus extends HttpServlet {

    /**
     * Class UID.
     */
    private static final long serialVersionUID = 7122701102472959729L;

    /**
     * Startup time.
     */
    private static Date startDate = new Date();
    
    /**
     * Number of calls.
     */
    private static ConcurrentHashMap<String, Long> callCount = new ConcurrentHashMap<String, Long>();
    
    /**
     * Number of cached returns.
     */
    private static ConcurrentHashMap<String, Long> callCacheCount = new ConcurrentHashMap<String, Long>();
    
    /**
     * Number of errors.
     */
    private static ConcurrentHashMap<String, Long> callErrorCount = new ConcurrentHashMap<String, Long>();
    
    /**
     * Error timestamp.
     */
    private static ConcurrentHashMap<String, Date> callErrorTimeStamp = new ConcurrentHashMap<String, Date>();
    
    /**
     * Time of last call.
     */
    private static ConcurrentHashMap<String, Date> callTimeStamp = new ConcurrentHashMap<String, Date>();
    
    /**
     * Time for calls to complete.
     */
    private static ConcurrentHashMap<String, Long> callTotalTime = new ConcurrentHashMap<String, Long>();
    
    /**
     * Minimum time for calls to complete.
     */
    private static ConcurrentHashMap<String, Long> callMinTime = new ConcurrentHashMap<String, Long>();
    
    /**
     * Maximum time for calls to complete.
     */
    private static ConcurrentHashMap<String, Long> callMaxTime = new ConcurrentHashMap<String, Long>();
    
    /**
     * Called when a get request needs to be processed. Used to check status.
     *
     * @param request request object
     * @param response response object
     * 
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response) {
        try {
            String rtc = getHTMLStatus();
            response.setContentType("text/html;charset=UTF-8");
            response.setCharacterEncoding("UTF-8");
            response.getWriter().write("<html><head><title>BHIERelayService Status</title></head><body><H1>BHIERelayService</H1><hr><br />" + rtc + "<br /></body></html>");
        } catch (Throwable ex) {
            try {
                response.getWriter().write("<html><head><title>ERROR</title></head><body><hr><br />ERROR<br /></body></html>");
            } catch (Throwable ex2) {
                ex2 = null; // gobble up exception
            }
        }
    }

    /**
     * Record the statistics of a call.
     * 
     * @param method method called
     * @param cache did return come from cache (true or false)
     * @param error error flag
     * @param millis number of milliseconds for call to complete
     */
    public static synchronized void updateStatistics(String method, String cache, boolean error, long millis) {
        Date current = new Date();
        Long cnt = callCount.get(method);
        if (cnt == null) {
            cnt = Long.valueOf(0);
        }
        cnt++;
        callCount.put(method, cnt);
        if ("true".equalsIgnoreCase(cache)) {
            Long ccnt = callCacheCount.get(method);
            if (ccnt == null) {
                ccnt = Long.valueOf(0);
            }
            ccnt++;
            callCacheCount.put(method, ccnt);
        } else {
            Long min = callMinTime.get(method);
            if (min == null) {
                min = Long.valueOf(99999999);
            }
            if (millis < min && millis > 0) {
                callMinTime.put(method, Long.valueOf(millis));
            }
            Long max = callMaxTime.get(method);
            if (max == null) {
                max = Long.valueOf(0);
            }
            if (millis > max && millis > 0) {
                callMaxTime.put(method, Long.valueOf(millis));
            }
            Long tcnt = callTotalTime.get(method);
            if (tcnt == null) {
                tcnt = Long.valueOf(0);
            }
            tcnt += millis;
            callTotalTime.put(method, tcnt);
        }
        if (error) {
            Long ecnt = callErrorCount.get(method);
            if (ecnt == null) {
                ecnt = Long.valueOf(0);
            }
            ecnt++;
            callErrorCount.put(method, ecnt);
            callErrorTimeStamp.put(method, current);
        }
        callTimeStamp.put(method, current);
    }
    
    /**
     * Clear Statistics.
     */
    public static void resetStatus() {
        callCount.clear();
        callMinTime.clear();
        callMaxTime.clear();
        callCacheCount.clear();
        callErrorCount.clear();
        callErrorTimeStamp.clear();
        callTimeStamp.clear();
        callTotalTime.clear();
    }
    
    /**
     * Get service status.
     * 
     * @return status
     */
    public static String getStatus() {
        StringBuilder rtc = new StringBuilder(1024);
        rtc.append("BHIERelayService - Running Since: " + startDate.toString() + "\n");
        rtc.append("Running on: " + getComputerId() + "\n\n");
        rtc.append("Call Type                Count     Cached    Errors    Min(ms)   Max(ms)   Avg(ms)   Date Last Call               Date Last Error\n");
        rtc.append("----------------------------------------------------------------------------------------------------------------------------------------------\n");
        if (callCount == null || callCount.isEmpty()) {
            rtc.append("No calls received.");
        } else {
            Object[] key = callCount.keySet().toArray();
            Arrays.sort(key);
            for (int i = 0; i < key.length; ++i) {
                StringBuilder row = new StringBuilder();
                String name = (String) key[i];
                Long count = callCount.get(name);
                Long cache = callCacheCount.get(name);
                if (cache == null) {
                    cache = Long.valueOf(0);
                }
                Long error = callErrorCount.get(name);
                if (error == null) {
                    error = Long.valueOf(0);
                }
                Long min = callMinTime.get(name);
                Long max = callMaxTime.get(name);
                
                Date lastCall = callTimeStamp.get(name);
                Date lastErrorCall = callErrorTimeStamp.get(name);
                Long totalTime = callTotalTime.get(name);
                if (totalTime == null) {
                    totalTime = Long.valueOf(0);
                }
                if (count == null || count.longValue() == 0L) {
                    row.append(name);
                    row.append(": N/A\n");
                } else {
                    row.append(name);
                    row.append(' ');
                    while (row.length() < 25) {
                        row.append(' ');
                    }
                    row.append(count.toString());
                    row.append(' ');
                    while (row.length() < 35) {
                        row.append(' ');
                    }
                    if (cache != null) {
                        row.append(cache.toString());                        
                    }
                    row.append(' ');
                    while (row.length() < 45) {
                        row.append(' ');
                    }
                    if (error != null) {
                        row.append(error.toString());                        
                    }
                    row.append(' ');
                    while (row.length() < 55) {
                        row.append(' ');
                    }
                    if (min != null) {
                        row.append(min.toString());                        
                    }
                    row.append(' ');
                    while (row.length() < 65) {
                        row.append(' ');
                    }
                    if (max != null) {
                        row.append(max.toString());                        
                    }
                    row.append(' ');
                    while (row.length() < 75) {
                        row.append(' ');
                    }
                    if (cache == null || count.longValue() - cache.longValue() <= 0) {
                        row.append('0');
                    } else {
                        int avg = (int) (totalTime.longValue() / (count.longValue() - cache.longValue()));
                        row.append(avg);
                    }
                    row.append(' ');
                    while (row.length() < 85) {
                        row.append(' ');
                    }
                    if (lastCall != null) {
                        row.append(lastCall.toString());
                    }
                    if (lastErrorCall != null) {
                        row.append(' ');
                        row.append(lastErrorCall.toString());                        
                    }
                    rtc.append(row.toString());
                    rtc.append('\n');
                }
            }
        }
        return rtc.toString();
    }

    /**
     * Get service status.
     * 
     * @return status
     */
    public static String getHTMLStatus() {
        StringBuilder rtc = new StringBuilder(1024);
        rtc.append("BHIERelayService - Running Since: " + startDate.toString() + "<br />");
        rtc.append("Running on: " + getComputerId() + "<br /><br />");
        if (callCount == null || callCount.isEmpty()) {
            rtc.append("<br />No calls received.<br />");
        } else {
            rtc.append("<table>");
            rtc.append("<thead>");
            rtc.append("<tr style=\"text-align:left\">");
            rtc.append("<th>");
            rtc.append("Call Type");
            rtc.append("</th>");
            rtc.append("<th>");
            rtc.append("Count&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
            rtc.append("</th>");
            rtc.append("<th>");
            rtc.append("Cached&nbsp;&nbsp;&nbsp;&nbsp;");
            rtc.append("</th>");
            rtc.append("<th>");
            rtc.append("Errors&nbsp;&nbsp;&nbsp;&nbsp;");
            rtc.append("</th>");
            rtc.append("<th>");
            rtc.append("Min(ms)&nbsp;&nbsp;&nbsp;");
            rtc.append("</th>");
            rtc.append("<th>");
            rtc.append("Max(ms)&nbsp;&nbsp;&nbsp;");
            rtc.append("</th>");
            rtc.append("<th>");
            rtc.append("Avg(ms)&nbsp;&nbsp;&nbsp;");
            rtc.append("</th>");
            rtc.append("<th>");
            rtc.append("Date Last Call");
            rtc.append("</th>");
            rtc.append("<th>");
            rtc.append("Date Last Error");
            rtc.append("</th>");
            rtc.append("</tr>");
            rtc.append("</thead>");
            rtc.append("<tbody>");
            Object[] key = callCount.keySet().toArray();
            Arrays.sort(key);
            for (int i = 0; i < key.length; ++i) {
                String name = (String) key[i];
                Long count = callCount.get(name);
                Long cache = callCacheCount.get(name);
                if (cache == null) {
                    cache = Long.valueOf(0);
                }
                Long error = callErrorCount.get(name);
                if (error == null) {
                    error = Long.valueOf(0);
                }
                Long min = callMinTime.get(name);
                Long max = callMaxTime.get(name);
                Date lastCall = callTimeStamp.get(name);
                Date lastError = callErrorTimeStamp.get(name);
                Long totalTime = callTotalTime.get(name);
                if (totalTime == null) {
                    totalTime = Long.valueOf(0);
                }
                if (count == null || count.longValue() == 0L) {
                    continue;
                }
                rtc.append("<tr>");
                rtc.append("<td>");
                rtc.append(name);
                rtc.append("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td>");

                rtc.append("<td>");
                rtc.append(count.toString());
                rtc.append("</td>");

                rtc.append("<td>");
                rtc.append(cache.toString());
                rtc.append("</td>");

                rtc.append("<td>");
                rtc.append(error.toString());
                rtc.append("</td>");

                if (min == null) {
                    rtc.append("<td>");
                    rtc.append("</td>");
                } else {
                    rtc.append("<td>");
                    rtc.append(min.toString());
                    rtc.append("</td>");
                }
                if (max == null) {
                    rtc.append("<td>");
                    rtc.append("</td>");
                } else {
                    rtc.append("<td>");
                    rtc.append(max.toString());
                    rtc.append("</td>");
                }
                if (count.longValue() - cache.longValue() <= 0) {
                    rtc.append("<td>");
                    rtc.append('0');
                    rtc.append("</td>");
                } else {
                    int avg = (int) (totalTime.longValue() / (count.longValue() - cache.longValue()));
                    rtc.append("<td>");
                    rtc.append(avg);
                    rtc.append("</td>");
                }
                if (lastCall == null) {
                    rtc.append("<td>");
                    rtc.append("</td>");
                } else {
                    rtc.append("<td>");
                    rtc.append(lastCall.toString());
                    rtc.append("</td>");
                }
                if (lastError == null) {
                    rtc.append("<td>");
                    rtc.append("</td>");
                } else {
                    rtc.append("<td>");
                    rtc.append(lastError.toString());
                    rtc.append("</td>");
                }
                rtc.append("</tr>\n");
            }
            rtc.append("</tbody>\n");
            rtc.append("</table>\n");
        }
        return rtc.toString();
    }

    /**
     * Get computer id.
     * 
     * @return id
     */
    private static String getComputerId() {
        StringBuilder idstring = new StringBuilder();

        String hostname = System.getProperty("COMPUTERNAME", System.getenv().get("COMPUTERNAME"));
        if (hostname == null || hostname.length() == 0) {
            hostname = System.getProperty("HOSTNAME", System.getenv().get("HOSTNAME"));
        }
        if (hostname != null && hostname.length() > 0) {
            idstring.append(hostname);
            idstring.append('|');
        }

        try {
            Enumeration<NetworkInterface> nets = NetworkInterface.getNetworkInterfaces();
            for (NetworkInterface netint : Collections.list(nets)) {
                if (netint.isUp() && (!netint.isLoopback()) && (!netint.isPointToPoint()) && (!netint.isVirtual())) {
                    Enumeration<InetAddress> inetAddresses = netint.getInetAddresses();
                    for (InetAddress inetAddress : Collections.list(inetAddresses)) {
                        if (inetAddress instanceof Inet4Address) {
                            if (idstring.length() == 0) {
                                idstring.append(inetAddress.getHostName());
                                idstring.append('|');
                            }
                            idstring.append(inetAddress.getHostAddress());
                            idstring.append('|');
                        }
                    }
                }
            }
        } catch (SocketException ex) {
            idstring.append('|');
        }
        return idstring.toString();
    }
}
