package gov.va.med.bhie.webservice;

import gov.va.med.bhie.beans.NoteImage;
import gov.va.med.bhie.common.AppConfig;
import org.apache.log4j.Logger;
import org.apache.pdfbox.pdmodel.PDDocument;

import java.io.*;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Locale;
import java.util.UUID;

import static java.lang.System.currentTimeMillis;

/**
 * Note fixer.
 *
 * AHLTA notes are not always formatted to be viewed by latest programs like wordpad or word.
 *
 * This class checks to see if the rtf document can be parsed. If it can't be parsed it calls an
 * external program which opens the document using the 3rd party control that AHLTA uses and then
 * saves it.
 *
 * @author <a href="mailto:kweller@hawaiirg.com">kweller</a>
 * @version 1.0
 */
public final class NoteFixer {

    /**
     * Logger.
     */
    private static final Logger LOGGER =
            Logger.getLogger(NoteFixer.class.getName());

    /**
     * Fix note files.
     *
     * @param note note
     * @param pdfout pdf output flag
     * @return good note
     */

    public static NoteImage fixNoteImage(NoteImage note, boolean pdfout) {
        NoteImage rtc = note;

        String osname = System.getProperty("os.name", "generic").toLowerCase();
        if (note != null &&
                ("application/octet-stream".equalsIgnoreCase(note.getContentType())
                        || "text/rtf".equalsIgnoreCase(note.getContentType())
                        || (note.getContentDisposition().contains(".rtf")))
                && osname.toUpperCase(Locale.getDefault()).startsWith("WIN") /* && !goodRTF(note.getNoteBytes()) */) {
            try {
                UUID uuid = UUID.randomUUID();
                File ofile = new File("c:/temp/" + uuid.toString() + ".rtf");
                if (!ofile.createNewFile()) {
                    LOGGER.debug("File already existed.");
                }
                FileOutputStream stream = new FileOutputStream(ofile);
                try {
                    stream.write(note.getNoteBytes());
                } catch (IOException ioe) {
                    LOGGER.error("EXCEPTION: " + ioe.getLocalizedMessage(), ioe);
                } finally {
                    try {
                        stream.flush();
                    } catch (Exception ex) {
                        LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
                    }
                    try {
                        stream.close();
                    } catch (IOException ex) {
                        LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
                    }
                }

                String boundary = "---------------------------" + currentTimeMillis();
                String crlf = "\r\n";

                URL url = new URL(AppConfig.getInstance().getPdfServiceEndpoint());

                HttpURLConnection httpCon = null;
                BufferedWriter httpRequestBodyWriter = null;
                OutputStream outputStreamToRequestBody = null;
                FileInputStream inputStreamFile = null;
                InputStream is = null;
                try {
                    httpCon = (HttpURLConnection) url.openConnection();
                    httpCon.setDoOutput(true);
                    httpCon.setDoInput(true);
                    httpCon.setUseCaches(false);
                    httpCon.setRequestMethod("POST");
                    httpCon.setRequestProperty("Connection", "close");
                    httpCon.addRequestProperty("Content-Type", "multipart/form-data; boundary=" + boundary);

                    outputStreamToRequestBody = httpCon.getOutputStream();
                    httpRequestBodyWriter = new BufferedWriter(new OutputStreamWriter(outputStreamToRequestBody));

                    httpRequestBodyWriter.write("--" + boundary);
                    httpRequestBodyWriter.write(crlf);
                    httpRequestBodyWriter.write("Content-Disposition: form-data; name=\"inputFile\"; filename=\"" + ofile);
                    httpRequestBodyWriter.write(crlf);
                    httpRequestBodyWriter.write("Content-Type: text/rtf");
                    httpRequestBodyWriter.write(crlf);
                    httpRequestBodyWriter.write(crlf);
                    httpRequestBodyWriter.flush();

                    // Write the actual file contents
                    inputStreamFile = new FileInputStream(ofile);

                    int bytesRead;
                    byte[] dataBuffer = new byte[4096];
                    while ((bytesRead = inputStreamFile.read(dataBuffer)) != -1) {
                        outputStreamToRequestBody.write(dataBuffer, 0, bytesRead);
                    }

                    outputStreamToRequestBody.flush();
                    httpRequestBodyWriter.write(crlf);
                    httpRequestBodyWriter.write("--" + boundary + "--");
                    httpRequestBodyWriter.write(crlf);
                    httpRequestBodyWriter.flush();
                    
                    is = httpCon.getInputStream();
                    int status = httpCon.getResponseCode();
                    if (status != HttpURLConnection.HTTP_OK) {
                        throw new IOException(" failed with HTTP status: " + status);
                    }

                    // Read response from web server, which will trigger the multipart HTTP request to be sent.
                    rtc.setContentType("application/pdf");
                    rtc.setContentDisposition(note.getContentDisposition().replaceAll("\\.rtf", ".pdf").replaceAll("\\.RTF", ".PDF"));
                    rtc.setNoteBytes(org.apache.commons.io.IOUtils.toByteArray(is));
                    is.close();
                    try {
                        if (!ofile.delete()) {
                            LOGGER.debug("File wasn't deleted.");
                        }
                    } catch (Exception ex) {
                        LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
                    }
                } catch (IOException ex) {
                    LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
                } finally {
                    if (httpCon != null) {
                        httpCon.disconnect();
                    }
                    // close the resources
                    if (httpRequestBodyWriter != null) {
                        httpRequestBodyWriter.close();
                    }
                    if (outputStreamToRequestBody != null) {
                        outputStreamToRequestBody.close();
                    }
                    if (inputStreamFile != null) {
                        inputStreamFile.close();
                    }
                    if (is != null) {
                        is.close();
                    }
                    
                }
            } catch (MalformedURLException ex) {
                LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
                rtc = note;
            } catch (FileNotFoundException ex) {
                LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
                rtc = note;
            } catch (IOException ex) {
                LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
                rtc = note;
            }
        }
        return rtc;
    }


    public static NoteImage flattenNoteForm(NoteImage note) {

        PDDocument pdDocument = null;
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        NoteImage rtc = note;

        try {
            pdDocument = PDDocument.load(note.getNoteBytes());
            pdDocument.setAllSecurityToBeRemoved(true);
            if(pdDocument.getDocumentCatalog() != null && pdDocument.getDocumentCatalog().getAcroForm() != null) {
                pdDocument.getDocumentCatalog().getAcroForm().flatten();
            }

            pdDocument.save(out);
            pdDocument.close();

            rtc.setNoteBytes(out.toByteArray());

        } catch (IOException e) {
            LOGGER.error("NoteFixer EXCEPTION: " + e.getLocalizedMessage(), e);
        }

        return rtc;

    }

}
