package gov.va.med.bhie.webservice;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import gov.va.med.bhie.beans.QueryBean;
import gov.va.med.bhie.common.AppConfig;
import gov.va.med.bhie.common.ClinicalDomainLoincCode;
import gov.va.med.bhie.resultcollection.V4ResultCollection;
import gov.va.med.foundations.utilities.ARMLog;

import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.ConcurrentHashMap;
//import org.codehaus.jackson.map.DeserializationConfig;
//import org.codehaus.jackson.map.ObjectMapper;

public class V4Limit {

    /**
     * Logger.
     */
    private static final Logger LOGGER =
        LoggerFactory.getLogger(V4Limit.class);

    /**
     * Tokens to limit number of concurrent calls.
     */
    private static ConcurrentHashMap<String, Date> tokens = new ConcurrentHashMap<String, Date>();

    /**
     * Cache.
     */
    private static ConcurrentHashMap<String, V4ResultCollection> cachemap = new ConcurrentHashMap<String, V4ResultCollection>();

    /**
     * Age of cache entries.
     */
    private static ConcurrentHashMap<String, Date> cacheagemap = new ConcurrentHashMap<String, Date>();

    /**
     * Time to live for cache entries.
     */
    private int expiretimesecs = 900; // 15 minute

    private V4Limit() {
        super();
        String timeout = AppConfig.getInstance().getProperties().getProperty("V4Limit.timeout", "900");
        try {
            expiretimesecs = Integer.parseInt(timeout);
        } catch (NumberFormatException ex) {
            expiretimesecs = 900;
        }
    }
    
    /**
     * Get singleton class.
     *
     * @return this
     */
    @SuppressWarnings("all")
    public static synchronized V4Limit getInstance() {
        return V4LimitHolder.INSTANCE;
    }

    public V4ResultCollection getV4Cache(ClinicalDomainLoincCode loinc, QueryBean queryBean) {
        return getV4Cache(loinc.getLoincCode(), queryBean);
    }
    
    public V4ResultCollection getV4Cache(String loinc, QueryBean queryBean) {
        V4ResultCollection rtc = null;
        if (queryBean.isForceRefresh()) {
            dropCache(loinc, queryBean);
        } else {
            rtc = (V4ResultCollection) getCache(loinc, queryBean);
        }
        return rtc;
    }

    /**
     * User for synchronization across multiple relay services.
     * 
     * @param key key
     * @param value value
     */
    public void putV4Cache(String key, String value) {
        try {
            ObjectMapper objectMapper = new ObjectMapper();
            objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
            V4ResultCollection collection = (V4ResultCollection) objectMapper.readValue(value, V4ResultCollection.class);
            Date cdate = new Date();
            cacheagemap.put(key, cdate);
            cachemap.put(key, collection);
        } catch (IOException ex) {
            LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);    
        }
    }
    
    public void putV4Cache(ClinicalDomainLoincCode loinc, QueryBean queryBean, V4ResultCollection collection) {
        putCache(loinc.getLoincCode(), queryBean, collection);
    }

    public void putV4Cache(String loinc, QueryBean queryBean, V4ResultCollection collection) {
        putCache(loinc, queryBean, collection);
    }
    
    private V4ResultCollection getCache(String loinc, QueryBean queryBean) {
        V4ResultCollection rtc = null;
        if ("true".equalsIgnoreCase(AppConfig.getInstance().getProperties().getProperty("V4Limit.disable", "false"))) {
            cleanMap();
        } else {
            String key = getKey(loinc, queryBean);
            if (cachemap.containsKey(key) && cacheagemap.containsKey(key)) {
                Date ldate = cacheagemap.get(key);
                Date cdate = new Date();
                if (ldate != null && (expiretimesecs == -1 || (Math.abs(cdate.getTime() - ldate.getTime()) <= (1000L * expiretimesecs)))) {
                    rtc = cachemap.get(key);
                    LOGGER.info("Returning cache for: {}", ARMLog.cleanString(key));
                } else {
                    cleanMap();
                }
            } else {
                cleanMap();
            }
        }
        return rtc;
    }

    private void putCache(String loinc, QueryBean queryBean, V4ResultCollection collection) {
        String key = getKey(loinc, queryBean);
        Date cdate = new Date();
        cacheagemap.put(key, cdate);
        cachemap.put(key, collection);
    }

    private void dropCache(String loinc, QueryBean queryBean) {
        if (queryBean != null) {
            if (queryBean.isForceRefresh()) {
                dropCache(queryBean);
            } else {
                String key = getKey(loinc, queryBean);
                cacheagemap.remove(key);
                cachemap.remove(key);
            }
        }
    }

    private void dropCache(QueryBean queryBean) {
        Enumeration<String> keys = cacheagemap.keys();
        while (keys != null && keys.hasMoreElements()) {
            String key = keys.nextElement();
            if (key != null && key.startsWith(queryBean.getPatientId() + "|")) {
                cacheagemap.remove(key);
                cachemap.remove(key);
            }
        }
    }

    public String getKey(String loinc, QueryBean queryBean) {
        StringBuilder rtc = new StringBuilder();
        if (queryBean != null) {
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd", Locale.getDefault());
            rtc.append(queryBean.getPatientId());
            rtc.append('|');
            rtc.append(loinc);
            rtc.append('|');
            if (queryBean.getStartDate() != null) {
                rtc.append(sdf.format(queryBean.getStartDate().getTime()));
            }
            rtc.append('|');
            if (queryBean.getEndDate() != null) {
                rtc.append(sdf.format(queryBean.getEndDate().getTime()));
            }
            rtc.append('|');
            if (queryBean.getIncludedSources() != null) {
                List<String> src = queryBean.getIncludedSources();
                if (CollectionUtils.isNotEmpty(src)) {
                    int srcSize = src.size();
                    for (int i = 0; i < srcSize; ++i) {
                        rtc.append(src.get(i));
                        rtc.append('|');
                    }
                }
            }
            if (queryBean.getExcludedSources() != null) {
                List<String> src = queryBean.getExcludedSources();
                if (CollectionUtils.isNotEmpty(src)) {
                    int srcSize = src.size();
                    for (int i = 0; i < srcSize; ++i) {
                        rtc.append(src.get(i));
                        rtc.append('|');
                    }
                }
            }
        }
        return rtc.toString();
    }

    private void cleanMap() {
        Enumeration<String> keys = cacheagemap.keys();
        Date cdate = new Date();
        while (keys != null && keys.hasMoreElements()) {
            String key = keys.nextElement();
            Date adate = tokens.get(key);
            if (adate != null && cdate != null && (Math.abs(cdate.getTime()
                    - adate.getTime()) > (1000L * (60 + expiretimesecs)))) {
                cacheagemap.remove(key);
                cachemap.remove(key);
            }
        }
    }
    
    /**
     * Disable cloning.
     *
     * @return nothing
     *
     * @throws CloneNotSupportedException exception
     */
    @SuppressWarnings("all")
    public Object clone() throws CloneNotSupportedException {
        throw new CloneNotSupportedException();
    }

    /**
     * Singleton helper class.
     *
     */
    private static class V4LimitHolder {
        /**
         * V4Limit.
         */
        private static final V4Limit INSTANCE = new V4Limit();
    }
}
