/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


import gov.va.med.common.HttpKeys
import gov.va.med.dao.DaoException
import gov.va.med.foundations.utilities.EncryptUtil
import gov.va.med.groovy.user.AppUser
import gov.va.med.guibeans.GUISite
import gov.va.med.guibeans.GUIUser
import gov.va.med.jmeadows.webservice.AuthUserInfo
import gov.va.med.jmeadows.webservice.IehrUserProfile
import grails.util.Metadata
import jlv.WhiteListOnlyLoginController
import org.apache.commons.lang3.StringUtils

import java.security.cert.CertificateExpiredException
import java.security.cert.CertificateNotYetValidException
import java.security.cert.X509Certificate
import java.util.regex.Pattern

/**
 *
 *
 */
class SecurityFilters {
    static final String cardIdDefault = "987654321";
    static long seqCardIdSelMinId;
    static long seqCardIdSelMaxId;
    static long seqCardIdSelCurId;
    static boolean seqCardIdSelInit = true;

    def appConfig
    def tokenUtils

    def userDao
    def patientDao

    def authUserDao;
    def auditDao

    def filters = {

        getAndPostOnly(controller: '*', action: '*')
                {
                    before = {
                        // To prevent unallowed HTTP methods
                        if (request.method != 'GET' && request.method != 'POST') {
                            response.sendError 405
                        }
                    }
                }

        /** check csrf token to filter potential csrf **/

        csrfCheck(controller: '*', action: '*')
                {
                    before = {

                        if (request.method != 'GET') { //only filter 'destructive' requests
                            def csrfTokenServer = session.getAttribute("X-CSRF-Token");
                            def csrfTokenClient = '';

                            if (controllerName == 'authenticate') //authenticate controller does not send token in header
                            {
                                csrfTokenClient = params['csrf-token']
                            } else if ((controllerName == 'patientSelect' || controllerName == 'context') && params['csrfToken'] != null) {
                                csrfTokenClient = params['csrfToken']
                            } else {
                                csrfTokenClient = request.getHeader("X-CSRF-Token");
                            }

                            if (!csrfTokenServer.equals(csrfTokenClient) && (csrfTokenServer != null && controllerName != 'AHLTALogin' && controllerName != 'login')) {
                                log.debug("server and client csrfTokens do not match.")
                                return false;
                            }
                        }
                    }
                }

        /**
         * If invalid session, send user to authenticate login action
         * or if rest action, send 403 error
         */
        sessionCheck(controller: '*', action: '*')
                {
                    before = {

                        if (controllerName != 'authenticate'
                                && controllerName != "assets"
                                && controllerName != 'whiteListOnlyLogin') {
                            if (!session.appUser) {
                                //redirect to login screen
                                if (actionName == "index" ||
                                        controllerName == "app" ||
                                        controllerName == "context" ||
                                        controllerName == "patientMedWeb" ||
                                        controllerName == "view") {
                                    log.debug("No appUser in session, redirecting to login screen")
                                    redirect(controller: 'authenticate', action: 'login')
                                    return false
                                }

                                // Only show if not coming from AHLTA link
                                if (controllerName != "AHLTALogin" && controllerName != "login" ) {
                                    //if rest controller/action, send 403 error
                                    log.debug("REST call: No appUser in session, sending 403 Forbidden")
                                    response.sendError 403;
                                    return false;
                                }
                            } else {

                                def acceptedAgreement = session.getAttribute("acceptedAgreement") || session.ahltaLoginFlag;
                                if (!acceptedAgreement) {
                                    redirect(controller: 'authenticate', action: 'disclaimer')
                                    return false;
                                }
                            }
                        }

                    }
                }

        /**
         * Do not allow browser caching for all controllers/actions
         */
        noCacheCheck(controller: '*', action: '*')
                {
                    before = {
                        response.setHeader('Cache-Control', 'no-cache, no-store')
                        response.setDateHeader('Expires', (new Date() - 1).time)
                        response.setHeader('Pragma', 'no-cache')
                    }
                }

        /**
         * If login action and a valid session exists, send
         * user to the provider page
         */
        loginRedirect(controller: "authenticate|whiteListOnlyLogin", action: "login")
                {
                    before = {
                        if (session.appUser) {
                            redirect(controller: 'app')
                            return false
                        }
                    }
                }

        /**
         * Filter parses and verifies patient and record security tokens.
         *
         */
        securityTokenCheck(controller: '*', action: '*')
                {
                    before = {

                        /**
                         * If valid patient token exists in request params,
                         * patient tied to the token will be placed in the request object (request.patient).
                         *
                         * A patient token is an AES encrypted base 64 string that consists
                         * of four parameters: patient id, patient lookup site (site associated with the patient id),
                         * user session id, and timestamp.
                         */
                        if (params[HttpKeys.PATIENT_TOKEN] && controllerName != "patientSelect") {
                            log.debug("Processing patient token")

                            try {
                                def patientToken = params[HttpKeys.PATIENT_TOKEN]
                                def appUser = session.appUser
                                def selectedPatientMap = session.appUser.selectedPatientMap

                                if (patientToken == null || patientToken.isEmpty()) {
                                    log.error("securityTokenCheck::Patient token null or empty")
                                    throw new RuntimeException("Invalid patient token")
                                } else if (appUser == null || selectedPatientMap == null) {
                                    log.error("securityTokenCheck::Provider or selected patient map is null")
                                    throw new RuntimeException("Uninitialized session")
                                }

                                if (log.debugEnabled) {
                                    log.debug("Decrypted patient token: ${tokenUtils.decryptTokenToString(session.appUser.securityKey, patientToken)}");
                                }

                                //first check to see if patient is already mapped
                                def selectedPatient = selectedPatientMap.get(patientToken)

                                if (selectedPatient == null) {
                                    log.debug("Cannot find patient token in map, calling select patient")

                                    if (selectedPatient == null) {
                                        log.error("securityTokenCheck::selected patient is null")
                                        throw new RuntimeException("Invalid patient token")
                                    }
                                }

                                request.patient = selectedPatient
                            }
                            catch (Exception e) {
                                response.sendError 500
                                return false;
                            }
                        }
                        /**
                         * If a valid record details exists in request params,
                         * necessary record variables (record query, isSessionCachedRecord)
                         * will be added to the request object.
                         *
                         * A record details token is tied to the user session.
                         */
                        if (params[HttpKeys.RECORD_DETAILS_TOKEN]) {
                            log.debug("Processing record details token")

                            try {
                                def recordDetailsToken = params[HttpKeys.RECORD_DETAILS_TOKEN]
                                def provider = session.appUser

                                if (recordDetailsToken == null || recordDetailsToken.isEmpty()) {
                                    log.error("securityTokenCheck::Record details token is null or empty")
                                    throw new RuntimeException("Invalid record details token")
                                } else if (provider == null) {
                                    log.error("securityTokenCheck::Provider is null")
                                    throw new RuntimeException("Uninitialized session")
                                }

                                if (log.isDebugEnabled()) {
                                    log.debug("Decrypted record details token: ${tokenUtils.decryptTokenToString(session.appUser.securityKey, recordDetailsToken)}")
                                }

                                def detailsParams = tokenUtils.decryptTokenToMap(session.appUser.securityKey, recordDetailsToken)

                                def paramRecordControllerName = detailsParams[HttpKeys.RECORD_CONTROLLER_NAME]
                                def paramSessionCachedRecord = detailsParams[HttpKeys.SESSION_CACHED_RECORD]

                                //record token security checks
                                if (controllerName != paramRecordControllerName) {
                                    log.error("securityTokenCheck::record details token's controller name is invalid.")
                                    throw new RuntimeException("Invalid record details token")
                                }

                                request.recordQuery = detailsParams

                                request.isSessionCachedRecord = Boolean.valueOf(paramSessionCachedRecord)

                                return true
                            }
                            catch (Exception e) {
                                response.sendError 500
                                return false;
                            }
                        }
                    }
                }

        smartCardCheck(controller: 'authenticate|whiteListOnlyLogin', action: 'login')
                {
                    before = {
                        // We only want to do the smart card check when requesting login page, not when logging into JLV.
                        if (request.method == 'POST') {
                            return;
                        }

                        if (appConfig.getSsoiEnabled()) {
                            String adEmail = request.getHeader("adEmail")

                            if (StringUtils.isEmpty(adEmail)) {

                                response.sendError(403, "UserAccessDenied")
                                return false

                            } else {

                                if (request.getScheme() != "https") {
                                    response.sendError(403, "SSLRequired")
                                    return false
                                }

                                String agency = "VA"

                                IehrUserProfile userProfile = null
                                if (session.userProfile == null) {

                                    try {
                                        userProfile = userDao.getIehrUserProfile(null, agency, adEmail)
                                        userProfile.setEmailAddress(adEmail)

                                        if (userProfile != null && StringUtils.isNotEmpty(userProfile.dbUserID)) {
                                            if (userProfile.flags == null) {
                                                userProfile.setFlags("")
                                            }
                                            if (userProfile.locID == null) {
                                                userProfile.setLocID("")
                                            }
                                            if (userProfile.cfg == null) {
                                                userProfile.setCfg("")
                                            }
                                            if (userProfile.emailAddress == null) {
                                                userProfile.setEmailAddress("")
                                            }
                                            if (userProfile.subjectDN == null) {
                                                userProfile.setSubjectDN("")
                                            }
                                        } else {
                                            userProfile.setDbUserID("")
                                            userProfile.setLocID("")
                                            userProfile.setCfg("")
                                            userProfile.setFlags("")
                                            userProfile.setSmartCardID("")
                                            userProfile.setSmartCardAgency(agency)
                                            userProfile.setEmailAddress(adEmail)
                                            userProfile.setSubjectDN("")
                                        }
                                        session.userProfile = userProfile
                                    }
                                    catch (Exception e) {
                                        log.info("Could not process JLV User Profile: ${e.getMessage()}")
                                        response.sendError(403, "ProfileRetrieve")
                                        return false
                                    }
                                }

                                GUIUser auditUser = new GUIUser()
                                auditUser.setAgency(agency)
                                auditUser.setEmail(adEmail)
                                if (appConfig.getVaUserAuthEnabled()) {  //check DB if the flag is on
                                    try {
                                        auditDao.auditDetailsQuery("SSOiLoginAttempt", auditUser, null, null, null, null, null)

                                        AuthUserInfo authUserInfo = new AuthUserInfo()
                                        authUserInfo.setCardId("")
                                        authUserInfo.setOrganization(agency)
                                        authUserInfo.setEmail(adEmail)
                                        authUserInfo = authUserDao.getAuthUser(authUserInfo)

                                        if (authUserInfo == null) {
                                            response.sendError(403, "UserAccessDenied")
                                            return false
                                        }

                                    } catch (DaoException e) {
                                        response.sendError(403, "UserAccessDenied")
                                        return false
                                    }
                                } else {
                                    auditDao.auditDetailsQuery("authNoDbVerification", auditUser, null, null, null, null, null)
                                }
                                return true
                            }
                        } else {

                            String subjectDN = "";
                            String certificate = "";
                            String agency = "";
                            String lastName = "";
                            String firstName = "";
                            String middleName = "";
                            String scId = "";
                            String scEmail = "";

                            // Could not have min/max values set at variable creation, so I created a static
                            // variable flag to ensure setting the min/max from appconfig properties file
                            // only happens once at application startup and first access to smartCardCheck.
                            if (seqCardIdSelInit) {
                                seqCardIdSelMinId = appConfig.getSeqCardIdSelMinId();
                                seqCardIdSelMaxId = appConfig.getSeqCardIdSelMaxId();

                                if (seqCardIdSelMinId == -1L) {
                                    seqCardIdSelMinId = Long.parseLong(cardIdDefault);
                                }

                                if (seqCardIdSelMaxId == -1L) {
                                    seqCardIdSelMaxId = Long.parseLong(cardIdDefault);
                                }

                                seqCardIdSelCurId = seqCardIdSelMinId;
                                seqCardIdSelInit = false;
                            }

                            //get person name and id from CAC x.509 certificate
                            X509Certificate[] certs = (X509Certificate[]) request.getAttribute("javax.servlet.request.X509Certificate");

                            if (appConfig.getSmartCardBypass() && (request.getScheme() != "https")) {
                                subjectDN = "CN=this,OU=is_a_test";
                                session.subjectDN = subjectDN;
                                agency = "VA";
                                lastName = "McLuvin";
                                firstName = "Dood";
                                middleName = "";
                                scEmail = "dood@mcluvin.com"

                                // Sequence Card ID Select is active
                                if (appConfig.getSeqCardIdSelEnabled()) {
                                    scId = seqCardIdSelCurId.toString();

                                    // Loop card ID to min once it exceeds the max value
                                    if (++seqCardIdSelCurId > seqCardIdSelMaxId) {
                                        seqCardIdSelCurId = seqCardIdSelMinId;
                                    }
                                } else {
                                    scId = appConfig.getTestSmartCard();

                                    // appconfig property file does NOT have ID set, use default
                                    if (StringUtils.isBlank(scId)) {
                                        scId = cardIdDefault  // default card ID
                                    }
                                }

                                session.scEmail = scEmail;
                                session.scId = scId;
                                session.scAgency = agency;

                            } else {

                                // App server will handle the ssl connection. If no cert, ssl is not enabled (http), else
                                // ssl is enabled (https) and the server has authenticated it by this point (cert exist).
                                if (request.getScheme() != "https") {
                                    response.sendError(403, "SSLRequired");
                                    return false;
                                }

                                if (certs == null) {
                                    response.sendError(403, "SmartCardRequired");
                                    return false;
                                }

                                if (log.debugEnabled) {
                                    for (X509Certificate cert : certs) {
                                        log.debug("cert::key usage: ${cert.getKeyUsage()}");
                                        log.debug("cert::extended key usage: ${cert.getExtendedKeyUsage()}");
                                        log.debug("cert::serial number: ${cert.getSerialNumber()}");
                                        log.debug("cert::subject DN: ${cert.getSubjectDN()}");
                                    }
                                }

                                X509Certificate cert = certs[0];

                                try {
                                    cert.checkValidity();
                                }
                                catch (CertificateExpiredException e) {
                                    log.info("Could not grant user access: ${e.getMessage()}");
                                    response.sendError(403, "UserAccessDenied-Expired")
                                    return false;
                                }
                                catch (CertificateNotYetValidException e) {
                                    log.info("Could not grant user access: ${e.getMessage()}");
                                    response.sendError(403, "UserAccessDenied-NotYetValid")
                                    return false;
                                }
                                certificate = javax.xml.bind.DatatypeConverter.printBase64Binary(cert.getEncoded());
                                session.certificate = certificate;

                                subjectDN = cert.getSubjectDN();
                                session.subjectDN = subjectDN;

                                // Check for card issuer's agency

                                if (Pattern.compile(Pattern.quote("ou=dod"), Pattern.CASE_INSENSITIVE).matcher(subjectDN).find() ||
                                        Pattern.compile(Pattern.quote("dc=dod"), Pattern.CASE_INSENSITIVE).matcher(subjectDN).find()) {
                                    agency = "DOD";
                                } else if (Pattern.compile(Pattern.quote("ou=va"), Pattern.CASE_INSENSITIVE).matcher(subjectDN).find() ||
                                        Pattern.compile(Pattern.quote("dc=va"), Pattern.CASE_INSENSITIVE).matcher(subjectDN).find()) {
                                    agency = "VA";
                                } else {
                                    response.sendError(403, "UserAccessDenied-Invalid")
                                    return false;
                                }
                                session.scAgency = agency;

                                //parse person name and id from the subjectDN of primary cert
                                String[] split = subjectDN.split(",");

                                int cnIdx = -1;
                                int emailIdx = -1;

                                for (int i = 0; i < split.length; i++) {
                                    if (split[i].contains("CN=") && cnIdx < 0) {
                                        cnIdx = i;
                                    }
                                    if (split[i].contains("@")) {
                                        emailIdx = i;
                                    }
                                }

                                if (cnIdx < 0) {
                                    response.sendError(403, "UserAccessDenied-Invalid")
                                    return false;
                                }

                                if (emailIdx > 0) {
                                    scEmail = split[emailIdx].substring(split[emailIdx].indexOf("UID=") + 4, split[emailIdx].length());
                                    session.scEmail = scEmail;
                                }

                                String cn = split[cnIdx];

                                cn = cn.substring(cn.indexOf("=") + 1);

                                lastName = "";
                                firstName = "";
                                middleName = "";
                                scId = "";

                                if (agency.equalsIgnoreCase("DoD")) {
                                    StringTokenizer tok = new StringTokenizer(cn, ".");

                                    while (tok.hasMoreTokens()) {
                                        String token = tok.nextToken();
                                        //assuming that person id is always the last item in string
                                        if (!tok.hasMoreTokens()) scId = token;
                                        else {
                                            if (lastName.isEmpty()) lastName = token;
                                            else if (firstName.isEmpty()) firstName = token;
                                            else if (middleName.isEmpty()) middleName = token;
                                        }
                                    }
                                } else {
                                    StringTokenizer tok = new StringTokenizer(cn, " ");

                                    while (tok.hasMoreTokens()) {
                                        String token = tok.nextToken();

                                        if (!token.matches("\\d+")) {
                                            if (firstName.isEmpty()) firstName = token;
                                            else if (lastName.isEmpty()) lastName = token;
                                            else if (middleName.isEmpty()) {
                                                middleName = lastName;
                                                lastName = token;
                                            }
                                        } else {
                                            scId = token;
                                            session.scId = scId;
                                            //token is the last item we want, exit while loop
                                            break;
                                        }
                                    }
                                }

                            }

                            log.debug("PersonName: ${lastName}, ${firstName} ${middleName} scId: ${scId}")
                            session.profileName = lastName + ", " + firstName + " " + middleName;

                            IehrUserProfile userProfile = null
                            if (session.userProfile == null) {

                                try {

                                    userProfile = userDao.getIehrUserProfile(scId, agency, scEmail);
                                    if (scEmail != null && !"".equals(scEmail)) {
                                        userProfile.setEmailAddress(scEmail);
                                    }
                                    if (userProfile != null && userProfile.dbUserID != null && userProfile.dbUserID != "") {
                                        if (userProfile.flags == null)
                                            userProfile.setFlags("");
                                        if (userProfile.locID == null)
                                            userProfile.setLocID("");
                                        if (userProfile.cfg == null)
                                            userProfile.setCfg("");
                                        if (userProfile.emailAddress == null)
                                            userProfile.setEmailAddress("");
                                        if (userProfile.subjectDN == null)
                                            userProfile.setSubjectDN("");
                                    } else {
                                        userProfile.setDbUserID("");
                                        userProfile.setLocID("");
                                        userProfile.setCfg("");
                                        userProfile.setFlags("");
                                        userProfile.setSmartCardID(scId);
                                        userProfile.setSmartCardAgency(agency);
                                        userProfile.setEmailAddress(scEmail);
                                        userProfile.setSubjectDN(subjectDN);
                                    }
                                    session.userProfile = userProfile;
                                }
                                catch (Exception e) {
                                    log.info("Could not process JLV User Profile: ${e.getMessage()}");
                                    response.sendError(403, "ProfileRetrieve")
                                    return false;
                                }
                            }

                            GUIUser auditUser = new GUIUser();
                            auditUser.setCardId(scId);
                            auditUser.setAgency(agency);
                            auditUser.setEmail(scEmail);
                            // If organization's access flag is enabled, then do database check to verify user
                            // is authorized to access JLV
                            if ((appConfig.getDodUserAuthEnabled() && agency.equalsIgnoreCase(appConfig.getAgencyDOD()))
                                    || (appConfig.getVaUserAuthEnabled() && agency.equalsIgnoreCase(appConfig.getAgencyVA()))) {
                                try {

                                    //audit login attempt
                                    StringBuilder auditAttempt = new StringBuilder()

                                    if (agency.equalsIgnoreCase(appConfig.getAgencyDOD())) {
                                        auditAttempt.append("SnareworksLoginAttempt");
                                    } else {
                                        auditAttempt.append("getAuthUser");
                                    }
//                                auditAttempt.append("^userEDIPI=" + scId)
//                                auditAttempt.append("^scEmail=" + scEmail)
                                    auditDao.auditDetailsQuery(auditAttempt.toString(), auditUser, null, null, null, null, null)

                                    AuthUserInfo authUserInfo = new AuthUserInfo();

                                    authUserInfo.setCardId(scId);
                                    authUserInfo.setOrganization(agency);
                                    authUserInfo.setEmail(scEmail);
                                    authUserInfo.setCertificate(certificate);
                                    if (WhiteListOnlyLoginController.JLV_WHITE_LIST_ONLY_APP_NAME.equals(Metadata.current.getApplicationName())) {
                                        authUserInfo.setSnareworksFlag(false);
                                    } else {
                                        authUserInfo.setSnareworksFlag(true);
                                    }
                                    authUserInfo = authUserDao.getAuthUser(authUserInfo);

                                    if (authUserInfo == null) {
                                        response.sendError(403, "UserAccessDenied");
                                        return false;
                                    }
//                                if (authUserInfo.getSnareFlag()) {
//                                    session.snareFlag = true;
//                                }
                                    if (WhiteListOnlyLoginController.JLV_WHITE_LIST_ONLY_APP_NAME.equals(Metadata.current.getApplicationName())) {
                                        authUserInfo.setNcid(scId);
                                    }
                                    if (authUserInfo.isSnareworksValid() ||
                                            WhiteListOnlyLoginController.JLV_WHITE_LIST_ONLY_APP_NAME.equals(Metadata.current.getApplicationName())) {
                                        // TODO: Fill in more fields?
                                        GUISite hostSite = new GUISite()
                                        hostSite.setAgency(agency)
                                        hostSite.setDmisId("0000")
                                        hostSite.setMoniker("AHLTA")
                                        hostSite.setName("AHLTA")
                                        hostSite.setSiteCode("0000")
                                        hostSite.setStatus("active")

                                        // TODO: Fill in more fields?
                                        GUIUser user = new GUIUser()
                                        user.setHostSite(hostSite)
                                        user.setAgency(agency)
                                        if (authUserInfo.isSnareworksValid()) {
                                            user.setName(authUserInfo.getName())
                                        } else {
                                            user.setName(lastName + "," + firstName + " " + middleName)
                                        }
                                        user.setUserIen(authUserInfo.getNcid())
                                        user.setEncryptedFederatedUid(EncryptUtil.encrypt("123456789"))
                                        user.setCardId(scId);
                                        user.setEmail(scEmail);

                                        // Audit Successful login audit
                                        StringBuilder auditSuccess = new StringBuilder()
                                        if (authUserInfo.isSnareworksValid()) {
                                            auditSuccess.append("SnareworksSuccess")
                                        } else {
                                            auditSuccess.append("WhiteListOnlySuccess")
                                        }
                                        auditSuccess.append("^userEDIPI=" + scId)
                                        auditSuccess.append("^userName=" + authUserInfo.getName())
                                        auditSuccess.append("^userNCID=" + authUserInfo.getNcid())
                                        auditDao.auditDetailsQuery("login", auditSuccess.toString(), user, null, null, null, null, agency)

                                        AppUser appUser = new AppUser(user)
                                        appUser.setSecurityKey(tokenUtils.generateSecurityKey())

                                        //check for valid profile
                                        if (session.userProfile.getDbUserID() == null || "".equalsIgnoreCase(session.userProfile.getDbUserID())) {
                                            //user doesn' have profile, let's create one.
                                            try {
                                                if (userDao.setIehrUserProfile(session.userProfile)) {
                                                    if (StringUtils.isEmpty(userProfile.getDbUserID())) {
                                                        userProfile = userDao.getIehrUserProfile(userProfile.smartCardID, userProfile.smartCardAgency, userProfile.getEmailAddress())
                                                        appUser.setUserId(userProfile.getDbUserID())
                                                        appUser.setCfg(userProfile.getCfg())
                                                    }
                                                }

                                            }
                                            catch (Exception e) {
                                                String errorText = "Could not save the User Profile."
                                                log.error("SecurityFilters::setIehrUserProfile", e)
                                                response.sendError(403, 'ErrorCreatingProfile')
                                                return
                                            }

                                        } else {
                                            appUser.setUserId(session.userProfile.getDbUserID())
                                            appUser.setCfg(session.userProfile.getCfg())
                                        }
                                        session.appUser = appUser
                                        redirect(controller: 'app')
                                    }
                                }
                                catch (DaoException e) {
                                    if (agency.equalsIgnoreCase(appConfig.getAgencyVA())) {
                                        response.sendError(403, "UserAccessDenied");
                                    } else {
                                        response.sendError(403, "SnareworksIssue");
                                    }

                                    return false;
                                }
                            } else {
                                auditDao.auditDetailsQuery("authNoDbVerification", auditUser, null, null, null, null, null)
                            }

                            return true;
                        }
                    }
                    after = { Map model ->

                    }
                    afterView = { Exception e ->

                    }
                }
    }
}

