/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package jlv

import gov.va.med.common.HttpKeys
import gov.va.med.dao.DaoException
import gov.va.med.guibeans.*
import gov.va.med.jmeadows.webservice.NoteImage
import gov.va.med.jmeadows.webservice.PatientDocument
import gov.va.med.jmeadows.webservice.Site
import gov.va.med.vhahon.common.Utils
import org.apache.commons.beanutils.BeanUtils
import org.apache.commons.lang.StringUtils

class PatientDocumentsController extends PatientNoteDetailBaseController {

    def documentDao
    def noteDao
    def auditDao
    def sensitiveRecordDao

    static allowedMethods = [list: 'POST']

    def beforeInterceptor = {

        if (request.patient == null) {
            response.sendError 500
            return false;
        }

        if (actionName == 'details' || actionName == 'getNoteDataStream') {
            if (request.recordQuery == null) {
                response.sendError 500
                return false;
            }
        } else dateRangeCheck();
    }

    def list = {

        def DOCUMENT_PROGRESS_NOTE = "Progress Note";
        def DOCUMENT_ENCOUNTER_NOTE = "Encounter Note";
        def DOCUMENT_OUTPATIENT_ENCOUNTER = "Outpatient Encounter";
        def DOCUMENT_CONSULT_REPORT = "Consult Report";
        def DOCUMENT_DEPLOYMENT_FORM = "Deployment Form";
        def String DOCUMENT_INPATIENT_NOTE = "Inpatient Note";
        def String DOCUMENT_QUESTIONNAIRE = "Questionnaires";
        def String DOCUMENT_RADIOLOGY = "Radiology Exams";
        def String DOCUMENT_HAIMS = "HAIMS";

        def guiDocuments = documentDao.getPatientDocuments(
                session.appUser,
                request.patient,
                Utils.getCal(request[HttpKeys.START_DATE]),
                Utils.getCal(request[HttpKeys.END_DATE]))

        String ipAddress = request.getHeader("X-FORWARDED-FOR");
        if (ipAddress == null) {
            ipAddress = request.getRemoteAddr();
        }

        if (guiDocuments != null) {
            for (def guiDocument : guiDocuments) {

                if (DOCUMENT_PROGRESS_NOTE.equalsIgnoreCase(guiDocument.getDocumentType())
                        || DOCUMENT_ENCOUNTER_NOTE.equalsIgnoreCase(guiDocument.getDocumentType())
                        || DOCUMENT_HAIMS.equalsIgnoreCase(guiDocument.getDocumentType())) {

                    if (StringUtils.isNotEmpty(guiDocument.getNoteId())) {
                        def queryMap = [(HttpKeys.NOTE_ID)            : guiDocument.getNoteId(),
                                        (HttpKeys.RECORD_DATE)        : guiDocument.getNoteDate(),
                                        (HttpKeys.RECORD_SITE_MONIKER): guiDocument.getSite().getMoniker(),
                                        (HttpKeys.RECORD_SITE_AGENCY) : guiDocument.getSite().getAgency()]


                        if (DOCUMENT_HAIMS.equalsIgnoreCase(guiDocument.getDocumentType()) &&
                                guiDocument.getContentType()
                                && (guiDocument.getContentType().indexOf("application/hpx") >= 0
                                || guiDocument.getContentType().indexOf("application/hta") >= 0
                                || guiDocument.getContentType().indexOf("application/pkix-crl") >= 0
                                || guiDocument.getContentType().indexOf("application/x-javascript") >= 0
                                || guiDocument.getContentType().indexOf("application/x-x509-ca-cert") >= 0
                                || guiDocument.getContentType().indexOf("message/rfc822") >= 0)) {
                            guiDocument.setRecordDetailsToken("NOT_SUPPORTED");
                            guiDocument.setSessionCachedRecord(false);
                            guiDocument.setComplexDataUrl("");

                        } else {
                            guiDocument.setRecordDetailsToken(
                                    generateRecordDetailsToken(
                                            params[HttpKeys.PATIENT_TOKEN],
                                            controllerName,
                                            queryMap,
                                            guiDocument.getSite(),
                                            guiDocument.getSourceProtocol(),
                                            guiDocument.isSessionCachedRecord()))
                        }

                        guiDocument.setGuiIpAddress(ipAddress)

                        if (guiDocument.isSessionCachedRecord()) {
                            //cache outpatient progress note in session
                            def progressNoteDetail = BeanUtils.cloneBean(guiDocument)

                            session.appUser.outpatientNotesMap.put(
                                    getCachedRecordKey(guiDocument.getRecordDetailsToken()), progressNoteDetail)

                            //don't send note text with note listings
                            guiDocument.setNoteText("")
                            guiDocument.setImages("")
                        }

                        // Don't pass out any sensitive data as part of the progress notes list
                        if (guiDocument.isSensitive()) {
                            if (guiDocument.getNoteText() != null && !guiDocument.getNoteText().isEmpty()) {
                                guiDocument.setNoteText("*** Sensitive ***");
                            } else {
                                guiDocument.setNoteText("");
                            }
                            if (guiDocument.getComplexDataUrl() != null && !guiDocument.getComplexDataUrl().isEmpty()) {
                                guiDocument.setComplexDataUrl("*** Sensitive ***");
                            } else {
                                guiDocument.setComplexDataUrl("");
                            }
                            guiDocument.setImages("");
                            guiDocument.setNoteType("");
                            guiDocument.setComplexDataUrl("");
                            guiDocument.setCodes(null);
                        }
                    }
                } else if (DOCUMENT_INPATIENT_NOTE.equalsIgnoreCase(guiDocument.getDocumentType())) {

                    if (StringUtils.isNotEmpty(guiDocument.getNoteId())) {
                        def queryMap = [(HttpKeys.NOTE_ID)            : guiDocument.getNoteId(),
                                        (HttpKeys.RECORD_DATE)        : guiDocument.getNoteDate(),
                                        (HttpKeys.RECORD_SITE_MONIKER): guiDocument.getSite().getMoniker(),
                                        (HttpKeys.RECORD_SITE_AGENCY) : guiDocument.getSite().getAgency()]



                        guiDocument.setRecordDetailsToken(
                                generateRecordDetailsToken(
                                        params[HttpKeys.PATIENT_TOKEN],
                                        "patientDischargeSummaries",
                                        queryMap,
                                        guiDocument.getSite(),
                                        guiDocument.getSourceProtocol(),
                                        guiDocument.isSessionCachedRecord()))

                        guiDocument.setGuiIpAddress(ipAddress)

                        if (guiDocument.isSessionCachedRecord()) {
                            //cache outpatient progress note in session
                            def progressNoteDetail = BeanUtils.cloneBean(guiDocument)

                            session.appUser.inpatientNotesMap.put(
                                    getCachedRecordKey(guiDocument.getRecordDetailsToken()), progressNoteDetail)

                            //don't send note text with note listings
                            guiDocument.setNoteText("")
                            guiDocument.setImages("")
                        }

                        // Don't pass out any sensitive data as part of the progress notes list
                        if (guiDocument.isSensitive()) {
                            if (guiDocument.getNoteText() != null && !guiDocument.getNoteText().isEmpty()) {
                                guiDocument.setNoteText("*** Sensitive ***");
                            } else {
                                guiDocument.setNoteText("");
                            }
                            if (guiDocument.getComplexDataUrl() != null && !guiDocument.getComplexDataUrl().isEmpty()) {
                                guiDocument.setComplexDataUrl("*** Sensitive ***");
                            } else {
                                guiDocument.setComplexDataUrl("");
                            }
                            guiDocument.setImages("");
                            guiDocument.setNoteType("");
                            guiDocument.setComplexDataUrl("");
                            guiDocument.setCodes(null);
                        }
                    }
                } else if (DOCUMENT_OUTPATIENT_ENCOUNTER.equalsIgnoreCase(guiDocument.getDocumentType())) {
                    GUIPatientEncounter guiEncounter = guiDocument.getOutpatientEncounter();

                    if (StringUtils.isNotEmpty(guiEncounter.getEncounterId())) {
                        String noteId = "";
                        if (guiEncounter.getDocuments() != null
                                && guiEncounter.getDocuments().size() > 0
                                && guiEncounter.getDocuments().get(0).getId() != null) {
                            noteId = guiEncounter.getDocuments().get(0).getId();
                            guiEncounter.setVaNote(true);
                        }

                        def queryMap = [(HttpKeys.ENCOUNTER_ID)       : guiEncounter.getEncounterId(),
                                        (HttpKeys.NOTE_ID)            : noteId,
                                        (HttpKeys.RECORD_DATE)        : guiEncounter.getApptDate(),
                                        (HttpKeys.RECORD_SITE_AGENCY) : guiEncounter.getSite().getAgency(),
                                        (HttpKeys.RECORD_SITE_MONIKER): guiEncounter.getSite().getMoniker()]

                        guiEncounter.setRecordDetailsToken(
                                generateRecordDetailsToken(
                                        params[HttpKeys.PATIENT_TOKEN],
                                        "patientEncounters",
                                        queryMap,
                                        guiEncounter.getSite(),
                                        guiEncounter.getSourceProtocol(),
                                        guiEncounter.isSessionCachedRecord()))

                        guiDocument.setRecordDetailsToken(guiEncounter.getRecordDetailsToken());
                        guiEncounter.setGuiIpAddress(ipAddress)
                        guiDocument.setGuiIpAddress(ipAddress)

                        if (guiEncounter.isSessionCachedRecord()) {
                            //cache encounter report in session
                            def encounterDetail = BeanUtils.cloneBean(guiEncounter)

                            session.appUser.encounterMap.put(
                                    getCachedRecordKey(guiEncounter.getRecordDetailsToken()), encounterDetail)

                            //don't send report text with note listings
                            //guiEncounter.setReport("")
                        }

                        // Don't pass out any sensitive data as part of the progress notes list
                        if (guiEncounter.isSensitive()) {
                            guiEncounter.setClinic("** Sensitive **");
                            guiEncounter.setReason("** Sensitive **");
                            guiEncounter.setDiagnosis(null);
                            guiEncounter.setProcedures(null);
                            guiEncounter.setEncounterNumber("");
                            guiEncounter.setEncounterId("");
                            if (guiEncounter.getComplexDataUrl() != null && !guiEncounter.getComplexDataUrl().isEmpty()) {
                                guiEncounter.setComplexDataUrl("** Sensitive **");
                            } else {
                                guiEncounter.setComplexDataUrl("");
                            }
                            guiEncounter.setReasonNarrative("");
                            guiEncounter.setDischargeDisposition("");
                            guiEncounter.setDocuments(null);
                            guiEncounter.setReasonCode("");
                            guiEncounter.setCodes(null);
                        }
                    }

                } else if (DOCUMENT_CONSULT_REPORT.equalsIgnoreCase(guiDocument.getDocumentType())) {
                    GUIConsult guiConsult = guiDocument.getConsult();
                    String consultId = "";
                    if (!StringUtils.isEmpty(guiConsult.getConsultId())) {
                        consultId = guiConsult.getConsultId();
                    } else if (!StringUtils.isEmpty(guiDocument.getEventId())) {
                        consultId = guiDocument.getEventId();
                    } else if (!StringUtils.isEmpty(guiDocument.getEventId())) {
                        consultId = guiDocument.getNoteId();
                    }

                    guiConsult.setConsultId(consultId);

                    def queryMap = [(HttpKeys.CONSULT_ID)         : consultId,
                                    (HttpKeys.RECORD_DATE)        : guiConsult.requestDate,
                                    (HttpKeys.RECORD_SITE_AGENCY) : guiConsult.site.agency,
                                    (HttpKeys.RECORD_SITE_MONIKER): guiConsult.site.moniker]

                    guiDocument.setRecordDetailsToken(
                            generateRecordDetailsToken(
                                    params[HttpKeys.PATIENT_TOKEN],
                                    "patientConsults",
                                    queryMap,
                                    guiDocument.getSite(),
                                    guiDocument.getSourceProtocol(),
                                    guiDocument.isSessionCachedRecord()))

                    guiConsult.setRecordDetailsToken(guiDocument.getRecordDetailsToken());

                    if (guiConsult.isSessionCachedRecord()) {
                        //cache consult report in session
                        def consultDetail = BeanUtils.cloneBean(guiConsult)

                        session.appUser.consultMap.put(
                                getCachedRecordKey(guiConsult.getRecordDetailsToken()), consultDetail)

                        //don't send report text with note listings
                        guiConsult.setReport("")
                    }

                } else if (DOCUMENT_QUESTIONNAIRE.equalsIgnoreCase(guiDocument.getDocumentType())
                        || DOCUMENT_DEPLOYMENT_FORM.equalsIgnoreCase(guiDocument.getDocumentType())) {

                    GUIPatientForm guiForm = guiDocument.getForm();
                    guiForm.setSessionCachedRecord(true);

                    if (guiForm.getFormId() == null) {
                        guiForm.setFormId(guiDocument.getNoteId());
                    }
                    if (StringUtils.isNotEmpty(guiForm.formId)) {
                        def queryMap = [(HttpKeys.FORM_ID)            : guiForm.getFormId(),
                                        (HttpKeys.RECORD_DATE)        : guiForm.getFormDate(),
                                        (HttpKeys.RECORD_SITE_AGENCY) : guiForm.site.getAgency(),
                                        (HttpKeys.RECORD_SITE_MONIKER): guiForm.site.getMoniker()]

                        guiForm.setRecordDetailsToken(
                                generateRecordDetailsToken(
                                        params[HttpKeys.PATIENT_TOKEN],
                                        "patientForms",
                                        queryMap,
                                        guiForm.getSite(),
                                        guiForm.getSourceProtocol(),
                                        guiForm.isSessionCachedRecord()))

                        guiDocument.setRecordDetailsToken(guiForm.getRecordDetailsToken());

                        if (guiDocument.isSessionCachedRecord()) {
                            //cache consult report in session
                            def formDetail = BeanUtils.cloneBean(guiForm)

                            session.appUser.formMap.put(
                                    getCachedRecordKey(guiForm.getRecordDetailsToken()), formDetail)

                            //don't send report text with note listings
                            guiForm.setDeploymentForm(null);
                            guiForm.setQuestionnaireBean(null);
                        }
                    }
                } else if (DOCUMENT_RADIOLOGY.equalsIgnoreCase(guiDocument.getDocumentType())) {

                    GUIRadiologyReport guiRad = guiDocument.getRadiologyReport();
                    //guiRad.setSessionCachedRecord(true);

                    if (guiRad.getCacheId() == null) {
                        guiRad.setCacheId(guiRad.getExamId());
                    }
                    if (StringUtils.isNotEmpty(guiRad.getCacheId())) {
                        def queryMap = [(HttpKeys.RAD_CACHE_ID)              : guiRad.getCacheId(),
                                        (HttpKeys.RAD_EXAM_ID)               : guiRad.getExamId(),
                                        (HttpKeys.RADIOLOGY_ACCESSION_NUMBER): guiRad.getAccessionNumber(),
                                        (HttpKeys.RECORD_SITE_MONIKER)       : guiRad.site.moniker,
                                        (HttpKeys.RECORD_SITE_AGENCY)        : guiRad.site.agency]

                        guiRad.setRecordDetailsToken(
                                generateRecordDetailsToken(
                                        params[HttpKeys.PATIENT_TOKEN],
                                        "patientRadiologyExams",
                                        queryMap,
                                        guiRad.getSite(),
                                        guiRad.getSourceProtocol(),
                                        guiRad.isSessionCachedRecord()))

                        guiDocument.setRecordDetailsToken(guiRad.getRecordDetailsToken());

                        if (guiRad.isSessionCachedRecord()) {
                            //cache radiology report in session
                            def guiRadDetail = BeanUtils.cloneBean(guiRad)

                            session.appUser.radiologyReportsMap.put(
                                    getCachedRecordKey(guiRad.getRecordDetailsToken()), guiRadDetail)

                            //don't send report text with rad exam listings
                            guiRad.setReportText("")
                        }
                    }
                }


            }
        }

        sendResponse(guiDocuments)
    }

    def details = {

        GUIProgressNote guiNote = null

        if (!request.isSessionCachedRecord) {
            guiNote = noteDao.getProgressNote(
                    session.appUser,
                    request.patient,
                    request.recordQuery[HttpKeys.NOTE_ID],
                    request.recordQuery[HttpKeys.RECORD_SITE_CODE],
                    request.recordQuery[HttpKeys.RECORD_DATE] + '^' + params[HttpKeys.TITLE]);

            def auditCategory = "ProgressNoteDetail";
            if (params[HttpKeys.WIDGET_NAME]) {
                auditCategory = params[HttpKeys.WIDGET_NAME] + '^' + auditCategory;
            }

            auditDao.auditDetailsQuery(auditCategory, request.recordQuery[HttpKeys.RECORD_DATE] + '^' +
                    params[HttpKeys.TITLE],
                    session.appUser, request.patient,
                    null,
                    request.recordQuery[HttpKeys.NOTE_ID],
                    request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

            sendResponse(guiNote)
            return true;
        }

        //retrieve cached note
        guiNote = session.appUser.outpatientNotesMap.get(
                getCachedRecordKey(params[HttpKeys.RECORD_DETAILS_TOKEN]))

        if (guiNote.sensitive) {
            String auditText = params[HttpKeys.AUDIT_TEXT]
            // audit with break the glass
            auditText = "btgPatientProgressNotes^" + auditText;
            String complexDataUrl = guiNote.getComplexDataUrl();

            String recordID = null;

            if (complexDataUrl != null) {
                String[] parsedComplex = complexDataUrl.split("/");
                recordID = parsedComplex[parsedComplex.length - 1];
            }

            sensitiveRecordDao.auditSensitiveRecordAccess(session.appUser, request.patient, auditText, recordID,
                    session.scId, session.scEmail, complexDataUrl, request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])
        }

        def auditCategory = "ProgressNoteDetail";
        if (params[HttpKeys.WIDGET_NAME]) {
            auditCategory = params[HttpKeys.WIDGET_NAME] + '^' + auditCategory;
        }

        auditDao.auditDetailsQuery(auditCategory, request.recordQuery[HttpKeys.RECORD_DATE], session.appUser,
                request.patient, null, request.recordQuery[HttpKeys.NOTE_ID],
                request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

        sendResponse(guiNote)
    }

    def auditDetail = {
        auditDao.auditDetailsQuery("ProgressNoteDetail", request.recordQuery[HttpKeys.RECORD_DATE], session.appUser,
                request.patient, null, request.recordQuery[HttpKeys.NOTE_ID],
                request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])
    }

    def getNoteDataStream = {

        //retrieve cached note
        GUIProgressNote guiNote = session.appUser.outpatientNotesMap.get(
                getCachedRecordKey(params[HttpKeys.RECORD_DETAILS_TOKEN]))

        def auditQueryName = "Documents^GetNoteDataStream"

        if ((guiNote == null) || !guiNote.noteAsBinaryStream) {
            response.sendError 500
            return false;
        }

        if (guiNote.sensitive) {
            String auditText = params[HttpKeys.AUDIT_TEXT]
            // audit with break the glass
            auditText = "btgPatientProgressNotes^" + auditText;

            String complexDataUrl = guiNote.getComplexDataUrl();

            String recordID = null;

            if (complexDataUrl != null) {
                String[] parsedComplex = complexDataUrl.split("/");
                recordID = parsedComplex[parsedComplex.length - 1];
            }

            sensitiveRecordDao.auditSensitiveRecordAccess(session.appUser, request.patient, auditText, recordID, session.scId,
                    session.scEmail, complexDataUrl, request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])
        }

        //stream share/essentris CDA notes
        //if (guiNote.getSite().siteCode.startsWith("2.16.840.1.113883.3.42.10002.100001.8")) {
        if (guiNote.contentType != null && ("CDA".equalsIgnoreCase(guiNote.contentType) || guiNote.contentType.indexOf("+xml") > 0 || guiNote.contentType.indexOf("/xml") > 0) && guiNote.sourceProtocol == appConfig.getProtocolDODAdapter()) {
            auditDao.auditDetailsQuery(auditQueryName, request.recordQuery[HttpKeys.RECORD_DATE] + '^' +
                    guiNote.getNoteTitle(), session.appUser,
                    request.patient, guiNote.getNoteId(),
                    request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

            makeCIS_CDAText(guiNote.getNoteText())
            return true;
        }
        //fetch dod note image
        else {

            String complexDataUrl = guiNote.getComplexDataUrl();

            NoteImage noteImage = null;

            if (complexDataUrl != null) {

                String[] parsedComplex = complexDataUrl.split("/");
                String recordID = parsedComplex[parsedComplex.length - 1];

                auditDao.auditDetailsQuery(auditQueryName, request.recordQuery[HttpKeys.RECORD_DATE] + '^' +
                        guiNote.getNoteTitle(), session.appUser, request.patient,
                        complexDataUrl,
                        recordID,
                        request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                        request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

                try {
                    noteImage = noteDao.getBHIENoteImage(session.appUser, complexDataUrl);
                }
                catch (DaoException e) {
                    log.error(e);
                }

            }

            response.setHeader("Pragma", "public");
            response.setHeader("Cache-Control", "maxage=1");

            if (noteImage != null && noteImage.getNoteBytes() != null) {
                response.setContentType(noteImage.getContentType());

                String contentDisposition = noteImage.getContentDisposition()
                response.setHeader("Content-Disposition", contentDisposition.replace("attachment", "inline"))


                response.getOutputStream().write(noteImage.getNoteBytes());
                response.getOutputStream().flush();
                response.getOutputStream().close();

                return true;
            }

            //if note attachment is null, send an error message

            String errorMsg = "An error occurred while attempting to retrieve note. - JLV";

            response.setContentType("text");
            response.setContentLength(errorMsg.length());
            response.getWriter() << errorMsg;
            response.getWriter().flush();

            response.sendError 500
            return false;
        }
    }

    def auditVixViewer = {
        auditDao.auditDetailsQuery(
                'DocumentVixViewer',
                session.appUser,
                request.patient,
                null,
                request.recordQuery[HttpKeys.NOTE_ID],
                request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                request.recordQuery[HttpKeys.RECORD_SITE_AGENCY]);
    }

    def getDocumentVixStudyViewerUrl = {
        def patientDocument = new PatientDocument();
        def site = new Site();
        site.setSiteCode(params.siteCode);
        patientDocument.setSite(site);
        patientDocument.setNoteId(params.noteId);
        patientDocument.setImages("1");

        def vixStudyViewerUrl = documentDao.getDocumentVixStudyViewerUrl(
                session.appUser,
                request.patient,
                Utils.getCal(request[HttpKeys.START_DATE]),
                Utils.getCal(request[HttpKeys.END_DATE]),
                patientDocument);
        sendResponse([vixStudyViewerUrl: vixStudyViewerUrl])
    }


}
