/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package jlv

import gov.va.med.common.HttpKeys
import gov.va.med.dao.DaoException
import gov.va.med.guibeans.GUIProgressNote
import gov.va.med.jmeadows.webservice.NoteImage
import gov.va.med.vhahon.common.Utils
import org.apache.commons.beanutils.BeanUtils
import org.apache.commons.lang.StringUtils

class PatientProgressNotesController extends PatientNoteDetailBaseController {
    def noteDao
    def auditDao
    def sensitiveRecordDao

    static allowedMethods = [list: 'POST']

    def beforeInterceptor = {

        if (request.patient == null) {
            response.sendError 500
            return false;
        }

        if (actionName == 'details' || actionName == 'getNoteDataStream') {
            if (request.recordQuery == null) {
                response.sendError 500
                return false;
            }
        } else dateRangeCheck();
    }

    def list = {

        def guiOutpatientNotes = noteDao.getPatientProgressNotes(
                session.appUser,
                request.patient,
                Utils.getCal(request[HttpKeys.START_DATE]),
                Utils.getCal(request[HttpKeys.END_DATE]))

        String ipAddress = request.getHeader("X-FORWARDED-FOR");
        if (ipAddress == null) {
            ipAddress = request.getRemoteAddr();
        }

        if (guiOutpatientNotes != null) {
            for (def guiOutpatientNote : guiOutpatientNotes) {
                if (StringUtils.isNotEmpty(guiOutpatientNote.getNoteId())) {
                    def queryMap = [(HttpKeys.NOTE_ID)            : guiOutpatientNote.getNoteId(),
                                    (HttpKeys.RECORD_DATE)        : guiOutpatientNote.getNoteDate(),
                                    (HttpKeys.RECORD_SITE_MONIKER): guiOutpatientNote.getSite().getMoniker(),
                                    (HttpKeys.RECORD_SITE_AGENCY) : guiOutpatientNote.getSite().getAgency()]

                    guiOutpatientNote.setRecordDetailsToken(
                            generateRecordDetailsToken(
                                    params[HttpKeys.PATIENT_TOKEN],
                                    controllerName,
                                    queryMap,
                                    guiOutpatientNote.getSite(),
                                    guiOutpatientNote.getSourceProtocol(),
                                    guiOutpatientNote.isSessionCachedRecord()))

                    guiOutpatientNote.setGuiIpAddress(ipAddress)

                    if (guiOutpatientNote.isSessionCachedRecord()) {
                        //cache outpatient progress note in session
                        def progressNoteDetail = BeanUtils.cloneBean(guiOutpatientNote)

                        session.appUser.outpatientNotesMap.put(
                                getCachedRecordKey(guiOutpatientNote.getRecordDetailsToken()), progressNoteDetail)

                        //don't send note text with note listings
                        guiOutpatientNote.setNoteText("")
                        guiOutpatientNote.setImages("")
                    }

                    // Don't pass out any sensitive data as part of the progress notes list
                    if (guiOutpatientNote.isSensitive()) {
                        if(guiOutpatientNote.getNoteText() != null && !guiOutpatientNote.getNoteText().isEmpty())
                        {
                            guiOutpatientNote.setNoteText("*** Sensitive ***");
                        } else {
                            guiOutpatientNote.setNoteText("");
                        }
                        if(guiOutpatientNote.getComplexDataUrl() != null && !guiOutpatientNote.getComplexDataUrl().isEmpty())
                        {
                            guiOutpatientNote.setComplexDataUrl("*** Sensitive ***");
                        } else {
                            guiOutpatientNote.setComplexDataUrl("");
                        }
                        guiOutpatientNote.setImages("");
                        guiOutpatientNote.setNoteType("");
                        guiOutpatientNote.setComplexDataUrl("");
                        guiOutpatientNote.setCodes(null);
                    }
                }
            }
        }

        sendResponse(guiOutpatientNotes)
    }

    //VA notes
    def details = {

        GUIProgressNote guiNote = null

        if (!request.isSessionCachedRecord) {
            guiNote = noteDao.getProgressNote(
                    session.appUser,
                    request.patient,
                    request.recordQuery[HttpKeys.NOTE_ID],
                    request.recordQuery[HttpKeys.RECORD_SITE_CODE],
                    request.recordQuery[HttpKeys.RECORD_DATE] + '^' + params[HttpKeys.TITLE]);

            def auditCategory = "ProgressNoteDetail";
            if (params[HttpKeys.WIDGET_NAME]) {
                auditCategory = params[HttpKeys.WIDGET_NAME] + '^' + auditCategory;
            }

            auditDao.auditDetailsQuery(auditCategory, request.recordQuery[HttpKeys.RECORD_DATE] + '^' +
                    params[HttpKeys.TITLE],
                    session.appUser, request.patient,
                    null,
                    request.recordQuery[HttpKeys.NOTE_ID],
                    request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

            sendResponse(guiNote)
            return true;
        }

        //retrieve cached note
        guiNote = session.appUser.outpatientNotesMap.get(
                getCachedRecordKey(params[HttpKeys.RECORD_DETAILS_TOKEN]))

        if (guiNote.sensitive) {
            String auditText = params[HttpKeys.AUDIT_TEXT]
            // audit with break the glass
            auditText = "btgPatientProgressNotes^" + auditText;

            sensitiveRecordDao.auditSensitiveRecordAccess(session.appUser, request.patient, auditText,
                    request.recordQuery[HttpKeys.NOTE_ID], session.scId, session.scEmail, "",
                    request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])
        }

        auditDao.auditDetailsQuery("ProgressNoteDetail", guiNote.getNoteDate() + '^' + guiNote.getNoteTitle(),
                session.appUser, request.patient, null,
                request.recordQuery[HttpKeys.NOTE_ID],
                request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

        sendResponse(guiNote)
    }

    //DOD Progress Notes
    def getNoteDataStream = {

        //retrieve cached note
        GUIProgressNote guiNote = session.appUser.outpatientNotesMap.get(
                getCachedRecordKey(params[HttpKeys.RECORD_DETAILS_TOKEN]))

        def auditQueryName = "ProgressNotes^GetNoteDataStream"

        if ((guiNote == null) || !guiNote.noteAsBinaryStream) {
            response.sendError 500
            return false;
        }

        if (guiNote.sensitive) {
            String auditText = params[HttpKeys.AUDIT_TEXT]
            // audit with break the glass
            auditText = "btgPatientProgressNotes^" + auditText;
            String complexDataUrl = guiNote.getComplexDataUrl();
            String recordID = null;

            if (complexDataUrl != null) {
                if (guiNote.getEventId() == null || guiNote.getEventId().length() == 0) {
                    String[] parsedComplex = complexDataUrl.split("/");
                    recordID = parsedComplex[parsedComplex.length - 1];
                } else {
                    recordID = guiNote.getEventId();
                }
            }

            sensitiveRecordDao.auditSensitiveRecordAccess(session.appUser, request.patient,
                    auditText, recordID, session.scId, session.scEmail, complexDataUrl,
                    request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])
        }

        //stream share/essentris CDA notes
        //if (guiNote.getSite().siteCode.startsWith("2.16.840.1.113883.3.42.10002.100001.8")) {
        if (guiNote.contentType != null && ("CDA".equalsIgnoreCase(guiNote.contentType) ||
                guiNote.contentType.indexOf("+xml") > 0 || guiNote.contentType.indexOf("/xml") > 0) &&
                guiNote.sourceProtocol == appConfig.getProtocolDODAdapter()) {
            auditDao.auditDetailsQuery(auditQueryName, request.recordQuery[HttpKeys.RECORD_DATE] + '^' + guiNote.getNoteTitle(),
                    session.appUser, request.patient, null,
                    request.recordQuery[HttpKeys.NOTE_ID],
                    request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

            makeCIS_CDAText(guiNote.getNoteText())
            return true;
        }
        //fetch dod note image
        else {

            String complexDataUrl = guiNote.getComplexDataUrl();

            NoteImage noteImage = null;

            if (complexDataUrl != null) {

                auditDao.auditDetailsQuery(auditQueryName, request.recordQuery[HttpKeys.RECORD_DATE] + '^' +
                        guiNote.getNoteTitle(),
                        session.appUser, request.patient,
                        complexDataUrl,
                        guiNote.getEventId(),
                        request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                        request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

                try {
                    noteImage = noteDao.getBHIENoteImage(session.appUser, complexDataUrl);
                }
                catch (DaoException e) {
                    log.error(e);
                }

            }

            response.setHeader("Pragma", "public");
            response.setHeader("Cache-Control", "maxage=1");

            if (noteImage != null && noteImage.getNoteBytes() != null) {
                response.setContentType(noteImage.getContentType());

                String contentDisposition = noteImage.getContentDisposition()
                response.setHeader("Content-Disposition", contentDisposition.replace("attachment", "inline"))


                response.getOutputStream().write(noteImage.getNoteBytes());
                response.getOutputStream().flush();
                response.getOutputStream().close();

                return true;
            }

            //if note attachment is null, send an error message

            String errorMsg = "An error occurred while attempting to retrieve note. - JLV";

            response.setContentType("text");
            response.setContentLength(errorMsg.length());
            response.getWriter() << errorMsg;
            response.getWriter().flush();

            response.sendError 500
            return false;
        }
    }

    def auditVixViewer = {
        auditDao.auditDetailsQuery(
                'ProgressNoteVixViewer',
                session.appUser,
                request.patient,
                null,
                request.recordQuery[HttpKeys.NOTE_ID],
                request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                request.recordQuery[HttpKeys.RECORD_SITE_AGENCY]);
    }
}
