/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package jlv

import gov.va.med.common.HttpKeys
import gov.va.med.guibeans.GUIKeyValuePair
import gov.va.med.jmeadows.webservice.Alert
import gov.va.med.jmeadows.webservice.Patient
import org.apache.commons.collections.CollectionUtils
import org.apache.commons.lang3.StringUtils

//import net.carefx.contextapi.ContextItem
//import net.carefx.contextapi.PublishResponse
//import net.carefx.contextapi.SurveyResult
//import net.carefx.contextapi.TypeDecision
//import net.carefx.contextmgt.cm.CMException

class PatientSelectController extends BaseController {

    def patientDao

    def tokenUtils

    static allowedMethods = [list: 'POST']

    def beforeInterceptor = {

        if (params[HttpKeys.PATIENT_TOKEN] == null) {
            response.sendError 500
            return false;
        }

        //clear session maps that contain cached patient data
        session.appUser.clearMaps();
    }

    def auditVARestrictedAccess = {
        def selectedPatientMap = session.appUser.selectedPatientMap;
        def patientToken = params[HttpKeys.PATIENT_TOKEN];
        def auditText = params["auditTxt"];
        def selectedPatient = null;

        if (selectedPatientMap != null && session.restrictedPat != null) {
            selectedPatient = session.restrictedPat;
            if (!selectedPatient.isSensitive()) {
                // Patient needs to go through another check (senstive audit) before allowing to add to map.
                selectedPatientMap.put(patientToken, selectedPatient);
            }
        }

        if (selectedPatient != null) {
            patientDao.auditVARestrictedAccess(session.appUser, selectedPatient, auditText);
        }

        sendResponse(["success": true])
    }

    def selectPatientForVASensitive = {
        def selectedPatientMap = session.appUser.selectedPatientMap;
        def patientToken = params[HttpKeys.PATIENT_TOKEN];
        def selectedPatient = null;

        if (selectedPatientMap != null) {
            if (session.restrictedPat != null) {
                selectedPatient = session.restrictedPat;
                selectedPatientMap.put(patientToken, selectedPatient);
            }
            else {
                selectedPatient = selectedPatientMap.get(patientToken);
            }
        }

        if (selectedPatient != null && CollectionUtils.isNotEmpty(selectedPatient.patientIens)) {
            patientDao.selectPatientForVASensitive(session.appUser, selectedPatient);
        }

        sendResponse(["success": true])
    }

    def select = {


        def patientToken = params[HttpKeys.PATIENT_TOKEN]
        def cmNotify = params[HttpKeys.CONTEXT_NOTIFY];
        def patName = params[HttpKeys.PATIENT_NAME];
        def patEdipi = params["patEdipi"];
        def patSsn = params["patSsn"];
        def patSponsorSsn = params["patSponsorSsn"];
        def patGender = params["patGender"];
        def patDob = params["patDob"];
        def patSensitive = params["patSensitive"];
        def patIcn = params["patIcn"];

        def provider = session.appUser
        def selectedPatientMap = session.appUser.selectedPatientMap

        def selectedPatient;
        String resultScript = ""
        def responseGUIPatientSelect;
        def userSelectedOwnRecord = StringUtils.startsWith(patientToken, "selfselectaccessdenied_");
        if (userSelectedOwnRecord)
            patientToken = StringUtils.substring(patientToken, "selfselectaccessdenied_".size());

        if (log.debugEnabled) {
            log.debug("Decrypted patient token: ${tokenUtils.decryptTokenToString(session.appUser.securityKey, patientToken)}");
        }

        def patientParams = tokenUtils.decryptTokenToMap(session.appUser.securityKey, patientToken)

        log.debug("Patient params from token: ${patientParams}")

        def paramPatId = patientParams[HttpKeys.PATIENT_ID]
        def paramIdType = patientParams[HttpKeys.ID_TYPE]
        def paramPatSiteCode = patientParams[HttpKeys.PATIENT_SITE_CODE]

        // Selected patient's info to be stored in the user's recently viewed patients table
        def patient = new Patient();
        patient.setName(patName);
        patient.setEDIPI(patEdipi);
        patient.setSSN(patSsn);
        patient.setFMP(patSponsorSsn);
        patient.setGender(patGender);
        patient.setDob(patDob);
        patient.setSensitive(Boolean.parseBoolean(patSensitive));
        patient.setICN(patIcn);

        responseGUIPatientSelect = patientDao.selectPatient(provider, patient, paramPatId, paramIdType, paramPatSiteCode)

        session.externalIDs = responseGUIPatientSelect.getExternalIDs();

        selectedPatient = responseGUIPatientSelect.getGUIPatient();

        if (selectedPatient == null) {
            log.error("selected patient is null")
            response.sendError 500
            return false
        }

        // Store patient SSN and name in selectedPatient as they are required along with the patient's EDIPI or
        // ICN for getting added to the user's recently viewed patient's list
        if (StringUtils.isNotBlank(patient.getSSN())) {
            selectedPatient.setSsn(patient.getSSN());
        }

        if (StringUtils.isNotBlank(patient.getName())) {
            selectedPatient.setName(patient.getName());
        }

        session.restrictedPat = null;
        if (responseGUIPatientSelect.isVARestricted() || selectedPatient.isSensitive()
                || (StringUtils.equalsIgnoreCase(provider.getAgency(), "VA")
                && StringUtils.equalsIgnoreCase(provider.getHostSite().getMoniker(), "VBA"))) {
            // Patient is not registered in VA, is VA sensitive, or user is VBA. Need to wait for audit confirmation
            // before adding to selectedPatientMap (waiting for audit confirmation was to fix the DHS burp exploit)
            session.restrictedPat = selectedPatient;
        } else {
            // Patient is not restricted in any way, continue to add to map.
            selectedPatientMap.put(patientToken, selectedPatient);
        }

        log.debug("Patient select success, mapping patient to patientToken: ${patientToken}")

        String ipAddress = request.getHeader("X-FORWARDED-FOR");
        if (ipAddress == null) {
            ipAddress = request.getRemoteAddr();
        }

        // Retrieve context management item values to be sent to the vault
        String cmName = "";
        String cmEdipi = "";
        String cmIcn = "";
        String cmVaIen = "";
        String cmVAIenSiteId = "";
        String sensitivePatientAlertText = "";
        String sensitivePatientAlertLevel = 0;

        if (selectedPatient != null) {
            ArrayList<Alert> alerts = new ArrayList();
            if (session.appUser != null
                    & session.appUser.encryptedFederatedUid == null) {
                Alert noFederatedUid = new Alert();
                noFederatedUid.setAlertId("4");
                noFederatedUid.setMessageText("Your SSN is missing from the NEW PERSON file. Contact your ADP Coordinator");
                alerts = [noFederatedUid];
            } else {
                if (selectedPatient.vistaSites != null) {
                    try {
                        alerts = responseGUIPatientSelect.getSensitivePatientAlerts();
                    } catch (Exception e) {
			log.error("Failed query: getVASensitivePatientAlerts: " + e.getMessage());
                        sensitivePatientAlertLevel =-1;
                        sensitivePatientAlertText = "An error occurred while attempting to select this patient. VA Data will not be displayed.";
                        //prevent vista queries
                        selectedPatient.vistaSites = null;
                        selectedPatient.patientIens = null;

                    }
                }
            }
            //In case MVI is down (and vista hosts tf are unknown, and cannot retrieve any alerts)
            if (userSelectedOwnRecord) {
                boolean sensitiveLevel3 = false;
                for (Alert a : alerts) {
                    if (a.alertId == 3) {
                        sensitiveLevel3 = true;
                    }
                }
                if (!sensitiveLevel3) { //same SSN, selecting own record, yet no level 3 alert, most likely MVI is down.
                    Alert selectOwnAlert = new Alert();
                    selectOwnAlert.setAlertId("3");
                    selectOwnAlert.setMessageText("Security regulations prohibit access to your own medical record.");
                    alerts.add(selectOwnAlert);
                }
            }

            Alert highestAlert = null;
            for (Alert a : alerts) {
                if (highestAlert == null || a.alertId > highestAlert.alertId)
                    highestAlert = a;
            }
            if (highestAlert != null && Integer.parseInt(highestAlert.alertId) > 0) {
                sensitivePatientAlertText = highestAlert.getMessageText();
                sensitivePatientAlertLevel = highestAlert.alertId;
                selectedPatient.setSensitive(true);
            }

            cmName = selectedPatient.getName();
            cmEdipi = selectedPatient.getEDIPI();
            cmIcn = selectedPatient.getICN();

            if (selectedPatient.getPatientIens() != null && provider != null &&
                    provider.getHostSite() != null) {
                def providerSiteId = provider.getHostSite().getSiteCode();

                for (GUIKeyValuePair kvp in selectedPatient.getPatientIens()) {
                    if (kvp.key == providerSiteId) {
                        cmVaIen = kvp.value;
                        cmVAIenSiteId = kvp.key;
                    }
                }
            }
        }

        sendResponse([
                "patSelected"                  : true,
                "resultScript"                 : resultScript,
                "isSuccess"                    : responseGUIPatientSelect.isSuccess(),
                "isVARestricted"               : responseGUIPatientSelect.isVARestricted(),
                "isAllowVAAccess"              : responseGUIPatientSelect.isAllowVAAccess(),
                "errorMsg"                     : responseGUIPatientSelect.getErrorMsg(),
                "hieSharingOptIn"              : responseGUIPatientSelect.getGUIPatient().isHieSharingOptIn(),
                "hieSharingOptInLastUpdateDate": responseGUIPatientSelect.getGUIPatient().getHieSharingOptInLastUpdateDate(),
                "isSensitive"                  : selectedPatient.isSensitive(),
                "sensitivePatientAlertText"    : sensitivePatientAlertText,
                "sensitivePatientAlertLevel"   : sensitivePatientAlertLevel,
                "providerName"                 : provider.getName(),
                "providerSite"                 : provider.getHostSite().getSiteCode(),
                "providerSiteMoniker"          : provider.getHostSite().getMoniker(),
                "ipAddress"                    : ipAddress,
                "patientName"                  : cmName,
                "patientEdipi"                 : cmEdipi,
                "patientIcn"                   : cmIcn,
                "patientVaIen"                 : cmVaIen,
                "patientVaIenSiteId"           : cmVAIenSiteId
        ])
    }
}
