/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package gov.va.med.dao;

import gov.va.med.guibeans.*;
import gov.va.med.jmeadows.webservice.*;
import gov.va.med.vhahon.common.Utils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

/**
 *
 *
 */
@Component
public class DocumentDao extends DaoBase {

    private static final Logger logger = Logger.getLogger(DocumentDao.class.getName());


    public List<GUIDocument> getPatientDocuments(GUIUser user,
                                                         GUIPatient patient, Calendar startDate, Calendar endDate) {
        try {
            PatientDocument[] notes = getServiceHandler().getPatientDocuments(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build());
            List<GUIDocument> guiNotes = new ArrayList<GUIDocument>();

            for (int i = 0; i < notes.length; i++) {
                PatientDocument note = notes[i];
                GUIDocument guiNote = mapToGUIDocument(note, i);

                //cache notes from the DoD adapter
                if (appConfig.getProtocolDODAdapter().equals(guiNote.getSourceProtocol())) {
                    guiNote.setSessionCachedRecord(true);
                    guiNote.setNoteId(appConfig.getCachedTag() + i);

                    if (guiNote.getComplexDataUrl() != null && StringUtils.isNotBlank(guiNote.getComplexDataUrl())) {
                        guiNote.setNoteAsBinaryStream(true);

                        guiNote.setNoteAsPlainText(StringUtils.isNotBlank(guiNote.getNoteText()) ||
                                StringUtils.isNotBlank(guiNote.getImages()));
                    }
                } else {
                    guiNote.setNoteId(note.getNoteId());
                }

                guiNotes.add(guiNote);
            }

            return guiNotes;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    private GUIDocument mapToGUIDocument(PatientDocument doc, int docId) {

        GUIDocument guiDoc = dozerBeanMapper.map(doc, GUIDocument.class);
        if(StringUtils.isNotEmpty(doc.getContentType()) && doc.getContentType().toLowerCase().indexOf("xml") < 0) {
            guiDoc.setNoteText(StringEscapeUtils.escapeHtml4(doc.getNoteText()));
        }
        guiDoc.setNoteDate(Utils.formatDisplayDateTime(doc.getNoteDate()));
        guiDoc.setVisitAdm(doc.getVisitDate());

        //check for DoD adapter sensitive note flag
        if (guiDoc.isSensitive() || (guiDoc.getNoteTitle() != null &&
                appConfig.getProtocolDODAdapter().equalsIgnoreCase(doc.getSourceProtocol()) &&
                guiDoc.getNoteTitle().startsWith("**"))) {
            guiDoc.setSensitive(true);
            guiDoc.setNoteTitle("** Sensitive **");
//            guiNote.setNoteText("");
//            guiNote.setImages("");
//            guiNote.setNoteType("");
//            guiNote.setComplexDataUrl("");
//            guiNote.setNoteId("");
//            guiNote.setCodes(null);
        }

        String providerName = "";

        if (doc.getProvider() != null) {
            providerName = doc.getProvider().getName();

        }

        guiDoc.setProviderName(providerName);
        if (doc.getProvider() != null) {
            guiDoc.setProvider(dozerBeanMapper.map(doc.getProvider(), GUIProvider.class));
        }

        guiDoc.setNoteAsPlainText(true);

        guiDoc.setEventId(doc.getCdrEventId());


        //specific Inpatient Notes (Discharge Summaries) mapping
        if ("Inpatient Note".equalsIgnoreCase(doc.getDocumentType())) {
            guiDoc.setLocation("");

            if ("UNKNOWN".equalsIgnoreCase(guiDoc.getNoteType())) {
                String noteType = guiDoc.getLocation();
                guiDoc.setLocation("");
                guiDoc.setNoteType(noteType);
            }

            //cache notes from the DoD adapter
            if (appConfig.getProtocolDODAdapter().equals(guiDoc.getSourceProtocol())) {
                {
                    if ("plaintext".equalsIgnoreCase(guiDoc.getContentType())) {
                        guiDoc.setNoteAsBinaryStream(false);
                        guiDoc.setNoteAsPlainText(true);
                    } else {
                        guiDoc.setNoteAsBinaryStream(true);
                    }
                }
            }

            guiDoc.setNoteId(appConfig.getCachedTag() + docId);
            guiDoc.setSessionCachedRecord(true);

        }

        //specific Outpatient Encounter mapping
        if("Outpatient Encounter".equalsIgnoreCase(doc.getDocumentType()) && doc.getOutpatientEncounter() != null){
            Encounter encounter = doc.getOutpatientEncounter();
            encounter.getCodes().addAll(doc.getCodes());
            AppointmentDao apptDao = new AppointmentDao();
            GUIPatientEncounter guiEncounter = dozerBeanMapper.map(encounter, GUIPatientEncounter.class);
            guiEncounter = apptDao.mapGUIEncounter(encounter, guiEncounter, docId);
            guiDoc.setOutpatientEncounter(guiEncounter);
        }

        //specific Consults mapping
        if("Consult Report".equalsIgnoreCase(doc.getDocumentType()) && doc.getConsult() != null){
            Consult consult = doc.getConsult();
            consult.getCodes().addAll(doc.getCodes());
            NoteDao noteDao = new NoteDao();
            GUIConsult guiConsult = dozerBeanMapper.map(consult, GUIConsult.class);
            guiConsult = noteDao.mapGUIConsult(consult, guiConsult, docId);
            guiDoc.setConsult(guiConsult);
        }

        //specific Radiology mapping
        if("Radiology Exams".equalsIgnoreCase(doc.getDocumentType()) && doc.getRadiologyReport() != null){
            RadiologyReport rad = doc.getRadiologyReport();
            RadiologyDao radDao = new RadiologyDao();
            GUIRadiologyReport guiRad = radDao.toGUIRadiologyReport(rad);
            guiDoc.setRadiologyReport(guiRad);
            if (guiRad.isHasImage()) {
                guiDoc.setHasImage(true);
            }
        }

        //specific Form mappings
        if("Deployment Form".equalsIgnoreCase(doc.getDocumentType()) && doc.getForm() != null){
            Form form = doc.getForm();
            GUIPatientForm guiForm = dozerBeanMapper.map(form, GUIPatientForm.class);

            if (form.getFormDate() != null)
                guiForm.setFormDate(Utils.formatDisplayDate(form.getFormDate()));
            if (appConfig.getProtocolDODAdapter().equals(form.getSourceProtocol())) {
                guiForm.setSessionCachedRecord(true);
                guiDoc.setSessionCachedRecord(true);
                guiForm.setFormId(doc.getCdrEventId());
            }

            if (guiForm.getDeploymentForm() != null) {
                guiForm.setNoteAsBinaryStream(true);
                guiDoc.setNoteAsBinaryStream(true);
                if (org.apache.commons.lang3.StringUtils.isNotEmpty(guiForm.getDeploymentForm().getNoteText())) {
                    guiForm.getDeploymentForm().setNoteFormat("CDA");
                } else if(org.apache.commons.lang3.StringUtils.isNotEmpty(guiForm.getDeploymentForm().getComplexDataUrl())){
                    guiForm.getDeploymentForm().setNoteFormat("PDF");
                }
            }
            guiDoc.setForm(guiForm);
        }

        //if DOD complex note
        if (appConfig.getProtocolDODAdapter().equalsIgnoreCase(doc.getSourceProtocol()) &&
                StringUtils.isNotBlank(guiDoc.getComplexDataUrl())) {
            guiDoc.setNoteAsBinaryStream(true);

            //DOD complex notes should not display plain text note
            guiDoc.setNoteAsPlainText(false);
        }

        // Map VIX study to GUI bean
        GUIVixStudy vixStudy = new GUIVixStudy();
        if (doc.getVixStudy() != null) {
            StudyDetail detail = doc.getVixStudy().getStudyDetail();
            if (detail != null && detail.getStudies() != null && detail.getStudies().size() > 0) {
                // Most likely only 1 study in the study detail
                StudyDetailStudy detailStudy = detail.getStudies().get(0);
                vixStudy = dozerBeanMapper.map(detailStudy, GUIVixStudy.class);
            }
            vixStudy.setViewerUrl(doc.getVixStudy().getViewerUrl());
        }

        if (StringUtils.isNotEmpty(doc.getImages())) {
            try {
                if (Integer.parseInt(doc.getImages()) > 0) {
                    guiDoc.setHasImage(true);
                    guiDoc.setImages(doc.getImages());
                }
            } catch (NumberFormatException nfe) {
                logger.debug("Failed to convert document images count - " + doc.getImages());
            }

        }

        guiDoc.setVixStudy(vixStudy);

        return guiDoc;
    }

    /**
     * Fetches a BHIE note image.
     *
     * @param user    GUI user
     * @param complexTransaction Note transaction
     * @return Note Image
     */
    public NoteImage getBHIENoteImage(GUIUser user, String complexTransaction) {
        try {
            return getServiceHandler().getBHIENoteImage(

                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .complexTransaction(complexTransaction)
                            .build());
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public String getDocumentVixStudyViewerUrl(GUIUser user, GUIPatient patient, Calendar startDate, Calendar endDate, PatientDocument patientDocument) {
        try {
            return getServiceHandler().getPatientDocumentVixStudyViewerUrl(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build()
                    , patientDocument);
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }
}
