/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */

package gov.va.med.jmeadows.common;

import gov.va.med.foundations.utilities.EncryptUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

public class AppConfig {

    private static final Logger logger = LoggerFactory.getLogger(AppConfig.class);
    /**
     * jMeadows version number *
     */
    private static final String APP_VERSION = "app.version";
    /**
     * jMeadows app environment *
     */
    private static final String APP_ENVIRONMENT = "app.environment";
//    /**
//     * log4j properties file path *
//     */
//    private static final ServiceType LOG4J_PROPERTIES = "LOG4J_PROPERTIES";
//
    /**
     * JMeadowsDataService endpoint  *
     */
    private static final String JMEADOWS_URL = "JMEADOWS_URL";
    /**
     * VistaDataService endpoint  *
     */
    private static final String VISTA_URL = "VISTA_URL";
    /**
     * MVI endpoint *
     */
    private static final String MVI_URL = "MVI_URL";
    /**
     * BHIERelayService endpoint *
     */
    private static final String BHIE_RELAY_SERVICE_URL = "BHIE_RELAY_SERVICE_URL";
    /**
     * PDWSService endpoint *
     */
    private static final String PDWS_URL = "PDWS_URL";
    /**
     * PDWSService HL7 Message Header Processing Code *
     */
    private static final String PDWS_PROC_CODE = "PDWS_PROC_CODE";
    /**
     * PDWSService HL7 Message //Sender/Device/id *
     */
    private static final String PDWS_SND_DEV_ID = "PDWS_SND_DEV_ID";
    private static final String PDWS_PATIENT_ID_TYPE = "PDWS_PATIENT_ID_TYPE";
    private static final String PDWS_PATIENT_ID = "PDWS_PATIENT_ID";
    private static final String PDWS_PATIENT_LAST_NAME = "PDWS_PATIENT_LAST_NAME";
    /**
     * BHIERelayService encrypted username  *
     */
    private static final String BHIE_RELAY_SERVICE_USERNAME = "BHIE_RELAY_SERVICE_USERNAME";
    /**
     * BHIERelayService encrypted pwd  *
     */
    private static final String BHIE_RELAY_SERVICE_PWD = "BHIE_RELAY_SERVICE_PWD";
    /**
     * true to utilize C3PO for as the sql connection pooling api.
     * If false, defaults to utilizing JNDI lookup for connection pool.
     */
    private static final String USE_C3PO = "USE_C3PO";
    /**
     * SQL server JDBC connection string (for use with C3PO) *
     */
    private static final String SQL_URL = "SQL_URL";
    /**
     * SQL server driver (for use with C3PO) *
     */
    private static final String SQL_DRIVER = "SQL_DRIVER";
    /**
     * JDNI connection pooling lookup name *
     */
    private static final String JNDI_JDBC = "JNDI_JDBC";
    /**
     * General service request timeout setting in milliseconds *
     */
    private static final String REQUEST_TIMEOUT_MS = "REQUEST_TIMEOUT_MS";
    /**
     * General service connection timeout setting in milliseconds *
     */
    private static final String CONNECTION_TIMEOUT_MS = "CONNECTION_TIMEOUT_MS";
    /**
     * General thread timeout in milliseconds *
     */
    private static final String THREAD_TIMEOUT_MS = "THREAD_TIMEOUT_MS";
    /**
     * Default error message string *
     */
    private static final String ERROR_MSG_CONNECTION_UNAVAILABLE = "ERROR_MSG_CONNECTION_UNAVAILABLE";
    /**
     * AHLTA custodian OID.
     */
    private static final String BHIE_CUSTODIAN_OID_AHLTA = "BHIE_CUSTODIAN_OID_AHLTA";
    /**
     * VA custodian OID.
     */
    private static final String BHIE_CUSTODIAN_OID_VA = "BHIE_CUSTODIAN_OID_VA";
    /**
     * VA agency name string Should be configured as VA
     */
    private static final String AGENCY_VA = "AGENCY_VA";
    /**
     * DOD agency name string. Should be configured as DOD
     */
    private static final String AGENCY_DOD = "AGENCY_DOD";
    /**
     * SiteMap's EHCache name
     */
    private static final String EHCACHE_NAME_SITEMAP = "EHCACHE_NAME_SITEMAP";
    /**
     * RPC endpoint environment (enterprise, hrg, etc.)
     */
    private static final String RPC_ENDPOINT_ENVIRONMENT = "RPC_ENDPOINT_ENVIRONMENT";
    /**
     * CACHE endpoint environment (enterprise, hrg, etc.)
     */
    private static final String CACHE_ENDPOINT_ENVIRONMENT = "CACHE_ENDPOINT_ENVIRONMENT";
    /**
     * Share endpoint environment
     */
    private static final String SHARE_ENDPOINT_ENVIRONMENT = "SHARE_ENDPOINT_ENVIRONMENT";
    /**
     * Medweb endpoint environment
     */
    private static final String MEDWEB_ENDPOINT_ENVIRONMENT = "MEDWEB_ENDPOINT_ENVIRONMENT";
    private static final String AHLTA_ENDPOINT_ENVIRONMENT = "AHLTA_ENDPOINT_ENVIRONMENT";
    private static final AppConfig INSTANCE = new AppConfig();
    private Properties appConfigProp;

    private AppConfig() {
        appConfigProp = new Properties();

        ClassLoader classLoader = AppConfig.class.getClassLoader();

        if (classLoader == null) {
            throw new NullPointerException("ClassLoader is null!");
        }

        InputStream inputStream = null;

        try {
            inputStream = classLoader.getResourceAsStream("application.properties");
            appConfigProp.load(inputStream);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
        finally {

            try {
                if (inputStream != null) {
                    inputStream.close();
                }
            }
            catch (IOException e) {
                logger.error("IOException in AppConfig");
            }
        }

        try {
            inputStream = classLoader.getResourceAsStream(String.format("appconfig-%s.properties", appConfigProp.get(APP_ENVIRONMENT)));
            appConfigProp.load(inputStream);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
        finally {

            try {
                if (inputStream != null) {
                    inputStream.close();
                }
            }
            catch (IOException e) {
                logger.error("IOException in AppConfig");
            }
        }
    }

    public static final AppConfig getInstance() {
        return INSTANCE;
    }

    /**
     * jMeadows version number *
     */
    public final String getAppVersion() {
        return appConfigProp.getProperty(APP_VERSION);
    }

    /**
     * jMeadows app environment *
     */
    public final String getAppEnvironment() {
        return appConfigProp.getProperty(APP_ENVIRONMENT);
    }
//
//    /**
//     * log4j properties file path *
//     */
//    public final ServiceType getLog4jProperties()
//    {
//        return appConfigProp.getProperty(LOG4J_PROPERTIES);
//    }

    /**
     * JMeadowsDataService endpoint  *
     */
    public final String getJMeadowsURL() {
        return appConfigProp.getProperty(JMEADOWS_URL);
    }

    /**
     * VistaDataService endpoint  *
     */
    public final String getVistaWsURL() {
        return appConfigProp.getProperty(VISTA_URL);
    }

    /**
     * MVI endpoint
     */
    public final String getMVIURL() {
        return appConfigProp.getProperty(MVI_URL);
    }

    /**
     * BHIERelayService endpoint *
     */
    public final String getBHIERelayServiceURL() {
        return appConfigProp.getProperty(BHIE_RELAY_SERVICE_URL);
    }

    /**
     * PDWSService endpoint *
     */
    public final String getPDWSURL() {
        return appConfigProp.getProperty(PDWS_URL);
    }

    /**
     * PDWSService endpoint *
     */
    public final String getPDWSProcCode() {
        return appConfigProp.getProperty(PDWS_PROC_CODE);
    }

    /**
     * PDWSService endpoint *
     */
    public final String getPDWSSndDevId() {
        return appConfigProp.getProperty(PDWS_SND_DEV_ID);
    }

    public final String getPDWSPatientIdType() {
        return appConfigProp.getProperty(PDWS_PATIENT_ID_TYPE);
    }

    public final String getPDWSPatientId() {
        return appConfigProp.getProperty(PDWS_PATIENT_ID);
    }

    public final String getPDWSPatientLastName() {
        return appConfigProp.getProperty(PDWS_PATIENT_LAST_NAME);
    }

    /**
     * BHIERelayService username  *
     */
    public final String getBHIERelayServiceUsername() {
        return EncryptUtil.decrypt(appConfigProp.getProperty(BHIE_RELAY_SERVICE_USERNAME));
    }

    /**
     * BHIERelayService pwd  *
     */
    public final String getBHIERelayServicePassword() {
        return EncryptUtil.decrypt(appConfigProp.getProperty(BHIE_RELAY_SERVICE_PWD));
    }

    /**
     * true to utilize C3PO for as the sql connection pooling api.
     * If false, defaults to utilizing JNDI lookup for connection pool.
     */
    public final boolean useC3PO() {
        return Boolean.valueOf(appConfigProp.getProperty(USE_C3PO));
    }

    /**
     * SQL server JDBC connection string (for use with C3PO) *
     */
    public final String getSQLURL() {
        return appConfigProp.getProperty(SQL_URL);
    }

    /**
     * SQL server driver (for use with C3PO) *
     */
    public final String getSQLDriver() {
        return appConfigProp.getProperty(SQL_DRIVER);
    }

    /**
     * JDNI connection pooling lookup name *
     */
    public final String getJDBC_JNDILookup() {
        return appConfigProp.getProperty(JNDI_JDBC);
    }

    /**
     * General service request timeout setting in milliseconds *
     */
    public final int getRequestTimeoutMS() {
        return Integer.parseInt(appConfigProp.getProperty(REQUEST_TIMEOUT_MS));
    }

    /**
     * General service connection timeout setting in milliseconds *
     */
    public final int getConnectionTimeoutMS() {
        return Integer.parseInt(appConfigProp.getProperty(CONNECTION_TIMEOUT_MS));
    }

    /**
     * General thread timeout in milliseconds *
     */
    public final long getThreadTimeoutMS() {
        return Long.parseLong(appConfigProp.getProperty(THREAD_TIMEOUT_MS));
    }

    /**
     * BHIE custodian OIDs. Each OID represents a particular data site that BHIE aggregates from.
     * These custodian OIDs are included as BHIE query included or excluded site parameters.
     */

    /**
     * Default error message string *
     */
    public final String getErrorMsgConnectionUnavailable() {
        return appConfigProp.getProperty(ERROR_MSG_CONNECTION_UNAVAILABLE);
    }

    /**
     * AHLTA custodian OID.
     */
    public final String getBHIECustodianOidAHLTA() {
        return appConfigProp.getProperty(BHIE_CUSTODIAN_OID_AHLTA);
    }

    /**
     * VA custodian OID.
     */
    public final String getBHIECustodianOidVA() {
        return appConfigProp.getProperty(BHIE_CUSTODIAN_OID_VA);
    }

    /**
     * VA agency name string Should be configured as VA
     */
    public final String getAgencyVA() {
        return appConfigProp.getProperty(AGENCY_VA);
    }

    /**
     * DOD agency name string. Should be configured as DOD
     */
    public final String getAgencyDOD() {
        return appConfigProp.getProperty(AGENCY_DOD);
    }

    public final String getEHCacheNameSiteMap() {
        return appConfigProp.getProperty(EHCACHE_NAME_SITEMAP);
    }

    public final String getRPCEndpointEnvironment() {
        return appConfigProp.getProperty(RPC_ENDPOINT_ENVIRONMENT);
    }

    public final String getCacheEndpointEnvironment() {
        return appConfigProp.getProperty(CACHE_ENDPOINT_ENVIRONMENT);
    }

    public final String getShareEndpointEnvironment() {
        return appConfigProp.getProperty(SHARE_ENDPOINT_ENVIRONMENT);
    }

    public final String getMedwebEndpointEnvironment() {
        return appConfigProp.getProperty(MEDWEB_ENDPOINT_ENVIRONMENT);
    }

    public final String getAhltaEndpointEnvironment() {
        return appConfigProp.getProperty(AHLTA_ENDPOINT_ENVIRONMENT);
    }

    public final long getServiceMonitorTaskTimerDelay() {
        return Long.parseLong(appConfigProp.getProperty("SERVICE_MONITOR_TASK_TIMER_DELAY"));
    }

    public final long getServiceMonitorTaskTimerPeriod() {
        return Long.parseLong(appConfigProp.getProperty("SERVICE_MONITOR_TASK_TIMER_PERIOD"));
    }

    public final long getVistaStatusServiceMonitorTaskTimerPeriod() {
        return Long.parseLong(appConfigProp.getProperty("VISTA_STATUS_SERVICE_MONITOR_TASK_TIMER_PERIOD"));
    }

    public final Boolean isServiceMonitorTestBRS() {
        return Boolean.parseBoolean(appConfigProp.getProperty("SERVICE_MONITOR_TEST_BRS"));
    }

    public final Boolean isServiceMonitorTestJMDS() {
        return Boolean.parseBoolean(appConfigProp.getProperty("SERVICE_MONITOR_TEST_JMDS"));
    }

    public final Boolean isServiceMonitorTestMVI() {
        return Boolean.parseBoolean(appConfigProp.getProperty("SERVICE_MONITOR_TEST_MVI"));
    }

    public final Boolean isServiceMonitorTestPDWS() {
        return Boolean.parseBoolean(appConfigProp.getProperty("SERVICE_MONITOR_TEST_PDWS"));
    }

    public final Boolean isServiceMonitorTestVDS() {
        return Boolean.parseBoolean(appConfigProp.getProperty("SERVICE_MONITOR_TEST_VDS"));
    }

    public final Boolean isServiceMonitorTestShare() {
        return Boolean.parseBoolean(appConfigProp.getProperty("SERVICE_MONITOR_TEST_SHARE"));
    }

    public final Boolean isServiceMonitorTestSnareworks() {
        return Boolean.parseBoolean(appConfigProp.getProperty("SERVICE_MONITOR_TEST_SNAREWORKS"));
    }

    public final Boolean isServiceMonitorTestVistaSites() {
        return Boolean.parseBoolean(appConfigProp.getProperty("SERVICE_MONITOR_TEST_VISTA_SITES"));
    }

    public final String getEmailUsername() {
        return EncryptUtil.decrypt(appConfigProp.getProperty("EMAIL_USERNAME"));
    }

    public final String getEmailPassword() {
        return EncryptUtil.decrypt(appConfigProp.getProperty("EMAIL_PASSWORD"));
    }

    public final String getEmailHost() {
        return appConfigProp.getProperty("EMAIL_HOST");
    }

    public final String getEmailSSLPort() {
        return appConfigProp.getProperty("EMAIL_SSL_PORT");
    }

    public final String getEmailFrom() {
        return appConfigProp.getProperty("EMAIL_FROM");
    }

    public final String getEmailTo() {
        return appConfigProp.getProperty("EMAIL_TO");
    }

    public final String getEmailSrcSys() {
        return appConfigProp.getProperty("EMAIL_SRC_SYS");
    }

    public final Boolean isEmailStatusChangeOnly() {
        return Boolean.parseBoolean(appConfigProp.getProperty("EMAIL_STATUS_CHANGE_ONLY"));
    }

    public final Boolean isEmailErrorAtEveryCheck() {
        return Boolean.parseBoolean(appConfigProp.getProperty("EMAIL_ERROR_AT_EVERY_CHECK"));
    }

    public final long getEmailSendTaskTimerPeriod() {
        return Long.parseLong(appConfigProp.getProperty("EMAIL_SEND_TASK_TIMER_PERIOD"));
    }

    public final String getUserName() {
        return EncryptUtil.decrypt(appConfigProp.getProperty("ENCRYPTED_USER_NAME"));
    }

    public final String getUserIEN() {
        return EncryptUtil.decrypt(appConfigProp.getProperty("ENCRYPTED_USER_IEN"));
    }

    public final String getEncryptedUserAVCode() {
        return appConfigProp.getProperty("ENCRYPTED_USER_AV_CODE");
    }

    public final String getEncryptedUserFederatedUID() {
        return appConfigProp.getProperty("ENCRYPTED_USER_FEDERATED_UID");
    }

    public final String getUserSiteCode() {
        return appConfigProp.getProperty("USER_SITE_CODE");
    }

    public final String getPatientEDIPIForMVI() {
        return appConfigProp.getProperty("PATIENT_EDIPI_MVI");
    }

    public final String getPatientEDIPIForBHIE() {
        return appConfigProp.getProperty("PATIENT_EDIPI_BHIE");
    }

    public final String getEnv()  {
        return appConfigProp.getProperty("ENV");
    }
}
