/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dataservicehandler;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.enterprise.beans.ResponsePatientSelect;
import gov.va.med.jmeadows.webservice.beans.JMeadowsQuery;
import gov.va.med.mvi.*;
import gov.va.med.vds.webservice.KeyValuePair;
import gov.va.med.vds.webservice.Patient;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;
import java.io.StringWriter;
import java.text.SimpleDateFormat;
import java.util.*;

public class MVIServiceHandler extends DataServiceHandler
{
    private static final Logger logger = LoggerFactory.getLogger(MVIServiceHandler.class);
    private VAIdM mviService;

    public MVIServiceHandler(String url)
    {
        super();

        serviceURL = url;
        serviceName = "MVI_Service";
        serviceNamespace = "http://webservice.mviws.med.va.gov";
        qServiceName = new QName(serviceNamespace, serviceName);

        mviService = new VAIdM();
    }

    private VAIdMPort getMVIPort()
    {
        VAIdMPort port = mviService.getVAIdMPort();

        BindingProvider bp = (BindingProvider) port;
        bp.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, serviceURL);
        bp.getRequestContext().put("com.sun.xml.internal.ws.connect.timeout", serviceConnectionTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.internal.ws.request.timeout", serviceRequestTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.ws.connect.timeout", serviceConnectionTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.ws.request.timeout", serviceRequestTimeoutMS);

        // SOAP message logging - Enable to show logging in output window
        if ("test".equalsIgnoreCase(APP_CONFIG.getAppEnvironment()) || "development".equalsIgnoreCase(APP_CONFIG.getAppEnvironment()))
        {
            bp.getBinding().getHandlerChain().add(new SOAPLoggingHandler());
            List<Handler> handlerList = bp.getBinding().getHandlerChain();
            handlerList.add(new SOAPLoggingHandler());
            bp.getBinding().setHandlerChain(handlerList);
        }

        return port;
    }

    public ResponsePatientSelect getPatientIds(JMeadowsQuery queryBean)
    {
        String errorMessage = "";

        ResponsePatientSelect responsePatientSelect = new ResponsePatientSelect();

        try
        {
            // Generate PRPAIN201309UV02 request message
            PRPAIN201309UV02 request = generateRequest(queryBean);

            PRPAIN201310UV02 response = getMVIPort().prpaIN201309UV02(request);

            // Parse response into patient IDs map
            List<String> patientIds = parseResponseToPatientIds(response);

            Patient patient = new Patient();

            Iterator<String> iterator = patientIds.iterator();
            while (iterator.hasNext())
            {
                // Corresponding IDs format: ID^TYPE^SOURCE^ISSUER^STATUS
                StringTokenizer tokenizer = new StringTokenizer(iterator.next(), "^");
                if (tokenizer.countTokens() == 5) // According to format, there should only be 5 tokens
                {
                    String id = tokenizer.nextToken();
                    String type = tokenizer.nextToken();
                    String source = tokenizer.nextToken();
                    String issuer = tokenizer.nextToken();

                    // Makes sure each relevant token is not null or blank
                    if (StringUtils.isNotBlank(id) && StringUtils.isNotBlank(type)
                            && StringUtils.isNotBlank(source) && StringUtils.isNotBlank(issuer))
                    {
                        if ("NI".equalsIgnoreCase(type))
                        {
                            // National patient identifier
                            if ("200DOD".equalsIgnoreCase(source) && "USDOD".equalsIgnoreCase(issuer))
                            {
                                // DoD EDIPI
                                patient.setEDIPI(id);
                            }
                            else if ("200M".equalsIgnoreCase(source) && "USVHA".equalsIgnoreCase(issuer))
                            {
                                // VA ICN
                                patient.setICN(id);
                            }
                        }
                        else if ("PI".equalsIgnoreCase(type))
                        {
                            // Local facility patient identifier
                            if ("USVHA".equalsIgnoreCase(issuer))
                            {
                                // Add patient's VA IEN and corresponding site code to key-value pair list
                                KeyValuePair kvp = new KeyValuePair();
                                kvp.setKey(source);
                                kvp.setValue(id);
                                patient.getPatientIens().add(kvp);
                            }
                        }
                    }
                }
                else
                {
                    logger.debug("Invalid corresponding ID string format");
                }
            }

            responsePatientSelect.setSuccess(true);
            responsePatientSelect.setPatient(patient);
        }
        catch (JMeadowsException e)
        {
            responsePatientSelect.setSuccess(false);
            responsePatientSelect.setErrorMsg(e.getMessage());
        }
        catch (Exception e)
        {
            responsePatientSelect.setSuccess(false);
            responsePatientSelect.setErrorMsg(e.getMessage());
        }

        return responsePatientSelect;
    }

    private PRPAIN201309UV02 generateRequest(JMeadowsQuery queryBean)
    {
        ObjectFactory factory;
        PRPAIN201309UV02 request;
        MCCIMT000100UV01Receiver receiver;
        MCCIMT000100UV01Device device;
        MCCIMT000100UV01Sender sender;
        PRPAIN201309UV02QUQIMT021001UV01ControlActProcess controlActProcess;
        PRPAMT201307UV02QueryByParameter queryByParameter;
        PRPAMT201307UV02ParameterList parameterList;
        PRPAMT201307UV02PatientIdentifier patientIdentifier;
        SimpleDateFormat simpleDateFormat;
        II ii;
        TS ts;
        CS cs;
        CD cd;
        Date date;
        String status;
        String oid;
        String id;
        String type;
        String source;
        String issuer;

        factory = new ObjectFactory();
        request = factory.createPRPAIN201309UV02();
        date = new Date();

        //id
        simpleDateFormat = new SimpleDateFormat("yyyyMMdd");
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.933");
        ii.setExtension("MCID_JANUS_DEV_T1_" + simpleDateFormat.format(date));
        request.setId(ii);

        //creationTime
        ts = factory.createTS();
        simpleDateFormat = new SimpleDateFormat("yyyyMMddhhmms");
        ts.setValue(simpleDateFormat.format(date));
        request.setCreationTime(ts);

        //interactionId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.1.6");
        ii.setExtension("PRPA_IN201309UV02");
        request.setInteractionId(ii);

        //processingCode
        cs = factory.createCS();
        if ("production".equalsIgnoreCase(APP_CONFIG.getAppEnvironment()))
        {
            cs.setCode("P");
        }
        else
        {
            cs.setCode("T");
        }
        request.setProcessingCode(cs);

        //processingModeCode
        cs = factory.createCS();
        cs.setCode("T");
        request.setProcessingModeCode(cs);

        //acceptAckCode
        cs = factory.createCS();
        cs.setCode("AL");
        request.setAcceptAckCode(cs);

        //receiver
        receiver = factory.createMCCIMT000100UV01Receiver();
        request.getReceiver().add(receiver);

        //receiver/typeCode
        receiver.setTypeCode(CommunicationFunctionType.RCV);

        //receiver/device
        device = factory.createMCCIMT000100UV01Device();
        receiver.setDevice(device);

        //receiver/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //receiver/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //receiver/device/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.4.349");
        ii.setExtension("200M");
        device.getId().add(ii);

        //sender
        sender = factory.createMCCIMT000100UV01Sender();
        request.setSender(sender);

        //sender/typeCode
        sender.setTypeCode(CommunicationFunctionType.SND);

        //sender/device
        device = factory.createMCCIMT000100UV01Device();
        sender.setDevice(device);

        //sender/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //sender/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //sender/device/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.4.349");
        ii.setExtension("200JNUS");
        device.getId().add(ii);

        //controlActProcess
        controlActProcess = factory.createPRPAIN201309UV02QUQIMT021001UV01ControlActProcess();
        request.setControlActProcess(controlActProcess);

        //controlActProcess/classCode
        controlActProcess.setClassCode(ActClassControlAct.CACT);

        //controlActProcess/moodCode
        controlActProcess.setMoodCode(XActMoodIntentEvent.EVN);

        //controlActProcess/code
        cd = factory.createCD();
        cd.setCode("PRPA_TE201309UV02");
        controlActProcess.setCode(cd);

        //controlActProcess/codeSystem
        cd.setCodeSystem("2.16.840.1.113883.1.6");

        //controlActProcess/queryByParameter
        queryByParameter = factory.createPRPAMT201307UV02QueryByParameter();
        controlActProcess.setQueryByParameter(factory.createPRPAIN201309UV02QUQIMT021001UV01ControlActProcessQueryByParameter(queryByParameter));

        //controlActProcess/queryByParameter/queryId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.933");
        ii.setExtension("33452");
        queryByParameter.setQueryId(ii);

        //controlActProcess/queryByParameter/statusCode
        cs = factory.createCS();
        cs.setCode("new");
        queryByParameter.setStatusCode(cs);

        //controlActProcess/queryByParameter/responsePriorityCode
        cs = factory.createCS();
        cs.setCode("I");
        queryByParameter.setResponsePriorityCode(cs);

        //controlActProcess/queryByParameter/parameterList
        parameterList = factory.createPRPAMT201307UV02ParameterList();
        queryByParameter.setParameterList(parameterList);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier
        patientIdentifier = factory.createPRPAMT201307UV02PatientIdentifier();
        parameterList.getPatientIdentifier().add(patientIdentifier);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier/value
        ii = factory.createII();
        status = queryBean.getStatus();
        if ("EDIPI".equalsIgnoreCase(status))
        {
            // Get identifiers by DoD EDIPI
            oid = "2.16.840.1.113883.3.364";
            id = queryBean.getItemId();
            type = "NI";
            source = "200DOD";
            issuer = "USDOD";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type + "^" + source + "^" + issuer);

            logger.debug("Getting corresponding patient identifiers via EDIPI");
        }
        else if ("ICN".equalsIgnoreCase(status))
        {
            // Get identifiers by VA ICN
            oid = "2.16.840.1.113883.4.349";
            id = queryBean.getItemId();
            type = "NI";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type);

            logger.debug("Getting corresponding patient identifiers via ICN");
        }
        else if ("VAIEN".equalsIgnoreCase(status))
        {
            // Get identifiers by VA IEN
            oid = "2.16.840.1.113883.4.349";
            id = queryBean.getItemId();
            type = "PI";
            source = queryBean.getRecordSiteCode();
            issuer = "USVHA";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type + "^" + source + "^" + issuer);

            logger.debug("Getting corresponding patient identifiers via VA IEN");
        }

        patientIdentifier.getValue().add(ii);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier/semanticsText
        patientIdentifier.setSemanticsText("Patient.id");

        return request;
    }

    private List<String> parseResponseToPatientIds(PRPAIN201310UV02 response) throws JMeadowsException
    {
        Iterator<MCCIMT000300UV01Acknowledgement> acknowledgementIterator;
        MCCIMT000300UV01Acknowledgement acknowledgement;
        Iterator<MCCIMT000300UV01AcknowledgementDetail> acknowledgementDetailIterator;
        MCCIMT000300UV01AcknowledgementDetail acknowledgementDetail;
        Iterator<PRPAIN201310UV02MFMIMT700711UV01Subject1> subjectIterator;
        PRPAIN201310UV02MFMIMT700711UV01Subject1 subject;
        PRPAMT201304UV02Patient patient;
        Iterator<II> iiIterator;
        II ii;

        List<String> patientIds = new ArrayList<String>();

        // Check for errors
        acknowledgementIterator = response.getAcknowledgement().iterator();
        while (acknowledgementIterator.hasNext())
        {
            acknowledgement = acknowledgementIterator.next();
            if (!"AA".equalsIgnoreCase(acknowledgement.getTypeCode().getCode()))
            {
                acknowledgementDetailIterator = acknowledgement.getAcknowledgementDetail().iterator();
                while (acknowledgementDetailIterator.hasNext())
                {
                    acknowledgementDetail = acknowledgementDetailIterator.next();
                    throw new JMeadowsException(acknowledgementDetail.getText());
                }
            }
        }

        // Parse out patient identifiers
        subjectIterator = response.getControlActProcess().getSubject().iterator();
        while (subjectIterator.hasNext())
        {
            subject = subjectIterator.next();
            patient = subject.getRegistrationEvent().getSubject1().getPatient();

            iiIterator = patient.getId().iterator();
            while (iiIterator.hasNext())
            {
                ii = iiIterator.next();
                patientIds.add(ii.getExtension());
            }
        }

        return patientIds;
    }
}
