/*
 * Janus 4.0 (c)
 * Copyright (c) 2012 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dataservicehandler;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.common.PDWSQueryBean;
import gov.va.med.jmeadows.enterprise.beans.ResponsePatientQuery;
import gov.va.med.pdws.*;
import gov.va.med.vds.webservice.Patient;
import gov.va.med.vds.webservice.User;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;
import java.io.Serializable;
import java.math.BigInteger;
import java.net.MalformedURLException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class PDWSServiceHandler extends DataServiceHandler
{
    private static final Logger logger = LoggerFactory.getLogger(PDWSServiceHandler.class);
    private static final String AGENCY_DOD = APP_CONFIG.getAgencyDOD();
    private static final String AGENCY_VA = APP_CONFIG.getAgencyVA();
    private EntityPatientInquiry pdwsService;

    public PDWSServiceHandler(String url)
    {
        serviceURL = url;
        serviceName = "PDWSService";
        serviceNamespace = "http://webservice.pdws.URL       /";
        qServiceName = new QName(serviceNamespace, serviceName);

        pdwsService = new EntityPatientInquiry();
    }

    private EntityPatientSearchPortType getEntityPatientSearchPort() throws MalformedURLException
    {
        EntityPatientSearchPortType port;

        port = pdwsService.getEntityPatientSearchPortSoap11();

        BindingProvider bp = (BindingProvider) port;
        bp.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, serviceURL);
        bp.getRequestContext().put("com.sun.xml.internal.ws.connect.timeout", serviceConnectionTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.internal.ws.request.timeout", serviceRequestTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.ws.connect.timeout", serviceConnectionTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.ws.request.timeout", serviceRequestTimeoutMS);

        // SOAP message logging - Enable to show logging in output window
        bp.getBinding().getHandlerChain().add(new SOAPLoggingHandler());
        List<Handler> handlerList = bp.getBinding().getHandlerChain();
        handlerList.add(new SOAPLoggingHandler());
        bp.getBinding().setHandlerChain(handlerList);

        return port;
    }

    /**
     * Get patient id. Returns empty string if id is not found.
     * Throws an exception if communications is down.
     *
     * @param queryBean Query parameters.
     * @return Map of lookup patients.
     */
    public ResponsePatientQuery lookupPatient(PDWSQueryBean queryBean) throws JMeadowsException, MalformedURLException
    {
        PRPAIN201305UV02 request = GenerateRequest(queryBean);
        PRPAIN201306UV02 response = null;

        try {
            response = getEntityPatientSearchPort().entityPatientSearch(request);

            if (response != null)
            {
                return ParseResponseToPatientList(response);
            }

            return null;
        }
        catch(Exception e) {
            throw new JMeadowsException(e.getMessage());
        }
    }

    private PRPAIN201305UV02 GenerateRequest(PDWSQueryBean queryBean)
    {
        ObjectFactory factory;
        PRPAIN201305UV02 request;
        MCCIMT000100UV01Receiver receiver;
        MCCIMT000100UV01Sender sender;
        MCCIMT000100UV01Device device;
        MCCIMT000100UV01Agent agent;
        MCCIMT000100UV01Organization organization;
        PRPAIN201305UV02QUQIMT021001UV01ControlActProcess controlActProcess;
        PRPAMT201306UV02QueryByParameter queryByParameter;
        PRPAMT201306UV02ParameterList parameterList;
        PRPAMT201306UV02LivingSubjectId livingSubjectId;
        PRPAMT201306UV02LivingSubjectAdministrativeGender adminGender;
        PRPAMT201306UV02LivingSubjectBirthTime birthTime;
        PRPAMT201306UV02LivingSubjectName name;
        PRPAMT201306UV02OtherIDsScopingOrganization otherIDsScopingOrganization;
        QUQIMT021001UV01DataEnterer dataEnterer;
        COCTMT090100UV01AssignedPerson assignedPerson;
        COCTMT090100UV01Person person1;
        SimpleDateFormat simpleDateFormat;
        EnFamily enFamily;
        EnGiven enGiven;
        IVLTS ivlts;
        INT i;
        II ii;
        CS cs;
        ST st;
        TS ts;
        CD cd;
        CE ce;
        EN en;

        User user = queryBean.getUser();
        String agency = user.getAgency();
        String dmisId = user.getHostSite().getDmisId();
        String siteId = user.getHostSite().getSiteCode();

        factory = new ObjectFactory();
        request = factory.createPRPAIN201305UV02();

        //id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.2017");
        ii.setExtension("DMDC");
        request.setId(ii);

        //creationTime
        ts = factory.createTS();
        simpleDateFormat = new SimpleDateFormat("yyyyMMddhhmmss");
        ts.setValue(simpleDateFormat.format(new Date()));
        request.setCreationTime(ts);

        //interactionId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.1.6");
        ii.setExtension("PRPA_IN201305UV02");
        request.setInteractionId(ii);

        //processingCode
        cs = factory.createCS();
        cs.setCode(APP_CONFIG.getPDWSProcCode());
        request.setProcessingCode(cs);

        //processingModeCode
        cs = factory.createCS();
        cs.setCode("I");
        request.setProcessingModeCode(cs);

        //acceptAckCode
        cs = factory.createCS();
        cs.setCode("AL");
        request.setAcceptAckCode(cs);

        //receiver
        receiver = factory.createMCCIMT000100UV01Receiver();
        request.getReceiver().add(receiver);

        //receiver/typeCode
        receiver.setTypeCode(CommunicationFunctionType.RCV);

        //receiver/device
        device = factory.createMCCIMT000100UV01Device();
        receiver.setDevice(device);

        //receiver/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //receiver/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //receiver/device/id
        ii = factory.createII();
        ii.setRoot("1.3.6.1.4.1.21367.2009.1.3.1.145");
        device.getId().add(ii);

        //sender
        sender = factory.createMCCIMT000100UV01Sender();
        request.setSender(sender);

        //sender/typeCode
        sender.setTypeCode(CommunicationFunctionType.SND);

        //sender/device
        device = factory.createMCCIMT000100UV01Device();
        sender.setDevice(device);

        //sender/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //sender/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //sender/device/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.2017.11.7");
        ii.setExtension(APP_CONFIG.getPDWSSndDevId());
        device.getId().add(ii);

        //sender/device/asAgent
        agent = factory.createMCCIMT000100UV01Agent();
        device.setAsAgent(factory.createMCCIMT000100UV01DeviceAsAgent(agent));

        //sender/device/asAgent/classCode
        agent.getClassCode().add("AGNT");

        //sender/device/asAgent/representedOrganization
        organization = factory.createMCCIMT000100UV01Organization();
        agent.setRepresentedOrganization(factory.createMCCIMT000100UV01AgentRepresentedOrganization(organization));

        //sender/device/asAgent/representedOrganization/classCode
        organization.setClassCode("ORG");

        //sender/device/asAgent/representedOrganization/determinerCode
        organization.setDeterminerCode("INSTANCE");

        //sender/device/asAgent/representedOrganization/id
        ii = factory.createII();
//        ii.setRoot("2.16.840.1.113883.3.42");
//        ii.setExtension(queryBean.getUser().getHostSite().getSiteCode());
        if (AGENCY_DOD.equalsIgnoreCase(agency))
        {
            ii.setRoot("2.16.840.1.113883.3.42.10002.100001.8");
            ii.setExtension(dmisId);
        }
        else if (AGENCY_VA.equalsIgnoreCase(agency))
        {
            ii.setRoot("2.16.840.1.113883.4.349");
            ii.setExtension(siteId);
        }
        organization.getId().add(ii);

        //controlActProcess
        controlActProcess = factory.createPRPAIN201305UV02QUQIMT021001UV01ControlActProcess();
        request.setControlActProcess(controlActProcess);

        //controlActProcess/classCode
        controlActProcess.setClassCode(ActClassControlAct.CACT);

        //controlActProcess/moodCode
        controlActProcess.setMoodCode(XActMoodIntentEvent.EVN);

        //controlActProcess/code
        cd = factory.createCD();
        cd.setCode("PRPA_TE201305UV02");
        controlActProcess.setCode(cd);

        //controlActProcess/codeSystem
        cd.setCodeSystem("2.16.840.1.113883.1.6");

        //controlActProcess/dataEnterer
        dataEnterer = factory.createQUQIMT021001UV01DataEnterer();
        controlActProcess.getDataEnterer().add(dataEnterer);

        //controlActProcess/dataEnterer/typeCode
        dataEnterer.getTypeCode().add("ENT");

        //controlActProcess/dataEnterer/contextControlCode
        dataEnterer.setContextControlCode("AP");

        //controlActProcess/dataEnterer/assignedPerson
        assignedPerson = factory.createCOCTMT090100UV01AssignedPerson();
        dataEnterer.setAssignedPerson(assignedPerson);

        //controlActProcess/dataEnterer/assignedPerson/classCode
        assignedPerson.setClassCode("ASSIGNED");

        //controlActProcess/dataEnterer/assignedPerson/id
        ii = factory.createII();
        if (AGENCY_DOD.equalsIgnoreCase(agency))
        {
            ii.setRoot("2.16.840.1.113883.3.42.127.100001.79");
            ii.setExtension(user.getUserIen() + "^" + dmisId);
        }
        else if (AGENCY_VA.equalsIgnoreCase(agency))
        {
            ii.setRoot("2.16.840.1.113883.4.349");
            ii.setExtension(user.getUserIen() + "^" + siteId);
        }
        assignedPerson.getId().add(ii);

        //controlActProcess/dataEnterer/assignedPerson/Person
        person1 = factory.createCOCTMT090100UV01Person();
        assignedPerson.setAssignedPerson(factory.createCOCTMT090100UV01AssignedPersonAssignedPerson(person1));

        //controlActProcess/dataEnterer/assignedPerson/Person/name
        en = factory.createEN();
        en.getContent().add(queryBean.getUser().getName());
        person1.getName().add(en);

        //controlActProcess/queryByParameter
        queryByParameter = factory.createPRPAMT201306UV02QueryByParameter();
        controlActProcess.setQueryByParameter(factory.createPRPAIN201305UV02QUQIMT021001UV01ControlActProcessQueryByParameter(queryByParameter));

        //controlActProcess/queryByParameter/queryId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.2017");
        ii.setExtension("333334433");
        queryByParameter.setQueryId(ii);

        //controlActProcess/queryByParameter/statusCode
        cs = factory.createCS();
        cs.setCode("new");
        queryByParameter.setStatusCode(cs);

        //controlActProcess/queryByParameter/initialQuantity
        i = factory.createINT();
        i.setValue(BigInteger.valueOf(10));
        queryByParameter.setInitialQuantity(i);

        //controlActProcess/queryByParameter/parameterList
        parameterList = factory.createPRPAMT201306UV02ParameterList();
        queryByParameter.setParameterList(parameterList);

        //controlActProcess/queryByParameter/parameterList/livingSubjectId
        if (!"".equals(queryBean.getPatientId()) && !"".equals(queryBean.getPatientIdType()))
        {
            ii = factory.createII();
            ii.setExtension(queryBean.getPatientId());

            String type = queryBean.getPatientIdType();
            if ("SSN".equalsIgnoreCase(type))
            {
                ii.setRoot("2.16.840.1.113883.4.1");
            }
            else if ("TIN".equalsIgnoreCase(type))
            {
                ii.setRoot("2.16.840.1.113883.3.42.10001.100001.225");
            }
            else if ("FIN".equalsIgnoreCase(type))
            {
                ii.setRoot("2.16.840.1.113883.3.42.10001.100001.224");
            }
            else if ("ITIN".equalsIgnoreCase(type))
            {
                ii.setRoot("2.16.840.1.113883.4.2");
            }
            else if ("PIN".equalsIgnoreCase(type))
            {
                ii.setRoot("2.16.840.1.113883.3.42.10001.100001.226");
            }

            livingSubjectId = factory.createPRPAMT201306UV02LivingSubjectId();
            livingSubjectId.getValue().add(ii);
            parameterList.getLivingSubjectId().add(livingSubjectId);

            //controlActProcess/queryByParameter/parameterList/livingSubjectId/semanticsText
            st = factory.createST();
            st.getContent().add("LivingSubject.id");
            livingSubjectId.setSemanticsText(st);
        }

        //controlActProcess/queryByParameter/parameterList/livingSubjectAdministrativeGender
        if (queryBean.getGender() != null)
        {
            adminGender = factory.createPRPAMT201306UV02LivingSubjectAdministrativeGender();
            ce = factory.createCE();
            ce.setCode(queryBean.getGender());
            adminGender.getValue().add(ce);
            parameterList.getLivingSubjectAdministrativeGender().add(adminGender);

            //controlActProcess/queryByParameter/parameterList/livingSubjectAdministrativeGender/semanticsText
            st = factory.createST();
            st.getContent().add("LivingSubject.administrativeGender");
            adminGender.setSemanticsText(st);
        }

        //controlActProcess/queryByParameter/parameterList/livingSubjectBirthTime
        birthTime = factory.createPRPAMT201306UV02LivingSubjectBirthTime();
        ivlts = factory.createIVLTS();
        ivlts.setValue(queryBean.getDob());
        birthTime.getValue().add(ivlts);
        parameterList.getLivingSubjectBirthTime().add(birthTime);

        //controlActProcess/queryByParameter/parameterList/livingSubjectBirthTime/semanticsText
        st = factory.createST();
        st.getContent().add("LivingSubject.birthTime");
        birthTime.setSemanticsText(st);

        //controlActProcess/queryByParameter/parameterList/livingSubjectName
        name = factory.createPRPAMT201306UV02LivingSubjectName();
        en = factory.createEN();

        if (queryBean.getFirstName() != null)
        {
            enGiven = factory.createEnGiven();
            enGiven.getContent().add(queryBean.getFirstName());
            en.getContent().add(factory.createENGiven(enGiven));
        }

        enFamily = factory.createEnFamily();
        enFamily.getContent().add(queryBean.getLastName());
        en.getContent().add(factory.createENFamily(enFamily));

        name.getValue().add(en);
        parameterList.getLivingSubjectName().add(name);

        //controlActProcess/queryByParameter/parameterList/livingSubjectName/semanticsText
        st = factory.createST();
        st.getContent().add("LivingSubject.name");
        name.setSemanticsText(st);

        //controlActProcess/queryByParameter/parameterList/otherIDsScopingOrganization
        otherIDsScopingOrganization = factory.createPRPAMT201306UV02OtherIDsScopingOrganization();
        parameterList.getOtherIDsScopingOrganization().add(otherIDsScopingOrganization);

        //controlActProcess/queryByParameter/parameterList/otherIDsScopingOrganization/value/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.4.1");
        otherIDsScopingOrganization.getValue().add(ii);

        //controlActProcess/queryByParameter/parameterList/otherIDsScopingOrganization/semanticsText
        st = factory.createST();
        st.getContent().add("OtherIDs.scopingOrganization.id");
        otherIDsScopingOrganization.setSemanticsText(st);

        return request;
    }

    private ResponsePatientQuery ParseResponseToPatientList(PRPAIN201306UV02 response) throws JMeadowsException
    {
        ResponsePatientQuery respPatQuery = new ResponsePatientQuery();

        // Initially set to failure with generic error message. Will only be
        // set to success if there are no acknowledgement errors and no
        // errors retrieving patient list, even if the list is empty
        respPatQuery.setSuccess(false);
        respPatQuery.setErrorMsg("Error retrieving patients."
                + "<br><br>Please try again."
                + "<br><br>If the problem persists, notify your system administrator or help desk.");

        // Create list of result patients
        List<Patient> resultPatList = new ArrayList<Patient>();

        if (response != null)
        {
            // Check the response acknowledgement for an application error
            List<MCCIMT000300UV01Acknowledgement> ackList = response.getAcknowledgement();
            if (ackList != null)
            {
                for (MCCIMT000300UV01Acknowledgement ack : ackList)
                {
                    CS cs = ack.getTypeCode();
                    if (cs != null)
                    {
                        if(!"AA".equalsIgnoreCase(cs.getCode()))
                        {
                            // Application error occurred
                            List<MCCIMT000300UV01AcknowledgementDetail> ackDetailList = ack.getAcknowledgementDetail();
                            if (ackDetailList != null)
                            {
                                for (MCCIMT000300UV01AcknowledgementDetail ackDetail : ackDetailList)
                                {
                                    CE ce = ackDetail.getCode();
                                    if ((ce != null) && (ce.getDisplayName() != null))
                                    {
                                        respPatQuery.setErrorMsg(ackDetail.getTypeCode().toString() + ": " + ce.getDisplayName());
                                        return respPatQuery;
                                    }
                                }
                            }
                        }
                    }
                }
            }

            // Parse message for list of patients
            PRPAIN201306UV02MFMIMT700711UV01ControlActProcess ctrlActProc = response.getControlActProcess();
            if (ctrlActProc != null)
            {
                // Iterate through each patient
                List<PRPAIN201306UV02MFMIMT700711UV01Subject1> subjList = ctrlActProc.getSubject();
                if (subjList != null)
                {
                    StringBuilder familyName = new StringBuilder();
                    StringBuilder givenName = new StringBuilder();
                    StringBuilder prefixName = new StringBuilder();
                    StringBuilder suffixName = new StringBuilder();
                    for (PRPAIN201306UV02MFMIMT700711UV01Subject1 subj : subjList)
                    {
                        familyName.setLength(0);
                        givenName.setLength(0);
                        prefixName.setLength(0);
                        suffixName.setLength(0);
                        String edipi = null;
                        String ssn = "N/A";
                        String gender = "N/A";
                        String dob = "N/A";

                        PRPAIN201306UV02MFMIMT700711UV01RegistrationEvent regEvent = subj.getRegistrationEvent();
                        if (regEvent != null)
                        {
                            PRPAIN201306UV02MFMIMT700711UV01Subject2 subj1 = regEvent.getSubject1();
                            if (subj1 != null)
                            {
                                PRPAMT201310UV02Patient pat = subj1.getPatient();
                                if (pat != null)
                                {
                                    // Retrieve patient EDIPI
                                    List<II> iiList = pat.getId();
                                    if ((iiList != null) && !iiList.isEmpty())
                                    {
                                        edipi = iiList.get(0).getExtension();
                                    }

                                    // Retrieve patient's demographics information (name, gender, dob, ssn)
                                    JAXBElement<PRPAMT201310UV02Person> jaxbElementPerson = pat.getPatientPerson();
                                    if (jaxbElementPerson != null)
                                    {
                                        PRPAMT201310UV02Person person = jaxbElementPerson.getValue();
                                        if (person != null)
                                        {
                                            // Retrieve patient gender
                                            CE ce = person.getAdministrativeGenderCode();
                                            if (ce != null)
                                            {
                                                gender = ce.getCode();
                                            }

                                            // Retrieve patient date of birth
                                            TS ts = person.getBirthTime();
                                            if (ts != null)
                                            {
                                                dob = ts.getValue();
                                            }

                                            // Retrieve patient name
                                            List<PN> pnList = person.getName();
                                            if (pnList != null)
                                            {
                                                for (PN pn : pnList)
                                                {
                                                    List<Serializable> serialList = pn.getContent();
                                                    if (serialList != null)
                                                    {
                                                        for (Serializable serial : serialList)
                                                        {
                                                            if (serial instanceof JAXBElement)
                                                            {
                                                                Object obj = ((JAXBElement) serial).getValue();
                                                                if (obj != null)
                                                                {
                                                                    if (obj instanceof EnFamily)
                                                                    {
                                                                        List<Serializable> serial2List = ((EnFamily) obj).getContent();
                                                                        if ((serial2List != null) && !serial2List.isEmpty())
                                                                        {
                                                                            familyName.append(serial2List.get(0));
                                                                            familyName.append(" ");
                                                                        }
                                                                    }
                                                                    else if (obj instanceof EnGiven)
                                                                    {
                                                                        List<Serializable> serial2List = ((EnGiven) obj).getContent();
                                                                        if ((serial2List != null) && !serial2List.isEmpty())
                                                                        {
                                                                            givenName.append(serial2List.get(0));
                                                                            givenName.append(" ");
                                                                        }
                                                                    }
                                                                    else if (obj instanceof EnPrefix)
                                                                    {
                                                                        List<Serializable> serial2List = ((EnPrefix) obj).getContent();
                                                                        if ((serial2List != null) && !serial2List.isEmpty())
                                                                        {
                                                                            prefixName.append(serial2List.get(0));
                                                                            prefixName.append(" ");
                                                                        }
                                                                    }
                                                                    else if (obj instanceof EnSuffix)
                                                                    {
                                                                        List<Serializable> serial2List = ((EnSuffix) obj).getContent();
                                                                        if ((serial2List != null) && !serial2List.isEmpty())
                                                                        {
                                                                            suffixName.append(serial2List.get(0));
                                                                            suffixName.append(" ");
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else
                                                    {
                                                        logger.error("List<Serializable> is null");
                                                    }
                                                }

                                                // Format: PrefixName FamilyName,GivenName SuffixName
                                                // (comma will be added when generating name string)

                                                // Add prefix before family name
                                                if (prefixName.length() > 0)
                                                {
                                                    familyName = prefixName.append(familyName);
                                                }

                                                // Add suffix after given name
                                                if (suffixName.length() > 0)
                                                {
                                                    givenName.append(suffixName);
                                                }
                                            }
                                            else
                                            {
                                                logger.error("List<PN> is null");
                                            }

                                            // Retrieve patient SSN
                                            List<PRPAMT201310UV02OtherIDs> otherIdsList = person.getAsOtherIDs();
                                            if (otherIdsList != null)
                                            {
                                                for (PRPAMT201310UV02OtherIDs otherIDs : otherIdsList)
                                                {
                                                    List<II> ii2List = otherIDs.getId();
                                                    if (ii2List != null)
                                                    {
                                                        for (II ii : ii2List)
                                                        {
                                                            CS cs = otherIDs.getStatusCode();
                                                            if (cs != null)
                                                            {
                                                                if ("2.16.840.1.113883.4.1".equalsIgnoreCase(ii.getRoot()) && "active".equalsIgnoreCase(cs.getCode()))
                                                                {
                                                                    ssn = ii.getExtension();
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        Patient resultPat = new Patient();

                        // Set patient information
                        resultPat.setEDIPI(edipi);
                        resultPat.setSSN(ssn);
                        resultPat.setName(familyName.toString().trim() + "," + givenName.toString().trim());
                        resultPat.setGender(gender);
                        resultPat.setDob(dob);

                        // Add to patient list
                        resultPatList.add(resultPat);
                    }
                }
            }
        }

        respPatQuery.setSuccess(true);
        respPatQuery.setErrorMsg(null);
        respPatQuery.setPatients(resultPatList);

        return respPatQuery;
    }
}
