/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.site;

import gov.va.med.jmeadows.common.AppConfig;
import gov.va.med.vds.webservice.Site;

import java.beans.PropertyVetoException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Ehcache;
import net.sf.ehcache.constructs.blocking.SelfPopulatingCache;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.mchange.v2.c3p0.ComboPooledDataSource;

public class Sites {
    private static final Logger LOGGER = LoggerFactory.getLogger(Sites.class);
    private static final AppConfig APP_CONFIG = AppConfig.getInstance();

    private static final Sites INSTANCE = new Sites();
    private static final String CACHE_KEY_SITES_MAP = "sitesMap";

    private Ehcache cache;
    private DataSource dataSource;


    private Sites() {
        if (APP_CONFIG.useC3PO()) {
            //setup c3po db connection pools
            try {

                LOGGER.info("jMeadows JDBC Connection pooling: using C3P0");

                ComboPooledDataSource c3poSQLConnectionPool = new ComboPooledDataSource();
                c3poSQLConnectionPool.setDriverClass(APP_CONFIG.getSQLDriver());
                c3poSQLConnectionPool.setJdbcUrl(APP_CONFIG.getSQLURL());
                dataSource = c3poSQLConnectionPool;
            } catch (PropertyVetoException ex) {
                LOGGER.error(ex.getMessage(), ex);
                throw new RuntimeException(ex);
            }
        } else {
            try {
                dataSource = InitialContext.doLookup(APP_CONFIG.getJDBC_JNDILookup());
            } catch (NamingException ex) {
                LOGGER.error(ex.getMessage(), ex);
                throw new RuntimeException(ex);
            }
        }

        this.cache = CacheManager.getInstance().getCache(APP_CONFIG.getEHCacheNameSiteMap());

        SiteCacheEntryFactory siteCacheEntryFactory =
                new SiteCacheEntryFactory(dataSource,
                        CACHE_KEY_SITES_MAP,
                        APP_CONFIG.getRPCEndpointEnvironment(),
                        APP_CONFIG.getCacheEndpointEnvironment(),
                        APP_CONFIG.getShareEndpointEnvironment(),
                        APP_CONFIG.getMedwebEndpointEnvironment(),
                        APP_CONFIG.getAhltaEndpointEnvironment());

        SelfPopulatingCache selfPopulatingCache = new SelfPopulatingCache(this.cache, siteCacheEntryFactory);
        CacheManager.getInstance().replaceCacheWithDecoratedCache(this.cache, selfPopulatingCache);
        this.cache = selfPopulatingCache;

    }

    public static final Sites getInstance() {
        return INSTANCE;
    }

    public Map<String, Site> getSiteMap() {
        return (Map<String, Site>) cache.get(CACHE_KEY_SITES_MAP).getObjectValue();
    }

    public Site getSite(String siteCode) {
        if (StringUtils.isBlank(siteCode)) {
            throw new IllegalArgumentException("siteCode is blank");
        }

        return getSiteMap().get(siteCode);
    }

    public List<Site> getSiteList() {
        return new ArrayList<Site>(getSiteMap().values());
    }

    public Site getSiteByDmisID(String dmisId) {
        Site rtc = null;
        if (StringUtils.isBlank(dmisId)) {
            throw new IllegalArgumentException("DMIS Id is blank");
        }

        List<Site> siteList = getSiteList();

        for (Site site : siteList) {
            if (dmisId.equals(site.getDmisId())) {
                rtc = site;
                break;
            }
        }

        return rtc;
    }


}
