package gov.va.med.monitor;

import com.mchange.v2.c3p0.ComboPooledDataSource;
import gov.va.med.jmeadows.common.AppConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;
import java.beans.PropertyVetoException;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

public class DatabaseHandler extends ServiceBase
{
    private static final Logger logger = LoggerFactory.getLogger(DatabaseHandler.class);
    private static final DatabaseHandler INSTANCE = new DatabaseHandler();
    protected static final AppConfig APP_CONFIG = AppConfig.getInstance();

    private DataSource dataSource;

    private DatabaseHandler()
    {
        super(ServiceType.DatabaseService);

        AppConfig appConfig = AppConfig.getInstance();

        // Set up database connection
        try
        {
            if (appConfig.useC3PO())
            {
                //setup c3po db connection pools
                ComboPooledDataSource c3poSQLConnectionPool = new ComboPooledDataSource();
                c3poSQLConnectionPool.setDriverClass(appConfig.getSQLDriver());
                c3poSQLConnectionPool.setJdbcUrl(appConfig.getSQLURL());

                dataSource = c3poSQLConnectionPool;
            }
            else
            {
                dataSource = InitialContext.doLookup(appConfig.getJDBC_JNDILookup());
            }
        }
        catch (PropertyVetoException e)
        {
            dataSource = null;
        }
        catch (NamingException e)
        {
            dataSource = null;
        }
    }

    public static DatabaseHandler getInstance()
    {
        return INSTANCE;
    }

    public void logReports(List<ServiceReport> reportList)
    {
        Connection connection = null;
        CallableStatement statement = null;

        String env = APP_CONFIG.getEnv();

        try
        {
            connection = dataSource.getConnection();
            statement = connection.prepareCall("{call addQoSReport(?,?,?,?)}");

            for (ServiceReport report : reportList)
            {
                statement.setString(1, report.getType());
                statement.setString(2, report.getStatus());
                statement.setString(3, report.getMessage());
                statement.setString(4, env);
                statement.executeUpdate();
            }

            status = ServiceStatus.OK;
            message.append(SERVICE_OK_MSG);

            logger.debug("Logging to database completed");
        }
        catch (SQLException e)
        {
            status = ServiceStatus.ERROR;
            message.append(e.getMessage());

            logger.error("Logging to database failed: " + e.getMessage());
        }
        finally
        {
            if (statement != null)
            {
                try {
                    statement.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db statement.");
                }
            }

            if (connection != null)
            {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db connection");
                }
            }
        }

        report(type.toString(), status.toString(), message.toString());
    }

    public List<ShareSiteInfo> getShareEndpoints()
    {
        Connection connection = null;
        CallableStatement statement = null;
        ResultSet resultSet = null;
        List<ShareSiteInfo> shareSiteInfoList = new ArrayList<ShareSiteInfo>();
        ShareSiteInfo shareSiteInfo;

        try
        {
            connection = dataSource.getConnection();
            statement = connection.prepareCall("{call getQoSShareEndpoints}");
            resultSet = statement.executeQuery();

            while (resultSet.next())
            {
                shareSiteInfo = new ShareSiteInfo();

                shareSiteInfo.setHostURL(resultSet.getString("host"));
                shareSiteInfo.setEncryptedUsername(resultSet.getString("uname"));
                shareSiteInfo.setEncryptedPassword(resultSet.getString("pword"));
                shareSiteInfo.setSiteCode(resultSet.getString("sitecode"));
                shareSiteInfo.setName(resultSet.getString("name"));

                shareSiteInfoList.add(shareSiteInfo);
            }

            logger.debug("Retrieved SHARE sites from database");
        }
        catch (SQLException e)
        {
            shareSiteInfoList = null;

            logger.error("Failed to retrieve SHARE sites: " + e.getMessage());
        }
        finally
        {
            if (resultSet != null)
            {
                try {
                    resultSet.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db resultSet");
                }
            }

            if (statement != null)
            {
                try {
                    statement.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db statement.");
                }
            }

            if (connection != null)
            {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db connection");
                }
            }
        }

        return shareSiteInfoList;
    }

    @Override
    public void run()
    {
        // Nothing to implement. Monitoring happens when it's time to log results to
        // the database. Service report is generated in the logReports() method above.
    }
}
