/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */
 
 
package gov.va.med.rpc;

import gov.va.med.vds.vistabeans.Site;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.lang.reflect.Method;
import java.util.concurrent.Callable;

/**
 *
 *
 */
public class QueryThread implements Callable<Object> {
    private static final Logger LOGGER = LoggerFactory.getLogger(QueryThread.class);

    private Object classObj;
    private String methodName;
    private Object[] args;
    private Site site;


    public QueryThread(Site site, Object classObj, String methodName, Object[] args) {
        this.site = site;
        this.classObj = classObj;
        this.methodName = methodName;
        if (args == null) {
            this.args = null;
        } else {
            this.args = args.clone();
        }

    }

    public Object call() {
        Object rtc;
        if (classObj == null) {
            rtc = null;
        } else {

            try {
                Class clazz = classObj.getClass();

                int argsLen;
                if (args == null) {
                    argsLen = 0;
                }
                else {
                    argsLen = args.length;
                }

                Class[] paramTypes = new Class[argsLen];
                for (int i = 0; i < argsLen; i++) {
                    paramTypes[i] = args[i].getClass();
                }
    
                Method method = getMethod(clazz, methodName, paramTypes);
    
                if (LOGGER.isDebugEnabled()) {
                    LOGGER.debug("siteCode: " + site.getSiteCode() + " class: " + clazz.getName() + " method: " + methodName);
                }

                rtc = method.invoke(classObj, args);

            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return rtc;
    }

    public Site getSite() {
        return site;
    }

    public String toString() {
        StringBuilder sbld = new StringBuilder();
        sbld.append("siteCode: ");
        sbld.append(site.getSiteCode());
        sbld.append(" classObj: ");
        if (classObj != null) {
            sbld.append(classObj.getClass().getName());
        }
        sbld.append(" methodName: ");
        sbld.append(methodName);
        sbld.append(" args count: ");
        if (args != null) {
            sbld.append(args.length);
        }

        return sbld.toString();
    }

    /**
     * Searches for and returns method within given class that matches 
     * name and parameter type list. This method is used in lieu of
     * Class.getMethod() because it matches parameter types based on
     * assignability instead of type name.
     * (i.e. can pass parameters that are a subclass of those
     * defined in method signature).
     * 
     * @param clazz Class to get method from.
     * @param methodName Name of method to get.
     * @param paramTypes List of parameter types.
     * @return Method if found.
     * @throws NoSuchMethodException If no method is found.
     */
    private Method getMethod(Class clazz, String methodName, Class[] paramTypes) throws NoSuchMethodException {
        Method[] methods = clazz.getMethods();

        for (Method method : methods) {
            if (method.getName().equals(methodName)) {
                Class[] methodParamTypes = method.getParameterTypes();
                if (methodParamTypes.length != paramTypes.length) {
                    continue;
                }

                boolean isParamMatch = true;
                for (int i = 0; i < paramTypes.length; i++) {
                    Class param = methodParamTypes[i];
                    if (!param.isAssignableFrom(paramTypes[i])) {
                        isParamMatch = false;
                    }
                }

                if (isParamMatch) {
                    return method;
                }
            }
        }

        throw new NoSuchMethodException();
    }
}
