/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.vds.base.dao.DaoBase;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.patient.dao.rpc.LabsRPC;
import gov.va.med.vds.vistabeans.FreeTextReport;
import gov.va.med.vds.vistabeans.LabOrder;
import gov.va.med.vds.vistabeans.LabResult;
import gov.va.med.vds.vistabeans.QueryBean;

/**
 *
 *
 */
public class LabsDao extends DaoBase {
    public LabsDao(RPCConnection cxn) {
        super(cxn);
    }

    public LabResult[] getLabOrderResult(QueryBean queryBean) throws RPCException, UnsupportedRPCMethodException {
        LabsRPC rpc = new LabsRPC(cxn);

        // in the query bean the startdate is the collection date and the itemId is the accession of the order
        return rpc.getLabOrderResult(queryBean.getStartDate(), queryBean.getItemId());
    }


    public LabResult[] getPatientLabTestResults(QueryBean queryBean) throws RPCException, UnsupportedRPCMethodException {
        LabsRPC rpc = new LabsRPC(cxn);

        // in the query bean the itemId is the LabResult.testName to match for the result list.
        return rpc.getPatientLabTestResults(queryBean.getStartDate(), queryBean.getEndDate(), queryBean.getItemId());
    }

    /**
     * This calls the RPC to get the list of Chem Lab Panels and results based on the given start
     * and end dates in the queryBean. This is a list of the lab panels of each order
     * embedded as a field in the LabOrder.
     *
     * @param queryBean needs to contain the startDate and endDate for the RPC
     * @return the list of LabOrders that have results in them
     * @throws RPCException
     */
    public LabOrder[] getPatientChemLabs(QueryBean queryBean) throws RPCException {
        LabsRPC labsRPC = new LabsRPC(cxn);

        return labsRPC.getPatientLabs(queryBean.getStartDate(), queryBean.getEndDate());

    }

    /**
     * This calls the ORWRP REPORT TEXT to get the list of Microbiology labs and results based on the given start
     * and end dates in the queryBean. This is a list of the microbiology lab panels with the lab results of each order
     * embedded as a field in the LabOrder.
     *
     * @param queryBean needs to contain the startDate and endDate for the RPC
     * @return the list of LabOrders that have results in them
     * @throws RPCException
     */
    public LabOrder[] getPatientMicroLabs(QueryBean queryBean) throws RPCException {
        LabsRPC labs = new LabsRPC(cxn);

        return labs.getPatientMicroLabOrders(queryBean.getStartDate(), queryBean.getEndDate());
    }

    /**
     * This calls the ORWRP REPORT TEXT to get the list of Anatomic Pathology labs and results.
     * This is a list of the AP labs with the lab results of each order embedded as a field in the LabOrder.
     *
     * @param queryBean needs to contain the startDate and endDate for the RPC
     * @return the list of LabOrders that have results in them
     * @throws RPCException
     */
    public LabOrder[] getPatientAnaPathLabs(QueryBean queryBean) throws RPCException {
        LabsRPC labs = new LabsRPC(cxn);

        return labs.getPatientAnaPathLabOrders(queryBean.getStartDate(), queryBean.getEndDate());
    }

    /**
     * This calls the VPR RPC to get the list of Lab results based on the given start and end dates in the
     * queryBean. This is only the lab results and does not include order information.
     *
     * @param queryBean needs to contain the startDate and endDate for the RPC
     * @return the array of LabResults in the date range
     * @throws RPCException
     */
    public LabResult[] getPatientLabResultsVPR(QueryBean queryBean) throws RPCException {
        LabsRPC rpc = new LabsRPC(cxn);

        return rpc.getVPRLabResults(queryBean.getStartDate(), queryBean.getEndDate());
    }

    public LabResult[] getPatientLabResultsSkinTests(QueryBean queryBean) throws RPCException {
        LabsRPC rpc = new LabsRPC(cxn);

        return rpc.getPatientLabResultsSkinTests(queryBean.getStartDate(), queryBean.getEndDate());
    }

    public LabResult[] getPatientLabResultsBloodTypeTests(QueryBean queryBean) throws RPCException {
        LabsRPC rpc = new LabsRPC(cxn);

        return rpc.getPatientLabResultsBloodTypeTests(queryBean.getStartDate(), queryBean.getEndDate());
    }

    public LabResult[] getPatientLabResultsAutopsy(QueryBean queryBean) throws RPCException {
        LabsRPC rpc = new LabsRPC(cxn);

        return rpc.getPatientLabResultsAutopsy(queryBean.getStartDate(), queryBean.getEndDate());
    }

    public LabOrder[] getPatientLabsAutopsy(QueryBean queryBean) throws RPCException {
        LabsRPC rpc = new LabsRPC(cxn);

        return rpc.getPatientLabsAutopsy(queryBean.getStartDate(), queryBean.getEndDate());
    }

    /**
     * This calls the ORWRP REPORT TEXT RPC (L:OR_MICL:MICROBIOLOGY~MI;ORDV05;38;999) to get the list of Microbiology
     * Lab results based on the given start and end dates in the queryBean. This is only the lab results and does not
     * include order information.
     *
     * @param queryBean needs to contain the startDate and endDate for the RPC
     * @return the array of LabResults in the date range
     * @throws RPCException
     */
    public LabResult[] getPatientMicroLabResults(QueryBean queryBean) throws RPCException {
        LabsRPC rpc = new LabsRPC(cxn);

        return rpc.getPatientMicroLabResults(queryBean.getStartDate(), queryBean.getEndDate());
    }

    /**
     * This calls the ORWRP REPORT TEXT RPC (L:OR_APL:ANATOMIC PATHOLOGY - ALL REPORTS~SP;ORDV02A;0;999) to get the
     * list of Anatomic Pathology Lab results. This is only the lab results and does include order information.
     *
     * @param queryBean needs to contain the startDate and endDate for the RPC
     * @return the array of LabResults in the date range
     * @throws RPCException
     */
    public LabResult[] getPatientAnaPathLabResults(QueryBean queryBean) throws RPCException {
        LabsRPC rpc = new LabsRPC(cxn);

        return rpc.getPatientAnaPathLabResults(queryBean.getStartDate(), queryBean.getEndDate());
    }

    public FreeTextReport getPatientLabReport(QueryBean queryBean) throws RPCException {
        LabsRPC labsRPC = new LabsRPC(cxn);

        return labsRPC.getLabReport(queryBean.getStartDate(), queryBean.getItemId());
    }

    public LabOrder[] getPatientLabOrdersVPR(QueryBean queryBean) throws RPCException {
        LabsRPC labsRPC = new LabsRPC(cxn);

        return labsRPC.getVPRLabOrders(queryBean.getStartDate(), queryBean.getEndDate());
    }

}
