/*
 * Janus 4.0 (c)
 * Copyright (c) 2014 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.Timestamp;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.Code;
import gov.va.med.vds.vistabeans.Diagnosis;
import gov.va.med.vds.vistabeans.Encounter;
import gov.va.med.vds.vistabeans.EncounterDocument;
import gov.va.med.vds.vistabeans.Procedure;
import gov.va.med.vds.vistabeans.ProgressNote;
import gov.va.med.vds.vistabeans.Provider;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 *
 *
 */
public class EncountersRPC extends BaseRPC {

    private static final Logger LOGGER = LoggerFactory.getLogger(EncountersRPC.class);

    public EncountersRPC(RPCConnection cxn) {
        super(cxn);
    }


    /**
     * Uses the VPR RPC with "visits" as the type.
     *
     * @param startDate  The start of the date range for the list of encounters
     * @param endDate    The end of the date range for the list of encounters
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public Encounter[] getPatientEncounters(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientEncounters(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "appointments" as the type.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of encounters
     * @param endDate    The end of the date range for the list of encounters
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public Encounter[] getPatientEncounters(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetEncountersRequest(patientIen, startDate, endDate);
        String response = (String) cxn.query(request);

        Encounter[] encounters = parseEncounters(response, false);

        if (encounters.length > 0) {
            // get the documents padding 3 months to end date. Cross reference against visits for hasImage
            NotesRPC notesRPC = new NotesRPC(cxn);
            Calendar docsEndDate = Calendar.getInstance();
            docsEndDate.setTime(endDate.getTime());
            docsEndDate.add(Calendar.MONTH, 3);
            ProgressNote[] notes = notesRPC.getProgressNotesVPR(startDate, docsEndDate);
            // put notes in a HashMap using the id as the key
            if (notes.length > 0) {
                HashMap<String, ProgressNote> notesMap = new HashMap<String, ProgressNote>();
                for (int noteNum = 0; noteNum < notes.length; noteNum++) {
                    notesMap.put(notes[noteNum].getNoteId(), notes[noteNum]);
                }
                // check each encounter document for hasImage
                for (Encounter encounter: encounters) {
                    if (encounter.getDocuments() != null) {
                        for (EncounterDocument encounterDocument : encounter.getDocuments()) {
                            if (notesMap.containsKey(encounterDocument.getId())) {
                                encounterDocument.setImages(notesMap.get(encounterDocument.getId()).getImages());
                            }
                        }
                    }
                }
            }
        }

        return encounters;
    }

    /**
     * Uses the underlying call to VPR visits used in getting the Encounters and filters for the Admissions.
     * Admissions in the VPR visits are noted by those that have an arrivalDate.
     *
     * @param startDate  The start of the date range for the list of encounters
     * @param endDate    The end of the date range for the list of encounters
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public Encounter[] getPatientAdmissionsFromEncounters(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientAdmissionsFromEncounters(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the underlying call to VPR visits used in getting the Encounters and filters for the Admissions.
     * Admissions in the VPR visits are noted by those that have an arrivalDate.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of encounters
     * @param endDate    The end of the date range for the list of encounters
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public Encounter[] getPatientAdmissionsFromEncounters(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetEncountersRequest(patientIen, startDate, endDate);
        String response = (String) cxn.query(request);

        Encounter[] encounters = parseEncounters(response, true);

        return encounters;
    }


    /**
     * Uses the underlying call to VPR visits used in getting the Encounters and filters for the Admissions.
     * Admissions in the VPR visits are noted by those that have an arrivalDate.
     *
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public Encounter getPatientAdmission(String encounterId) throws RPCException {
        return getPatientAdmission(cxn.getPatientIen(), encounterId);
    }

    /**
     * Uses the underlying call to VPR visits used in getting the Encounters and filters for the Admissions.
     * Admissions in the VPR visits are noted by those that have an arrivalDate.
     *
     * @param patientIen The patient IEN
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public Encounter getPatientAdmission(String patientIen, String encounterId) throws RPCException {
        Encounter rtc;
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, VPRUtil.VPRType.VISITS, encounterId);
        String response = (String) cxn.query(request);

        Encounter[] encounters = parseEncounters(response, true);
        if (encounters != null && encounters.length > 0) {
            rtc = encounters[0];
        } else {
            rtc = null;
        }
        return rtc;
    }


    /**
     * Uses the underlying call to VPR visits used in getting the Encounters and filters for the non-Admissions.
     * Admissions in the VPR visits are noted by those that have an arrivalDate. The non-admissions are all other
     * encounters.
     *
     * @param startDate  The start of the date range for the list of encounters
     * @param endDate    The end of the date range for the list of encounters
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public Encounter[] getPatientNonAdmissionsFromEncounters(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientNonAdmissionsFromEncounters(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the underlying call to VPR visits used in getting the Encounters and filters for the non-Admissions.
     * Admissions in the VPR visits are noted by those that have an arrivalDate. The non-admissions are all other
     * encounters.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of encounters
     * @param endDate    The end of the date range for the list of encounters
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public Encounter[] getPatientNonAdmissionsFromEncounters(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        Encounter[] encounters = getPatientEncounters(patientIen, startDate, endDate);

        List<Encounter> encounterList = new ArrayList<Encounter>(Arrays.asList(encounters));
        Iterator<Encounter> encounterIterator = encounterList.iterator();
        while (encounterIterator.hasNext()) {
            Encounter encounter = encounterIterator.next();
            if (encounter.getArrivalDateTime() != null) {
                encounterIterator.remove();
            }
        }

        return encounterList.toArray(new Encounter[0]);
    }


    /**
     * Builds the VPR RPC request string using "appointments" as the type. The problems contain the list of
     * appointments. The RPC will fail if the "VPR APPLICATION PROXY" context is not set. That is set when
     * setting up the connection.
     *
     * @param patientIen Patient IEN
     * @param startDate  Start date of range
     * @param endDate    End date of range
     * @return The RPC response message string
     */
    private String buildGetEncountersRequest(String patientIen, Calendar startDate, Calendar endDate) {
        return VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.VISITS);
    }

    private Encounter[] parseEncounters(String response, boolean admissionsOnly) throws RPCException {
        Encounter[] rtc;
        if ("".equals(response)) {
            rtc = new Encounter[0];
        } else {
    
            StringBuilder sbld = new StringBuilder(response);
            sbld.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    
            ArrayList<Encounter> encountersList = new ArrayList<Encounter>();
    
            final String XPATH_VISIT = "//visit";
    
            Document document = null;
            XPathExpression exprVisit = null;
    
            try {
                document = parseXMLDocument(sbld.toString());
                XPathFactory xPathFactory = XPathFactory.newInstance();
                XPath xpath = xPathFactory.newXPath();
                exprVisit = xpath.compile(XPATH_VISIT);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            NodeList visitsNodeList = null;
            try {
                visitsNodeList = (NodeList) exprVisit.evaluate(document, XPathConstants.NODESET);
            } catch (XPathExpressionException e) {
                throw new RPCException(e);
            }
    
            List<VPRVisitsParser> visitsParserList = new ArrayList<VPRVisitsParser>();
            int visitNodeListLength = visitsNodeList.getLength();
            for (int visitNum = 0; visitNum < visitNodeListLength; visitNum++) {
                Node visitNode = visitsNodeList.item(visitNum);
                visitNode.getParentNode().removeChild(visitNode);
    
                visitsParserList.add(new VPRVisitsParser(visitNode, admissionsOnly));
            }

            ExecutorService executorService = cxn.getParserThreadPool().getThreadPool();
            List<Future<Encounter>> futures = null;
            try {
                futures = executorService.invokeAll(visitsParserList, cxn.getParserThreadPool().getPoolTimeoutSeconds(), TimeUnit.SECONDS);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse visits(encounters) results : visit node parser futures list is null.");
                }
            } catch (InterruptedException e) {
                RPCUtil.logErrorMessage(LOGGER, e);
                throw new RPCException(e);
            }
    
            Iterator<Future<Encounter>> futureIterator = futures.iterator();
    
            while (futureIterator.hasNext()) {
                try {
                    Future<Encounter> future = futureIterator.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR Visit(Encounter) Data");
                    } else {
                        Encounter encounter = future.get();
    
                        // exclude null results
                        if (encounter != null) {
                            encountersList.add(encounter);
                        }
                    }
                } catch (InterruptedException ie) {
                    RPCUtil.logErrorMessage(LOGGER, ie);
                } catch (ExecutionException ee) {
                    RPCUtil.logErrorMessage(LOGGER, ee);
                }
            }
    
            rtc = encountersList.toArray(new Encounter[0]);
        }
        return rtc;
    }

    private class VPRVisitsParser implements Callable<Encounter> {
        private Node visitNode;
        private boolean admissionsOnly;

        public VPRVisitsParser(Node visitNode, boolean admissionsOnly) {
            this.visitNode = visitNode;
            this.admissionsOnly = admissionsOnly;
        }

        public Encounter call() throws Exception {
            Encounter rtc;
            // Admissions
            final String XPATH_ARRIVAL_DATETIME = "./arrivalDateTime/@value";
            final String XPATH_DEPARTURE_DATETIME = "./departureDateTime/@value";
            final String XPATH_ROOMBED = "./roomBed/@value";
            final String XPATH_SPECIALTY = "./specialty/@value";
            final String XPATH_ADMISSION = "./admission/@value";

            XPathExpression exprArrivalDateTime = null;
            XPathExpression exprDepartureDateTime = null;
            XPathExpression exprRoomBed = null;
            XPathExpression exprSpecialty = null;
            XPathExpression exprAdmission = null;

            XPathFactory xPathFactory = XPathFactory.newInstance();
            XPath xpath = xPathFactory.newXPath();

            try {
                // Admission expressions
                exprArrivalDateTime = xpath.compile(XPATH_ARRIVAL_DATETIME);
                exprDepartureDateTime = xpath.compile(XPATH_DEPARTURE_DATETIME);
                exprRoomBed = xpath.compile(XPATH_ROOMBED);
                exprSpecialty = xpath.compile(XPATH_SPECIALTY);
                exprAdmission = xpath.compile(XPATH_ADMISSION);
            } catch (XPathExpressionException e) {
                throw new RPCException(e);
            }

            // Admission parsing
            String arrivalDateTime = null;
            Calendar arrivalDate = null;
            try {
                arrivalDateTime = (String) exprArrivalDateTime.evaluate(visitNode, XPathConstants.STRING);
                if (arrivalDateTime != null) {
                    arrivalDate = Timestamp.toDateTime(arrivalDateTime);
                }
            } catch (XPathExpressionException e) {
                RPCUtil.logErrorMessage(LOGGER, e);
            }
            String departureDateTime = null;
            Calendar departureDate = null;
            try {
                departureDateTime = (String) exprDepartureDateTime.evaluate(visitNode, XPathConstants.STRING);
                if (departureDateTime != null) {
                    departureDate = Timestamp.toDateTime(departureDateTime);
                }
            } catch (XPathExpressionException e) {
                RPCUtil.logErrorMessage(LOGGER, e);
            }
            String roomBed = RPCUtil.evaluateNodeToString(visitNode, exprRoomBed, LOGGER);
            String specialty = RPCUtil.evaluateNodeToString(visitNode, exprSpecialty, LOGGER);
            String admission = RPCUtil.evaluateNodeToString(visitNode, exprAdmission, LOGGER);

            // If not an admission and we are looking for admissionOnly, we don't need to continue
            if (admissionsOnly && arrivalDate == null) {
                rtc = null;
            } else {
    
                final String XPATH_DATETIME = "./dateTime/@value";
    
    //            final String XPATH_FACILITY_CODE = "./facility/@code";
    //            final String XPATH_FACILITY_NAME = "./facility/@name";
                final String XPATH_ID = "./id/@value";
                final String XPATH_LOCATION = "./location/@value";
                final String XPATH_PATIENT_CLASS = "./patientClass/@value";
                final String XPATH_SERVICE = "./service/@value";
                final String XPATH_SERVICE_CAT_CODE = "./serviceCategory/@code";
                final String XPATH_SERVICE_CAT_NAME = "./serviceCategory/@name";
                final String XPATH_VISIT_STRING = "./visitString/@value";
                final String XPATH_CREDIT_STOP_CODE_CODE = "./creditStopCode/@code";
                final String XPATH_CREDIT_STOP_CODE_NAME = "./creditStopCode/@name";
                final String XPATH_STOP_CODE_CODE = "./stopCode/@code";
                final String XPATH_STOP_CODE_NAME = "./stopCode/@name";
    //            final String XPATH_TYPE_CODE = "./type/@code";
                final String XPATH_TYPE_NAME = "./type/@name";
                final String XPATH_REASON_CODE = "./reason/@code";
                final String XPATH_REASON_NAME = "./reason/@name";
                final String XPATH_REASON_NARRATIVE = "./reason/@narrative";
    
                final String XPATH_PROVIDERS = "./providers";
                final String XPATH_PROVIDER = "./provider";
                final String XPATH_PROVIDER_CODE = "./@code";
                final String XPATH_PROVIDER_NAME = "./@name";
                final String XPATH_PROVIDER_ROLE = "./@role";
    //            final String XPATH_PROVIDER_PRIMARY = "./@primary";
                final String XPATH_PROVIDER_TAXONOMY_CODE = "./@taxonomyCode";
                final String XPATH_PROVIDER_PROVIDER_TYPE = "./@providerType";
                final String XPATH_PROVIDER_CLASSIFICATION = "./@classification";
                final String XPATH_PROVIDER_SPECIALIZATION = "./@specialization";
                final String XPATH_SERVICE_ATTR = "./@service";
                final String XPATH_OFFICE_PHONE = "./@officePhone";
    //            final String XPATH_FAX = "./@fax";
                final String XPATH_ANALOG_PAGER = "./@analogPager";
                final String XPATH_EMAIL = "./@email";
    
                final String XPATH_CPTS = "./cpts";
                final String XPATH_CPT = "./cpt";
                final String XPATH_CPT_NAME = "./@name";
                final String XPATH_CPT_CODE = "./@code";
    
                final String XPATH_ICDS = "./icds";
                final String XPATH_ICD = "./icd";
                final String XPATH_ICD_NAME = "./@name";
                final String XPATH_ICD_CODE = "./@code";
                final String XPATH_ICD_NARRATIVE = "./@narrative";
                final String XPATH_ICD_RANKING = "./ranking";
    
                final String XPATH_DOCUMENTS = "./documents";
                final String XPATH_DOCUMENT = "./document";
                final String XPATH_DOCUMENT_ID = "./@id";
                final String XPATH_DOCUMENT_VUID = "./@vuid";
                final String XPATH_DOCUMENT_LOCAL_TITLE = "./@localTitle";
                final String XPATH_DOCUMENT_NATIONAL_TITLE = "./@nationalTitle";
                final String XPATH_DOCUMENT_CONTENT = "./content";
    
                XPathExpression exprDateTime = null;
                XPathExpression exprId = null;
                XPathExpression exprLocation = null;
                XPathExpression exprPatientClass = null;
                XPathExpression exprService = null;
                XPathExpression exprServiceCatCode = null;
                XPathExpression exprServiceCatName = null;
                XPathExpression exprVisitString = null;
                XPathExpression exprCreditStopCode = null;
                XPathExpression exprCreditStopName = null;
                XPathExpression exprStopCodeCode = null;
                XPathExpression exprStopCodeName = null;
                XPathExpression exprReasonCode = null;
                XPathExpression exprReasonName = null;
                XPathExpression expreReasonNarrative = null;
    //                    exprTypeCode = null,
                XPathExpression exprTypeName = null;
                XPathExpression exprProviders = null;
                XPathExpression exprProvider = null;
                XPathExpression exprProviderCode = null;
                XPathExpression exprProviderName = null;
                XPathExpression exprProviderRole = null;
                XPathExpression exprProviderTaxonomyCode = null;
                XPathExpression exprServiceAttr = null;
                XPathExpression exprOfficePhone = null;
                XPathExpression exprFax = null;
                XPathExpression exprAnalogPager = null;
                XPathExpression exprEmail = null;
    //                    exprProviderPrimary = null,
                XPathExpression exprProviderProviderType = null;
                XPathExpression exprProviderClassification = null;
                XPathExpression exprProviderSpecialization = null;
                XPathExpression exprCpts = null;
                XPathExpression exprCpt = null;
                XPathExpression exprCptName = null;
                XPathExpression exprCptCode = null;
                XPathExpression exprIcds = null;
                XPathExpression exprIcd = null;
                XPathExpression exprIcdName = null;
                XPathExpression exprIcdCode = null;
                XPathExpression exprIcdNarrative = null;
                XPathExpression exprIcdRanking = null;
                XPathExpression exprDocuments = null;
                XPathExpression exprDocumentId = null;
                XPathExpression exprDocumentLocalTitle = null;
                XPathExpression exprDocument = null;
                XPathExpression exprDocumentVuid = null;
                XPathExpression exprDocumentNationalTitle = null;
                XPathExpression exprDocumentContent = null;
    
                try {
                    exprDateTime = xpath.compile(XPATH_DATETIME);
    //                exprFacilityCode = xpath.compile(XPATH_FACILITY_CODE);
    //                exprFacilityName = xpath.compile(XPATH_FACILITY_NAME);
                    exprId = xpath.compile(XPATH_ID);
                    exprLocation = xpath.compile(XPATH_LOCATION);
                    exprPatientClass = xpath.compile(XPATH_PATIENT_CLASS);
                    exprService = xpath.compile(XPATH_SERVICE);
                    exprServiceCatCode = xpath.compile(XPATH_SERVICE_CAT_CODE);
                    exprServiceCatName = xpath.compile(XPATH_SERVICE_CAT_NAME);
                    exprVisitString = xpath.compile(XPATH_VISIT_STRING);
    //                exprTypeCode = xpath.compile(XPATH_TYPE_CODE);
                    exprTypeName = xpath.compile(XPATH_TYPE_NAME);
                    exprCreditStopCode = xpath.compile(XPATH_CREDIT_STOP_CODE_CODE);
                    exprCreditStopName = xpath.compile(XPATH_CREDIT_STOP_CODE_NAME);
                    exprStopCodeCode = xpath.compile(XPATH_STOP_CODE_CODE);
                    exprStopCodeName = xpath.compile(XPATH_STOP_CODE_NAME);
                    exprReasonCode = xpath.compile(XPATH_REASON_CODE);
                    exprReasonName = xpath.compile(XPATH_REASON_NAME);
                    expreReasonNarrative = xpath.compile(XPATH_REASON_NARRATIVE);
                    exprProviders = xpath.compile(XPATH_PROVIDERS);
                    exprProvider = xpath.compile(XPATH_PROVIDER);
                    exprProviderCode = xpath.compile(XPATH_PROVIDER_CODE);
                    exprProviderName = xpath.compile(XPATH_PROVIDER_NAME);
                    exprProviderRole = xpath.compile(XPATH_PROVIDER_ROLE);
    //                exprProviderPrimary = xpath.compile(XPATH_PROVIDER_PRIMARY);
                    exprProviderTaxonomyCode = xpath.compile(XPATH_PROVIDER_TAXONOMY_CODE);
                    exprProviderProviderType = xpath.compile(XPATH_PROVIDER_PROVIDER_TYPE);
                    exprProviderClassification = xpath.compile(XPATH_PROVIDER_CLASSIFICATION);
                    exprProviderSpecialization = xpath.compile(XPATH_PROVIDER_SPECIALIZATION);
                    exprServiceAttr = xpath.compile(XPATH_SERVICE_ATTR);
                    exprOfficePhone = xpath.compile(XPATH_OFFICE_PHONE);
                    exprFax = xpath.compile(XPATH_PROVIDER_SPECIALIZATION);
                    exprAnalogPager = xpath.compile(XPATH_ANALOG_PAGER);
                    exprEmail = xpath.compile(XPATH_EMAIL);
                    exprCpts = xpath.compile(XPATH_CPTS);
                    exprCpt = xpath.compile(XPATH_CPT);
                    exprCptName = xpath.compile(XPATH_CPT_NAME);
                    exprCptCode = xpath.compile(XPATH_CPT_CODE);
                    exprIcds = xpath.compile(XPATH_ICDS);
                    exprIcd = xpath.compile(XPATH_ICD);
                    exprIcdName = xpath.compile(XPATH_ICD_NAME);
                    exprIcdCode = xpath.compile(XPATH_ICD_CODE);
                    exprIcdNarrative = xpath.compile(XPATH_ICD_NARRATIVE);
                    exprIcdRanking = xpath.compile(XPATH_ICD_RANKING);
                    exprDocuments = xpath.compile(XPATH_DOCUMENTS);
                    exprDocument = xpath.compile(XPATH_DOCUMENT);
                    exprDocumentId = xpath.compile(XPATH_DOCUMENT_ID);
                    exprDocumentLocalTitle = xpath.compile(XPATH_DOCUMENT_LOCAL_TITLE);
                    exprDocumentNationalTitle = xpath.compile(XPATH_DOCUMENT_NATIONAL_TITLE);
                    exprDocumentContent = xpath.compile(XPATH_DOCUMENT_CONTENT);
                    exprDocumentVuid = xpath.compile(XPATH_DOCUMENT_VUID);
    
                } catch (XPathExpressionException e) {
                    throw new RPCException(e);
                }
    
                Encounter encounter = new Encounter();
                fillCommonDataBeanValues(encounter);
    
                // Other encounter parsing
                String dateTime = null;
                Calendar visitDate = null;
                try {
                    dateTime = (String) exprDateTime.evaluate(visitNode, XPathConstants.STRING);
                    visitDate = Timestamp.toDateTime(dateTime);
                } catch (XPathExpressionException e) {
                    RPCUtil.logErrorMessage(LOGGER, e);
                }
                String id = RPCUtil.evaluateNodeToString(visitNode, exprId, LOGGER);
                String location = RPCUtil.evaluateNodeToString(visitNode, exprLocation, LOGGER);
                String patientClass = RPCUtil.evaluateNodeToString(visitNode, exprPatientClass, LOGGER);
                String serviceCategoryCode = RPCUtil.evaluateNodeToString(visitNode, exprServiceCatCode, LOGGER);
                String serviceCategoryName = RPCUtil.evaluateNodeToString(visitNode, exprServiceCatName, LOGGER);
                String visitString = RPCUtil.evaluateNodeToString(visitNode, exprVisitString, LOGGER);
    //            String typeCode = RPCUtil.evaluateNodeToString(visitNode, exprTypeCode, logger);
                String typeName = RPCUtil.evaluateNodeToString(visitNode, exprTypeName, LOGGER);
                String service = RPCUtil.evaluateNodeToString(visitNode, exprService, LOGGER);
                String creditStopCode = RPCUtil.evaluateNodeToString(visitNode, exprCreditStopCode, LOGGER);
                String creditStopName = RPCUtil.evaluateNodeToString(visitNode, exprCreditStopName, LOGGER);
                String stopCodeCode = RPCUtil.evaluateNodeToString(visitNode, exprStopCodeCode, LOGGER);
                String stopCodeName = RPCUtil.evaluateNodeToString(visitNode, exprStopCodeName, LOGGER);
                String reasonCode = RPCUtil.evaluateNodeToString(visitNode, exprReasonCode, LOGGER);
                String reasonName = RPCUtil.evaluateNodeToString(visitNode, exprReasonName, LOGGER);
                String reasonNarrative = RPCUtil.evaluateNodeToString(visitNode, expreReasonNarrative, LOGGER);
    
                Node providersNode = RPCUtil.evaluateNodeToNode(visitNode, exprProviders, LOGGER);
                if (providersNode != null) {
                    NodeList providerNodeList = RPCUtil.evaluateNodeToNodeList(providersNode, exprProvider, LOGGER);
    
                    if (providerNodeList != null) {
                        int providersNodeListLength = providerNodeList.getLength();
    
                        for (int providerNum = 0; providerNum < providersNodeListLength; providerNum++) {
                            Provider provider = new Provider();
                            Node providerNode = providerNodeList.item(providerNum);
                            providerNode.getParentNode().removeChild(providerNode);
    
                            String providerRole = RPCUtil.evaluateNodeToString(providerNode, exprProviderRole, LOGGER);
                            if ("P".equals(providerRole)) {
                                provider.setPrimary(true);
                            } else if (arrivalDate == null) {
                                // this is not an admission so we can continue to look for only the primary provider
                                continue;
                            }
    
                            String providerCode = RPCUtil.evaluateNodeToString(providerNode, exprProviderCode, LOGGER);
                            String providerName = RPCUtil.evaluateNodeToString(providerNode, exprProviderName, LOGGER);
                            String providerTaxonomyCode = RPCUtil.evaluateNodeToString(providerNode, exprProviderTaxonomyCode, LOGGER);
                            String providerType = RPCUtil.evaluateNodeToString(providerNode, exprProviderProviderType, LOGGER);
                            String providerClassification = RPCUtil.evaluateNodeToString(providerNode, exprProviderClassification, LOGGER);
                            String providerSpecialization = RPCUtil.evaluateNodeToString(providerNode, exprProviderSpecialization, LOGGER);
                            String providerService = RPCUtil.evaluateNodeToString(providerNode, exprServiceAttr, LOGGER);
                            String providerOfficePhone = RPCUtil.evaluateNodeToString(providerNode, exprOfficePhone, LOGGER);
                            String providerAnalogPager = RPCUtil.evaluateNodeToString(providerNode, exprAnalogPager, LOGGER);
                            String providerFax = RPCUtil.evaluateNodeToString(providerNode, exprFax, LOGGER);
                            String providerEmail = RPCUtil.evaluateNodeToString(providerNode, exprEmail, LOGGER);
    
                            provider.setCode(providerCode);
                            provider.setName(providerName);
                            provider.setRole(providerRole);
                            provider.setTaxonomyCode(providerTaxonomyCode);
                            provider.setProviderType(providerType);
                            provider.setClassification(providerClassification);
                            provider.setSpecialization(providerSpecialization);
                            provider.setService(providerService);
                            provider.setPhone(providerOfficePhone);
                            provider.setAnalogPager(providerAnalogPager);
                            provider.setFax(providerFax);
                            provider.setEmail(providerEmail);
    
                            if (provider.getName() != null) {
                                if (provider.isPrimary()) {
                                    encounter.setProvider(provider);
                                    if (arrivalDate == null) {
                                        // we found the primary provider for the non-admission so we can leave the for loop
                                        break;
                                    }
                                } else if (arrivalDate != null && "A".equalsIgnoreCase(providerRole)) {
                                    // this is encounter is an admission and the current provider is the admitting provider
                                    encounter.setAdmittingProvider(provider);
                                }
                            }
                        }
    
                    }
                }
    
                List<Code> codes = new ArrayList<Code>();
    
                Node cptsNode = RPCUtil.evaluateNodeToNode(visitNode, exprCpts, LOGGER);
                if (cptsNode != null) {
                    NodeList cptNodeList = RPCUtil.evaluateNodeToNodeList(cptsNode, exprCpt, LOGGER);
    
                    if (cptNodeList != null) {
                        int cptNodeListLength = cptNodeList.getLength();
                        List<Code> cptList = new ArrayList<Code>();
                        List<Procedure> procedureList = new ArrayList<Procedure>();
                        for (int cptNum = 0; cptNum < cptNodeListLength; cptNum++) {
                            Node cptNode = cptNodeList.item(cptNum);
                            cptNode.getParentNode().removeChild(cptNode);
    
                            Code cpt = new Code();
    
                            cpt.setSystem("CPT");
    
                            String cptName = RPCUtil.evaluateNodeToString(cptNode, exprCptName, LOGGER);
                            String cptCode = RPCUtil.evaluateNodeToString(cptNode, exprCptCode, LOGGER);
    
                            cpt.setDisplay(cptName);
                            cpt.setCode(cptCode);
    
                            Procedure proc = new Procedure();
                            fillCommonDataBeanValues(proc);
                            proc.setCode(cptCode);
                            proc.setLocation(location);
                            proc.setCodeSystem("CPT");
                            proc.setDescription(cptName);
                            proc.setProvider(encounter.getProvider());
                            proc.setEncounterId(id);
    
                            // If this is an admission, set as inpatient procedure, else set as out
                            if (arrivalDate != null) {
                                proc.setInOutpatient("I");
                            } else {
                                proc.setInOutpatient("O");
                            }
    
                            proc.setDate(encounter.getApptDate());
    
                            procedureList.add(proc);
                            cptList.add(cpt);
                        }
    
                        if (!cptList.isEmpty()) {
                            codes.addAll(cptList);
                            encounter.setProcedures(procedureList.toArray(new Procedure[0]));
                        }
                    }
                }
    
                Node icdsNode = RPCUtil.evaluateNodeToNode(visitNode, exprIcds, LOGGER);
                if (icdsNode != null) {
                    NodeList icdNodeList = RPCUtil.evaluateNodeToNodeList(icdsNode, exprIcd, LOGGER);
    
                    if (icdNodeList != null) {
                        int icdNodeListLength = icdNodeList.getLength();
                        List<Code> icdList = new ArrayList<Code>();
                        List<Diagnosis> diagnosisList = new ArrayList<Diagnosis>();
                        for (int icdNum = 0; icdNum < icdNodeListLength; icdNum++) {
                            Node icdNode = icdNodeList.item(icdNum);
                            icdNode.getParentNode().removeChild(icdNode);
    
                            Code icd = new Code();
                            icd.setSystem("ICD9");
    
                            String icdName = RPCUtil.evaluateNodeToString(icdNode, exprIcdName, LOGGER);
                            String icdCode = RPCUtil.evaluateNodeToString(icdNode, exprIcdCode, LOGGER);
                            String icdNarrative = RPCUtil.evaluateNodeToString(icdNode, exprIcdNarrative, LOGGER);
                            String icdRanking = RPCUtil.evaluateNodeToString(icdNode, exprIcdRanking, LOGGER);
    
                            icd.setDisplay(icdName);
                            icd.setCode(icdCode);
                            icd.setNarrative(icdNarrative);
                            icd.setRanking(icdRanking);
    
                            Diagnosis diag = new Diagnosis();
                            fillCommonDataBeanValues(diag);
                            diag.setCode(icdCode);
                            diag.setDescription(icdName);
                            diag.setCodeSystem("ICD9");
    
    
                            icdList.add(icd);
                            diagnosisList.add(diag);
                        }
    
                        if (!icdList.isEmpty()) {
                            codes.addAll(icdList);
                            encounter.setDiagnosis(diagnosisList.toArray(new Diagnosis[0]));
                        }
                    }
                }
    
                Node documentsNode = RPCUtil.evaluateNodeToNode(visitNode, exprDocuments, LOGGER);
                if (documentsNode != null) {
                    NodeList documentNodeList = RPCUtil.evaluateNodeToNodeList(documentsNode, exprDocument, LOGGER);
    
                    if (documentNodeList != null) {
                        int documentNodeListLength = documentNodeList.getLength();
                        List<EncounterDocument> documentList = new ArrayList<EncounterDocument>();
                        for (int documentNum = 0; documentNum < documentNodeListLength; documentNum++) {
                            Node documentNode = documentNodeList.item(documentNum);
                            documentNode.getParentNode().removeChild(documentNode);
    
                            EncounterDocument document = new EncounterDocument();
                            fillCommonDataBeanValues(document);
    
                            String documentId = RPCUtil.evaluateNodeToString(documentNode, exprDocumentId, LOGGER);
                            String documentLocalTitle = RPCUtil.evaluateNodeToString(documentNode, exprDocumentLocalTitle, LOGGER);
                            String documentNationalTitle = RPCUtil.evaluateNodeToString(documentNode, exprDocumentNationalTitle, LOGGER);
                            String documentContent = RPCUtil.evaluateNodeToString(documentNode, exprDocumentContent, LOGGER);
                            String documentVuid = RPCUtil.evaluateNodeToString(documentNode, exprDocumentVuid, LOGGER);
    
                            document.setId(documentId);
                            document.setLocalTitle(documentLocalTitle);
                            document.setNationalTitle(documentNationalTitle);
                            document.setContent(documentContent);
                            document.setVuid(documentVuid);
    
                            documentList.add(document);
                        }
    
                        if (!documentList.isEmpty()) {
                            encounter.setDocuments(documentList.toArray(new EncounterDocument[0]));
                        }
                    }
                }
    
                encounter.setApptDate(visitDate);
                encounter.setApptId(id);
                encounter.setEncounterId(id);
                encounter.setClinic(location);
                encounter.setApptType(typeName);
                encounter.setPatientClass(patientClass);
                encounter.setService(service);
                encounter.setServiceCategoryCode(serviceCategoryCode);
                encounter.setServiceCategoryName(serviceCategoryName);
                encounter.setVisitString(visitString);
                encounter.setClinicStopCode(stopCodeCode);
                encounter.setClinicStopName(stopCodeName);
                encounter.setCreditStopCode(creditStopCode);
                encounter.setCreditStopName(creditStopName);
                encounter.setCodes(codes.toArray(new Code[0]));
                encounter.setReasonCode(reasonCode);
                encounter.setReason(reasonName);
                encounter.setReasonNarrative(reasonNarrative);
    
                // Admission fields
                encounter.setArrivalDateTime(arrivalDate);
                encounter.setDepartureDateTime(departureDate);
                encounter.setRoomBed(roomBed);
                encounter.setSpecialty(specialty);
                encounter.setAdmissionRegistrationNum(admission);
    
    
                rtc = encounter;
            }
            return rtc;
        }
    }

}
