/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.Code;
import gov.va.med.vds.vistabeans.Immunization;
import gov.va.med.vds.vistabeans.PCE;
import gov.va.med.vhahon.common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 *
 *
 */
public class ImmunizationsRPC extends BaseRPC {
    private static final Logger LOGGER = LoggerFactory.getLogger(ImmunizationsRPC.class);

    public ImmunizationsRPC(RPCConnection cxn) {
        super(cxn);
    }

    public Immunization[] getImmunizationList() throws RPCException {
        return getImmunizationList(cxn.getPatientIen());
    }

    //SB: missing reaction and series
    public Immunization[] getImmunizationList(String patientIen) throws RPCException {
        String request = buildGetImmunizationsListRequest(patientIen);
        String response = (String) cxn.query(request);
        return toImmunizationList(response/*, patientIen*/);
    }

    private String buildGetImmunizationsListRequest(String patientIen) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORQQPX IMMUN LIST");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        return vquery.buildMessage();
    }

    private Immunization[] toImmunizationList(String response/*, String patientIen*/) {
        Immunization[] rtc;
        if ("".equals(response)) {
            rtc = new Immunization[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
    
            ArrayList<Immunization> lst = new ArrayList<Immunization>();
            Immunization rec = null;
            for (int i = 0; i < lines.length; i++) {
                if ("".equals(lines[i])) {
                    continue;
                }
                String[] flds = StringUtils.split(lines[i], StringUtils.CARET);
    
                rec = new Immunization();
                rec.setId(flds[0]);
                rec.setName(flds[1]);
                Calendar dateTimeCal = Utils.getCal(Timestamp.toUtcString(flds[2]));
                rec.setDateTime(Utils.formatDisplayDate(dateTimeCal));
                fillCommonDataBeanValues(rec);
    
                lst.add(rec);
            }
    
            rtc = lst.toArray(new Immunization[0]);
        }
        return rtc;
    }

    /**
     * Uses the VPR RPC with "immunizations" as the type.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of immunizations
     * @param endDate    The end of the date range for the list of immunizations
     * @return The list of immunizations for the specified date range
     * @throws RPCException
     */
    public Immunization[] getImmunizationsVPR(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.IMMUNIZATIONS);
        String response = (String) cxn.query(request);

        Immunization[] immunizations = parseImmunizationsFromVPR(response);
        ArrayList<Immunization> rtc = new ArrayList<Immunization>(Arrays.asList(immunizations));

//        // Add the skinTests
//        PCERPC pcerpc = new PCERPC(cxn);
//        PCE[] skinTests = pcerpc.getSkinTests(patientIen, startDate, endDate);
//        if (skinTests != null && skinTests.length > 0) {
//            // map each skinTest into an Immunization
//            for (PCE skinTest : skinTests) {
//                Immunization immunization = new Immunization();
//                fillCommonDataBeanValues(immunization);
//
//                immunization.setName(skinTest.getName());
//                String administered = Timestamp.fromDateTime(skinTest.getReadDate());
//                String formattedAdministered = Timestamp.formatDisplayDateNoBadMonthOrDay(administered);
//                immunization.setDateTime(formattedAdministered);
//                immunization.setResult(skinTest.getResult());
//                immunization.setUnits(skinTest.getReading());
//
//                rtc.add(immunization);
//            }
//        }

        return rtc.toArray(new Immunization[0]);
    }

    /**
     * Retrieves Skin Tests
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of immunizations
     * @param endDate    The end of the date range for the list of immunizations
     * @return The list of immunizations for the specified date range
     * @throws RPCException
     */
    public Immunization[] getImmunizationSkinTests(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        ArrayList<Immunization> rtc = new ArrayList<Immunization>();

        // Add the skinTests
        PCERPC pcerpc = new PCERPC(cxn);
        PCE[] skinTests = pcerpc.getSkinTests(patientIen, startDate, endDate);
        if (skinTests != null && skinTests.length > 0) {
            // map each skinTest into an Immunization
            for (PCE skinTest : skinTests) {
                Immunization immunization = new Immunization();
                fillCommonDataBeanValues(immunization);

                immunization.setName(skinTest.getName());
                String administered = Timestamp.fromDateTime(skinTest.getEnteredDate());
                String formattedAdministered = Timestamp.formatDisplayDateNoBadMonthOrDay(administered);
                immunization.setDateTime(formattedAdministered);
                StringBuilder immunizationResultSB = new StringBuilder();
                immunizationResultSB.append(skinTest.getReading());
                immunizationResultSB.append("  ");
                immunizationResultSB.append(skinTest.getResult());
                immunization.setResult(immunizationResultSB.toString());
                immunization.setUnits(skinTest.getReading());

                rtc.add(immunization);
            }
        }

        return rtc.toArray(new Immunization[0]);
    }


    /**
     * Uses the VPR RPC with "immunizations" as the type.
     *
     * @param startDate The start of the date range for the list of immunizations
     * @param endDate   The end of the date range for the list of immunizations
     * @return The list of immunizations for the specified date range
     * @throws RPCException
     */
    public Immunization[] getImmunizationsVPR(Calendar startDate, Calendar endDate) throws RPCException {
        return getImmunizationsVPR(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Retrieves Skin Tests.
     *
     * @param startDate The start of the date range for the list of immunizations
     * @param endDate   The end of the date range for the list of immunizations
     * @return The list of immunizations for the specified date range
     * @throws RPCException
     */
    public Immunization[] getImmunizationSkinTests(Calendar startDate, Calendar endDate) throws RPCException {
        return getImmunizationSkinTests(cxn.getPatientIen(), startDate, endDate);
    }

    private Immunization[] parseImmunizationsFromVPR(String response) throws RPCException {
        Immunization[] rtc;
        if ("".equals(response)) {
            rtc = new Immunization[0];
        } else {
    
            List<Immunization> immunizationList = new ArrayList<Immunization>();
    
            final String XPATH_IMMUNIZATION = "//immunization";
    
            Document doc = null;
    
            try {
                doc = parseXMLDocument(response);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();
    
            XPathExpression exprImmunization;
    
            try {
                exprImmunization = xpath.compile(XPATH_IMMUNIZATION);
    
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            NodeList immunizationsNodeList = null;
            try {
                immunizationsNodeList = (NodeList) exprImmunization.evaluate(doc, XPathConstants.NODESET);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            List<VPRImmunizationParser> immunizationParserList = new ArrayList<VPRImmunizationParser>();
            int listLength = immunizationsNodeList.getLength();
            for (int immunizationNum = 0; immunizationNum < listLength; immunizationNum++) {
                Node immunizationNode = immunizationsNodeList.item(immunizationNum);
                immunizationNode.getParentNode().removeChild(immunizationNode);
    
                immunizationParserList.add(new VPRImmunizationParser(immunizationNode));
            }

            ExecutorService executorService = cxn.getParserThreadPool().getThreadPool();
            List<Future<Immunization>> futures = null;
            try {
                futures = executorService.invokeAll(immunizationParserList, cxn.getParserThreadPool().getPoolTimeoutSeconds(), TimeUnit.SECONDS);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse immunizations results : accession node parser futures list is null.");
                }
    
            } catch (InterruptedException e) {
                LOGGER.error(e.getMessage(), e);
                throw new RPCException(e);
            }
    
            Iterator<Future<Immunization>> futuresItr = futures.iterator();
    
            while (futuresItr.hasNext()) {
                try {
                    Future<Immunization> future = futuresItr.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR Immunization Data");
                    } else {
                        immunizationList.add(future.get());
                    }
    
                } catch (InterruptedException e) {
                    LOGGER.error(e.getMessage(), e);
                } catch (ExecutionException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
    
            rtc = immunizationList.toArray(new Immunization[0]);
        }
        return rtc;
    }

    private class VPRImmunizationParser implements Callable<Immunization> {
        private Node immunizationNode;

        public VPRImmunizationParser(Node immunizationNode) {
            this.immunizationNode = immunizationNode;
        }

        public Immunization call() throws RPCException {
            final String XPATH_ADMINISTERED = "./administered/@value";
            final String XPATH_CONTRAINDICATED = "./contraindicated/@value";
            final String XPATH_CPT_CODE = "./cpt/@code";
            final String XPATH_CPT_NAME = "./cpt/@name";
            final String XPATH_ID = "./id/@value";
            final String XPATH_NAME = "./name/@value";
            final String XPATH_PROVIDER_NAME = "./provider/@name";
//            final String XPATH_PROVIDER_CODE = "./provider/@code";
            final String XPATH_REACTION = "./reaction/@value";
            final String XPATH_SERIES = "./series/@value";

            final String XPATH_BODY_SITE = "./bodySite/@value";
            final String XPATH_CVX = "./cvx/@value";
            final String XPATH_DOCUMENTED_BY_NAME = "./documentedBy/@name";
            final String XPATH_DOSE = "./dose/@value";
            final String XPATH_EXPIRATION_DATE = "./expirationDate/@value";
            final String XPATH_LOT = "./lot/@value";
            final String XPATH_MANUFACTURER = "./manufacturer/@value";
            final String XPATH_ORDERING_PROVIDER = "./orderingProvider/@name";
            final String XPATH_ROUTE = "./route/@value";
            final String XPATH_SOURCE_NAME = "./source/@name";
            final String XPATH_LOCATION = "./location/@value";

            final String XPATH_COMMENT = "./comment";

//            final String XPATH_VIS = "./vis";


            Immunization immunization = new Immunization();

            XPathExpression exprName = null;
            XPathExpression exprId = null;
            XPathExpression exprAdministered = null;
            XPathExpression exprContraindicated = null;
            XPathExpression exprCptCode = null;
            XPathExpression exprCptName = null;
            XPathExpression exprProviderName = null;
                    //exprProviderCode = null,
            XPathExpression exprReaction = null;
            XPathExpression exprSeries = null;

            XPathExpression exprBodySite = null;
            XPathExpression exprCvx = null;
            XPathExpression exprDocumentedByName = null;
            XPathExpression exprDose = null;
            XPathExpression exprExpirationDate = null;
            XPathExpression exprLot = null;
            XPathExpression exprManufacturer = null;
            XPathExpression exprOrderingProvider = null;
            XPathExpression exprRoute = null;
            XPathExpression exprSourceName = null;
            XPathExpression exprComment = null;
            XPathExpression exprLocation = null;

            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();

            try {
                exprName = xpath.compile(XPATH_NAME);
                exprId = xpath.compile(XPATH_ID);
                exprAdministered = xpath.compile(XPATH_ADMINISTERED);
                exprContraindicated = xpath.compile(XPATH_CONTRAINDICATED);
                exprCptCode = xpath.compile(XPATH_CPT_CODE);
                exprCptName = xpath.compile(XPATH_CPT_NAME);
                exprProviderName = xpath.compile(XPATH_PROVIDER_NAME);
                //exprProviderCode = xpath.compile(XPATH_PROVIDER_CODE);
                exprReaction = xpath.compile(XPATH_REACTION);
                exprSeries = xpath.compile(XPATH_SERIES);

                exprBodySite = xpath.compile(XPATH_BODY_SITE);
                exprCvx = xpath.compile(XPATH_CVX);
                exprDocumentedByName = xpath.compile(XPATH_DOCUMENTED_BY_NAME);
                exprDose = xpath.compile(XPATH_DOSE);
                exprExpirationDate = xpath.compile(XPATH_EXPIRATION_DATE);
                exprLot = xpath.compile(XPATH_LOT);
                exprManufacturer = xpath.compile(XPATH_MANUFACTURER);
                exprOrderingProvider = xpath.compile(XPATH_ORDERING_PROVIDER);
                exprRoute = xpath.compile(XPATH_ROUTE);
                exprSourceName = xpath.compile(XPATH_SOURCE_NAME);
                exprComment = xpath.compile(XPATH_COMMENT);
                exprLocation = xpath.compile(XPATH_LOCATION);

            } catch (Exception e) {
                throw new RPCException(e);
            }

            String immunizationName = null;
            String immunizationIen = null;
            String contraindicatedString = null;
            String cptCode = null;
            String cptName = null;
            String providerName = null;
            //providerCode = null,
            String reaction = null;
            String series = null;
            String administered = null;

            try {
                immunizationName = (String) exprName.evaluate(immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                immunizationName = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                immunizationIen = (String) exprId.evaluate(immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                immunizationIen = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                contraindicatedString = (String) exprContraindicated.evaluate(immunizationNode, XPathConstants.STRING);

            } catch (XPathExpressionException e) {
                contraindicatedString = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                cptCode = (String) exprCptCode.evaluate(immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                cptCode = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                cptName = (String) exprCptName.evaluate(immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                cptName = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                providerName = (String) exprProviderName.evaluate(immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                providerName = null;
                LOGGER.error(e.getMessage(), e);
            }
/*
            try {
                providerCode = (String) exprProviderCode.evaluate( immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                providerCode = null;
                logger.error(e.getMessage(), e);
            }
*/
            try {
                reaction = (String) exprReaction.evaluate(immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                reaction = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                series = (String) exprSeries.evaluate(immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                series = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                administered = (String) exprAdministered.evaluate(immunizationNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                administered = null;
                LOGGER.error(e.getMessage(), e);
            }

            String bodySite = RPCUtil.evaluateNodeToString(immunizationNode, exprBodySite, LOGGER);
            String cvx = RPCUtil.evaluateNodeToString(immunizationNode, exprCvx, LOGGER);
            String documentedByName = RPCUtil.evaluateNodeToString(immunizationNode, exprDocumentedByName, LOGGER);
            String dose = RPCUtil.evaluateNodeToString(immunizationNode, exprDose, LOGGER);
            String expirationDateString = RPCUtil.evaluateNodeToString(immunizationNode, exprExpirationDate, LOGGER);
            String lot = RPCUtil.evaluateNodeToString(immunizationNode, exprLot, LOGGER);
            String manufacturer = RPCUtil.evaluateNodeToString(immunizationNode, exprManufacturer, LOGGER);
            String orderingProvider = RPCUtil.evaluateNodeToString(immunizationNode, exprOrderingProvider, LOGGER);
            String route = RPCUtil.evaluateNodeToString(immunizationNode, exprRoute, LOGGER);
            String sourceName = RPCUtil.evaluateNodeToString(immunizationNode, exprSourceName, LOGGER);
            String comment = RPCUtil.evaluateNodeToString(immunizationNode, exprComment, LOGGER);
            String location = RPCUtil.evaluateNodeToString(immunizationNode, exprLocation, LOGGER);

            immunization.setBodySite(bodySite);
            immunization.setCvx(cvx);
            immunization.setDocumentedBy(documentedByName);
            immunization.setDose(dose);
            Calendar expirationDate = null;
            if (expirationDateString != null) {
                expirationDate = Timestamp.toDateTime(expirationDateString);
            }
            immunization.setLotExpirationDate(expirationDate);
            immunization.setLot(lot);
            immunization.setManufacturerName(manufacturer);
            immunization.setOrderingProvider(orderingProvider);
            immunization.setRoute(route);
            immunization.setSource(sourceName);
            immunization.setComment(comment);
            immunization.setVisitLocation(location);

            immunization.setName(immunizationName);
            immunization.setId(immunizationIen);
            immunization.setEncounterProvider(providerName);
            immunization.setSeries(series);
            immunization.setReaction(reaction);
            //placeholder for DoD results
            immunization.setResult("");
            if (contraindicatedString != null) {
                if ("1".equals(contraindicatedString)) {
                    immunization.setContraindicated(true);
                } else {
                    immunization.setContraindicated(false);
                }
            }

            if (!StringUtils.isEmpty(administered)) {
//                Calendar administeredDate;
//                administeredDate = Timestamp.formatDisplayDateNoBadMonthOrDay(administered);
                String formattedAdministered = Timestamp.formatDisplayDateNoBadMonthOrDay(administered);
                immunization.setDateTime(formattedAdministered);
            }
            List<Code> codeList = new ArrayList<Code>();
            if (!StringUtils.isEmpty(cptCode)) {
                Code immunizationCode = new Code();
                immunizationCode.setCode(cptCode);
                immunizationCode.setDisplay(cptName);
                immunizationCode.setSystem("CPT");
                codeList.add(immunizationCode);
            }

            if (!StringUtils.isEmpty(cvx)) {
                Code immunizationCode = new Code();
                immunizationCode.setCode(cvx);
                immunizationCode.setDisplay("");
                immunizationCode.setSystem("CVX");
                codeList.add(immunizationCode);
            }

            Code immunizationCode = new Code();
            immunizationCode.setCode(immunizationName);
            immunizationCode.setDisplay(immunizationName);
            immunizationCode.setSystem("TEXT");
            codeList.add(immunizationCode);

            immunization.setCodes(codeList.toArray(new Code[0]));

            fillCommonDataBeanValues(immunization);

            return immunization;
        }
    }

}
