/*
 * Janus 4.0 (c)
 * Copyright (c) 2014 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.Code;
import gov.va.med.vds.vistabeans.InsuranceBean;
import gov.va.med.vds.vistabeans.Telecom;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 *
 *
 */
public class InsuranceRPC extends BaseRPC {

    private static final Logger LOGGER = LoggerFactory.getLogger(InsuranceRPC.class);

    public InsuranceRPC(RPCConnection cxn) {
        super(cxn);
    }

    /**
     * Uses the VPR RPC with "insurancePolicies" as the type.
     *
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public InsuranceBean[] getPatientInsurance(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientInsurance(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "insurancePolicies" as the type.
     *
     * @param patientIen The patient IEN
     * @return The list of appointments for the specified date range
     * @throws gov.va.med.rpc.RPCException
     */
    public InsuranceBean[] getPatientInsurance(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetEncountersRequest(patientIen/*, startDate, endDate*/);
        String response = (String) cxn.query(request);

        InsuranceBean[] policies = parseInsurance(response);

        return policies;
    }

    /**
     * Builds the VPR RPC request string using "insurancePolicies" as the type. The problems contain the list of
     * appointments. The RPC will fail if the "VPR APPLICATION PROXY" context is not set. That is set when
     * setting up the connection.
     *
     * @param patientIen Patient IEN
     * @param startDate  Start date of range
     * @param endDate    End date of range
     * @return The RPC response message string
     */
    private String buildGetEncountersRequest(String patientIen/*, Calendar startDate, Calendar endDate*/) {
        return VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, VPRUtil.VPRType.INSURANCE_POLICIES);
    }

    private InsuranceBean[] parseInsurance(String response) throws RPCException {
        InsuranceBean[] rtc;
        if ("".equals(response)) {
            rtc = new InsuranceBean[0];
        } else {
    
            StringBuilder sbld = new StringBuilder(response);
            sbld.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    
            ArrayList<InsuranceBean> insuranceList = new ArrayList<InsuranceBean>();
    
            final String XPATH_INSURANCE_POLICY = "//insurancePolicy";
    
            Document document = null;
            XPathExpression exprInsurancePolicy = null;
    
            try {
                document = parseXMLDocument(sbld.toString());
                XPathFactory xPathFactory = XPathFactory.newInstance();
                XPath xpath = xPathFactory.newXPath();
                exprInsurancePolicy = xpath.compile(XPATH_INSURANCE_POLICY);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            NodeList policiesNodeList = null;
            try {
                policiesNodeList = (NodeList) exprInsurancePolicy.evaluate(document, XPathConstants.NODESET);
            } catch (XPathExpressionException e) {
                throw new RPCException(e);
            }
    
            List<VPRInsurancePoliciesParser> policiesParserList = new ArrayList<VPRInsurancePoliciesParser>();
            int policiesNodeListLength = policiesNodeList.getLength();
            for (int policyNum = 0; policyNum < policiesNodeListLength; policyNum++) {
                Node policyNode = policiesNodeList.item(policyNum);
                policyNode.getParentNode().removeChild(policyNode);
    
                policiesParserList.add(new VPRInsurancePoliciesParser(policyNode));
            }
    
            ExecutorService executorService = cxn.getParserThreadPool().getThreadPool();
            List<Future<InsuranceBean>> futures = null;
            try {
                futures = executorService.invokeAll(policiesParserList, cxn.getParserThreadPool().getPoolTimeoutSeconds(), TimeUnit.SECONDS);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse insurance results : insurance node parser futures list is null.");
                }
            } catch (InterruptedException e) {
                RPCUtil.logErrorMessage(LOGGER, e);
                throw new RPCException(e);
            }
    
            Iterator<Future<InsuranceBean>> futureIterator = futures.iterator();
    
            while (futureIterator.hasNext()) {
                try {
                    Future<InsuranceBean> future = futureIterator.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR Insurance Data");
                    } else {
                        InsuranceBean policy = future.get();
    
                        // exclude null results
                        if (policy != null) {
                            insuranceList.add(policy);
                        }
                    }
                } catch (InterruptedException ie) {
                    RPCUtil.logErrorMessage(LOGGER, ie);
                } catch (ExecutionException ee) {
                    RPCUtil.logErrorMessage(LOGGER, ee);
                }
            }
    
            rtc = insuranceList.toArray(new InsuranceBean[0]);
        }
        return rtc;
    }

    private class VPRInsurancePoliciesParser implements Callable<InsuranceBean> {
        private static final String ALPHA_ONLY_REGEX = "[a-zA-Z]+";
        private static final String NUMERIC_ONLY_REGEX = "[0-9]+";
        private Node policyNode;
        private Pattern numbersOnlyPattern =  Pattern.compile(NUMERIC_ONLY_REGEX);
        private Pattern alphaOnlyPattern = Pattern.compile(ALPHA_ONLY_REGEX);

        public VPRInsurancePoliciesParser(Node policyNode) {
            this.policyNode = policyNode;
        }

        public InsuranceBean call() throws Exception {
            final String XPATH_EFFECTIVE_DATE = "./effectiveDate/@value";
            final String XPATH_EXPIRATION_DATE = "./expirationDate/@value";
            final String XPATH_ID = "./id/@value";
            final String XPATH_COMPANY = "./company";
            final String XPATH_SUBSCRIBER = "./subscriber";

            final String XPATH_CODE = "./@code";
            final String XPATH_NAME = "./@name";
            final String XPATH_ADDRESS = "./address";
            final String XPATH_DOB = "./dob";
            final String XPATH_TELECOM_LIST = "./telecomList";
            final String XPATH_STREET_LINE1 = "./@streetLine1";
            final String XPATH_STREET_LINE2 = "./@streetLine2";
            final String XPATH_STREET_LINE3 = "./@streetLine3";
            final String XPATH_CITY = "./@city";
            final String XPATH_STATE_PROVINCE = "./@stateProvince";
            final String XPATH_POSTAL_CODE = "./@postalCode";
            final String XPATH_TELECOM = "./telecom";
            final String XPATH_USAGE_TYPE = "./@usageType";
            final String XPATH_VALUE = "./@value";

            final String XPATH_GROUP_NAME = "./groupName/@value";
            final String XPATH_GROUP_NUMBER = "./groupNumber/@value";
            final String XPATH_INSURANCE_TYPE_CODE = "./insuranceType/@code";
            final String XPATH_INSURANCE_TYPE_NAME = "./insuranceType/@name";
            final String XPATH_MEMBER_ID = "./memberID/@value";
            final String XPATH_RELATIONSHIP = "./relationship/@value";

            XPathExpression exprEffectiveDate = null;
            XPathExpression exprExpirationDate = null;
            XPathExpression exprId = null;
            XPathExpression exprMemberId = null;
            XPathExpression exprCompany = null;
            XPathExpression exprSubscriber = null;
            XPathExpression exprName = null;
            XPathExpression exprCode = null;
            XPathExpression exprDob = null;
            XPathExpression exprAddress = null;
            XPathExpression exprStreetLine1 = null;
            XPathExpression exprStreetLine2 = null;
            XPathExpression exprStreetLine3 = null;
            XPathExpression exprCity = null;
            XPathExpression exprStateProvince = null;
            XPathExpression exprPostalCode = null;
            XPathExpression exprTelecom = null;
            XPathExpression exprUsageType = null;
            XPathExpression exprValue = null;
            XPathExpression exprTelecomList = null;
            XPathExpression exprGroupName = null;
            XPathExpression exprGroupNumber = null;
            XPathExpression exprInsuranceTypeCode = null;
            XPathExpression exprInsuranceTypeName = null;
            XPathExpression exprRelationship = null;


            XPathFactory xPathFactory = XPathFactory.newInstance();
            XPath xpath = xPathFactory.newXPath();

            try {
                exprEffectiveDate = xpath.compile(XPATH_EFFECTIVE_DATE);
                exprExpirationDate = xpath.compile(XPATH_EXPIRATION_DATE);
                exprId = xpath.compile(XPATH_ID);
                exprCompany = xpath.compile(XPATH_COMPANY);
                exprSubscriber = xpath.compile(XPATH_SUBSCRIBER);

                exprCode = xpath.compile(XPATH_CODE);
                exprName = xpath.compile(XPATH_NAME);
                exprDob = xpath.compile(XPATH_DOB);
                exprAddress = xpath.compile(XPATH_ADDRESS);
                exprTelecomList = xpath.compile(XPATH_TELECOM_LIST);
                exprStreetLine1 = xpath.compile(XPATH_STREET_LINE1);
                exprStreetLine2 = xpath.compile(XPATH_STREET_LINE2);
                exprStreetLine3 = xpath.compile(XPATH_STREET_LINE3);
                exprCity = xpath.compile(XPATH_CITY);
                exprStateProvince = xpath.compile(XPATH_STATE_PROVINCE);
                exprPostalCode = xpath.compile(XPATH_POSTAL_CODE);
                exprTelecom = xpath.compile(XPATH_TELECOM);
                exprUsageType = xpath.compile(XPATH_USAGE_TYPE);
                exprValue = xpath.compile(XPATH_VALUE);

                exprGroupName = xpath.compile(XPATH_GROUP_NAME);
                exprGroupNumber = xpath.compile(XPATH_GROUP_NUMBER);
                exprInsuranceTypeCode = xpath.compile(XPATH_INSURANCE_TYPE_CODE);
                exprInsuranceTypeName = xpath.compile(XPATH_INSURANCE_TYPE_NAME);
                exprMemberId = xpath.compile(XPATH_MEMBER_ID);
                exprRelationship = xpath.compile(XPATH_RELATIONSHIP);

            } catch (XPathExpressionException e) {
                throw new RPCException(e);
            }

            InsuranceBean policy = new InsuranceBean();
            fillCommonDataBeanValues(policy);

            String effectiveDateString = null;
            Calendar effectiveDateCalendar = null;
            try {
                effectiveDateString = (String) exprEffectiveDate.evaluate(policyNode, XPathConstants.STRING);
                if (effectiveDateString != null) {
                    effectiveDateCalendar = Timestamp.toDateTime(effectiveDateString);
                }
            } catch (XPathExpressionException e) {
                RPCUtil.logErrorMessage(LOGGER, e);
            }
            String expirationDateString = null;
            Calendar expirationDateCalendar = null;
            try {
                expirationDateString = (String) exprExpirationDate.evaluate(policyNode, XPathConstants.STRING);
                if (expirationDateString != null) {
                    expirationDateCalendar = Timestamp.toDateTime(expirationDateString);
                }
            } catch (XPathExpressionException e) {
                RPCUtil.logErrorMessage(LOGGER, e);
            }

//            String id = RPCUtil.evaluateNodeToString(policyNode, exprId);
            String groupName = RPCUtil.evaluateNodeToString(policyNode, exprGroupName, LOGGER);
            String groupNumber = RPCUtil.evaluateNodeToString(policyNode, exprGroupNumber, LOGGER);
            String insuranceTypeCode = RPCUtil.evaluateNodeToString(policyNode, exprInsuranceTypeCode, LOGGER);
            String insuranceTypeName = RPCUtil.evaluateNodeToString(policyNode, exprInsuranceTypeName, LOGGER);
            String memberId = RPCUtil.evaluateNodeToString(policyNode, exprMemberId, LOGGER);
            String relationship = RPCUtil.evaluateNodeToString(policyNode, exprRelationship, LOGGER);

            Node subscriberNode = RPCUtil.evaluateNodeToNode(policyNode, exprSubscriber, LOGGER);
            String subscriberName = null;
            String subscriberId = null;
            String subscriberDobString = null;
            Calendar subscriberDob = null;
            if (subscriberNode != null) {
                subscriberName = RPCUtil.evaluateNodeToString(subscriberNode, exprName, LOGGER);
                subscriberId = RPCUtil.evaluateNodeToString(subscriberNode, exprCode, LOGGER);
                subscriberDobString = RPCUtil.evaluateNodeToString(subscriberNode, exprDob, LOGGER);
                if (subscriberDobString != null) {
                    subscriberDob = Timestamp.toDateTime(subscriberDobString);
                }
            }

            Node companyNode = RPCUtil.evaluateNodeToNode(policyNode, exprCompany, LOGGER);
            String companyName = null;
            String companyId = null;
            String companyStreetLine1 = null;
            String companyStreetLine2 = null;
            String companyStreetLine3 = null;
            String companyCity = null;
            String companyStateProvince = null;
            String companyPostalCode = null;
            if (companyNode != null) {
                companyName = RPCUtil.evaluateNodeToString(companyNode, exprName, LOGGER);
                companyId = RPCUtil.evaluateNodeToString(companyNode, exprId, LOGGER);

                Node companyAddressNode = RPCUtil.evaluateNodeToNode(companyNode, exprAddress, LOGGER);
                if (companyAddressNode != null) {
                    companyStreetLine1 = RPCUtil.evaluateNodeToString(companyAddressNode, exprStreetLine1, LOGGER);
                    companyStreetLine2 = RPCUtil.evaluateNodeToString(companyAddressNode, exprStreetLine2, LOGGER);
                    companyStreetLine3 = RPCUtil.evaluateNodeToString(companyAddressNode, exprStreetLine3, LOGGER);
                    companyCity = RPCUtil.evaluateNodeToString(companyAddressNode, exprCity, LOGGER);
                    companyStateProvince = RPCUtil.evaluateNodeToString(companyAddressNode, exprStateProvince, LOGGER);
                    companyPostalCode = RPCUtil.evaluateNodeToString(companyAddressNode, exprPostalCode, LOGGER);

                    // Due to a bug in the VPR RPC, the postalCode is the State name and the
                    // stateProvince is the IEN for the state. For now we'll "fix" the data by putting the
                    // postalCode as the state and have no postalCode.
                    Matcher stateProvAsNumbersMatcher = numbersOnlyPattern.matcher(companyStateProvince);
                    Matcher postalCodeAsAlphaMatcher = alphaOnlyPattern.matcher(companyPostalCode);
                    if (stateProvAsNumbersMatcher.matches() && postalCodeAsAlphaMatcher.matches()) {
                        companyStateProvince = companyPostalCode;
                        companyPostalCode = "";
                    }

                    policy.setCompanyAddressLine1(companyStreetLine1);
                    policy.setCompanyAddressLine2(companyStreetLine2);
                    policy.setCompanyAddressLine3(companyStreetLine3);
                    policy.setCompanyCity(companyCity);
                    policy.setCompanyStateProvince(companyStateProvince);
                    policy.setCompanyPostalCode(companyPostalCode);
                }

                Node telecomListNode = RPCUtil.evaluateNodeToNode(companyNode, exprTelecomList, LOGGER);
                if (telecomListNode != null) {
                    NodeList telecomNodeList = RPCUtil.evaluateNodeToNodeList(telecomListNode, exprTelecom, LOGGER);

                    if (telecomNodeList != null) {
                        int telecomNodeListLength = telecomNodeList.getLength();
                        List<Telecom> telecomList = new ArrayList<Telecom>();
                        for (int telecomNum = 0; telecomNum < telecomNodeListLength; telecomNum++) {
                            Node telecomNode = telecomNodeList.item(telecomNum);
                            telecomNode.getParentNode().removeChild(telecomNode);

                            String telecomValue = RPCUtil.evaluateNodeToString(telecomNode, exprValue, LOGGER);
                            String telecomUsageType = RPCUtil.evaluateNodeToString(telecomNode, exprUsageType, LOGGER);

                            Telecom telecom = new Telecom();
                            telecom.setValue(telecomValue);
                            telecom.setUsageType(telecomUsageType);

                            telecomList.add(telecom);
                        }

                        if (!telecomList.isEmpty()) {
                            policy.setCompanyTelecom(telecomList.toArray(new Telecom[0]));
                        }
                    }
                }

                policy.setCompanyName(companyName);
                policy.setCompanyIen(companyId);

            }

            policy.setEffectiveDate(effectiveDateCalendar);
            policy.setExpirationDate(expirationDateCalendar);

            policy.setPlanName(groupName);
            policy.setGroupNumber(groupNumber);
            policy.setInsuranceType(insuranceTypeName);
            policy.setInsuranceTypeId(insuranceTypeCode);
            policy.setRelationShip(relationship);
            policy.setSubscriberId(subscriberId);
            policy.setSubscriberDOB(subscriberDob);
            policy.setPolicyHolder(subscriberName);
            policy.setMemberId(memberId);

            List<Code> codeList = new ArrayList<Code>();

            if (insuranceTypeCode != null && !StringUtils.isEmpty(insuranceTypeCode)) {
                Code insType = new Code();
                insType.setSystem("VUID");
                insType.setCode(insuranceTypeCode);
                codeList.add(insType);
            }

            if(insuranceTypeName != null && !StringUtils.isEmpty(insuranceTypeName)){
                Code insType = new Code();
                insType.setSystem("TEXT");
                insType.setCode(insuranceTypeName);
                codeList.add(insType);
            }

            if(codeList.size() > 0) {
                policy.setCodes(codeList.toArray(new Code[codeList.size()]));
            }


            return policy;
        }
    }

}
