/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.ObjectMapper;
import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.PCE;
import gov.va.med.vhahon.common.Utils;
import org.codehaus.jackson.JsonParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * SkinTest RPC data access object.
 */
public class PCERPC extends BaseRPC {
    private static final Logger LOGGER = LoggerFactory.getLogger(PCERPC.class);

    public PCERPC(RPCConnection cxn) {
        super(cxn);
    }

    /**
     * Uses the VPR RPC with "skinTests" as the type.
     *
     * @param startDate The start of the date range for the list of skin tests
     * @param endDate   The end of the date range for the list of skin tests
     * @return The list of skin tests for the specified date range
     * @throws RPCException
     */
    public PCE[] getSkinTests(Calendar startDate, Calendar endDate) throws RPCException {
        return getSkinTests(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "skinTests" as the type.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of skin tests
     * @param endDate    The end of the date range for the list of skin tests
     * @return The list of skin tests for the specified date range
     * @throws RPCException
     */
    public PCE[] getSkinTests(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
//        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.SKIN_TESTS);
//        String request = VPRUtil.buildVPRJSONRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.SKIN);
        String request = buildSkinTestReportRequest(patientIen);
        String response = (String) cxn.query(request);
        PCE[] rtc;
        PCE[] allSkinTests = toPCESkinTests(response);

        // filter for start and end date on admin(entered) date
        ArrayList <PCE> filteredSkinTests = new ArrayList<PCE>();
        for (PCE pce : allSkinTests) {
            Calendar adminDate = pce.getEnteredDate();
            if (adminDate.compareTo(startDate) >= 0 && adminDate.compareTo(endDate) <= 0) {
                filteredSkinTests.add(pce);
            }
        }
        if (filteredSkinTests.isEmpty()) {
            rtc = new PCE[0];
        } else {
            rtc = filteredSkinTests.toArray(new PCE[0]);
        }
        return rtc;

//        return parseJSONSkinTests(response);
//        return parsePCEs(response, "skinTest");
    }

    private String buildSkinTestReportRequest (String patientIen) {
        VistaQuery vistaQuery = vistaQueryFactory.getVistaQuery("ORWRP REPORT TEXT");
        vistaQuery.addParameter(VistaQuery.LITERAL, patientIen);
        vistaQuery.addParameter(VistaQuery.LITERAL, "OR_ST:SKIN TESTS~;;208;101");
        vistaQuery.addParameter(VistaQuery.LITERAL, "");
        vistaQuery.addParameter(VistaQuery.LITERAL, "50000");
        vistaQuery.addParameter(VistaQuery.LITERAL, "");
        vistaQuery.addParameter(VistaQuery.LITERAL, "0");
        vistaQuery.addParameter(VistaQuery.LITERAL, "0");

        return vistaQuery.buildMessage();
    }

    private static final String DATE_MM_DD_YYYY_REGEX = "(0[1-9]|1[012])[- /.](0[1-9]|[12][0-9]|3[01])[- /.](19|20)\\d\\d";
    private static final String READING_VALUE_REGEX = "\\d{1,2} mm";
    private static final String START_TABLE_MARKER = "Skin Test   Reading  Results     Admin.      Reading     Facility";
    private static final String END_TABLE_MARKER = "*** END ***************  CONFIDENTIAL  SUMMARY   pg. 1 ***********************";
    private static final int START_OF_READING_INDEX = 12;
    private static final int START_OF_RESULTS_INDEX = 21;
    private static final int START_OF_ADMIN_DATE_INDEX = 33;
    private static final int START_OF_READING_DATE_INDEX = 45;
    private static final int START_OF_FACILITY_INDEX = 57;
    private PCE[] toPCESkinTests(String skinTestReport) {
        ArrayList<PCE> rtc = new ArrayList<PCE>();
        Pattern datePattern = Pattern.compile(DATE_MM_DD_YYYY_REGEX);
//        Pattern readingPattern = Pattern.compile(READING_VALUE_REGEX);
        SimpleDateFormat sdf = new SimpleDateFormat("MM/dd/yyyy");
        int startOfLineIndex = skinTestReport.indexOf(START_TABLE_MARKER);
        // skip the table header
        startOfLineIndex = skinTestReport.indexOf("\n", startOfLineIndex) + 1;
        // skip the first line
        startOfLineIndex = skinTestReport.indexOf("\n", startOfLineIndex) + 1;
        int endOfLineIndex = skinTestReport.indexOf("\n", startOfLineIndex);
        String nextLine = null;
        boolean inTable = true;
        String testName = null;
        while (inTable) {
            nextLine = skinTestReport.substring(startOfLineIndex, endOfLineIndex);
            startOfLineIndex = endOfLineIndex + 1;
            endOfLineIndex = skinTestReport.indexOf("\n", startOfLineIndex);
            if (nextLine.length() < 2 || nextLine.contains(END_TABLE_MARKER)) {
                inTable = false;
            }
            // if line has a reading value and an Admin Date then it is not a comment
            if (nextLine.length() > START_OF_READING_DATE_INDEX) {
                String adminDateString = nextLine.substring(START_OF_ADMIN_DATE_INDEX, START_OF_READING_DATE_INDEX).trim();
                Matcher dateMatcher = datePattern.matcher(adminDateString);
                if (dateMatcher.matches()) {
                    // check if has reading
                    String readingString = nextLine.substring(START_OF_READING_INDEX, START_OF_RESULTS_INDEX).trim();

                    // this is a line to parse
                    String testNameString = nextLine.substring(0, START_OF_READING_INDEX).trim();
                    if (testNameString.length() > 0) {
                        testName = testNameString;
                    }
                    String resultsString = nextLine.substring(START_OF_RESULTS_INDEX, START_OF_ADMIN_DATE_INDEX).trim();
                    String readingDateString = nextLine.substring(START_OF_READING_DATE_INDEX, START_OF_FACILITY_INDEX).trim();

                    PCE skinTestPCE = new PCE();
                    fillCommonDataBeanValues(skinTestPCE);
                    skinTestPCE.setName(testName);
                    skinTestPCE.setReading(readingString);
                    Date adminDate = null;
                    try {
                        adminDate = sdf.parse(adminDateString);
                    } catch (ParseException pe) {
                        LOGGER.error("Error parsing skin test admin date: " + pe.getMessage());
                    }
                    if (adminDate != null) {
                        Calendar adminCalendar = Calendar.getInstance();
                        adminCalendar.setTime(adminDate);
                        skinTestPCE.setEnteredDate(adminCalendar);
                    }
                    Matcher readingDateMatcher = datePattern.matcher(readingDateString);
                    if (readingDateMatcher.matches()) {
                        Date readingDate = null;
                        try {
                            readingDate = sdf.parse(readingDateString);
                        } catch (ParseException pe) {
                            LOGGER.error("Error parsing skin test reading date: " + pe.getMessage());
                        }
                        if (readingDate != null) {
                            Calendar readingCalendar = Calendar.getInstance();
                            readingCalendar.setTime(readingDate);
                            skinTestPCE.setReadDate(readingCalendar);
                        }
                    }
                    skinTestPCE.setResult(resultsString);
                    // add the skin test report as the comment
                    skinTestPCE.setComment(skinTestReport);

                    rtc.add(skinTestPCE);
                }
            }
        }

        return rtc.toArray(new PCE[0]);
    }

    /**
     * Uses the VPR RPC with "exams" as the type.
     *
     * @param startDate The start of the date range for the list of exams
     * @param endDate   The end of the date range for the list of exams
     * @return The list of exams for the specified date range
     * @throws RPCException
     */
    public PCE[] getExams(Calendar startDate, Calendar endDate) throws RPCException {
        return getExams(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "exams" as the type.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of exams
     * @param endDate    The end of the date range for the list of exams
     * @return The list of exams for the specified date range
     * @throws RPCException
     */
    public PCE[] getExams(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.EXAMS);
        String response = (String) cxn.query(request);

        return parsePCEs(response, "exam");
    }

    /**
     * Uses the VPR RPC with "educationTopics" as the type.
     *
     * @param startDate The start of the date range for the list of education topics
     * @param endDate   The end of the date range for the list of education topics
     * @return The list of education topics for the specified date range
     * @throws RPCException
     */
    public PCE[] getEducationTopics(Calendar startDate, Calendar endDate) throws RPCException {
        return getEducationTopics(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "educationTopics" as the type.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of education topics
     * @param endDate    The end of the date range for the list of education topics
     * @return The list of education topics for the specified date range
     * @throws RPCException
     */
    public PCE[] getEducationTopics(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.EDUCATION_TOPICS);
        String response = (String) cxn.query(request);

        return parsePCEs(response, "educationTopic");
    }


    private PCE[] parsePCEs(String response, String type) throws RPCException {
        PCE[] rtc;
        if ("".equals(response)) {
            rtc = new PCE[0];
        } else {

            StringBuilder sbld = new StringBuilder(Utils.stripInvalidXMLCharacters(response));
            sbld.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");

            List<PCE> pceList = new ArrayList<PCE>();
    
            final String XPATH_PCE = "//" + type;

            Document doc = null;
    
            try {
                doc = parseXMLDocument(sbld.toString());
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();
    
            XPathExpression exprPce;
    
            try {
                exprPce = xpath.compile(XPATH_PCE);
    
            } catch (Exception e) {
                LOGGER.error(e.getMessage(), e);
                throw new RPCException(e);
            }
    
            NodeList pcesNodeList = null;
            try {
                pcesNodeList = (NodeList) exprPce.evaluate(doc, XPathConstants.NODESET);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            List<VPRPCEParser> pceParserList = new ArrayList<VPRPCEParser>();
            int listLength = pcesNodeList.getLength();
            for (int pceNum = 0; pceNum < listLength; pceNum++) {
                Node pceNode = pcesNodeList.item(pceNum);
                pceNode.getParentNode().removeChild(pceNode);
    
                pceParserList.add(new VPRPCEParser(pceNode));
            }
    
            ExecutorService executorService = cxn.getParserThreadPool().getThreadPool();
            List<Future<PCE>> futures = null;
            try {
                futures = executorService.invokeAll(pceParserList, cxn.getParserThreadPool().getPoolTimeoutSeconds(), TimeUnit.SECONDS);
    
                if (futures == null) {
                    StringBuilder exceptionMessage = new StringBuilder("Failed to parse pce results: ");
                    exceptionMessage.append(type);
                    exceptionMessage.append(" node parser futures list is null.");
                    throw new RPCException(exceptionMessage.toString());
                }
    
            } catch (InterruptedException e) {
                LOGGER.error(e.getMessage(), e);
                throw new RPCException(e);
            }
    
            Iterator<Future<PCE>> futuresItr = futures.iterator();
    
            while (futuresItr.hasNext()) {
                try {
                    Future<PCE> future = futuresItr.next();
                    if (future.isCancelled()) {
                        StringBuilder exceptionMessage = new StringBuilder("Failed to parse ");
                        exceptionMessage.append(type);
                        exceptionMessage.append(" Data.");
                        throw new RPCException(exceptionMessage.toString());
                    } else {
                        pceList.add(future.get());
                    }
    
                } catch (InterruptedException e) {
                    LOGGER.error(e.getMessage(), e);
                } catch (ExecutionException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
    
            rtc = pceList.toArray(new PCE[0]);
        }
        return rtc;
    }

    private class VPRPCEParser implements Callable<PCE> {
        private Node pceNode;

        public VPRPCEParser(Node pceNode) {
            this.pceNode = pceNode;
        }

        public PCE call() throws RPCException {
            final String XPATH_NAME = "./name/@value";
            final String XPATH_ID = "./id/@value";
            final String XPATH_ENCOUNTER = "./encounter/@value";
            final String XPATH_DATE_TIME = "./dateTime/@value";
            final String XPATH_COMMENT = "./comment/@value";
            final String XPATH_RESULT = "./result/@value";

            PCE pce = new PCE();
            fillCommonDataBeanValues(pce);

            XPathExpression exprName = null;
            XPathExpression exprId = null;
            XPathExpression exprEncounter = null;
            XPathExpression exprDateTime = null;
            XPathExpression exprComment = null;
            XPathExpression exprResult = null;

            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();

            try {
                exprName = xpath.compile(XPATH_NAME);
                exprId = xpath.compile(XPATH_ID);
                exprEncounter = xpath.compile(XPATH_ENCOUNTER);
                exprDateTime = xpath.compile(XPATH_DATE_TIME);
                exprComment = xpath.compile(XPATH_COMMENT);
                exprResult = xpath.compile(XPATH_RESULT);
            } catch (Exception e) {
                throw new RPCException(e);
            }

            String name = RPCUtil.evaluateNodeToString(pceNode, exprName, LOGGER);
            String id = RPCUtil.evaluateNodeToString(pceNode, exprId, LOGGER);
            String encounter = RPCUtil.evaluateNodeToString(pceNode, exprEncounter, LOGGER);
            String dateTimeString = RPCUtil.evaluateNodeToString(pceNode, exprDateTime, LOGGER);
            String comment = RPCUtil.evaluateNodeToString(pceNode, exprComment, LOGGER);
            String result = RPCUtil.evaluateNodeToString(pceNode, exprResult, LOGGER);

            pce.setName(name);
            pce.setIen(id);
            pce.setEncounterIen(encounter);
            pce.setComment(comment);
            pce.setResult(result);

            Calendar dateTimeCalendar = null;
            if (dateTimeString != null) {
                dateTimeCalendar = Timestamp.toDateTime(dateTimeString);
            }
            pce.setEnteredDate(dateTimeCalendar);

            return pce;
        }
    }

    private PCE[] parseJSONSkinTests (String response) throws RPCException {
        JsonParser jsonParser;
        ObjectMapper objectMapper = new ObjectMapper();
//        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        List<PCE> skinTests = new ArrayList<PCE>();

        try {
            SkinTestJSON skinTestJSON = objectMapper.readValue(response, SkinTestJSON.class);

            if (skinTestJSON != null && skinTestJSON.getData() != null && skinTestJSON.getData().getItems() != null && !skinTestJSON.getData().getItems().isEmpty()) {
                StringBuilder readingSB = new StringBuilder();
                for (ItemJSON itemJSON : skinTestJSON.getData().getItems()) {
                    PCE pce = new PCE();
                    fillCommonDataBeanValues(pce);

                    pce.setName(itemJSON.getName());
                    if (!StringUtils.isEmpty(itemJSON.getReading())) {
                        readingSB.setLength(0);
                        readingSB.append(itemJSON.getReading());
                        readingSB.append(" mm");
                        pce.setReading(readingSB.toString());
                    }
                    pce.setResult(itemJSON.getResult());
                    pce.setComment(itemJSON.getComment());
                    pce.setIen(itemJSON.getLocalId());

                    if (itemJSON.getEncounterUid() != null) {
                        int lastColonIndex = itemJSON.getEncounterUid().lastIndexOf(':');
                        if (lastColonIndex > 0) {
                            pce.setEncounterIen(itemJSON.getEncounterUid().substring(lastColonIndex + 1));
                        }
                    }

                    if (itemJSON.getEntered() != null) {
                        pce.setEnteredDate(Timestamp.toDateFromVPRJSONTimestamp(itemJSON.getEntered()));
                    }

                    if (itemJSON.getDateRead() != null) {
                        pce.setReadDate(Timestamp.toDateFromVPRJSONTimestamp(itemJSON.getDateRead()));
                    }

                    skinTests.add(pce);
                }

            }

        } catch (Exception jpe) {
            String errorString = "Cannot parse JSON VPR skin tests, siteCode: " + cxn.getSiteCode();
            LOGGER.error(errorString);
            throw new RPCException(errorString);
        }


        return skinTests.toArray(new PCE[0]);
    }


    @JsonIgnoreProperties(ignoreUnknown = true)
    private static class SkinTestJSON {
        @JsonProperty("data")
        private DataJSON data;

        public DataJSON getData() {
            return data;
        }

        public void setData(DataJSON data) {
            this.data = data;
        }
    }

    @JsonIgnoreProperties(ignoreUnknown = true)
    private static class DataJSON {
        @JsonProperty("items")
        private List<ItemJSON> items;

        public List<ItemJSON> getItems() {
            return items;
        }

        public void setItems(List<ItemJSON> items) {
            this.items = items;
        }
    }

    @JsonIgnoreProperties(ignoreUnknown = true)
    private static class ItemJSON {
        @JsonProperty("comment")
        private String comment;
        @JsonProperty("entered")
        private String entered;
        @JsonProperty("dateRead")
        private String dateRead;
        @JsonProperty("localId")
        private String localId;
        @JsonProperty("locationName")
        private String locationName;
        @JsonProperty("locationUid")
        private String locationUid;
        @JsonProperty("name")
        private String name;
        @JsonProperty("encounterUid")
        private String encounterUid;
        @JsonProperty("encounterName")
        private String encounterName;
        @JsonProperty("reading")
        private String reading;
        @JsonProperty("result")
        private String result;

        public String getComment() {
            return comment;
        }

        public void setComment(String comment) {
            this.comment = comment;
        }

        public String getEntered() {
            return entered;
        }

        public void setEntered(String entered) {
            this.entered = entered;
        }

        public String getDateRead() {
            return dateRead;
        }

        public void setDateRead(String dateRead) {
            this.dateRead = dateRead;
        }

        public String getLocalId() {
            return localId;
        }

        public void setLocalId(String localId) {
            this.localId = localId;
        }

        public String getLocationName() {
            return locationName;
        }

        public void setLocationName(String locationName) {
            this.locationName = locationName;
        }

        public String getLocationUid() {
            return locationUid;
        }

        public void setLocationUid(String locationUid) {
            this.locationUid = locationUid;
        }

        public String getName() {
            return name;
        }

        public void setName(String name) {
            this.name = name;
        }

        public String getEncounterUid() {
            return encounterUid;
        }

        public void setEncounterUid(String encounterUid) {
            this.encounterUid = encounterUid;
        }

        public String getEncounterName() {
            return encounterName;
        }

        public void setEncounterName(String encounterName) {
            this.encounterName = encounterName;
        }

        public String getReading() {
            return reading;
        }

        public void setReading(String reading) {
            this.reading = reading;
        }

        public String getResult() {
            return result;
        }

        public void setResult(String result) {
            this.result = result;
        }
    }

}
