/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.Code;
import gov.va.med.vds.vistabeans.RadiologyExam;
import gov.va.med.vds.vistabeans.RadiologyReport;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 *
 */
public class RadiologyRPC extends BaseRPC {

    private static final Logger LOGGER = LoggerFactory.getLogger(RadiologyRPC.class);

    public RadiologyRPC(RPCConnection cxn) {
        super(cxn);
    }

    public RadiologyReport[] getRadiologyExams(Calendar startDate, Calendar endDate) throws RPCException {
        return getRadiologyExams(cxn.getPatientIen(), startDate, endDate);
    }

    //SB: Missing exam id
    public RadiologyReport[] getRadiologyExams(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetExamListRequest(patientIen);
        String response = (String) cxn.query(request);
        return filterRadsByDate(toRadiologyExam(response/*, patientIen*/), startDate, endDate);
    }

    private RadiologyReport[] filterRadsByDate(RadiologyReport[] rads, Calendar startDate, Calendar endDate) {
        ArrayList<RadiologyReport> radsFiltered = new ArrayList<RadiologyReport>();

        //filter by date range
        for (RadiologyReport rad : rads) {
            Calendar radDate = rad.getExamDate();

            if (radDate.compareTo(startDate) >= 0 && radDate.compareTo(endDate) <= 0) {
                radsFiltered.add(rad);
            }
        }

        return radsFiltered.toArray(new RadiologyReport[0]);
    }

    private String buildGetExamListRequest(String patientIen) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWRA IMAGING EXAMS1");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        return vquery.buildMessage();
    }

    public RadiologyReport getRadiologyReport(String examId) throws RPCException {
        return getRadiologyReport(cxn.getPatientIen(), examId);
    }

    public RadiologyReport getRadiologyReport(String patientIen, String examId) throws RPCException {
        String request = buildGetReportRequest(patientIen, (String) examId);
        String reportText = (String) cxn.query(request);

        RadiologyReport report = new RadiologyReport();
        report.setReportText(reportText);

        fillCommonDataBeanValues(report);

        return report;
    }

    private String buildGetReportRequest(String patientIen, String id) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWRP REPORT TEXT");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        vquery.addParameter(VistaQuery.LITERAL, "18:IMAGING (LOCAL ONLY)~");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, id);
        vquery.addParameter(VistaQuery.LITERAL, "0");
        vquery.addParameter(VistaQuery.LITERAL, "0");
        return vquery.buildMessage();
    }

    private RadiologyReport[] toRadiologyExam(String response/*, String patientIen*/) {
        RadiologyReport[] rtc;
        if ("".equals(response)) {
            rtc = new RadiologyReport[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            ArrayList<RadiologyExam> lst = new ArrayList<RadiologyExam>();
            RadiologyReport rec = null;
            for (int i = 0; i < lines.length; i++) {
                if ("".equals(lines[i])) {
                    continue;
                }
                String[] flds = StringUtils.split(lines[i], StringUtils.CARET);
    
                for (int j = 0; j < flds.length; j++) {
                    String fldItem = flds[j];
                    if (fldItem != null) {
                        switch (j) {
                            //site
                            case 0:
                                rec = new RadiologyReport();
//                                String[] parts = StringUtils.split(fldItem, StringUtils.SEMICOLON);
    
                                fillCommonDataBeanValues(rec);
                                break;
                            //exam id
                            case 1:
                                if (rec != null) {
                                    rec.setExamId(fldItem);
                                }
                                break;
                            //exam date and time
                            case 2:
                                if (rec != null) {
                                    rec.setExamDate(Timestamp.toDateTime(fldItem));
                                }
                                break;
                            //Procedure name
                            case 3:
                                if (rec != null) {
                                    rec.setStudy(fldItem);
                                }
                                break;
                            //Case Number (used to build accession number)
                            case 4:
                                if (rec != null) {
                                    rec.setCaseNumber(fldItem);
                                }
                                break;
                            //status
                            case 5:
                                if (rec != null) {
                                    rec.setResultCode(fldItem);
                                }
                                break;
                            //severity
                            case 6:
                                if (rec != null) {
                                    rec.setSeverity(fldItem);
                                }
                                break;
                            case 7:
                                if (rec != null) {
                                    rec.setReportIEN(fldItem);
                                }
                                break;
                            default:
                                break;
                        }
                    }
                }
    
                if (rec != null) {
                    rec.setAccessionNumber(getRadAccessionNumber(rec.getExamDate(), rec.getCaseNumber()));
    
                    List<Code> codeList = new ArrayList<Code>();
                    Code radCode = new Code();
                    radCode.setSystem("TEXT");
                    radCode.setCode(rec.getStudy());
                    radCode.setDisplay(rec.getStudy());
                    codeList.add(radCode);
                    rec.setCodes(codeList.toArray(new Code[0]));
                }
    
                lst.add(rec);
            }
    
            rtc = lst.toArray(new RadiologyReport[0]);
        }
        return rtc;
    }

    private String getRadAccessionNumber(Calendar examDate, String caseNumber) {
        StringBuilder accessionNumber = new StringBuilder();

        // add the accession number derived as done in gov.va.med.vds.patient.dao.cache.model.RadiologyData
        if (examDate != null && caseNumber != null) {
            String month = Integer.toString(examDate.get(Calendar.MONTH) + 1);
            String day = Integer.toString(examDate.get(Calendar.DAY_OF_MONTH));
            String year = (Integer.toString(examDate.get(Calendar.YEAR))).substring(2, 4);

            if (month.length() < 2) {
                accessionNumber.append("0");
            }
            accessionNumber.append(month);
            if (day.length() < 2) {
                accessionNumber.append("0");
            }
            accessionNumber.append(day);
            accessionNumber.append(year);
            accessionNumber.append("-");
            accessionNumber.append(caseNumber);
        }

        return accessionNumber.toString();
    }

    /**
     * Uses the VPR RPC with "radiologyExams" as the type.
     *
     * @param startDate The start of the date range for the list of problems
     * @param endDate   The end of the date range for the list of problems
     * @return The list of radiology exams for the specified date range
     * @throws RPCException
     */
    public RadiologyReport[] getRadiologyVPR(Calendar startDate, Calendar endDate) throws RPCException {
        return getRadiologyVPR(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "radiologyExams" as the type.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of problems
     * @param endDate    The end of the date range for the list of problems
     * @return The list of radiology exams for the specified date range
     * @throws RPCException
     */
    public RadiologyReport[] getRadiologyVPR(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.RADIOLOGY_EXAMS);
        String response = (String) cxn.query(request);

        return parseRadiologyFromVPR(response);
    }

    public RadiologyReport[] parseRadiologyFromVPR(String response) throws RPCException {
        RadiologyReport[] rtc;
        if ("".equals(response)) {
            rtc = new RadiologyReport[0];
        } else {
    
            List<RadiologyReport> radsList = new ArrayList<RadiologyReport>();
    
            final String XPATH_RAD = "//radiology";
    
            Document doc = null;
    
            try {
                doc = parseXMLDocument(response);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();
    
            XPathExpression exprRad;
    
            try {
                exprRad = xpath.compile(XPATH_RAD);
    
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            NodeList radsNodeList = null;
            try {
                radsNodeList = (NodeList) exprRad.evaluate(doc, XPathConstants.NODESET);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            List<VPRRadiologyParser> radParserList = new ArrayList<VPRRadiologyParser>();
            int listLength = radsNodeList.getLength();
            for (int radNum = 0; radNum < listLength; radNum++) {
                Node radNode = radsNodeList.item(radNum);
                radNode.getParentNode().removeChild(radNode);
    
                radParserList.add(new VPRRadiologyParser(radNode));
            }
    
            ExecutorService executorService = cxn.getParserThreadPool().getThreadPool();
            List<Future<RadiologyReport>> futures = null;
            try {
                futures = executorService.invokeAll(radParserList, cxn.getParserThreadPool().getPoolTimeoutSeconds(), TimeUnit.SECONDS);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse radiologyExams results : accession node parser futures list is null.");
                }
    
            } catch (InterruptedException e) {
                LOGGER.error(e.getMessage(), e);
                throw new RPCException(e);
            }
    
            Iterator<Future<RadiologyReport>> futuresItr = futures.iterator();
    
            while (futuresItr.hasNext()) {
                try {
                    Future<RadiologyReport> future = futuresItr.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR Radiology Data");
                    } else {
                        radsList.add(future.get());
                    }
    
                } catch (InterruptedException e) {
                    LOGGER.error(e.getMessage(), e);
                } catch (ExecutionException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
    
            rtc = radsList.toArray(new RadiologyReport[0]);
        }
        return rtc;
    }

    private class VPRRadiologyParser implements Callable<RadiologyReport> {
        private Node radNode;

        public VPRRadiologyParser(Node radNode) {
            this.radNode = radNode;
        }

        public RadiologyReport call() throws RPCException {
            final String XPATH_EXAM_ID = "./id/@value";
            final String XPATH_EXAM_DATETIME = "./dateTime/@value";
            final String XPATH_PROCEDURE_NAME = "./name/@value";
            final String XPATH_CASE_NUMBER = "./case/@value";
            final String XPATH_STATUS = "./documents/document/@status";
            final String XPATH_RAD_ORDER_ID = "radOrderID/@value";
            final String XPATH_URGENCY = "urgency/@value";
            final String XPATH_TYPE_CODE = "./type/@code";
            final String XPATH_TYPE_NAME = "./type/@name";
            final String XPATH_INTERPRETATION = "./interpretation/@value";
            final String XPATH_HAS_IMAGES = "./hasImages/@value";
            final String XPATH_FACILITY_NAME = "./facility/@name";

            XPathExpression exprExamId = null;
            XPathExpression exprExamDatetime = null;
            XPathExpression exprProcedureName = null;
            XPathExpression exprCaseNumber = null;
            XPathExpression exprStatus = null;
            XPathExpression exprTypeCode = null;
            XPathExpression exprTypeName = null;
            XPathExpression exprInterpretation = null;
            XPathExpression exprRadOrderID = null;
            XPathExpression exprUrgency = null;
            XPathExpression exprHasImages = null;
            XPathExpression exprFacilityName = null;

            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();

            try {
                exprExamId = xpath.compile(XPATH_EXAM_ID);
                exprExamDatetime = xpath.compile(XPATH_EXAM_DATETIME);
                exprProcedureName = xpath.compile(XPATH_PROCEDURE_NAME);
                exprCaseNumber = xpath.compile(XPATH_CASE_NUMBER);
                exprStatus = xpath.compile(XPATH_STATUS);
                exprTypeCode = xpath.compile(XPATH_TYPE_CODE);
                exprTypeName = xpath.compile(XPATH_TYPE_NAME);
                exprRadOrderID = xpath.compile(XPATH_RAD_ORDER_ID);
                exprUrgency = xpath.compile(XPATH_URGENCY);
                exprInterpretation = xpath.compile(XPATH_INTERPRETATION);
                exprHasImages = xpath.compile(XPATH_HAS_IMAGES);
                exprFacilityName = xpath.compile(XPATH_FACILITY_NAME);
            } catch (Exception e) {
                throw new RPCException(e);
            }

            String examId;
            String examDateTimeStr;
            String procedureName;
            String caseNumber;
            String status;
            String typeCode;
            String typeName;
            Calendar examDate;

            try {
                examId = (String) exprExamId.evaluate(radNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                examId = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                examDateTimeStr = (String) exprExamDatetime.evaluate(radNode, XPathConstants.STRING);
                examDate = Timestamp.toDateTime(examDateTimeStr);
            } catch (XPathExpressionException e) {
                examDate = null;
                LOGGER.error(e.getMessage(), e);
            }

            try {
                procedureName = (String) exprProcedureName.evaluate(radNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                procedureName = null;
                LOGGER.error(e.getMessage(), e);
            }

            try {
                caseNumber = (String) exprCaseNumber.evaluate(radNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                caseNumber = null;
                LOGGER.error(e.getMessage(), e);
            }

            try {
                status = (String) exprStatus.evaluate(radNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                status = null;
                LOGGER.error(e.getMessage(), e);
            }

            try {
                typeCode = (String) exprTypeCode.evaluate(radNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                typeCode = null;
                LOGGER.error(e.getMessage(), e);
            }

            try {
                typeName = (String) exprTypeName.evaluate(radNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                typeName = null;
                LOGGER.error(e.getMessage(), e);
            }

            String radOrderID = RPCUtil.evaluateNodeToString(radNode, exprRadOrderID, LOGGER);
            String urgency = RPCUtil.evaluateNodeToString(radNode, exprUrgency, LOGGER);
            String interpretation = RPCUtil.evaluateNodeToString(radNode, exprInterpretation, LOGGER);
            String hasImagesString = RPCUtil.evaluateNodeToString(radNode, exprHasImages, LOGGER);
            String facilityNameString = RPCUtil.evaluateNodeToString(radNode, exprFacilityName, LOGGER);

            RadiologyReport rad = new RadiologyReport();
            rad.setExamId(examId);
            rad.setExamDate(examDate);
            rad.setStudy(procedureName);
            rad.setCaseNumber(caseNumber);
            rad.setStatus(status);
            rad.setPriority(urgency);
            rad.setOrderNumber(radOrderID);
            rad.setAbnormal(interpretation);
            rad.setFacilityName(facilityNameString);
            if (hasImagesString != null && "1".equals(hasImagesString)) {
                rad.setHasImages(true);
            }

            rad.setAccessionNumber(getRadAccessionNumber(examDate, caseNumber));

            List<Code> codeList = new ArrayList<Code>();
            Code normCode = new Code();
            normCode.setSystem("CPT");
            normCode.setCode(typeCode);
            normCode.setDisplay(typeName);
            codeList.add(normCode);
            rad.setCodes(codeList.toArray(new Code[0]));

            fillCommonDataBeanValues(rad);

            return rad;
        }
    }
}
