/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.vds.patient.dao.rpc.utils;

import gov.va.med.rpc.DictionaryHashList;
import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaConstants;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.rpc.VistaQueryFactory;
import gov.va.med.vds.provider.dao.rpc.UserRPC;

import java.util.Calendar;

/**
 * Utility class with utility methods for VPR RPC calls
 */
public class VPRUtil {

    public static String buildVPRRequest(VistaQueryFactory vistaQueryFactory, String patientIen, Calendar startDate, Calendar endDate, VPRType vprType) {
        VistaQuery vquery = buildVPRRequestVistaQuery(vistaQueryFactory, patientIen, startDate, endDate, vprType);

        return vquery.buildMessage();
    }

    public static String buildVPRRequest(VistaQueryFactory vistaQueryFactory, String patientIen, Calendar startDate, Calendar endDate, VPRType vprType, String category, boolean text) {
        VistaQuery vquery = buildVPRRequestVistaQuery(vistaQueryFactory, patientIen, startDate, endDate, vprType);

        vquery.addParameter(VistaQuery.LITERAL, "0");
        vquery.addParameter(VistaQuery.LITERAL, "0");

        DictionaryHashList paramList = new DictionaryHashList();
        paramList.add("\"category\"", category);
        if (text) {
            paramList.add("\"text\"", "1");
        }
        vquery.addParameter(VistaQuery.LIST, paramList);

        return vquery.buildMessage();
    }

    public static String buildVPRRequest(VistaQueryFactory vistaQueryFactory, String patientIen, VPRType vprType, String id) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("VPR GET PATIENT DATA");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        vquery.addParameter(VistaQuery.LITERAL, vprType.value());
        // start
        vquery.addParameter(VistaQuery.LITERAL, "0");
        // stop
        vquery.addParameter(VistaQuery.LITERAL, "0");
        // max
        vquery.addParameter(VistaQuery.LITERAL, "0");
        vquery.addParameter(VistaQuery.LITERAL, id);

        return vquery.buildMessage();
    }

    public static String buildVPRRequest(VistaQueryFactory vistaQueryFactory, String patientIen, VPRType vprType) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("VPR GET PATIENT DATA");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        vquery.addParameter(VistaQuery.LITERAL, vprType.value());

        return vquery.buildMessage();
    }

    public static String buildVPRRequestWithTypeFilter(VistaQueryFactory vistaQueryFactory, String patientIen, Calendar startDate, Calendar endDate, VPRType vprType, String typeFilter) {
        VistaQuery vquery = buildVPRRequestVistaQuery(vistaQueryFactory, patientIen, startDate, endDate, vprType);

        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");

        DictionaryHashList paramList = new DictionaryHashList();
        paramList.add("\"type\"", typeFilter);

        vquery.addParameter(VistaQuery.LIST, paramList);

        return vquery.buildMessage();

    }

    public static String buildVPRRequestWithTypeFilter(VistaQueryFactory vistaQueryFactory, String patientIen, Calendar startDate, Calendar endDate, VPRType vprType, String typeFilter, boolean text) {
        VistaQuery vquery = buildVPRRequestVistaQuery(vistaQueryFactory, patientIen, startDate, endDate, vprType);

        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");

        DictionaryHashList paramList = new DictionaryHashList();
        paramList.add("\"type\"", typeFilter);
        if (text) {
            paramList.add("\"text\"", "1");
        }

        vquery.addParameter(VistaQuery.LIST, paramList);

        return vquery.buildMessage();

    }

    private static VistaQuery buildVPRRequestVistaQuery(VistaQueryFactory vistaQueryFactory, String patientIen, Calendar startDate, Calendar endDate, VPRType vprType) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("VPR GET PATIENT DATA");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        vquery.addParameter(VistaQuery.LITERAL, vprType.value());
        String startDateString = Timestamp.fromDateTime(startDate);
        String[] flds = StringUtils.split(startDateString, StringUtils.PERIOD);
        startDateString = flds[0];
        String endDateString = Timestamp.fromDateTime(endDate);
        flds = StringUtils.split(endDateString, StringUtils.PERIOD);
        endDateString = flds[0];
        vquery.addParameter(VistaQuery.LITERAL, startDateString);
        vquery.addParameter(VistaQuery.LITERAL, endDateString);

        return vquery;
    }

    public static String buildVPRJSONRequest(VistaQueryFactory vistaQueryFactory, String patientIen, Calendar startDate, Calendar endDate, VPRType vprType) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("VPR GET PATIENT DATA JSON");
        DictionaryHashList paramList = new DictionaryHashList();
        paramList.add("\"patientId\"", patientIen);
        paramList.add("\"domain\"", vprType.value());

        String[] flds;
        if (startDate != null) {
            String startDateString = Timestamp.fromDateTime(startDate);
            flds = StringUtils.split(startDateString, StringUtils.PERIOD);
            startDateString = flds[0];
            paramList.add("\"start\"", startDateString);
        }
        if (endDate != null) {
            String endDateString = Timestamp.fromDateTime(endDate);
            flds = StringUtils.split(endDateString, StringUtils.PERIOD);
            endDateString = flds[0];
            paramList.add("\"stop\"", endDateString);
        }
//        if (maxReturnItems != null) {
//            paramList.add("\"max\"", maxReturnItems);
//        }
        vquery.addParameter(VistaQuery.LIST, paramList);

        return vquery.buildMessage();
    }


    /**
     * VPR RPC type
     */
    public static enum VPRType {
        ACCESSIONS("accessions"),
        REACTIONS("reactions"),
        APPOINTMENTS("appointments"),
        CLINICAL_PROCEDURES("clinicalProcedures"),
        CONSULTS("consults"),
        DEMOGRAPHICS("demographics"),
        DOCUMENTS("documents"),
        HEALTH_FACTORS("healthFactors"),
        FLAGS("flags"),
        IMMUNIZATIONS("immunizations"),
        SKIN_TESTS("skinTests"),
        SKIN("skin"),
        EXAMS("exams"),
        EDUCATION_TOPICS("educationTopics"),
        INSURANCE_POLICIES("insurancePolicies"),
        LABS("labs"),
        PANELS("panels"),
        MEDS("meds"),
        OBSERVATIONS("observations"),
        ORDERS("orders"),
        PROBLEMS("problems"),
        PROCEDURES("procedures"),
        SURGERIES("surgeries"),
        VISITS("visits"),
        VITALS("vitals"),
        RADIOLOGY_EXAMS("radiologyExams"),
        PATIENTS("patients"),
        CPT("cpt"),
        ICD("icd"),
        REMINDERS("reminders");

        private final String text;

        /**
         * @param text
         */
        private VPRType(final String text) {
            this.text = text;
        }

        public String value() {
            return text;
        }
    }

    /**
     * Utility method to set user context once initial context has been established through UserRPC visit.
     *
     * @param context Context to set (VistaConstants.CPRS_CONTEXT
     * @return true if context was successfully set
     * @throws gov.va.med.rpc.RPCException if an error occurs.
     */
    public static boolean setUserContext(String context, RPCConnection cxn) throws RPCException {
        UserRPC userRPC = new UserRPC(cxn);

        boolean isContextSuccess = false;

        try {
            isContextSuccess = userRPC.setContext(VistaConstants.CPRS_CONTEXT);

        } catch (RPCException e) {
            isContextSuccess = false;
        }

        if (!isContextSuccess) {
            userRPC.addMenuOption(VistaConstants.CPRS_CONTEXT, cxn.getUserIen());
            isContextSuccess = userRPC.setContext(VistaConstants.CPRS_CONTEXT);
        }

        return isContextSuccess;
    }

}
