/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.vds.patient.dataservice;

import gov.va.med.rpc.RPCException;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.base.dataservice.VistaRPCConnectionManager;
import gov.va.med.vds.vistabeans.PCE;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.vds.vistabeans.Site;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 *
 *
 */
public class PCEDataService extends DataServiceBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(PCEDataService.class.getName());

    /**
     * Retrieves list of patient skin tests.
     * @param queryBean Query parameters.
     *                  Required params: User, Patient, startDate, endDate
     * @return A list of patient skin tests.
     * @throws VDSException if there was an error..
     */
    public List<PCE> getPatientSkinTests(QueryBean queryBean) throws VDSException {
        validateQueryBeanDateRange(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> skinTestsMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_PCE,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_SKIN_TESTS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = skinTestsMap.entrySet();
            List<PCE> skinTestList = new ArrayList<PCE>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object skinTests = entry.getValue();
                if (skinTests instanceof Exception) {
                    skinTests = generateConnectionUnavailableBean(entry.getKey(), (Exception) skinTests);
                }
                skinTestList.addAll(Arrays.asList((PCE[]) skinTests));
            }

            return skinTestList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    /**
     * Retrieves list of patient PCE exams.
     * @param queryBean Query parameters.
     *                  Required params: User, Patient, startDate, endDate
     * @return A list of patient PCE exams.
     * @throws VDSException if there was an error..
     */
    public List<PCE> getPatientExams(QueryBean queryBean) throws VDSException {
        validateQueryBeanDateRange(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupVPRContextMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> examsMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_PCE,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_EXAMS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = examsMap.entrySet();
            List<PCE> examList = new ArrayList<PCE>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object exams = entry.getValue();
                if (exams instanceof Exception) {
                    exams = generateConnectionUnavailableBean(entry.getKey(), (Exception) exams);
                }
                examList.addAll(Arrays.asList((PCE[]) exams));
            }

            return examList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    /**
     * Retrieves list of patient PCE education topics.
     * @param queryBean Query parameters.
     *                  Required params: User, Patient, startDate, endDate
     * @return A list of patient PCE education topics.
     * @throws VDSException if there was an error..
     */
    public List<PCE> getPatientEducationTopics(QueryBean queryBean) throws VDSException {
        validateQueryBeanDateRange(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupVPRContextMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> educationTopicsMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_PCE,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_EDUCATION_TOPICS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = educationTopicsMap.entrySet();
            List<PCE> educationTopicList = new ArrayList<PCE>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object educationTopics = entry.getValue();
                if (educationTopics instanceof Exception) {
                    educationTopics = generateConnectionUnavailableBean(entry.getKey(), (Exception) educationTopics);
                }
                educationTopicList.addAll(Arrays.asList((PCE[]) educationTopics));
            }

            return educationTopicList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private PCE[] generateConnectionUnavailableBean(Site site, Exception except) {
        
        PCE pce = new PCE();
        pce.setName(CONNECTION_UNAVAILABLE);
        pce.setSite(site);
        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            pce.setName(UNABLE_TO_RETRIEVE);
        } else {
            if(except.getCause() != null && except.getCause().getCause() != null
                    && except.getCause().getCause() instanceof InvocationTargetException
                    && ((InvocationTargetException) except.getCause().getCause()).getTargetException() != null ){
                LOGGER.error("VDS Error InvocationTargetException: " + ((InvocationTargetException) except.getCause().getCause()).getTargetException().toString());
            } else {
                LOGGER.error(except.getLocalizedMessage(), except);        //log if any other exception type
            }
        }
        return new PCE[]{pce};
    }
}
