/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.vds.patient.dataservice;

import gov.va.med.rpc.RPCException;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.base.dataservice.VistaRPCConnectionManager;
import gov.va.med.vds.vistabeans.Diagnosis;
import gov.va.med.vds.vistabeans.Procedure;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.vds.vistabeans.Site;
import gov.va.med.vhahon.common.Utils;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 *
 *
 */
public class PatientApptDataService extends DataServiceBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(PatientApptDataService.class.getName());

    public List<Diagnosis> getVisitDiagnosis(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> diagMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_PATIENT_APPT,
                    VistaRPCConnectionManager.EXECUTION_METHOD_VISIT_DIAGNOSIS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = diagMap.entrySet();
            List<Diagnosis> diagList = new ArrayList<Diagnosis>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object diag = entry.getValue();
                if (diag instanceof Exception) {
                    diag = generateConnectionUnavailableBeanDiag(entry.getKey(), (Exception) diag);
                }
                diagList.addAll(Arrays.asList((Diagnosis[]) diag));
            }

            return diagList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public List<Procedure> getVisitProcedures(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> procMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_PATIENT_APPT,
                    VistaRPCConnectionManager.EXECUTION_METHOD_VISIT_PROCEDURES, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = procMap.entrySet();
            List<Procedure> procList = new ArrayList<Procedure>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object proc = entry.getValue();
                if (proc instanceof Exception) {
                    proc = generateConnectionUnavailableBeanProc(entry.getKey(), (Exception) proc);
                }
                procList.addAll(Arrays.asList((Procedure[]) proc));
            }

            return procList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private Procedure[] generateConnectionUnavailableBeanProc(Site site, Exception except)
    {
        
        Procedure proc = new Procedure();
        proc.setDescription(CONNECTION_UNAVAILABLE);
        proc.setDate(Utils.backDate().toGregorianCalendar());
        proc.setSite(site);

        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            proc.setDescription(UNABLE_TO_RETRIEVE);
        } else {
            if(except.getCause() != null && except.getCause().getCause() != null
                    && except.getCause().getCause() instanceof InvocationTargetException
                    && ((InvocationTargetException) except.getCause().getCause()).getTargetException() != null ){
                LOGGER.error("VDS Error InvocationTargetException: " + ((InvocationTargetException) except.getCause().getCause()).getTargetException().toString());
            } else {
                LOGGER.error(except.getLocalizedMessage(), except);        //log if any other exception type
            }
        }


        return new Procedure[] {proc};
    }

    private Diagnosis[] generateConnectionUnavailableBeanDiag(Site site, Exception except)
    {
        //log exception
        LOGGER.error(except.getMessage(), except);

        Diagnosis diag = new Diagnosis();
        diag.setDescription(CONNECTION_UNAVAILABLE);
        diag.setDate(Utils.backDate().toGregorianCalendar());
        diag.setSite(site);

        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            diag.setDescription(UNABLE_TO_RETRIEVE);
        } else {
            if(except.getCause() != null && except.getCause().getCause() != null
                    && except.getCause().getCause() instanceof InvocationTargetException
                    && ((InvocationTargetException) except.getCause().getCause()).getTargetException() != null ){
                LOGGER.error("VDS Error InvocationTargetException: " + ((InvocationTargetException) except.getCause().getCause()).getTargetException().toString());
            } else {
                LOGGER.error(except.getLocalizedMessage(), except);        //log if any other exception type
            }
        }


        return new Diagnosis[] {diag};
    }
}
