/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.vds.patient.dataservice;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.base.dataservice.VistaRPCConnectionManager;
import gov.va.med.vds.patient.dao.PatientDao;
import gov.va.med.vds.vistabeans.*;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class PatientDataService extends DataServiceBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(PatientDataService.class.getName());

    public Patient[] lookupPatient(QueryBean queryBean) throws VDSException {
        if (queryBean == null) {
            throw new IllegalArgumentException("QueryBean parameter is null");
        } else if (queryBean.getUser() == null) {
            throw new IllegalArgumentException("User is null.");
        } else if (queryBean.getRecordSite().getSiteCode() == null) {
            throw new IllegalArgumentException("Record site code is null or empty: Must supply patient lookup site code.");
        } else if (queryBean.getItemId() == null) {
            throw new IllegalArgumentException("Lookup name is null.");
        }

        RPCConnection cxn = null;

        try {
            Site site = queryBean.getRecordSite();
            cxn = setupCPRSContextSingleRemoteSourceQuery(queryBean.getUser(), site);
            PatientDao patientDao = new PatientDao(cxn);
            Patient[] matches = patientDao.lookupPatient(queryBean);

            return matches;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxn != null) {
                try {
                    cxn.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    /**
     * Calls patient select RPC at each site specified in patient's vistaSites list.
     * @param queryBean QueryBean
     *    Required Params:
     *                  Fully qualified User and Patient objects.
     * @return true if patient select was successfully invoked at EACH vista site,
     *                  false otherwise (even if patient select was successful at some sites).
     * @throws VDSException if an error occurred.
     */
    public boolean selectPatient(QueryBean queryBean) throws VDSException {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> patSelectMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_PATIENT,
                    VistaRPCConnectionManager.EXECUTION_METHOD_SELECT_PATIENT, new Object[]{});

            Set<Map.Entry<Site, Object>> entries = patSelectMap.entrySet();
            List<Patient> patientList = new ArrayList<Patient>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object patient = entry.getValue();
                if (patient instanceof Patient) {
                    patientList.add((Patient) patient);
                }
            }

            return queryBean.getPatient().getVistaSites().size() == patientList.size();
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }

    }
    
    public List<PatientDemographics> getPatientDemographics(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupVPRContextMultiSourceQuery(queryBean);
            LinkedHashMap<Site, RPCConnection> connectionTable = msq.getConnectionTable();

            List<PatientDemographics> demosList = new ArrayList<PatientDemographics>();
            LinkedHashMap<Site, Object> demosMap = VistaRPCConnectionManager.execute(connectionTable,
                    VistaRPCConnectionManager.EXECUTION_CLASS_PATIENT,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_DEMOGRAPHICS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = demosMap.entrySet();
            for (Map.Entry<Site, Object> entry : entries) {
                Object demos = entry.getValue();
                if (demos instanceof Exception) {
                    demos = generateConnectionUnavailableBeanDemos(entry.getKey(), (Exception) demos);
                }
                demosList.add((PatientDemographics) demos);
            }

            return demosList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public PatientDemographics getPatientDemographicsForSite(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        if (queryBean.getRecordSite().getSiteCode() == null || queryBean.getRecordSite().getSiteCode().isEmpty())
        {
            throw new IllegalArgumentException("record site code cannot be null.");
        }

        RPCConnection cxn = null;

        try {
            cxn = setupSingleRemoteSourceQuery(queryBean);

            PatientDao patientDao = new PatientDao(cxn);
            PatientDemographics patDemo = patientDao.getPatientDemographics(queryBean);

            return patDemo;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxn != null) {
                try {
                    cxn.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public List<PatientDemographicsDetail> getPatientDemographicsDetail(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, RPCConnection> connectionTable = msq.getConnectionTable();

            List<PatientDemographicsDetail> demosDetailsList = new ArrayList<PatientDemographicsDetail>();
            LinkedHashMap<Site, Object> demoDetailsMap = VistaRPCConnectionManager.execute(connectionTable,
                    VistaRPCConnectionManager.EXECUTION_CLASS_PATIENT,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_DEMOGRAPHICS_DETAIL, new Object[]{queryBean});
            Set<Map.Entry<Site, Object>> entries = demoDetailsMap.entrySet();

            for (Map.Entry<Site, Object> entry : entries) {
                Object demoDetails = entry.getValue();
                if (demoDetails instanceof Exception) {
                    demoDetails = generateConnectionUnavailableBeanDemoDetails(entry.getKey(), (Exception) demoDetails);
                }
                demosDetailsList.add((PatientDemographicsDetail) demoDetails);
            }

            return demosDetailsList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public List<Alert> getSensitivePatientAlerts(QueryBean queryBean) throws VDSException {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, RPCConnection> connectionTable = msq.getConnectionTable();
            LinkedHashMap<Site, Object> patSensitiveMap = VistaRPCConnectionManager.execute(connectionTable,
                    VistaRPCConnectionManager.EXECUTION_CLASS_PATIENT,
                    VistaRPCConnectionManager.EXECUTION_METHOD_SENSITIVE_PATIENT, new Object[]{queryBean});
            Set<Map.Entry<Site, Object>> entries = patSensitiveMap.entrySet();
            List<Alert> rtc = new ArrayList<Alert>();
            for (Map.Entry<Site, Object> entry : entries) {
                Object alert = entry.getValue();
                if (alert instanceof Exception) {
                    alert = generateConnectionUnavailableBeanAlert(entry.getKey(), (Exception) alert);
                }
                rtc.add((Alert) alert);
            }

            return rtc;
        } catch (Exception e) {
            throw new VDSException(e);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public List<ReferralBean> getDRReferrals(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            if (StringUtils.isEmpty(queryBean.getStatus())) {
                throw new IllegalArgumentException("Problem type is null or empty");
            }
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> drrMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_PATIENT,
                    VistaRPCConnectionManager.EXECUTION_METHOD_DR_REFERRALS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = drrMap.entrySet();
            List<ReferralBean> drrList = new ArrayList<ReferralBean>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object drr = entry.getValue();
                if (drr instanceof Exception) {
                    drr = generateConnectionUnavailableBeanReferral(entry.getKey(), (Exception) drr);
                }
                drrList.addAll(Arrays.asList((ReferralBean[]) drr));
            }

            return drrList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private PatientDemographics generateConnectionUnavailableBeanDemos(Site site, Exception except)
    {
        
        PatientDemographics patDemo = new PatientDemographics();
        patDemo.setName(CONNECTION_UNAVAILABLE);
        patDemo.setSite(site);
        patDemo.setAge("");
        patDemo.setClinic("");
        patDemo.setDob("");
        patDemo.setGender("");
//        patDemo.setPatientId("");
        patDemo.setPrimaryProvider("");
        patDemo.setPhone1("");
        patDemo.setPhone2("");
        patDemo.setSSN("");
        patDemo.setCodeGreen("0");

        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            patDemo.setName(UNABLE_TO_RETRIEVE);
        } else {
            if(except.getCause() != null && except.getCause().getCause() != null
                    && except.getCause().getCause() instanceof InvocationTargetException
                    && ((InvocationTargetException) except.getCause().getCause()).getTargetException() != null ){
                LOGGER.error("VDS Error InvocationTargetException: " + ((InvocationTargetException) except.getCause().getCause()).getTargetException().toString());
            } else {
                LOGGER.error(except.getLocalizedMessage(), except);        //log if any other exception type
            }
        }


        return patDemo;
    }

    private PatientDemographicsDetail generateConnectionUnavailableBeanDemoDetails(Site site, Exception except)
    {
        
        PatientDemographicsDetail patDemoDetail = new PatientDemographicsDetail();
        patDemoDetail.setReport(CONNECTION_UNAVAILABLE);
        patDemoDetail.setSite(site);

        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            patDemoDetail.setReport(UNABLE_TO_RETRIEVE);
        } else {
            if(except.getCause() != null && except.getCause().getCause() != null
                    && except.getCause().getCause() instanceof InvocationTargetException
                    && ((InvocationTargetException) except.getCause().getCause()).getTargetException() != null ){
                LOGGER.error("VDS Error InvocationTargetException: " + ((InvocationTargetException) except.getCause().getCause()).getTargetException().toString());
            } else {
                LOGGER.error(except.getLocalizedMessage(), except);        //log if any other exception type
            }
        }


        return patDemoDetail;
    }

    private Alert generateConnectionUnavailableBeanAlert(Site site, Exception except)
    {
        Alert alert = new Alert();
        alert.setAlertId("-2");
        alert.setMessageText(CONNECTION_UNAVAILABLE);
        alert.setSite(site);

        if(except.getCause() != null && except.getCause().getCause() != null
                && except.getCause().getCause() instanceof InvocationTargetException
                && ((InvocationTargetException) except.getCause().getCause()).getTargetException() != null ){
            LOGGER.error("VDS Error InvocationTargetException: " + ((InvocationTargetException) except.getCause().getCause()).getTargetException().toString());
        } else {
            LOGGER.error(except.getLocalizedMessage(), except);        //log if any other exception type
        }

        return alert;
    }

    private ReferralBean[] generateConnectionUnavailableBeanReferral(Site site, Exception except)
    {
        
        ReferralBean ref = new ReferralBean();
        ref.setClinic(CONNECTION_UNAVAILABLE);
        ref.setSite(site);

        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            ref.setClinic(UNABLE_TO_RETRIEVE);
        } else {
            if(except.getCause() != null && except.getCause().getCause() != null
                    && except.getCause().getCause() instanceof InvocationTargetException
                    && ((InvocationTargetException) except.getCause().getCause()).getTargetException() != null ){
                LOGGER.error("VDS Error InvocationTargetException: " + ((InvocationTargetException) except.getCause().getCause()).getTargetException().toString());
            } else {
                LOGGER.error(except.getLocalizedMessage(), except);        //log if any other exception type
            }
        }


        return new ReferralBean[]{ref};
    }

}
