/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.provider.dataservice;

import gov.va.med.rpc.RPCConnectionManager;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.VistaConstants;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.provider.dao.UserDao;
import gov.va.med.vds.vistabeans.Alert;
import gov.va.med.vds.vistabeans.ProviderAppointmentBean;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.vds.vistabeans.Site;
import gov.va.med.vds.vistabeans.User;

import java.util.Calendar;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class UserDataService extends DataServiceBase {

    private static final Logger LOGGER = LoggerFactory.getLogger(UserDataService.class.getName());

    /**
     * VistA user login method.
     *
     * @param site       VistA ehr site.
     * @param accessCode User access code.
     * @param verifyCode User verify code.
     * @return A valid user bean.
     * @throws VDSException if there was an error..
     */
    public User login(Site site, String accessCode, String verifyCode) throws VDSException {
        return login(site, accessCode, verifyCode, false);
    }

    /**
     * VistA user login method.
     *
     * @param site         VistA ehr site.
     * @param accessCode   User access code.
     * @param verifyCode   User verify code.
     * @param defaultToRPC Use VistA impl to perform login (instead of defaulting to CACHE routine if it exists for site).
     * @return A valid user bean.
     * @throws VDSException if there was an error..
     */
    public User login(Site site, String accessCode, String verifyCode, boolean defaultToRPC) throws VDSException {

        RPCConnectionManager cxnMgr = null;

        try {
            if (site == null) {
                throw new IllegalArgumentException("Missing site");
            }
            if (StringUtils.isEmpty(accessCode)) {
                throw new IllegalArgumentException("Missing access code");
            } else if (StringUtils.isEmpty(verifyCode)) {
                throw new IllegalArgumentException("Missing verify code");
            }

            cxnMgr = connectToSite(site);

            String context = VistaConstants.CPRS_CONTEXT;

            //set VBA users context to DVBA CAPRI GUI
            if ("VBA".equalsIgnoreCase(site.getMoniker())) {
                context = VistaConstants.CAPRI_CONTEXT;
            }

            UserDao userDao = new UserDao(cxnMgr.loginConnection());
            User user = userDao.login(accessCode, verifyCode, context, defaultToRPC);
            //user.setHostSiteId((String) cxnMgr.loginConnection().getSiteCode().getKey());
            return user;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxnMgr != null) {
                try {
                    cxnMgr.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }

    }

    public User[] lookupUser(Site site, String lookupstr, String userId, int maxReturn) throws VDSException {
        RPCConnectionManager cxnMgr = null;

        try {
            if (site == null) {
                throw new IllegalArgumentException("Missing site");
            }

            cxnMgr = connectToSite(site);

            if (StringUtils.isEmpty(lookupstr)) {
                throw new IllegalArgumentException("Missing lookup string.");
            } else if (StringUtils.isEmpty(userId)) {
                throw new IllegalArgumentException("Missing userId");
            }
//            String context = VistaConstants.CPRS_CONTEXT;
            UserDao userDao = new UserDao(cxnMgr.loginConnection());
            User[] users = userDao.lookupUser(lookupstr, userId, maxReturn);

            return users;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxnMgr != null) {
                try {
                    cxnMgr.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public User getProviderProfile(Site site, String userIen, String providerNpi) throws VDSException {
        RPCConnectionManager cxnMgr = null;

        if (site == null) {
            throw new IllegalArgumentException("Missing site");
        } else if (StringUtils.isEmpty(userIen)) {
            throw new IllegalArgumentException("Missing user ID");
        }

        try {
            cxnMgr = connectToSite(site);
            UserDao userDao = new UserDao(cxnMgr.loginConnection());
            return userDao.getProviderProfile(userIen, providerNpi);
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxnMgr != null) {
                try {
                    cxnMgr.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public ProviderAppointmentBean[] getProviderWorkload(QueryBean queryBean) throws VDSException {
        if (queryBean == null) {
            throw new IllegalArgumentException("QueryBean is null.");
        } else if (queryBean.getUser() == null) {
            throw new IllegalArgumentException("User is null.");
        }

        RPCConnectionManager cxnMgr = null;

        try {
            String itemId = queryBean.getItemId();
            Site site = queryBean.getUser().getHostSite();
            if (itemId == null || itemId.length() <= 0) {
                throw new IllegalArgumentException("Item Id cannot be null or empty");
            }

            cxnMgr = connectToSite(site);
            UserDao userDao = new UserDao(cxnMgr.loginConnection());
            ProviderAppointmentBean[] providerWorkload = userDao.getProviderWorkload(queryBean);

            return providerWorkload;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxnMgr != null) {
                try {
                    cxnMgr.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public Alert[] getUserAlerts(QueryBean queryBean) throws VDSException {
        if (queryBean == null) {
            throw new IllegalArgumentException("QueryBean is null.");
        } else if (queryBean.getUser() == null) {
            throw new IllegalArgumentException("User is null.");
        }

        RPCConnectionManager cxnMgr = null;

        try {
            Calendar startDate = queryBean.getStartDate();
            Calendar endDate = queryBean.getEndDate();
            Site site = queryBean.getUser().getHostSite();
            if (startDate == null) {
                throw new IllegalArgumentException("Start Date is null");
            } else if (endDate == null) {
                throw new IllegalArgumentException("End Date is null");
            }
            cxnMgr = connectToSite(site);
            UserDao userDao = new UserDao(cxnMgr.loginConnection());
            Alert[] alerts = userDao.getUserAlerts(queryBean);

            return alerts;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxnMgr != null) {
                try {
                    cxnMgr.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }
}
