package gov.va.med.vds.webservice;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import gov.va.med.foundations.utilities.ARMLog;
import gov.va.med.vds.common.AppConfig;
import gov.va.med.vds.das.beans.Subscription;
import gov.va.med.vds.vistabeans.QueryBean;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by melcolm on 1/5/17.
 */
public class DASLimit {

    /**
     * Logger.
     */
    private static final Logger LOGGER =
            LoggerFactory.getLogger(DASLimit.class);

    /**
     * Subscription map.
     */
    private static ConcurrentHashMap<String, Subscription> subscriptions = new ConcurrentHashMap<String, Subscription>();

//    /**
//     * Cache.
//     */
//    private static ConcurrentHashMap<String, Subscription> cachemap = new ConcurrentHashMap<String, Subscription>();

    /**
     * Age of cache entries.
     */
    private static ConcurrentHashMap<String, Date> cacheagemap = new ConcurrentHashMap<String, Date>();

    /**
     * Time to live for cache entries.
     */
    private int expiretimesecs = 900; // 15 minute

    private DASLimit() {
        super();
        String timeout = AppConfig.getInstance().getProperties().getProperty("DASLimit.timeout", "900");
        try {
            expiretimesecs = Integer.parseInt(timeout);
        } catch (Exception ex) {
            expiretimesecs = 900;
        }
    }

    /**
     * Get singleton class.
     *
     * @return this
     */
    @SuppressWarnings("all")
    public static synchronized DASLimit getInstance() {
        return DASLimitHolder.INSTANCE;
    }

    public Subscription getSubscription(String key) {
        return subscriptions.get(key);
    }

    public void putSubscription(String key, Subscription subscription) {
        try {
            Date cdate = new Date();
            subscriptions.put(key, subscription);
            cacheagemap.put(key, cdate);
        } catch (Exception ex) {
            LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
        }
    }

    private void dropSubscription(String key) {
        Enumeration<String> keys = cacheagemap.keys();
        while (keys != null && keys.hasMoreElements()) {
            String cachedKey = keys.nextElement();
            if (cachedKey != null && cachedKey.equals(key)) {
                cacheagemap.remove(cachedKey);
                subscriptions.remove(cachedKey);
            }
        }
    }

    /**
     * User for synchronization across multiple relay services.
     *
     * @param key          key
     * @param subscription value
     */
    public void putDASCache(String key, Subscription subscription) {
        try {
            Date cdate = new Date();
            cacheagemap.put(key, cdate);
            subscriptions.put(key, subscription);
        } catch (Exception ex) {
            LOGGER.error("EXCEPTION: " + ex.getLocalizedMessage(), ex);
        }
    }

    public void putDASCache(String domain, QueryBean queryBean, Subscription subscription) {
        putCache(domain, queryBean, subscription);
    }

    private Subscription getCache(String domain, QueryBean queryBean) {
        Subscription rtc = null;
        if ("true".equalsIgnoreCase(AppConfig.getInstance().getProperties().getProperty("VDSLimit.disable", "false"))) {
            cleanMap();
        } else {
            String key = getKey(domain, queryBean);
            if (subscriptions.containsKey(key) && cacheagemap.containsKey(key)) {
                Date ldate = cacheagemap.get(key);
                Date cdate = new Date();
                if (ldate != null && (expiretimesecs == -1 || (Math.abs(cdate.getTime() - ldate.getTime()) <= (1000L * expiretimesecs)))) {
                    rtc = subscriptions.get(key);
                    LOGGER.info("Returning cache for: {}", ARMLog.cleanString(key));
                } else {
                    cleanMap();
                }
            } else {
                cleanMap();
            }
        }
        return rtc;
    }

    private void putCache(String domain, QueryBean queryBean, Subscription subscription) {
        String key = getKey(domain, queryBean);
        Date cdate = new Date();
        cacheagemap.put(key, cdate);
        subscriptions.put(key, subscription);
    }

//    private void dropCache(String domain, QueryBean queryBean) {
//        if (queryBean != null) {
//            if (queryBean.isForceRefresh()) {
//                dropCache(queryBean);
//            } else {
//                String key = getKey(domain, queryBean);
//                cacheagemap.remove(key);
//                subscriptions.remove(key);
//            }
//        }
//    }
//
//    private void dropCache(QueryBean queryBean) {
//        Enumeration<String> keys = cacheagemap.keys();
//        while (keys != null && keys.hasMoreElements()) {
//            String key = keys.nextElement();
//            if (key != null && key.startsWith(queryBean.getPatient().getICN() + "|")) {
//                cacheagemap.remove(key);
//                subscriptions.remove(key);
//            }
//        }
//    }

    public String getKey(String loinc, QueryBean queryBean) {
        StringBuilder rtc = new StringBuilder();
        if (queryBean != null) {
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd", Locale.getDefault());
            rtc.append(queryBean.getPatient().getICN());
            rtc.append('|');
            rtc.append(loinc);
            rtc.append('|');
            if (queryBean.getStartDate() != null) {
                rtc.append(sdf.format(queryBean.getStartDate().getTime()));
            }
            rtc.append('|');
            if (queryBean.getEndDate() != null) {
                rtc.append(sdf.format(queryBean.getEndDate().getTime()));
            }
            rtc.append('|');
        }
        return rtc.toString();
    }

    private void cleanMap() {
        Enumeration<String> keys = cacheagemap.keys();
        Date cdate = new Date();
        while (keys != null && keys.hasMoreElements()) {
            String key = keys.nextElement();
            Date adate = cacheagemap.get(key);
            if (adate != null && cdate != null && (Math.abs(cdate.getTime()
                    - adate.getTime()) > (1000L * (60 + expiretimesecs)))) {
                cacheagemap.remove(key);
                subscriptions.remove(key);
            }
        }
    }

    /**
     * Disable cloning.
     *
     * @return nothing
     * @throws CloneNotSupportedException exception
     */
    @SuppressWarnings("all")
    public Object clone() throws CloneNotSupportedException {
        throw new CloneNotSupportedException();
    }

    /**
     * Singleton helper class.
     */
    private static class DASLimitHolder {
        /**
         * VDSLimit.
         */
        private static final DASLimit INSTANCE = new DASLimit();
    }
}
