/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.des.dataservicehandler;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.ClientResponse;
import com.sun.jersey.api.client.WebResource;
import gov.va.med.des.beans.HostStatus;
import gov.va.med.des.common.AppConfig;
import gov.va.med.des.common.JMeadowsException;
import gov.va.med.monitor.DatabaseHandler;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import sun.misc.BASE64Encoder;

import java.util.List;


/**
 * DESRestServiceHandler
 */
public class SiteMonitorHandler extends DataServiceHandler {
    /**
     * Logger.
     */
    private static final Logger LOGGER =
            Logger.getLogger(SiteMonitorHandler.class.getName());

    private static final String SITE_STATUS_PATH = AppConfig.getInstance().getSiteStatusPath();

    private static final int CONNECTION_TIMEOUT_MS = AppConfig.getInstance().getConnectionTimeoutMS();
    private static final int READ_TIMEOUT_MS = AppConfig.getInstance().getRestReadTimeoutMs();

    private static final SiteMonitorHandler INSTANCE =
            new SiteMonitorHandler(SITE_STATUS_PATH);

    private String endpoint;
    private ObjectMapper objectMapper;
    private Client restClient;

    /**
     * Creates a new DESRestServiceHandler instance.
     *
     * @param endpoint  DESRestService end point
     */
    public SiteMonitorHandler(String endpoint) {

        super();

        this.endpoint = endpoint;

        if (endpoint != null && '/' != endpoint.charAt(endpoint.length() - 1)) {
            this.endpoint += "/";
        }

        objectMapper = new ObjectMapper();

        //allow unknown data fields
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        this.restClient = initRestClient(CONNECTION_TIMEOUT_MS, READ_TIMEOUT_MS);
    }

    public static SiteMonitorHandler getInstance() {
        return INSTANCE;
    }


    /**
     * Initializes REST client.
     *
     * @param connectionTimeout  REST client connection timeout
     * @param readTimeout  REST client read timeout
     * @return Jersey REST Client
     */
    private Client initRestClient(Integer connectionTimeout, Integer readTimeout) {
        Client client = Client.create();
        client.setConnectTimeout(connectionTimeout);
        client.setReadTimeout(readTimeout);

        return client;
    }

    /**
     * Sends a query request to BHIE REST interface and
     * then polls the BHIE REST comet interface with the provided queryID.
     *
     * When the long poller receives valid response, it will bind the json response to a ResultCollection object
     *  and then add the ResultCollection to a List.
     *
     * @return A List of ResultCollection objects.
     * @throws JMeadowsException If there is a failure.
     */
    @SuppressWarnings("unchecked")
    public Boolean querySiteMonitor() throws JMeadowsException {
        boolean result = false;


        try {
            WebResource queryResource = restClient.resource(SITE_STATUS_PATH);
            String queryResp = "";

            if(APP_CONFIG.isSiteStatusAuth()) {
                StringBuilder authString = new StringBuilder();
                authString.append(APP_CONFIG.getSiteStatusEntity());
                authString.append(":");
                authString.append(APP_CONFIG.getSiteStatusEntityToken());
                String authStringEnc = new BASE64Encoder().encode(authString.toString().getBytes());

                ClientResponse resp = queryResource.accept("application/json")
                        .header("Authorization", "Basic " + authStringEnc)
                        .get(ClientResponse.class);

                if (resp.getStatus() != 200) {
                    LOGGER.error("SITE_STATUS: Unable to connect to the server");
                }

                queryResp = resp.getEntity(String.class);
            } else {
                queryResp = queryResource.get(String.class);
            }

            List<HostStatus> statuses = objectMapper.readValue(queryResp, new TypeReference<List<HostStatus>>(){});

            if(!statuses.isEmpty() && !StringUtils.isEmpty(statuses.get(0).getHostUri())) {
                DatabaseHandler.getInstance().logStatuses(statuses);
                result = true;
            }
        } catch (Exception e) {
            LOGGER.error(e.getMessage());
            throw new JMeadowsException(e);
        }

        return result;
    }



}
