/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.des.dataservicehandler;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.ClientResponse;
import com.sun.jersey.api.client.WebResource;
import gov.va.med.des.beans.SystemInfos;
import gov.va.med.des.common.AppConfig;
import gov.va.med.des.common.JMeadowsException;
import gov.va.med.monitor.DatabaseHandler;
import org.apache.log4j.Logger;
import sun.misc.BASE64Encoder;


/**
 * DESRestServiceHandler
 */
public class SystemMonitorHandler extends DataServiceHandler {
    /**
     * Logger.
     */
    private static final Logger LOGGER =
            Logger.getLogger(SystemMonitorHandler.class.getName());

    private static final String SYSTEM_MONITOR_PATH = AppConfig.getInstance().getSystemMonitorPath();

    private static final int CONNECTION_TIMEOUT_MS = AppConfig.getInstance().getConnectionTimeoutMS();
    private static final int READ_TIMEOUT_MS = AppConfig.getInstance().getRestReadTimeoutMs();

    private static final SystemMonitorHandler INSTANCE =
            new SystemMonitorHandler(SYSTEM_MONITOR_PATH);

    private String endpoint;
    private ObjectMapper objectMapper;
    private Client restClient;

    /**
     * Creates a new DESRestServiceHandler instance.
     *
     * @param endpoint DESRestService end point
     */
    public SystemMonitorHandler(String endpoint) {

        super();

        this.endpoint = endpoint;

        if (endpoint != null && '/' != endpoint.charAt(endpoint.length() - 1)) {
            this.endpoint += "/";
        }

        objectMapper = new ObjectMapper();

        //allow unknown data fields
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        this.restClient = initRestClient(CONNECTION_TIMEOUT_MS, READ_TIMEOUT_MS);
    }

    public static SystemMonitorHandler getInstance() {
        return INSTANCE;
    }


    /**
     * Initializes REST client.
     *
     * @param connectionTimeout REST client connection timeout
     * @param readTimeout       REST client read timeout
     * @return Jersey REST Client
     */
    private Client initRestClient(Integer connectionTimeout, Integer readTimeout) {
        Client client = Client.create();
        client.setConnectTimeout(connectionTimeout);
        client.setReadTimeout(readTimeout);

        return client;
    }

    /**
     * Sends a query request to BHIE REST interface and
     * then polls the BHIE REST comet interface with the provided queryID.
     * <p/>
     * When the long poller receives valid response, it will bind the json response to a ResultCollection object
     * and then add the ResultCollection to a List.
     *
     * @return A List of ResultCollection objects.
     * @throws JMeadowsException If there is a failure.
     */
    @SuppressWarnings("unchecked")
    public Boolean logSystemInfo() throws JMeadowsException {
        boolean result = false;


        try {
            WebResource queryResource = restClient.resource(SYSTEM_MONITOR_PATH);
            String queryResp ="";

            if(APP_CONFIG.isSystemMonitorAuth()) {
                StringBuilder authString = new StringBuilder();
                authString.append(APP_CONFIG.getSystemMonitorEntity());
                authString.append(":");
                authString.append(APP_CONFIG.getSystemMonitorEntityToken());
                String authStringEnc = new BASE64Encoder().encode(authString.toString().getBytes());

                ClientResponse resp = queryResource.accept("application/json")
                        .header("Authorization", "Basic " + authStringEnc)
                        .get(ClientResponse.class);

                if (resp.getStatus() != 200) {
                    LOGGER.error("SYSTEM_MONITOR: Unable to connect to the server");
                }

                queryResp = resp.getEntity(String.class);
            } else {
                queryResp = queryResource.get(String.class);
            }

            SystemInfos sysInfos = objectMapper.readValue(queryResp, SystemInfos.class);

            if (sysInfos.getSystemInfo() != null && !sysInfos.getSystemInfo().isEmpty()) {
                DatabaseHandler.getInstance().logSystemMetrics(sysInfos);
                result = true;
            }
        } catch (Exception e) {
            LOGGER.error(e.getMessage());
            throw new JMeadowsException(e);
        }

        return result;
    }


}
