/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.des.dataservicehandler;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.ClientResponse;
import com.sun.jersey.api.client.WebResource;
import gov.va.med.des.beans.Transactions;
import gov.va.med.des.common.AppConfig;
import gov.va.med.des.common.JMeadowsException;
import gov.va.med.monitor.DatabaseHandler;
import org.apache.log4j.Logger;
import sun.misc.BASE64Encoder;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;


/**
 * TransactionMonitorHandler
 */
public class TransactionMonitorHandler extends DataServiceHandler {
    /**
     * Logger.
     */
    private static final Logger LOGGER =
            Logger.getLogger(TransactionMonitorHandler.class.getName());

    private static final String DES_TRANSACTION_SERVICE_PATH = AppConfig.getInstance().getDesTransactionServicePath();

    private static final int CONNECTION_TIMEOUT_MS = AppConfig.getInstance().getConnectionTimeoutMS();
    private static final int READ_TIMEOUT_MS = AppConfig.getInstance().getRestReadTimeoutMs();

    private static final TransactionMonitorHandler INSTANCE =
            new TransactionMonitorHandler(DES_TRANSACTION_SERVICE_PATH);

    private String endpoint;
    private ObjectMapper objectMapper;
    private Client restClient;

    /**
     * Creates a new DESRestServiceHandler instance.
     *
     * @param endpoint DESRestService end point
     */
    public TransactionMonitorHandler(String endpoint) {

        super();

        this.endpoint = endpoint;

        if (endpoint != null && '/' != endpoint.charAt(endpoint.length() - 1)) {
            this.endpoint += "/";
        }

        objectMapper = new ObjectMapper();

        //allow unknown data fields
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        this.restClient = initRestClient(CONNECTION_TIMEOUT_MS, READ_TIMEOUT_MS);
    }

    public static TransactionMonitorHandler getInstance() {
        return INSTANCE;
    }


    /**
     * Initializes REST client.
     *
     * @param connectionTimeout REST client connection timeout
     * @param readTimeout       REST client read timeout
     * @return Jersey REST Client
     */
    private Client initRestClient(Integer connectionTimeout, Integer readTimeout) {
        Client client = Client.create();
        client.setConnectTimeout(connectionTimeout);
        client.setReadTimeout(readTimeout);

        return client;
    }

    /**
     * Sends a query request to BHIE REST interface and
     * then polls the BHIE REST comet interface with the provided queryID.
     * <p/>
     * When the long poller receives valid response, it will bind the json response to a ResultCollection object
     * and then add the ResultCollection to a List.
     *
     * @return A List of ResultCollection objects.
     * @throws JMeadowsException If there is a failure.
     */
    @SuppressWarnings("unchecked")
    public Boolean queryTransactionMonitor() throws JMeadowsException {
        boolean result = false;

        long currentTime = System.currentTimeMillis();
        long queryPeriod = AppConfig.getInstance().getServiceMonitorTaskTimerPeriod();

        DateFormat formatter = new SimpleDateFormat("yyyyMMddHHmmss");
        formatter.setTimeZone(TimeZone.getTimeZone("UTC"));

//        String oldStartDate =  new SimpleDateFormat("yyyyMMddHHmmss").format(new Date(currentTime - queryPeriod));
//        String oldEndDate = new SimpleDateFormat("yyyyMMddHHmmss").format(new Date(currentTime));

        String startDate =  formatter.format(new Date(currentTime - queryPeriod));
        String endDate = formatter.format(new Date(currentTime));

        try {
            String serviceEndpoint = DES_TRANSACTION_SERVICE_PATH;
            if ("/".equals(serviceEndpoint.charAt(serviceEndpoint.length() - 1))) {
                serviceEndpoint = serviceEndpoint.substring(serviceEndpoint.length() - 1);
            }
            WebResource queryResource = restClient.resource(serviceEndpoint + "?queryStartDate=" + startDate + "&queryEndDate=" + endDate );

            String queryResp = "";

            if(APP_CONFIG.isDesTransactionServiceAuth()) {
                StringBuilder authString = new StringBuilder();
                authString.append(APP_CONFIG.getDesTransactionServiceEntity());
                authString.append(":");
                authString.append(APP_CONFIG.getDesTransactionServiceEntityToken());
                String authStringEnc = new BASE64Encoder().encode(authString.toString().getBytes());

                ClientResponse resp = queryResource.accept("application/json")
                        .header("Authorization", "Basic " + authStringEnc)
                        .get(ClientResponse.class);

                if (resp.getStatus() != 200) {
                    LOGGER.error("DES_TRANSACTION_SERVICE: Unable to connect to the server");
                }

                queryResp = resp.getEntity(String.class);
            } else {
               queryResp = queryResource.get(String.class);
            }

            Transactions transactions = objectMapper.readValue(queryResp, Transactions.class);

            if (transactions.getDetailList() != null && !transactions.getDetailList().isEmpty()) {
                DatabaseHandler.getInstance().logTransactions(transactions);
                DatabaseHandler.getInstance().logTransactionErrors(transactions);
                result = true;
            }
        } catch (Exception e) {
            LOGGER.error(e.getMessage());
            throw new JMeadowsException(e);
        }

        return result;
    }


}
