package gov.va.med.monitor;


import com.mchange.v2.c3p0.ComboPooledDataSource;
import gov.va.med.des.beans.*;
import gov.va.med.des.common.AppConfig;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;
import java.beans.PropertyVetoException;
import java.io.IOException;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;


public class DatabaseHandler extends ServiceBase {
    private static final Logger logger = LoggerFactory.getLogger(DatabaseHandler.class);
    private static final DatabaseHandler INSTANCE = new DatabaseHandler();
    protected static final AppConfig APP_CONFIG = AppConfig.getInstance();

    private static final String ERROR_STATUS = "DMIX_ERRORS";
    private DataSource dataSource;
    private ObjectMapper objectMapper;

    private DatabaseHandler() {
        super(ServiceType.DatabaseService);

        AppConfig appConfig = AppConfig.getInstance();

        objectMapper = new ObjectMapper();
        //allow unknown data fields
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        // Set up database connection
        try {
            if (appConfig.useC3PO()) {
                //setup c3po db connection pools
                ComboPooledDataSource c3poSQLConnectionPool = new ComboPooledDataSource();
                c3poSQLConnectionPool.setDriverClass(appConfig.getSQLDriver());
                c3poSQLConnectionPool.setJdbcUrl(appConfig.getSQLURL());

                dataSource = c3poSQLConnectionPool;
            } else {
                dataSource = InitialContext.doLookup(appConfig.getJDBC_JNDILookup());
            }
        } catch (PropertyVetoException e) {
            dataSource = null;
        } catch (NamingException e) {
            dataSource = null;
        }
    }

    public static DatabaseHandler getInstance() {
        return INSTANCE;
    }

    public void logStatuses(List<HostStatus> hostStatuses) {
        Connection connection = null;
        CallableStatement statement = null;

        try {
            connection = dataSource.getConnection();
            statement = connection.prepareCall("{call addDESSiteMonitors(?,?,?,?,?,?,?,?)}");

            for (HostStatus report : hostStatuses) {
                String hostName = report.getHostUri();
                int hostDuration = 0;

                if(!StringUtils.isEmpty(report.getDuration())){
                    hostDuration =  Integer.parseInt(report.getDuration());
                }


                for(Status status : report.getStatusList()){
                    int duration = 0;
                    if(!StringUtils.isEmpty(status.getDuration())){
                        duration =  Integer.parseInt(status.getDuration());
                    }

                    statement.setString(1, hostName);
                    statement.setLong(2, hostDuration);
                    statement.setTimestamp(3, bhieDateToSQL(status.getStartTime()));
                    statement.setTimestamp(4, bhieDateToSQL(status.getEndTime()));
                    statement.setLong(5,duration);
                    statement.setString(6, status.getSiteName());
                    statement.setString(7, status.getSiteStatusCode());
                    statement.setString(8, status.getSiteStatusDescription());
                    statement.addBatch();
                }
            }

            statement.executeBatch();

            status = ServiceStatus.OK;
            message.append(SERVICE_OK_MSG);

            logger.debug("Logging to database completed");
        } catch (SQLException e) {
            status = ServiceStatus.ERROR;
            message.append(e.getMessage());

            logger.error("Logging to database failed: " + e.getMessage());
        } finally {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db statement.");
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db connection");
                }
            }
        }

//        report(type.toString(), status.toString(), message.toString());
    }

    public void logTransactions(Transactions transactions) {
        Connection connection = null;
        CallableStatement statement = null;

        try {
            connection = dataSource.getConnection();
            statement = connection.prepareCall("{call addDESServiceMeasures(?,?,?,?,?,?,?)}");

            String site = "";
            String serviceStatus = "";

            if (transactions != null) {
//                if (transactions.getAdapterInfo() != null
//                        && !StringUtils.isEmpty(transactions.getAdapterInfo().getHostUri())) {
//                    host = transactions.getAdapterInfo().getHostUri();
//                }

                if (transactions.getStatus() != null) {
                    site = transactions.getStatus().getSiteName();
                    serviceStatus = transactions.getStatus().getSiteStatusCode();
                }


            }

            if(transactions != null && transactions.getDetailList() !=null && transactions.getDetailList().size() > 0) {
                for (Detail report : transactions.getDetailList()) {

                    String host = "";
                    String measureName = "";
                    String appName = "";

                    if (!report.getName().equals(ERROR_STATUS)) {
                        java.sql.Timestamp runDate = null;

                        if (!StringUtils.isEmpty(report.getRunDate())) {
                            runDate = bhieDateToSQL(report.getRunDate());
                        }
                        if (!StringUtils.isEmpty(report.getName())) {
                            measureName = report.getName();
                        }
                        if(!StringUtils.isEmpty(report.getComment())){
                            String[] comments = report.getComment().split("-", 2);
                            if(comments.length > 0 && !StringUtils.isEmpty(comments[0])){
                                host = comments[0];
                            }
                            if(comments.length > 1 && !StringUtils.isEmpty(comments[1])){
                                appName = comments[1];
                            }
                        }

                        statement.setTimestamp(1, runDate); // format into date
                        statement.setString(2, host);
                        statement.setString(3, site);
                        statement.setString(4, serviceStatus);
                        statement.setString(5, measureName);
                        statement.setInt(6, report.getValue());
                        statement.setString(7, appName);
                        statement.addBatch();
                    }
                }

                statement.executeBatch();
            }

            status = ServiceStatus.OK;
            message.append(SERVICE_OK_MSG);

            logger.debug("Logging to database completed");
        } catch (SQLException e) {
            status = ServiceStatus.ERROR;
            message.append(e.getMessage());

            logger.error("Logging to database failed: " + e.getMessage());
        } finally {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db statement.");
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db connection");
                }
            }
        }

//        report(type.toString(), status.toString(), message.toString());
    }

    public void logTransactionErrors(Transactions transactions) {
        Connection connection = null;
        CallableStatement statement = null;

        try {
            connection = dataSource.getConnection();
            statement = connection.prepareCall("{call addDESLogErrors(?,?,?,?,?,?)}");

            String host = "";
            String site = "";
            String serviceStatus = "";

            if (transactions != null) {
                if (transactions.getAdapterInfo() != null
                        && !StringUtils.isEmpty(transactions.getAdapterInfo().getHostUri())) {
                    host = transactions.getAdapterInfo().getHostUri();
                }

                if (transactions.getStatus() != null) {
                    site = transactions.getStatus().getSiteName();
                    serviceStatus = transactions.getStatus().getSiteStatusCode();
                }


            }

            if(transactions != null && transactions.getDetailList() != null && transactions.getDetailList().size() > 0) {
                for (Detail report : transactions.getDetailList()) {

                    if (report.getName().equals(ERROR_STATUS)) {
                        java.sql.Timestamp runDate = null;
                        String comment = "";

                        if (!StringUtils.isEmpty(report.getRunDate())) {
                            runDate = bhieDateToSQL(report.getRunDate());
                        }
                        if (!StringUtils.isEmpty(report.getComment())) {
                            comment = report.getComment();
                        }

                        statement.setTimestamp(1, runDate); // format into date
                        statement.setString(2, host);
                        statement.setString(3, site);
                        statement.setString(4, serviceStatus);
                        statement.setString(5, comment);
                        statement.setInt(6, report.getValue());
                        statement.addBatch();
                    }
                }

                statement.executeBatch();
            }

            status = ServiceStatus.OK;
            message.append(SERVICE_OK_MSG);

            logger.debug("Logging to database completed");
        } catch (SQLException e) {
            status = ServiceStatus.ERROR;
            message.append(e.getMessage());

            logger.error("Logging to database failed: " + e.getMessage());
        } finally {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db statement.");
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db connection");
                }
            }
        }

//        report(type.toString(), status.toString(), message.toString());
    }

    public void logSystemMetrics(SystemInfos sysInfos) {
        Connection connection = null;
        CallableStatement statement = null;

        try {
            connection = dataSource.getConnection();
            statement = connection.prepareCall("{call addDESSystemMetrics(?,?,?,?,?,?)}");

            for (SystemInfo sysInfo : sysInfos.getSystemInfo()) {

                java.sql.Timestamp captureDate = null;
                String host = "";

                float cpu;
                long usedMemory;
                long totalMemory;
                long freeMemory;

                SystemData sysData = null;

                if (sysInfo != null) {
                    if (sysInfo.getHost() != null) {
                        host = sysInfo.getHost();
                    }
                    if (!StringUtils.isEmpty(sysInfo.getCaptureDate())) {
                        captureDate = bhieDateToSQL(sysInfo.getCaptureDate());
                    }

                }

                try {
                    if(sysInfo != null && sysInfo.getSystemData() != null) {
                        sysData = objectMapper.readValue(sysInfo.getSystemData(), SystemData.class);
                    }

                    if (sysData != null) {
                        cpu = sysData.getCpuPercent();
                        usedMemory = sysData.getUsedBytes();
                        totalMemory = sysData.getTotalBytes();
                        freeMemory = sysData.getFreeBytes();

                        statement.setTimestamp(1, captureDate); // format into date
                        statement.setString(2, host);
                        statement.setFloat(3, cpu);
                        statement.setLong(4, usedMemory);
                        statement.setLong(5, totalMemory);
                        statement.setLong(6, freeMemory);
                        statement.addBatch();

                        logDiskMetrics(captureDate, host, sysData.getUsedDisk());

                    }

                    statement.executeBatch();

                } catch (IOException e) {
                    logger.error("Error while mapping from JSON: " + e.getMessage());
                } catch (Exception e) {
                    logger.error("Error while mapping from JSON: " + e.getMessage());
                }


            }

            status = ServiceStatus.OK;
            message.append(SERVICE_OK_MSG);

            logger.debug("Logging to database completed");
        } catch (SQLException e) {
            status = ServiceStatus.ERROR;
            message.append(e.getMessage());

            logger.error("Logging to database failed: " + e.getMessage());
        } finally {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db statement.");
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db connection");
                }
            }
        }

//        report(type.toString(), status.toString(), message.toString());
    }

    public void logDiskMetrics(java.sql.Timestamp timepoint, String host, List<DiskData> diskData) {
        Connection connection = null;
        CallableStatement statement = null;

        try {
            connection = dataSource.getConnection();
            statement = connection.prepareCall("{call addDESDiskMetrics(?,?,?,?,?,?)}");

            for (DiskData data : diskData) {

                String diskName;
                long usedBytes;
                long totalBytes;
                long freeBytes;

                if (data != null) {
                    diskName = data.getMountPoint();
                    usedBytes = data.getUsedBytes();
                    totalBytes = data.getTotalBytes();
                    freeBytes = data.getFreeBytes();

                    statement.setTimestamp(1, timepoint); // format into date
                    statement.setString(2, host);
                    statement.setString(3, diskName);
                    statement.setLong(4, freeBytes);
                    statement.setLong(5, totalBytes);
                    statement.setLong(6, usedBytes);
                    statement.addBatch();
                }
                statement.executeBatch();
            }


            status = ServiceStatus.OK;
            message.append(SERVICE_OK_MSG);

            logger.debug("Logging to database completed");
        } catch (SQLException e) {
            status = ServiceStatus.ERROR;
            message.append(e.getMessage());

            logger.error("Logging to database failed: " + e.getMessage());
        } finally {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db statement.");
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db connection");
                }
            }
        }

//        report(type.toString(), status.toString(), message.toString());
    }

    public boolean purgeLogs() {
        boolean result = false;

        Connection connection = null;
        CallableStatement statement = null;

        try {
            connection = dataSource.getConnection();
            statement = connection.prepareCall("{call purgeDESMonitoring(?)}");

            if(APP_CONFIG.getPurgeInterval() > 0){

                Calendar cal = Calendar.getInstance();
                cal.add(Calendar.DATE, -APP_CONFIG.getPurgeInterval());

                Timestamp ts = new Timestamp(cal.getTime().getTime());

                statement.setTimestamp(1, ts); // format into date
                statement.execute();

            status = ServiceStatus.OK;
            message.append(SERVICE_OK_MSG);

            logger.debug("purging of entries completed");
                result = true;
            }
        } catch (SQLException e) {
            status = ServiceStatus.ERROR;
            message.append(e.getMessage());

            logger.error("Purging of entries failed: " + e.getMessage());
        } finally {
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db statement.");
                }
            }

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    logger.error("Problem closing db connection");
                }
            }
        }

        return result;
    }

    @Override
    public void run() {
        // Nothing to implement. Monitoring happens when it's time to log results to
        // the database. Service report is generated in the logReports() method above.
    }

    /**
     * GregorianCalendar representation of date format string.
     *
     * @param dateStr Date format string
     * @return GregorianCalendar representation of date format string.
     */
    protected static java.sql.Timestamp bhieDateToSQL(String dateStr) {

        Date date;
        String format;
        java.sql.Timestamp retDate = null;
        SimpleDateFormat df;
        boolean isUTC = false;

        if (dateStr != null && dateStr.length() != 0) {
            try {

                if (dateStr.indexOf("/") > 0) {
                    format = "MM/dd/yy";
                } else if (dateStr.indexOf(" ") == 2) {
                    format = "dd MMM yyyy";
                } else if (dateStr.indexOf("T") > 0 && dateStr.indexOf("Z") > 0) {
                    format = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
                } else if (dateStr.indexOf("T") > 0 && dateStr.indexOf(":") > 0) {
                    dateStr = dateStr.replace("-00:00", "");
                    format = "yyyy-MM-dd'T'HH:mm:ss";
                } else if (dateStr.indexOf("-") > 0 && dateStr.indexOf(":") > 0) {
                    format = "yyyy-MM-dd HH:mm:ss";
                } else if (dateStr.indexOf("-") > 0 && dateStr.indexOf("@") > 0) {
                    format = "dd-MMM-yyyy@HHmm";
                } else if (dateStr.indexOf(",") > 0 && dateStr.indexOf("@") > 0) {
                    format = "MMM dd, yyyy@HH:mm";
                } else if (dateStr.indexOf("@") < 0 && dateStr.indexOf("-") == 4) {
                    format = "yyyy-MM-dd";
                } else if (dateStr.indexOf("@") > 0 && dateStr.indexOf("-") < 0) {
                    format = "ddMMMyy@HHmm";
                } else if (dateStr.length() == 8 && dateStr.indexOf("-") < 0) {
                    format = "yyyyMMdd";
                } else if (dateStr.length() == 12) {
                    format = "yyyyMMddHHmm";
                } else if (dateStr.length() == 14) {
                    format = "yyyyMMddHHmmss";
                } else if (dateStr.indexOf(".") > 12 && dateStr.length() > 14 && dateStr.length() < 19) {
                    //assumeUTC = false;
                    format = "yyyyMMddHHmmss.SSS";
                } else if (dateStr.indexOf(".") > 12 && dateStr.length() > 18) {
                    //assumeUTC = false;
                    format = "yyyyMMddHHmmss.SSSZ";
                } else if (dateStr.length() == 7) {
                    format = "ddMMMyy";
                } else if (dateStr.length() == 19 && (dateStr.contains("AM") || dateStr.contains("PM"))) {
                    format = "MMM dd yyyy hh:mma";
                } else if (dateStr.length() == 20) {
                    dateStr = dateStr.substring(0, 17);
                    format = "yyyyMMddHHmmssSSS";
                    isUTC = true;
                } else {
                    return null;
                }
                if (isUTC) {
                    df = new SimpleDateFormat(format);
                    df.setTimeZone(TimeZone.getTimeZone("UTC"));
                } else {
                    df = new SimpleDateFormat(format, Locale.getDefault());
                }

                date = df.parse(dateStr);

                GregorianCalendar retCal = new GregorianCalendar();

                retCal.setTime(date);
                retCal.setTimeZone(df.getTimeZone());

                retDate = new java.sql.Timestamp(retCal.getTimeInMillis());

            } catch (ParseException e) {
                logger.error(e.getMessage());
            }
        }

        return retDate;

    }

    /**
     * GregorianCalendar representation of date format string.
     *
     * @param durationStr Date format string
     * @return GregorianCalendar representation of date format string.
     */
    protected static long bhieDurationToMS(String durationStr) {

        Date date1;
        Date date2;
        String format = "H:m:s.S";
        long retVal = 0;
        SimpleDateFormat df;

        if (durationStr != null && durationStr.length() != 0) {
            try {

                df = new SimpleDateFormat(format, Locale.getDefault());

                date1 = df.parse("00:00:00.00");
                date2 = df.parse(durationStr);

                retVal = date2.getTime() - date1.getTime();

            } catch (ParseException e) {
                logger.error(e.getMessage());
            }
        }

        return retVal;

    }
}
