/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.PatientDocument;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.*;
import gov.va.med.jmeadows.smartbeans.patient.SmartDocument;
import gov.va.med.vhahon.common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 *
 * Patient notes data access object.
 */
public class DocumentDao extends DaoBase {

    /**
     * Retrieves a list of patient documents from available patient locations.
     *
     * @return A list of patient documents
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientDocument> getPatientDocuments() throws JMeadowsException {
        return getPatientDocuments();
    }

    /**
     * Retrieves a list of patient documents from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T).
     *
     * @return A list of patient documents
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientDocument> getPatientDocuments(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        List<PatientDocument> resultList = new ArrayList<PatientDocument>();

        auditQuery("PatDocuments", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                //ProgressNotes & Consults
                MultiLocationQueryConfig patNotesConfig = new PatientMultiLocQueryCfgBuilder()
                        .dataServiceHandler(sm_bhieRelayServiceHandler)
                        .dataServiceMethod("getPatientNotes")
                        .dataServiceMethodArgs(bhieQueryBean)
                        .smartBeanMappingMethod("mapProgressNotesFromBHIE")
                        .location(bhieLoc)
                        .build();

//                //ProgressNotes-encounterNotes
//                MultiLocationQueryConfig patEncounterNotesConfig = new PatientMultiLocQueryCfgBuilder()
//                        .dataServiceHandler(sm_bhieRelayServiceHandler)
//                        .dataServiceMethod("getEncounterNotes")
//                        .dataServiceMethodArgs(bhieQueryBean)
//                        .smartBeanMappingMethod("mapEncounterNotesFromBHIE")
//                        .location(bhieLoc)
//                        .build();

//                //Outpatient Encounters
                MultiLocationQueryConfig patEncountersConfig = new PatientMultiLocQueryCfgBuilder()
                        .dataServiceHandler(sm_bhieRelayServiceHandler)
                        .dataServiceMethod("getPatientEncounters")
                        .dataServiceMethodArgs(bhieQueryBean)
                        .smartBeanMappingMethod("mapEncountersFromBHIE")
                        .location(bhieLoc)
                        .build();

//                //Discharge/Essentris Notes
                MultiLocationQueryConfig patDischargeSummariesConfig =
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getPatientDischargeSummaries")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapPatientDischargeSummariesFromBHIE")
                                .location(bhieLoc)
                                .build();

//                //Questionnaires
                MultiLocationQueryConfig patQuestionnairesConfig =
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getPatientQuestionnaires")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapQuestionnairesFromBHIE")
                                .location(bhieLoc)
                                .build();

                //DeploymentForms
                MultiLocationQueryConfig patDeploymentFormsConfig =
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getDeploymentForms")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapDeploymentFormsFromBHIE")
                                .location(bhieLoc)
                                .build();

                //Radiology Exams
                MultiLocationQueryConfig patRadiologyExamsConfig =
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getPatientRadiologyReports")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapRadiologyExamFromBHIE")
                                .location(bhieLoc)
                                .build();

                //HAIMS Documents
                MultiLocationQueryConfig patHAIMSDocumentsConfig = new PatientMultiLocQueryCfgBuilder()
                        .dataServiceHandler(sm_bhieRelayServiceHandler)
                        .dataServiceMethod("getHAIMSDocuments")
                        .dataServiceMethodArgs(bhieQueryBean)
                        .smartBeanMappingMethod("mapDocumentsFromBHIE")
                        .location(bhieLoc)
                        .build();

                configs.add(patNotesConfig);
//                configs.add(patEncounterNotesConfig);
                configs.add(patEncountersConfig);
                configs.add(patDischargeSummariesConfig);
                configs.add(patQuestionnairesConfig);
                configs.add(patDeploymentFormsConfig);
                configs.add(patRadiologyExamsConfig);
                configs.add(patHAIMSDocumentsConfig);

            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientProgressNotes")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .smartBeanMappingMethod("mapProgressNotesFromVistA")
                                .location(loc)
                                .build()
                );
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientEncounters")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .smartBeanMappingMethod("mapEncountersFromVistA")
                                .location(loc)
                                .build()
                );
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientConsultRequests")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .smartBeanMappingMethod("mapConsultsFromVistA")
                                .location(loc)
                                .build()
                );
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientRads")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .smartBeanMappingMethod("mapRadiologyFromVistA")
                                .location(loc)
                                .build()
                );
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientDischargeSummaries")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .smartBeanMappingMethod("mapInpatientNotesFromVistA")
                                .location(loc)
                                .build()
                );

            }
        }

        Object[] docObjs = MultiLocationQuery.execute(
                configs, new SmartDocument(), queryBean);

        List<PatientDocument> documents = null;

        //TODO: reference OrdersDao for mapping multiple domains
        if (docObjs != null) {
            documents = Arrays.asList((PatientDocument[]) docObjs);
//            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, documents, CodingSystem.NOTES);
            List<PatientDocument> notesList = new ArrayList<PatientDocument>();
            List<PatientDocument> radList = new ArrayList<PatientDocument>();
            List<PatientDocument> encounters = new ArrayList<PatientDocument>();
            List<PatientDocument> othersList = new ArrayList<PatientDocument>();
            for(PatientDocument patDoc : documents){
                if(SmartDocument.DOCUMENT_PROGRESS_NOTE.equalsIgnoreCase(patDoc.getDocumentType())
                   || SmartDocument.DOCUMENT_ENCOUNTER_NOTE.equalsIgnoreCase(patDoc.getDocumentType())
                   || SmartDocument.DOCUMENT_INPATIENT_NOTE.equalsIgnoreCase(patDoc.getDocumentType())){
                    notesList.add(patDoc);
                }
                else if(SmartDocument.DOCUMENT_OUTPATIENT_ENCOUNTER.equalsIgnoreCase(patDoc.getDocumentType())){
                    encounters.add(patDoc);
                }
                else if(SmartDocument.DOCUMENT_RADIOLOGY.equalsIgnoreCase(patDoc.getDocumentType())){
                    radList.add(patDoc);
                }
                else {
                    othersList.add(patDoc);
                }
            }
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, notesList, CodingSystem.NOTES);
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, encounters, CodingSystem.SNOMED);
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, radList, CodingSystem.RADIOLOGY);

            resultList.addAll(notesList);
            resultList.addAll(encounters);
            resultList.addAll(radList);
            resultList.addAll(othersList);
        }

        return resultList;
    }
}
