/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartInsurance;
import gov.va.med.vds.webservice.InsuranceBean;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Patient insurance data access object
 */
public class InsuranceDao extends DaoBase {
//    private static final Logger LOGGER = LoggerFactory.getLogger(InsuranceDao.class);

    /**
     * Retrieves a list of patient insurances from all available patient locations
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @return A list of patient insurances
     * @throws JMeadowsException
     */
    public List<InsuranceBean> getPatientInsurances(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatInsurances", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                // Exclude date range to retrieve entire data set
                bhieQueryBean.setStartDate(null);
                bhieQueryBean.setEndDate(null);

                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                        .dataServiceHandler(sm_bhieRelayServiceHandler)
                        .dataServiceMethod("getPatientInsurances")
                        .dataServiceMethodArgs(bhieQueryBean)
                        .smartBeanMappingMethod("mapInsurancesFromBHIE")
                        .location(bhieLoc)
                        .build()
                );

                if (queryBean.isIncludeVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_bhieRelayServiceHandler)
                                    .dataServiceMethod("getVLERPatientInsurances")
                                    .dataServiceMethodArgs(bhieQueryBean)
                                    .smartBeanMappingMethod("mapInsurancesFromBHIE")
                                    .location(bhieLoc)
                                    .build()
                    );
                }
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                        .dataServiceHandler(sm_vistaHandler)
                        .dataServiceMethod("getPatientInsurances")
                        .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                        .location(loc)
                        .build()
                );
            }
        }

        Object[] insObjs = MultiLocationQuery.execute(
                configs, new SmartInsurance(), queryBean);

        List<InsuranceBean> insurances = null;

        if (insObjs != null) {
            insurances = Arrays.asList((InsuranceBean[]) insObjs);

            // Map insurance concept codes
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, insurances, CodingSystem.X12);
        }

        return insurances;
    }
}
