/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartMedication;
import gov.va.med.vds.webservice.FreeTextReport;
import gov.va.med.vds.webservice.Medication;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vhahon.common.Utils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeConstants;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

/**
 * Patient medications data access object
 */
public class MedicationsDao extends DaoBase {

    private static final Logger LOGGER = LoggerFactory.getLogger(MedicationsDao.class);

    /**
     * Retrieves a list of patient medications from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * status (possible options: O - outpatient, I - inpatient, defaults to "O"
     * active (defaults to 1)
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T).
     *
     * @return List of patient medications.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Medication> getPatientMedications(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatMedications", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        String medType = queryBean.getStatus();

        if (medType == null) {
            medType = "O";
        }

        if (queryBean.getActive() == null) {
            queryBean.setActive(1);
        }

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                if ("I".equals(medType)) {
                    PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                    List<String> includedSources = new ArrayList<String>();
                    includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                    gov.va.med.bhie.QueryBean bhieQueryBean =
                            BHIERelayBeanFactory.create(queryBean, includedSources, null);

                    // JLV-942 and JLV-943 (Clinical Relevant Date) pad 18 months to start date and filter back
                    try {
                        DatatypeFactory datatypeFactory = DatatypeFactory.newInstance();
                        bhieQueryBean.getStartDate().add(datatypeFactory.newDuration(false, 0, 18, 0, 0, 0, 0));
                    } catch (DatatypeConfigurationException e) {
                        LOGGER.error(e.getMessage());
                    }

                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_bhieRelayServiceHandler)
                                    .dataServiceMethod("getPatientMedications")
                                    .dataServiceMethodArgs(bhieQueryBean)
                                    .smartBeanMappingMethod("getInpatientMedicationsFromBHIE")
                                    .location(bhieLoc)
                                    .build()
                    );
                } else {
                    PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                    List<String> includedSources = new ArrayList<String>();
                    includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());
                    includedSources.add(APP_CONFIG.getBHIECustodianOidSHARE());

                    gov.va.med.bhie.QueryBean bhieQueryBean =
                            BHIERelayBeanFactory.create(queryBean, includedSources, null);

                    // JLV-942 and JLV-943 (Clinical Relevant Date) pad 18 months to start date and filter back
                    try {
                        DatatypeFactory datatypeFactory = DatatypeFactory.newInstance();
                        bhieQueryBean.getStartDate().add(datatypeFactory.newDuration(false, 0, 18, 0, 0, 0, 0));
                    } catch (DatatypeConfigurationException e) {
                        LOGGER.error(e.getMessage());
                    }

                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_bhieRelayServiceHandler)
                                    .dataServiceMethod("getPatientMedications")
                                    .dataServiceMethodArgs(bhieQueryBean)
                                    .smartBeanMappingMethod("getOutpatientMedicationsFromBHIE")
                                    .location(bhieLoc)
                                    .build()
                    );
                }
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientMedications")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .location(loc)
                                .build()
                );
            }
        }

        SmartMedication smartMedication = new SmartMedication();

        Object[] medObjs = MultiLocationQuery.execute(
                configs, smartMedication, queryBean);

        List<Medication> medications = null;

        if (medObjs != null) {
            medications = Arrays.asList((Medication[]) medObjs);

            //map medication concept codes
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, medications, CodingSystem.RXNORM);

            // JLV-942 and JLV-943 DoD filter Outpatient Meds on lastFillDate and Inpatient Meds on startDate
            // filter out for date range
            XMLGregorianCalendar startDate = null;
            XMLGregorianCalendar endDate = null;

            try {
                DatatypeFactory datatypeFactory = DatatypeFactory.newInstance();
                Date startDateDate = queryBean.getStartDate().getTime();
                GregorianCalendar startDateGregorianCalendar = new GregorianCalendar();
                startDateGregorianCalendar.setTime(startDateDate);
                startDate = datatypeFactory.newXMLGregorianCalendar(startDateGregorianCalendar);
                Date endDateDate = queryBean.getEndDate().getTime();
                GregorianCalendar endDateGregorianCalendar = new GregorianCalendar();
                endDateGregorianCalendar.setTime(endDateDate);
                endDate = datatypeFactory.newXMLGregorianCalendar(endDateGregorianCalendar);
            } catch (DatatypeConfigurationException e) {
                LOGGER.error(e.getMessage());
            }

            ArrayList<Medication> filteredList = new ArrayList<Medication>();

            for (Medication medication : medications) {
                if ((medication.getSite() != null && !StringUtils.isEmpty(medication.getSite().getMoniker())
                        && BHIE_SOURCE_STATUS_REPORT_FLAG.equals(medication.getSite().getMoniker()))
                        || medication.getDrugName().toLowerCase().indexOf("connection unavailable") > -1) {
                    filteredList.add(medication);

                } else if (startDate != null && endDate != null) {
                    // filter on Outpatient lastFilledDate
                    if (!"I".equals(medType)) {
                        XMLGregorianCalendar lastFilledDate = medication.getLastFilledDate();
                        if (lastFilledDate == null ||
                                (lastFilledDate.compare(startDate) == DatatypeConstants.GREATER || lastFilledDate.compare(startDate) == DatatypeConstants.EQUAL)
                                        && (lastFilledDate.compare(endDate) == DatatypeConstants.LESSER || lastFilledDate.compare(endDate) == DatatypeConstants.EQUAL)) {
                            filteredList.add(medication);
                        }
                    }
                    // filter on Inpatient startDate
                    else {
                        XMLGregorianCalendar medStartDate = medication.getStartDate();
                        if (medStartDate == null ||
                                (medStartDate.compare(startDate) == DatatypeConstants.GREATER || medStartDate.compare(startDate) == DatatypeConstants.EQUAL)
                                        && (medStartDate.compare(endDate) == DatatypeConstants.LESSER || medStartDate.compare(endDate) == DatatypeConstants.EQUAL)) {
                            filteredList.add(medication);
                        }
                    }
                }

            }
            return filteredList;
        }

        return null;
    }

    /**
     * Retrieves medication record report from the provided VistA location.
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     * recordSiteCode
     * itemId
     *
     * @return Inpatient medication record report.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public FreeTextReport getMedicationDetail(JMeadowsQuery queryBean) throws JMeadowsException {

        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.PATIENT,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);

        auditQuery("MedicationDetail", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        FreeTextReport medDetail = null;

        Site recordSite = getSite(queryBean.getRecordSiteCode());

        for (PatientLocation loc : locs) {

            // get vista data
            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())
                    && LocationHelper.vistaLocationsContainSiteCode(recordSite,
                    queryBean.getPatient().getVistaSites())) {

                medDetail = sm_vistaHandler.getMedicationDetail(VistaBeanFactory.create(queryBean));
                break;
            }
        }

        return medDetail;
    }
}
