/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartOutpatientProcedure;
import gov.va.med.vds.webservice.Procedure;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Outpatient Procedures data access object
 */
public class OutpatientProcedureDao extends DaoBase {

    /**
     * Retrieves a list of outpatient procedures from all available patient locations
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @return a list of outpatient procedures.
     * @throws JMeadowsException
     */
    public List<Procedure> getOutpatientProcedures(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("OutpatientProcedures", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getOutpatientProcedures")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapOutpatientProceduresFromBHIE")
                                .location(bhieLoc)
                                .build()
                );

                if (queryBean.isIncludeVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_bhieRelayServiceHandler)
                                    .dataServiceMethod("getVLERPatientProcedures")
                                    .dataServiceMethodArgs(bhieQueryBean)
                                    .smartBeanMappingMethod("mapOutpatientProceduresFromBHIE")
                                    .location(bhieLoc)
                                    .build()
                    );
                }
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientProcedures")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .location(loc)
                                .build()
                );
                if (queryBean.isIncludeDASVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_vistaVLERHandler)
                                    .dataServiceMethod("getDASVLERProcedures")
                                    .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                    .smartBeanMappingMethod("mapDASVLERProcedures")
                                    .location(loc)
                                    .build()
                    );
                }
            }
        }

        Object[] insObjs = MultiLocationQuery.execute(
                configs, new SmartOutpatientProcedure(), queryBean);

        List<Procedure> outpatientProcedures = null;

        if (insObjs != null) {
            outpatientProcedures = Arrays.asList((Procedure[]) insObjs);

            // Map outpatient procedure concept code
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, outpatientProcedures, CodingSystem.CPT);
        }

        return outpatientProcedures;
    }
}
