/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.jmeadows.dao.patient;

import gov.va.med.bhie.DemographicV4;
import gov.va.med.bhie.RecordIdentifierV4;
import gov.va.med.bhie.V4DataRecord;
import gov.va.med.bhie.V4ResultCollection;
import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.beans.PDWSQueryBean;
import gov.va.med.jmeadows.dao.beans.ResponsePatientQuery;
import gov.va.med.jmeadows.dao.beans.ResponsePatientSelect;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.UserLocation;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.heartbeat.ServiceErrors;
import gov.va.med.jmeadows.smartbeans.patient.*;
import gov.va.med.vds.webservice.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.datatype.XMLGregorianCalendar;

/**
 * Patient data access object.
 */
public class PatientDao extends DaoBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(PatientDao.class);

    private void mapDemographicsToRecentlyViewedPatients(Patient patient, PatientDemographics demographics)
    {
        if (patient != null && demographics != null) {
            if (StringUtils.isBlank(patient.getName())) {
                patient.setName(demographics.getName());
            }
            if (StringUtils.isBlank(patient.getEDIPI())) {
                patient.setEDIPI(demographics.getEDIPI());
            }
            if (StringUtils.isBlank(patient.getSSN())) {
                patient.setSSN(demographics.getSSN());
            }
            if (StringUtils.isBlank(patient.getFMP())) {
                patient.setFMP(demographics.getSponsorSSN());
            }
            if (StringUtils.isBlank(patient.getGender())) {
                patient.setGender(demographics.getGender());
            }
            if (StringUtils.isBlank(patient.getDob())) {
                patient.setDob(demographics.getDob());
            }
        }
    }

    /**
     * Performs a patient lookup against the provided site location.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * recordSiteCode - site to perform search against
     * itemId - patient name
     *
     * @return A list of patients that match search criteria.
     * @throws JMeadowsException if there is an error.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     * @deprecated Use lookupPatientPDWS
     */
    @Deprecated
    public List<Patient> lookupPatient(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean,
                QueryBeanParams.USER,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);

        String requestingApp = queryBean.getRequestingApp();

//        Patient bean = null;
//        org.hui.chcs.QueryBean qb_chcs;

        UserLocation loc = LocationHelper.getUserLocation(queryBean.getUser());

        AuditDao auditDao = new AuditDao();

        User user = queryBean.getUser();

        String userIen = user.getUserIen();

        //append providerId to userIen if one exists
        if (StringUtils.isNotBlank(user.getProviderIen())) {
            userIen += "^" + user.getProviderIen();
        }

        auditDao.setAudit((String) queryBean.getUser().getHostSite().getSiteCode(),
                queryBean.getSiteMoniker(), queryBean.getSiteAgency(),
                queryBean.getItemId(),
                userIen,
                "",
                //queryBean.getUser().getUserNPI(),
                queryBean.getUser().getName(),
                queryBean.getItemId(),
                "PatientLookup", requestingApp, "", "", "");

        List<Patient> pats = null;

        if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {

            pats = sm_vistaHandler.lookupPatient(VistaBeanFactory.create(queryBean));

        }

        if (pats != null && queryBean.getMax() != null
                && queryBean.getMax() < pats.size()) {
            pats = pats.subList(0, queryBean.getMax());
        }

        return pats;
    }

    /**
     * Performs a patient lookup against the DMDC Patient Discovery Web Service (PDWS).
     *
     * @param queryBean PDWS Query parameters.
     * @return A list of patients that match search criteria.
     * @throws JMeadowsException if there is an error.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public ResponsePatientQuery lookupPatientPDWS(PDWSQueryBean queryBean) throws JMeadowsException {
/*      TODO: PDWSQueryBean parameter check
        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.DATE_RANGE,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);
*/
        AuditDao auditDao = new AuditDao();
        User user = queryBean.getUser();
        String siteId = "";
        String userIen = "";
        String userNPI = "";
        String userName = "";
        String info = "";
        String requestingApp = "";
        //In the future be able to grab this information from PDWS
        String siteMoniker = "";
        String siteAgency = "";
        String recordID = "";
        String complexTransaction = "";

        siteId = user.getHostSite().getSiteCode();
        userIen = user.getUserIen();
        userNPI = user.getNPI();
        userName = user.getName();
        info = queryBean.getPatientId()
                + "^" + queryBean.getLastName()
                + "^" + queryBean.getFirstName()
                + "^" + queryBean.getDob()
                + "^" + queryBean.getGender();
        requestingApp = queryBean.getRequestingApp();

        auditDao.setAudit(siteId, siteMoniker, siteAgency, recordID, userIen, userNPI, userName, queryBean.getPatientId(), "lookupPatientPDWS", requestingApp, "", "", info, user.getCardId(), user.getEmail(), complexTransaction);

        ResponsePatientQuery responsePatientQuery = null;
        try {
            responsePatientQuery = sm_pdwsHandler.lookupPatient(queryBean);

            if (!responsePatientQuery.isSuccess()) {
                auditDao.setAudit(siteId, siteMoniker, siteAgency, recordID, userIen, userNPI, userName, queryBean.getPatientId(), "Error: PDWS Service", requestingApp, "", "", info, user.getCardId(), user.getEmail(), complexTransaction);
                LOGGER.error(responsePatientQuery.getErrorMsg());
            } else {
                // if user has site restricted list, need to correlate with MVI and filter for sites.
                List<Object> restrictedSiteCodesList = user.getRestrictedSiteCodesList();
                // assume it was EDIPI search and only one patient came from the PDWS search with that EDIPI
                if (CollectionUtils.isNotEmpty(restrictedSiteCodesList)
                    && CollectionUtils.isNotEmpty(responsePatientQuery.getPatients())) {
                    
                    JMeadowsQuery jMeadowsQueryBean = new JMeadowsQuery();
                    jMeadowsQueryBean.setUser(user);
                    jMeadowsQueryBean.setItemId(queryBean.getPatientId()); // set EDIPI
                    jMeadowsQueryBean.setStatus("EDIPI"); // idType
                    ResponsePatientSelect result = sm_mviHandler.getPatientIds(jMeadowsQueryBean, SITES.getSiteMap());
                    if (result.isSuccess() && result.getPatient() != null) {
                        // check the patient against the sites list
                        boolean allowPatient = false;
                        // since there is a restricted list, at first assume this patient is not allowed to be added to lookup
                        for (Site vistaSite : result.getPatient().getVistaSites()) {
                            // patient's vistaSites added to patient.getVistaSites() during parsing of identifiers; check those sites against restricted list
                            if (restrictedSiteCodesList.contains(vistaSite.getSiteCode())) {
                                // this patient has an id at a vista site in the site list so set this patient back to unrestricted
                                allowPatient = true;
                                break;
                            }
                        }
                        // add only unrestricted patients to lookup list
                        if (!allowPatient) {
                            if (APP_CONFIG.useEnhancedDebug()) {
                                StringBuilder debugString = new StringBuilder("!!! VBAUSER SITERESTRICTED: ");
                                // Log that a patient was filtered out.
                                debugString.append("Removing from PDWS Lookup PATIENT: ");
                                debugString.append(result.getPatient().getName());
                                debugString.append("   Restricted SITES: ");
                                for (Object restrictedSiteCode : restrictedSiteCodesList) {
                                    debugString.append(restrictedSiteCode);
                                    debugString.append(" ");
                                }
                                debugString.append("  Patient SITES: ");
                                for (Site site : result.getPatient().getVistaSites()) {
                                    debugString.append(site.getSiteCode());
                                    debugString.append(" ");
                                }
                                LOGGER.debug(debugString.toString());
                            }
                            // the patient from PDWS is not allowed since he does not have a IEN at a site on the
                            // restricted site list, so set the search response to return no patients.
                            ArrayList<Patient> patientArrayList = new ArrayList<Patient>();
                            responsePatientQuery.setPatients(patientArrayList);
                        }
                    }
                }
            }
        } catch (Exception e) {
            // Audit error
            auditDao.setAudit(siteId, siteMoniker, siteAgency, recordID, userIen, userNPI, userName, queryBean.getPatientId(), "Error: PDWS Service", requestingApp, "", "", info, user.getCardId(), user.getEmail(), complexTransaction);
            throw new JMeadowsException(e);
        }

        return responsePatientQuery;
    }

    /**
     * Performs a patient lookup against the MVI Web Service (MVI).
     *
     * @param queryBean PDWSQueryBean - MVI Query parameters.
     * @return A list of patients that match search criteria.
     * @throws JMeadowsException if there is an error.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public ResponsePatientQuery lookupPatientMVI(PDWSQueryBean queryBean) throws JMeadowsException {
/*      TODO: PDWSQueryBean - MVI parameter check
        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.DATE_RANGE,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);
*/
        AuditDao auditDao = new AuditDao();
        User user = queryBean.getUser();
        String siteId = "";
        String userIen = "";
        String userNPI = "";
        String userName = "";
        String info = "";
        String requestingApp = "";
        //In the future be able to grab this information from PDWS
        String siteMoniker = "";
        String siteAgency = "";
        String recordID = "";
        String complexTransaction = "";

        siteId = user.getHostSite().getSiteCode();
        userIen = user.getUserIen();
        userNPI = user.getNPI();
        userName = user.getName();
        info = queryBean.getPatientId()
                + "^" + queryBean.getLastName()
                + "^" + queryBean.getFirstName()
                + "^" + queryBean.getDob()
                + "^" + queryBean.getGender();
        requestingApp = queryBean.getRequestingApp();

        auditDao.setAudit(siteId, siteMoniker, siteAgency, recordID, userIen, userNPI, userName, queryBean.getPatientId(), "lookupPatientMVI", requestingApp, "", "", info, user.getCardId(), user.getEmail(), complexTransaction);

        ResponsePatientQuery responsePatientQuery = null;
        try {
            responsePatientQuery = sm_mviHandler.lookupPatient(queryBean, SITES.getSiteMap());

            if (!responsePatientQuery.isSuccess()) {
                auditDao.setAudit(siteId, siteMoniker, siteAgency, recordID, userIen, userNPI, userName, queryBean.getPatientId(), "Error: MVI Service", requestingApp, "", "", info, user.getCardId(), user.getEmail(), complexTransaction);
                LOGGER.error(responsePatientQuery.getErrorMsg());
            }
        } catch (Exception e) {
            // Audit error
            auditDao.setAudit(siteId, siteMoniker, siteAgency, recordID, userIen, userNPI, userName, queryBean.getPatientId(), "Error: MVI Service", requestingApp, "", "", info, user.getCardId(), user.getEmail(), complexTransaction);
            throw new JMeadowsException(e);
        }

        return responsePatientQuery;
    }

    /**
     * Performs patient selection against the MVI service using either EDIPI, ICN or VA IEN.
     *
     * MVI service-based patient selection details:
     *
     *      MVI service based patient selection will query the MVI service to
     *      retrieve available corresponding VA identifiers.
     *
     *      The following Patient object fields if the MVI service-based selection method is utilized:
     *          edipi
     *          icn
     *          patientIens
     *
     *       The edipi, icn or va ien will be set to the itemId parameter.
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * itemId - The edipi, icn or va ien identifier of the patient to select
     * recordSiteCode - the site code of the identifier if it's a va ien
     *
     * @return The selected patient response.
     * @throws JMeadowsException if there is an error.
     */
    public ResponsePatientSelect selectPatientMVI(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.ITEM_ID);

        auditQuery("SelectPatientMVI", queryBean);

        LOGGER.debug("selectPatientMVI::Performing MVI based patient select");

        ResponsePatientSelect result = sm_mviHandler.getPatientIds(queryBean, SITES.getSiteMap());

        if (result != null && result.getPatient() != null) {
            Patient patient = result.getPatient();

            queryBean.setPatient(patient);

            List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);
            List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

            for (PatientLocation loc : locs) {
                if (StringUtils.equalsIgnoreCase(SYSTEM_TYPE_DOD_ADAPTER, loc.getSystemType())) {
                    // DES Demographics (for AHLTA Unit Number, FMP/SSN, Sponsor Name, and HIE Sharing Opt In)
                    if (StringUtils.isNotBlank(patient.getEDIPI())) {
                        PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);
                        gov.va.med.bhie.QueryBean bhieQueryBean = new gov.va.med.bhie.QueryBean();

                        StringBuilder patientId = new StringBuilder();
                        patientId.append("EDIPI:");
                        patientId.append(patient.getEDIPI());

                        bhieQueryBean.setUserId(queryBean.getUser().getUserId());
                        bhieQueryBean.setPatientId(patientId.toString());

                        configs.add(
                                new PatientMultiLocQueryCfgBuilder()
                                        .dataServiceHandler(sm_bhieRelayServiceHandler)
                                        .dataServiceMethod("getPatientDemographics")
                                        .dataServiceMethodArgs(bhieQueryBean)
                                        .smartBeanMappingMethod("mapPatientDemographicsFromDes")
                                        .location(bhieLoc)
                                        .build()
                        );
                    }
                }
                else if (StringUtils.equalsIgnoreCase(SYSTEM_TYPE_VISTA, loc.getSystemType())) {
                    QueryBean vistaQueryBean = new QueryBean();

                    vistaQueryBean.setUser(queryBean.getUser());
                    vistaQueryBean.setPatient(patient);

                    // Patient has VA IENs
                    if (CollectionUtils.isNotEmpty(patient.getPatientIens())) {
                        // Check if patient is VA sensitive
                        configs.add(
                                new PatientMultiLocQueryCfgBuilder()
                                        .dataServiceHandler(sm_vistaHandler)
                                        .dataServiceMethod("getSensitivePatientAlerts")
                                        .dataServiceMethodArgs(vistaQueryBean)
                                        .smartBeanMappingMethod("mapSensitivePatientAlertsFromVista")
                                        .location(loc)
                                        .build()
                        );
                    }
                }
            }

            Object[] respObjs = MultiLocationQuery.execute(configs, new SmartPatientSelect(), queryBean);

            if (respObjs != null) {
                List<ResponsePatientSelect> responses = Arrays.asList((ResponsePatientSelect[])respObjs);
                Patient respPatient;

                for (ResponsePatientSelect response : responses) {
                    if (response == null) { continue; }

                    respPatient = response.getPatient();

                    if (response.getResponseType() == ResponsePatientSelect.ResponseType.DEMOGRAPHICS_DES) {
                        // AHLTA unit number
                        patient.setAhltaUnitNumber(respPatient.getAhltaUnitNumber());

                        // Sponsor FMP/SSN and name
                        patient.setSponsorSSN(respPatient.getSponsorSSN());
                        patient.setSponsorName(respPatient.getSponsorName());
                        patient.setFMP(respPatient.getFMP());

                        // HIE Sharing option
                        patient.setHIESharingOptIn(respPatient.isHIESharingOptIn());
                        patient.setHIESharingOptInLastUpdateDate(respPatient.getHIESharingOptInLastUpdateDate());
                    }
                    else if (response.getResponseType() == ResponsePatientSelect.ResponseType.SENSITIVE_ALERTS) {
                        // VA sensitve flag and alerts levels
                        patient.setSensitive(respPatient.isSensitive());
                        result.setSensitivePatientAlerts(response.getSensitivePatientAlerts());
                    }
                }

                result.setPatient(patient);
            }
        }

        return result;
    }

    /**
     * Performs patient selection to VistA for a sensitive patient after user has accepted sensitive patient warning.
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @throws JMeadowsException if there is an error.
     */
    public void selectPatientForVASensitive(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("selectPatientForVASensitive", queryBean);

        QueryBean qbean = new QueryBean();

        qbean.setUser(queryBean.getUser());
        qbean.setPatient(queryBean.getPatient());

        sm_vistaHandler.selectPatient(qbean);
    }

    /**
     * Performs request to VistA for a list of sensitive patient alerts from each site.
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @throws JMeadowsException if there is an error.
     */
    public List<Alert> getVASensitivePatientAlerts(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("getVASensitivePatientAlerts", queryBean);

        QueryBean qbean = new QueryBean();

        qbean.setUser(queryBean.getUser());
        qbean.setPatient(queryBean.getPatient());

        return sm_vistaHandler.getSensitivePatientAlerts(qbean);
    }


    /**
     * Performs audit on VA user who accepts to access a DoD-only patient record
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @throws JMeadowsException if there is an error.
     */
    public void auditVARestrictedAccess(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQueryEncrypted("auditVARestrictedAccess", queryBean);
    }

    /**
     * Performs audit on VA user who accepts to access a sensitive patient record
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @throws JMeadowsException if there is an error.
     */
    public void auditSensitiveRecordAccess(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

//        StringBuilder auditTextSB = new StringBuilder(queryBean.getItemId());
//        queryBean.setItemId(auditTextSB.toString());

        auditQueryEncrypted("sensitiveRecordAccess", queryBean);
    }


    /**
     * Retrieves a list of patient demographics from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     *
     * @return A list of patient demographics.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientDemographics> getPatientDemographics(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatDemographics", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();
        //for DoD Flags call
        List<MultiLocationQueryConfig> configFlags = new ArrayList<MultiLocationQueryConfig>();
        boolean callDODFlags = false;

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                callDODFlags= true;

                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                // Leaving below code in case BDA demographics needs to be called -Jimi
                List<String> includedSources = new ArrayList<String>();
//                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                //SB: Exclude data range to retrieve entire data set
                bhieQueryBean.setStartDate(null);
                bhieQueryBean.setEndDate(null);

                if (queryBean.isIncludeVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_bhieRelayServiceHandler)
                                    .dataServiceMethod("getVLERPatientDemographics")
                                    .dataServiceMethodArgs(bhieQueryBean)
                                    .smartBeanMappingMethod("mapPatientDemographicsFromBHIE")
                                    .location(bhieLoc)
                                    .build()
                    );
                }

                PDWSQueryBean pdwsQueryBean = new PDWSQueryBean();
                pdwsQueryBean.setUser(queryBean.getUser());
                pdwsQueryBean.setPatientIdType("EDIPI");
                pdwsQueryBean.setPatientId(queryBean.getPatient().getEDIPI());
                pdwsQueryBean.setInquiryType("PR");
                pdwsQueryBean.setRequestingApp(queryBean.getRequestingApp());

                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_pdwsHandler)
                                .dataServiceMethod("lookupPatient")
                                .dataServiceMethodArgs(pdwsQueryBean)
                                .smartBeanMappingMethod("mapPatientDemographicsFromPDWS")
                                .location(bhieLoc)
                                .build()
                );

                configFlags.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getPatientManagement")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapPatientFlagsFromBHIE")
                                .location(bhieLoc)
                                .build()
                );

            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientDemographics")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .location(loc)
                                .build()
                );
                if (queryBean.isIncludeDASVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_vistaVLERHandler)
                                    .dataServiceMethod("getDASVLERDemographics")
                                    .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                    .smartBeanMappingMethod("mapDASVLERDemographics")
                                    .location(loc)
                                    .build()
                    );
                }
            }
        }

        Object[] demographicObjs = MultiLocationQuery.execute(
                configs, new SmartDemographics(), queryBean);

        Object[] flagObjs = null;

        if(callDODFlags) {
            flagObjs = MultiLocationQuery.execute(configFlags, new SmartFlag(), queryBean);
        }

        List<PatientDemographics> demographics = null;

        if (demographicObjs != null) {
            demographics = new ArrayList(Arrays.asList((PatientDemographics[]) demographicObjs));

            ConceptCodeMapper.getInstance().mapDemographics(hddDao, demographics);

            Patient patient = queryBean.getPatient();
            String edipi = "";
            String fmp = "";
            String sponsorName = "";
            String sponsorSSN = "";
            boolean hieSharingOptIn = false;
            XMLGregorianCalendar hieSharingOptInLastUpdateDate = null;
            if (patient != null) {
                if (StringUtils.isNotBlank(patient.getEDIPI())) {
                    edipi = patient.getEDIPI();
                }
                // PDWS does not return FMP, sponsor SSN/Name, therefore FMP and sponsor SSN/Name are retrieved from the
                // BDA demographics call during MVI patient select when querying for the patient's AHLTA unit number
                if (StringUtils.isNotBlank(patient.getFMP())) {
                    fmp = patient.getFMP(); // Format: FMP/SponsorSSN (FMP 'forward slash' Sponsor SSN)
                }
                if (StringUtils.isNotBlank(patient.getSponsorName())) {
                    sponsorName = patient.getSponsorName();
                }
                if (StringUtils.isNotBlank(patient.getSponsorSSN())) {
                    sponsorSSN = patient.getSponsorSSN();
                }
                hieSharingOptIn = patient.isHIESharingOptIn();
                hieSharingOptInLastUpdateDate = patient.getHIESharingOptInLastUpdateDate();
            }

            String userAgency = queryBean.getUser().getAgency();
            boolean isMappedToRecentlyViewedPatient = false;
            boolean dodDemoExists = false;
            for (PatientDemographics patDemo : demographics) {
                if ("DOD".equalsIgnoreCase(patDemo.getSite().getAgency())) {
                    dodDemoExists = true;
                    patDemo.setFMP(fmp);
                    patDemo.setSponsorName(sponsorName);
                    patDemo.setSponsorSSN(sponsorSSN);

                    if(flagObjs != null && flagObjs.length > 0){
                        List<Flag> dodFlags = null;
                        dodFlags = new ArrayList(Arrays.asList((Flag[]) flagObjs));
                        patDemo.getFlags().addAll(dodFlags);
                    }

                    if (StringUtils.equalsIgnoreCase(userAgency, AGENCY_DOD)) {
                        mapDemographicsToRecentlyViewedPatients(patient, patDemo);
                        isMappedToRecentlyViewedPatient = true;
                    }

                    if(patDemo.getName() != null && patDemo.getName().toLowerCase().indexOf("connection unavailable") > -1){
                        patDemo.getSite().setMoniker("PDWS");
                    }
                }
                else {
                    if (StringUtils.equalsIgnoreCase(userAgency, AGENCY_VA)
                            && StringUtils.equalsIgnoreCase(queryBean.getUser().getHostSite().getSiteCode(),
                                patDemo.getSite().getSiteCode())) {
                        mapDemographicsToRecentlyViewedPatients(patient, patDemo);
                        isMappedToRecentlyViewedPatient = true;
                    }
                }

                // Set patient EDIPI
                if(patDemo.getSite() != null && !"DASVLER".equals(patDemo.getSite().getAgency())) {
                    patDemo.setEDIPI(edipi);
                } else {
                    patDemo.setEDIPI(patDemo.getICN());
                }
                patDemo.setHIESharingOptIn(hieSharingOptIn);
                patDemo.setHIESharingOptInLastUpdateDate(hieSharingOptInLastUpdateDate);
            }

            // Wasn't able to map based on user site, try to map to first VA demographics
            if (!isMappedToRecentlyViewedPatient) {
                for (PatientDemographics patDemo : demographics) {
                    if (StringUtils.equalsIgnoreCase(patDemo.getSite().getAgency(), AGENCY_VA)) {
                        mapDemographicsToRecentlyViewedPatients(patient, patDemo);
                        break;
                    }
                }
            }

            // Add user's selected patient to recently viewed patients table
            if ((patient != null)
                    && StringUtils.isNotBlank(patient.getName())
                    && (StringUtils.isNotBlank(patient.getEDIPI()) || StringUtils.isNotBlank(patient.getICN()))
                    && StringUtils.isNotBlank(patient.getSSN())) {
                UserDao userDao = new UserDao();
                userDao.addRecentlyViewedPatient(queryBean.getUser().getUserId(), patient);
            }

            if (StringUtils.isNotBlank(edipi) && !dodDemoExists) {

                Site dodSite = new Site();
                dodSite.setAgency(AGENCY_DOD);
                dodSite.setSiteCode(AGENCY_DOD);
                dodSite.setName(AGENCY_DOD);
                dodSite.setMoniker(AGENCY_DOD);
                dodSite.setAgency(AGENCY_DOD);

                PatientDemographics dodDemo = new PatientDemographics();
                dodDemo.setName(ERROR_MSG_CONNECTION_UNAVAILABLE);
                dodDemo.setAge("");
                dodDemo.setClinic("");
                dodDemo.setDob("");
                dodDemo.setGender("");
                dodDemo.setPrimaryProvider("");
                dodDemo.setPhone1("");
                dodDemo.setPhone2("");
                dodDemo.setSSN("");
                dodDemo.setCodeGreen("0");
                dodDemo.setSite(dodSite);
                dodDemo.getAddress().clear();
                demographics.add(dodDemo);
            }

        }

        return demographics;
    }

    /**
     * Retrieves a list of patient demographics details from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     *
     * @return A list of patient demographics details.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientDemographicsDetail> getPatientDemographicsDetail(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatDemographicsDetail", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, false);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                //SB: Exclude data range to retrieve entire data set
                bhieQueryBean.setStartDate(null);
                bhieQueryBean.setEndDate(null);

//                configs.add(
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_bhieRelayServiceHandler)
//                                .dataServiceMethod("getPatientDemographics")
//                                .dataServiceMethodArgs(bhieQueryBean)
//                                .smartBeanMappingMethod("mapPatientDemographicsDetailsFromBHIE")
//                                .location(bhieLoc)
//                                .build()
//
//                );
                PDWSQueryBean pdwsQueryBean = new PDWSQueryBean();
                pdwsQueryBean.setUser(queryBean.getUser());
                pdwsQueryBean.setPatientIdType("EDIPI");
                pdwsQueryBean.setPatientId(queryBean.getPatient().getEDIPI());
                pdwsQueryBean.setInquiryType("PR");
                pdwsQueryBean.setRequestingApp(queryBean.getRequestingApp());

                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_pdwsHandler)
                                .dataServiceMethod("lookupPatient")
                                .dataServiceMethodArgs(pdwsQueryBean)
                                .smartBeanMappingMethod("mapPatientDemographicsDetailsFromPDWS")
                                .location(bhieLoc)
                                .build()
                );
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientDemographicsDetail")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .location(loc)
                                .build()
                );
            }
        }

        Object[] demographicDetailObjs = MultiLocationQuery.execute(
                configs, new SmartDemographicsDetail(), queryBean);

        List<PatientDemographicsDetail> demographicsDetail = null;

        if (demographicDetailObjs != null) {
            demographicsDetail = Arrays.asList((PatientDemographicsDetail[]) demographicDetailObjs);
        }

        return demographicsDetail;
    }
}
