/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.beans.VLERCCDA;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartCCDA;
import gov.va.med.vds.webservice.VlerDocument;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 *
 * Patient notes data access object.
 */
public class VLERDao extends DaoBase {

    /**
     * Retrieves a list of patient c32 document from the NwHIN VLER interface.
     *
     * @param queryBean Query parameters.
     *
     * {@link gov.va.med.vds.webservice.QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @return an array of VlerDocument
     * @throws gov.va.med.jmeadows.common.JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<VlerDocument> getVLERDocumentList(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatVLERDocQuery", queryBean);

        return sm_vistaVLERHandler.getVLERDocumentList(VistaBeanFactory.create(queryBean));
    }

    /**
     * Retrieves a patient c32 document from the NwHIN VLER interface.
     *
     * @param vDoc VLER c32 document location information
     * @param queryBean Query parameters.
     *
     * {@link gov.va.med.vds.webservice.QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @return a VlerDocument
     * @throws gov.va.med.jmeadows.common.JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public VlerDocument getVLERDocument(VlerDocument vDoc, JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatVLERDocRetrieve", queryBean);

        return sm_vistaVLERHandler.getVLERDocument(vDoc, VistaBeanFactory.create(queryBean));
    }


    public List<VLERCCDA> getDODVLERPatientCCDA(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatDODVLERCCDA", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {

                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
//                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());
//                includedSources.add(APP_CONFIG.getBHIECustodianOidSHARE());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                //SB: Exclude data range to retrieve entire data set
                bhieQueryBean.setStartDate(null);
                bhieQueryBean.setEndDate(null);

                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getVLERPatientCCDA")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapVLERCCDAFromBHIE")
                                .location(bhieLoc)
                                .build()
                );
            }
        }

        SmartCCDA smartCCDA = new SmartCCDA();

        Object[] vlerObjs = MultiLocationQuery.execute(
                configs, smartCCDA, queryBean);

        List<VLERCCDA> vlerCCDAs = null;

        if (vlerObjs != null) {
            vlerCCDAs = Arrays.asList((VLERCCDA[]) vlerObjs);

        }

        return vlerCCDAs;
    }

    public List<VLERCCDA> getDASVLERData(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatDASVLERData", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                //do nothing
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
//                configs.add(
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_vistaVLERHandler)
//                                .dataServiceMethod("getDASVLERAllergies")
//                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
//                                .smartBeanMappingMethod("mapDASVLERData")
//                                .location(loc)
//                                .build()
//                );
//                configs.add(
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_vistaVLERHandler)
//                                .dataServiceMethod("getDASVLERProblems")
//                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
//                                .smartBeanMappingMethod("mapDASVLERData")
//                                .location(loc)
//                                .build()
//                );
//
//                configs.add(
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_vistaVLERHandler)
//                                .dataServiceMethod("getDASVLERVitals")
//                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
//                                .smartBeanMappingMethod("mapDASVLERData")
//                                .location(loc)
//                                .build()
//                );
//
//                configs.add(
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_vistaVLERHandler)
//                                .dataServiceMethod("getDASVLERProcedures")
//                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
//                                .smartBeanMappingMethod("mapDASVLERData")
//                                .location(loc)
//                                .build()
//                );
//
//                configs.add(
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_vistaVLERHandler)
//                                .dataServiceMethod("getDASVLERImmunizations")
//                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
//                                .smartBeanMappingMethod("mapDASVLERData")
//                                .location(loc)
//                                .build()
//                );
//
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaVLERHandler)
                                .dataServiceMethod("getDASVLERDemographics")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .smartBeanMappingMethod("mapDASVLERData")
                                .location(loc)
                                .build()
                );
            }
        }

        SmartCCDA smartCCDA = new SmartCCDA();

        Object[] vlerObjs = MultiLocationQuery.execute(
                configs, smartCCDA, queryBean);

        List<VLERCCDA> vlerCCDAs = null;

        if (vlerObjs != null) {
            vlerCCDAs = Arrays.asList((VLERCCDA[]) vlerObjs);

        }

        return vlerCCDAs;
    }
}
