/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartVitals;
import gov.va.med.vds.webservice.Vitals;
import gov.va.med.vhahon.common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Vitals data access object.
 */
public class VitalsDao extends DaoBase {

    /**
     * Retrieves a list of patient vitals from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T).
     *
     * @return A list of patient vitals.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Vitals> getPatientVitals(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatVitals", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {

                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_bhieRelayServiceHandler)
                            .dataServiceMethod("getPatientVitals")
                            .dataServiceMethodArgs(bhieQueryBean)
                            .smartBeanMappingMethod("mapVitalsFromBHIEV4")
                            .location(bhieLoc)
                            .build()
                );

                if (queryBean.isIncludeVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_bhieRelayServiceHandler)
                                    .dataServiceMethod("getVLERPatientVitals")
                                    .dataServiceMethodArgs(bhieQueryBean)
                                    .smartBeanMappingMethod("mapVitalsFromBHIEV4")
                                    .location(bhieLoc)
                                    .build()
                    );
                }

            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getPatientVitals")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );

                if (queryBean.isIncludeDASVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_vistaVLERHandler)
                                    .dataServiceMethod("getDASVLERVitals")
                                    .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                    .smartBeanMappingMethod("mapDASVLERVitals")
                                    .location(loc)
                                    .build()
                    );
                }
            }
        }

        SmartVitals smartVitals = new SmartVitals();

        Object[] vitalObjs = MultiLocationQuery.execute(
                configs, smartVitals, queryBean);

        List<Vitals> vitalsList = null;

        if (vitalObjs != null) {
            vitalsList = Arrays.asList((Vitals []) vitalObjs);

            //map vitals concept codes
            //ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, vitalsList, CodingSystem.LOINC);
            //support non-HDD
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, vitalsList, CodingSystem.VITALS);
        }

        return vitalsList;
    }

    /**
     * Retrieves a list of patient's current vitals from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     *
     * @return A list of patient's current current vitals.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Vitals> getPatientCurrentVitals(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatCurrentVitals", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);


        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getPatientCurrentVitals")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );
            }
        }

        Object[] vitalObjs = MultiLocationQuery.execute(
                configs, new SmartVitals(), queryBean);

        List<Vitals> vitalsList = null;

        if (vitalObjs != null) {
            vitalsList = Arrays.asList((Vitals []) vitalObjs);
        }

        return vitalsList;
    }
}
