package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.beans.PatientDocument;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartDocument;
import gov.va.med.jmeadows.smartbeans.patient.SmartVixStudyQuery;
import gov.va.med.vds.webservice.Encounter;
import gov.va.med.vds.webservice.EncounterDocument;
import gov.va.med.vds.webservice.KeyValuePair;
import gov.va.med.vds.webservice.Patient;
import gov.va.med.vds.webservice.Procedure;
import gov.va.med.vds.webservice.QueryBean;
import gov.va.med.vds.webservice.RadiologyReport;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vds.webservice.StudyDetail;
import gov.va.med.vds.webservice.StudyDetailStudy;
import gov.va.med.vds.webservice.StudyQuery;
import gov.va.med.vds.webservice.StudyQueryStudy;
import gov.va.med.vhahon.common.Utils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

public class VixDao extends DaoBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(VixDao.class);
    private static final String STATUS_CODE_OK = "200";
    private static final String DOD_SITE_ID = "200";
    private static final String CONNECTION_UNAVAILABLE = APP_CONFIG.getErrorMsgConnectionUnavailable();

    public VixDao()
    {
        super();
    }

    private String getPatientIen(Patient patient, String siteCode)
    {
        String patIen = null;

        // Find matching patient site code in patient ien/site pair list with site code and return patient ien
        if (patient != null && patient.getPatientIens() != null && StringUtils.isNotBlank(siteCode)) {
            for (KeyValuePair siteIenPair : patient.getPatientIens()) {
                if (StringUtils.equals(siteCode, siteIenPair.getKey().toString())) {
                    patIen = siteIenPair.getValue().toString();
                    break;
                }
            }
        }

        return patIen;
    }

    private String generateVaPatientDocumentContextId(Patient patient, PatientDocument document)
    {
        StringBuilder contextId = new StringBuilder();

        if (patient != null && document != null && document.getSite() != null) {
            String patIen = getPatientIen(patient, document.getSite().getSiteCode());
            String docIen = document.getNoteId();
            String docHasImg = "0";

            if (StringUtils.isNotEmpty(document.getImages())) {
                try {
                    if (Integer.parseInt(document.getImages()) > 0) {
                        docHasImg = "1";
                    }
                }
                catch (NumberFormatException e) {
                    LOGGER.debug("Failed to convert document images count to integer");
                }
            }

            if (StringUtils.isNotBlank(patIen) && StringUtils.isNotBlank(docIen) && StringUtils.equals("1", docHasImg)) {
                // ContextId format: RPT^CPRS^[patientIen]^TIU^[noteIen]^^^^^^^^[hasImage]
                // Example: RPT^CPRS^123456^TIU^1234^^^^^^^^1
                contextId.append("RPT^CPRS^");
                contextId.append(patIen);
                contextId.append("^TIU^");
                contextId.append(docIen);
                contextId.append("^^^^^^^^");
                contextId.append(docHasImg);

                // Append site code
                contextId.append(":");
                contextId.append(document.getSite().getSiteCode());
            }
        }

        return contextId.toString();
    }

    private String generateVaEncounterDocumentContextId(Patient patient, String documentId, String documentSiteId)
    {
        StringBuilder contextId = new StringBuilder();

        if (patient != null && documentSiteId != null) {
            String patIen = getPatientIen(patient, documentSiteId);

            if (StringUtils.isNotBlank(patIen) && StringUtils.isNotBlank(documentId)) {
                // ContextId format: RPT^CPRS^[patientIen]^TIU^[noteIen]^^^^^^^^[hasImage]
                // Example: RPT^CPRS^123456^TIU^1234^^^^^^^^1
                contextId.append("RPT^CPRS^");
                contextId.append(patIen);
                contextId.append("^TIU^");
                contextId.append(documentId);
                contextId.append("^^^^^^^^1");

                // Append site code
                contextId.append(":");
                contextId.append(documentSiteId);
            }
        }

        return contextId.toString();
    }

    private String generateVaRadiologyReportContextId(Patient patient, RadiologyReport radiology)
    {
        StringBuilder contextId = new StringBuilder();

        if (patient != null && radiology != null && radiology.getSite() != null && radiology.isHasImages()) {
            String patIen = getPatientIen(patient, radiology.getSite().getSiteCode());
            String radCaseNum = radiology.getCaseNumber();
            String radExamId = radiology.getExamId();
            String radSiteName = radiology.getFacilityName();
            String radHasImg = radiology.isHasImages() ? "1" : "0";

            if (StringUtils.isNotBlank(patIen) && StringUtils.isNotBlank(radCaseNum)
                    && StringUtils.isNotBlank(radExamId) && StringUtils.isNotBlank(radSiteName)) {
                // ContextId format: RPT^CPRS^[patientDfn]^[category]^[case#]^[localId]^[facilityName]^^^^^^[hasImage]
                // Example: RPT^CPRS^123456^RA^5^1234567.1234-1^TEST SITE^^^^^^0
                contextId.append("RPT^CPRS^");
                contextId.append(patIen);
                contextId.append("^RA^");
                contextId.append(radExamId);
                contextId.append("^");
                contextId.append(radCaseNum);
                contextId.append("^");
                contextId.append(radSiteName);
                contextId.append("^^^^^^");
                contextId.append(radHasImg);

                // Append site code
                contextId.append(":");
                contextId.append(radiology.getSite().getSiteCode());
            }
        }

        return contextId.toString();
    }

    private String generateDodRadiologyContextId(Patient patient, RadiologyReport radiology)
    {
        StringBuilder contextId = new StringBuilder();

        // TODO: Generate DoD radiology contextId

        return contextId.toString();
    }

    private void getAllPatientStudyQueries(JMeadowsQuery jMeadowsQuery,
                                           Map<String, List<StudyQueryStudy>> queryStudyMap)
    {
        if (jMeadowsQuery == null || queryStudyMap == null) {
            return;
        }

        Patient pat = jMeadowsQuery.getPatient();
        List<PatientLocation> locs = LocationHelper.getPatientLocations(jMeadowsQuery, true);
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();
        QueryBean qb;

        // Set up multi location query for each VistA site to call VIX service
        for (PatientLocation loc : locs) {
            if (!SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                continue;
            }

            for (Site site : pat.getVistaSites()) {
                // Skip if site is null
                if (site == null) {
                    continue;
                }

                try {
                    qb = VistaBeanFactory.create(jMeadowsQuery);
                }
                catch (JMeadowsException e) {
                    LOGGER.error("Failed to create vista query bean");
                    continue;
                }

                qb.setItemId(site.getSiteCode());

                configs.add(new PatientMultiLocQueryCfgBuilder().dataServiceHandler(sm_vistaHandler)
                        .dataServiceMethod("getVixStudyQueryFromSite").dataServiceMethodArgs(qb)
                        .location(loc).build());
            }
        }

        Object[] queryObjs = null;
        try {
            queryObjs = MultiLocationQuery.execute(configs, new SmartVixStudyQuery(), jMeadowsQuery);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to get study queries by sites");
        }

        if (queryObjs != null) {
            StudyQuery query;
            StudyQueryStudy queryStudy;
            StudyDetail detail;
            StudyDetailStudy detailStudy;
            SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
            Date startDate = jMeadowsQuery.getStartDate().getTime();
            Date endDate = jMeadowsQuery.getEndDate().getTime();
            Date studyDate;

            for (Object queryObj : queryObjs) {
                query = (StudyQuery)queryObj;

                if (query == null || query.getStudies() == null
                        || StringUtils.equalsIgnoreCase(query.getPatientIcn(), CONNECTION_UNAVAILABLE)) {
                    continue;
                }

                for (Iterator<StudyQueryStudy> queryStudyItr = query.getStudies().iterator(); queryStudyItr.hasNext(); ) {
                    queryStudy = queryStudyItr.next();

                    detail = queryStudy.getStudyDetail();
                    if (detail == null || detail.getStudies() == null || detail.getStudies().size() < 1) {
                        queryStudyItr.remove();
                        continue;
                    }

                    detailStudy = detail.getStudies().get(0);
                    if (detailStudy == null) {
                        queryStudyItr.remove();
                        continue;
                    }

                    try {
                        studyDate = sdf.parse(detailStudy.getStudyDate());
                    }
                    catch (ParseException e) {
                        LOGGER.error("Failed to parse study date");
                        queryStudyItr.remove();
                        continue;
                    }

                    // Remove query study from list if outside of date range
                    if (studyDate.compareTo(startDate) < 0 || studyDate.compareTo(endDate) > 0) {
                        queryStudyItr.remove();
                    }
                }

                // Put remaining list in map after out of range studies were removed
                queryStudyMap.put(query.getSiteNumber(), query.getStudies());
            }
        }
    }

    private String generatePatientDocumentContextIdMap(JMeadowsQuery jMeadowsQuery, List<PatientDocument> patientDocuments,
                                                       Map<String, PatientDocument> documentMap)
    {
        if (jMeadowsQuery == null || patientDocuments == null || documentMap == null) {
            return "";
        }

        QueryBean qb;
        try {
            qb = VistaBeanFactory.create(jMeadowsQuery);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to create vista query bean");
            return "";
        }

        StringBuilder itemId = new StringBuilder();
        String contextId;
        String mapKey;
        int index;

        // Map contextId and patient documents
        for (PatientDocument document : patientDocuments) {
            if (document == null || document.getSite() == null) {
                continue;
            }

            // Ignore non-VA documents
            if (!StringUtils.equalsIgnoreCase(document.getSite().getAgency(), APP_CONFIG.getAgencyVA())) {
                continue;
            }

            contextId = "";

            // Only interested in radiology reports and progress notes documents
            if (SmartDocument.DOCUMENT_RADIOLOGY.equalsIgnoreCase(document.getDocumentType())) {
                contextId = generateVaRadiologyReportContextId(qb.getPatient(), document.getRadiologyReport());
            }
            else if (SmartDocument.DOCUMENT_PROGRESS_NOTE.equalsIgnoreCase(document.getDocumentType())) {
                contextId = generateVaPatientDocumentContextId(qb.getPatient(), document);
            }

            if (StringUtils.isNotBlank(contextId)) {
                mapKey = contextId;
                index = StringUtils.lastIndexOf(mapKey, ":");

                if (index != -1) {
                    mapKey = StringUtils.substring(mapKey, 0, index);
                }

                documentMap.put(mapKey, document);

                itemId.append(contextId);
                itemId.append(";"); // ';' delimited string for each contextId
            }
        }

        return itemId.toString();
    }

    private void updateMaps(StudyQuery documentQuery, Map<String, PatientDocument> documentMap,
                            Map<String, List<StudyQueryStudy>> allQueryStudyMap)
    {
        if (documentQuery == null || documentMap == null || allQueryStudyMap == null) {
            return;
        }

        PatientDocument doc;
        StudyQueryStudy allQueryStudy;
        StudyDetail docDetail;
        StudyDetail allDetail;
        StudyDetailStudy docDetailStudy;
        StudyDetailStudy allDetailStudy;
        List<StudyQueryStudy> allQueryStudies;

        for (StudyQueryStudy docQueryStudy : documentQuery.getStudies()) {
            if (docQueryStudy == null) {
                continue;
            }

            if (!StringUtils.equals(docQueryStudy.getStatusCode(), STATUS_CODE_OK)) {
                continue;
            }

            doc = documentMap.get(docQueryStudy.getContextId());
            if (doc == null) {
                continue;
            }

            // Map study to document
            doc.setVixStudy(docQueryStudy);

            docDetail = docQueryStudy.getStudyDetail();
            if (docDetail == null || docDetail.getStudies() == null || docDetail.getStudies().size() < 1) {
                continue;
            }

            for (Iterator<StudyDetailStudy> docStudyItr = docDetail.getStudies().iterator(); docStudyItr.hasNext(); ) {
                docDetailStudy = docStudyItr.next();
                if (docDetailStudy == null) {
                    continue;
                }

                // Check if queryStudy is in allQueryStudyMap to be removed as already mapped to document
                allQueryStudies = allQueryStudyMap.get(docQueryStudy.getSiteNumber());
                if (allQueryStudies == null) {
                    continue;
                }

                for (Iterator<StudyQueryStudy> queryStudyItr = allQueryStudies.iterator(); queryStudyItr.hasNext(); ) {
                    allQueryStudy = queryStudyItr.next();

                    allDetail = allQueryStudy.getStudyDetail();
                    if (allDetail == null || allDetail.getStudies() == null || allDetail.getStudies().size() < 1) {
                        continue;
                    }

                    allDetailStudy = allDetail.getStudies().get(0);
                    if (allDetailStudy == null) {
                        continue;
                    }

                    // Remove matching document queryStudy from allQueryStudyMap
                    if (StringUtils.equalsIgnoreCase(docDetailStudy.getStudyId(), allDetailStudy.getStudyId())) {
                        queryStudyItr.remove();
                    }
                }
            }
        }
    }

    private void generateUnmappedStudiesPatientDocuments(List<PatientDocument> patientDocuments,
                                                         Map<String, List<StudyQueryStudy>> allQueryStudyMap)
    {
        PatientDocument document;
        List<StudyQueryStudy> queryStudies;
        StudyDetail detail;
        StudyDetailStudy detailStudy;
        Site site;
        SimpleDateFormat simpleDateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        Date date;

        for (Map.Entry<String, List<StudyQueryStudy>> entry : allQueryStudyMap.entrySet()) {
            site = SITES.getSite(entry.getKey());
            queryStudies = entry.getValue();

            // Map studies to new patient documents
            for (StudyQueryStudy tmpStudyQuery : queryStudies) {
                if (tmpStudyQuery == null) {
                    continue;
                }

                detail = tmpStudyQuery.getStudyDetail();
                if (detail == null || detail.getStudies() == null || detail.getStudies().size() < 1) {
                    continue;
                }

                detailStudy = detail.getStudies().get(0);

                try {
                    date = simpleDateFormat.parse(detailStudy.getStudyDate());

                    document = new PatientDocument();

                    document.setNoteDate(Utils.toXMLGregorianCalendar(date));
                    document.setNoteTitle(detailStudy.getStudyDescription());
                    document.setVixStudy(tmpStudyQuery);
                    document.setSite(site);

                    patientDocuments.add(document);
                }
                catch (ParseException e) {
                    LOGGER.error("Failed to parse study date");
                }
            }
        }
    }
/*
    public void mapPatientDocuments(JMeadowsQuery jMeadowsQuery, List<PatientDocument> patientDocuments)
    {
        // Step 1: Retrieve all VIX studies for all patient's VistA sites within query date range
        // allQueryStudyMap: Key=VistA site code. Value=list of study query studies at the VistA site in key
        Map<String, List<StudyQueryStudy>> allQueryStudyMap = new HashMap<String, List<StudyQueryStudy>>();
        getAllPatientStudyQueries(jMeadowsQuery, allQueryStudyMap);

        // Step 2: Map only VA radiology report and VA progress note documents with their VIX CPRS contextId
        // documentMap: Key=VIX context ID. Value=patient document
        Map<String, PatientDocument> documentMap = new HashMap<String, PatientDocument>();
        String contextIds = generatePatientDocumentContextIdMap(jMeadowsQuery, patientDocuments, documentMap);

        // Step 3: Get VIX study query for mapped (documentMap) documents
        StudyQuery query = null;
        QueryBean queryBean;
        try {
            queryBean = VistaBeanFactory.create(jMeadowsQuery);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to create vista query bean");
            return;
        }

        if (StringUtils.isNotBlank(contextIds)) {
            queryBean.setItemId(contextIds);

            try {
                query = sm_vistaHandler.getVixStudyQueryWithDetails(queryBean);
            }
            catch (JMeadowsException e) {
                LOGGER.error("Failed to get study query");
            }
        }

        // Step 4: Map VIX query studies to documents and remove matching query study from allQueryStudyMap
        if (query != null) {
            updateMaps(query, documentMap, allQueryStudyMap);
        }

        // Step 5: Generate generic patient documents for remaining query studies in allQueryStudyMap
        // that weren't mapped to a document in the documentMap
        generateUnmappedStudiesPatientDocuments(patientDocuments, allQueryStudyMap);
    }

    public void mapOutpatientEncounters(JMeadowsQuery jMeadowsQuery, List<Encounter> encounters)
    {
        // Invalid arguments passed in
        if (jMeadowsQuery == null || encounters == null) {
            LOGGER.error("Could not map images: jMeadowsQuery AND/OR encounters list is null");
            return;
        }

        QueryBean qb;
        try {
            qb = VistaBeanFactory.create(jMeadowsQuery);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to create vista query bean");
            return;
        }

        Map<String, EncounterDocument> documentMap = new HashMap<String, EncounterDocument>();
        Site site;
        StringBuilder itemId = new StringBuilder();
        String contextId;
        String mapKey;
        int index;

        for (Encounter encounter : encounters) {
            if (encounter == null || encounter.getDocuments() == null) {
                continue;
            }

            // Generate context ID to send to VIX for VA encounters only
            site = encounter.getSite();
            if (site == null || !StringUtils.equalsIgnoreCase(site.getAgency(), APP_CONFIG.getAgencyVA())) {
                continue;
            }

            for (EncounterDocument document : encounter.getDocuments()) {
                contextId = generateVaEncounterDocumentContextId(qb.getPatient(), document.getId(), site.getSiteCode());

                if (StringUtils.isNotBlank(contextId)) {
                    mapKey = contextId;
                    index = StringUtils.lastIndexOf(mapKey, ":");

                    if (index != -1) {
                        mapKey = StringUtils.substring(mapKey, 0, index);
                    }

                    documentMap.put(mapKey, document);

                    itemId.append(contextId);
                    itemId.append(";");
                }
            }
        }

        qb.setItemId(itemId.toString());

        StudyQuery studyQuery;
        try {
            studyQuery = sm_vistaHandler.getVixStudyQueryWithDetails(qb);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to get study query", e);
            return;
        }

        if (studyQuery == null || studyQuery.getStudies() == null) {
            LOGGER.debug("Failed to get study query AND/OR list of query studies");
            return;
        }

        EncounterDocument document;
        String code;

        for (StudyQueryStudy study : studyQuery.getStudies()) {
            code = study.getStatusCode();

            if (!StringUtils.equals(code, STATUS_CODE_OK)) {
                continue;
            }

            document = documentMap.get(study.getContextId());
            if (document != null) {
                document.setVixStudy(study);
            }
        }
    }

    public void mapProcedures(JMeadowsQuery jMeadowsQuery, List<Procedure> procedures)
    {
        // Invalid arguments passed in
        if (jMeadowsQuery == null || procedures == null) {
            LOGGER.error("Could not map images: jMeadowsQuery AND/OR procedures list is null");
            return;
        }

        QueryBean qb;
        try {
            qb = VistaBeanFactory.create(jMeadowsQuery);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to create vista query bean");
            return;
        }

        Map<String, EncounterDocument> documentMap = new HashMap<String, EncounterDocument>();
        Site site;
        StringBuilder itemId = new StringBuilder();
        String contextId;
        String mapKey;
        int index;

        for (Procedure procedure : procedures) {
            if (procedure == null || procedure.getDocuments() == null) {
                continue;
            }

            // Generate context ID to send to VIX for VA procedures only
            site = procedure.getSite();
            if (site == null || !StringUtils.equalsIgnoreCase(site.getAgency(), APP_CONFIG.getAgencyVA())) {
                continue;
            }

            for (EncounterDocument doc : procedure.getDocuments()) {
                contextId = generateVaEncounterDocumentContextId(qb.getPatient(), doc.getId(), site.getSiteCode());

                if (StringUtils.isNotBlank(contextId)) {
                    mapKey = contextId;
                    index = StringUtils.lastIndexOf(mapKey, ":");

                    if (index != -1) {
                        mapKey = StringUtils.substring(mapKey, 0, index);
                    }

                    documentMap.put(mapKey, doc);

                    itemId.append(contextId);
                    itemId.append(";");
                }
            }
        }

        qb.setItemId(itemId.toString());

        StudyQuery studyQuery;
        try {
            studyQuery = sm_vistaHandler.getVixStudyQueryWithDetails(qb);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to get study query", e);
            return;
        }

        if (studyQuery == null || studyQuery.getStudies() == null) {
            LOGGER.debug("Failed to get study query AND/OR list of query studies");
            return;
        }

        EncounterDocument document;
        String code;

        for (StudyQueryStudy study : studyQuery.getStudies()) {
            code = study.getStatusCode();

            if (!StringUtils.equals(code, STATUS_CODE_OK)) {
                continue;
            }

            document = documentMap.get(study.getContextId());
            if (document != null) {
                document.setVixStudy(study);
            }
        }
    }

    public void mapRadiologyReports(JMeadowsQuery jMeadowsQuery, List<RadiologyReport> radiologyReports) {
        // Invalid arguments passed in
        if (jMeadowsQuery == null || radiologyReports == null) {
            LOGGER.error("Could not map images: jMeadowsQuery AND/OR radiologyReports list is null");
            return;
        }

        // Step 1: Create DoD and VA Radiology maps and populate maps
        Map<String, RadiologyReport> dodRadReportMap = new HashMap<String, RadiologyReport>();
        Map<String, RadiologyReport> vaRadReportMap = new HashMap<String, RadiologyReport>();
        Site site;
        String contextId;
        String mapKey;
        int index;
        StringBuilder itemId = new StringBuilder();

        for (RadiologyReport radReport : radiologyReports) {
            if (radReport == null) {
                continue;
            }

            site = radReport.getSite();
            if (site == null || site.getAgency() == null) {
                continue;
            }

            if (StringUtils.equalsIgnoreCase(site.getAgency(), APP_CONFIG.getAgencyDOD())) {
                if (StringUtils.isNotBlank(radReport.getExamId())) {
                    // Mapping DoD radiology exam ID with radiology report
                    dodRadReportMap.put(radReport.getExamId(), radReport);
                }
            }
            else if (StringUtils.equalsIgnoreCase(site.getAgency(), APP_CONFIG.getAgencyVA())) {
                // Generate context ID to send to VIX
                contextId = generateVaRadiologyReportContextId(jMeadowsQuery.getPatient(), radReport);

                // Mapping VIX context ID with radiology report
                if (StringUtils.isNotBlank(contextId)) {
                    mapKey = contextId;
                    index = StringUtils.lastIndexOf(mapKey, ":");

                    // Strip out site ID from the context ID to be used as the map's key value
                    if (index != -1) {
                        mapKey = StringUtils.substring(mapKey, 0, index);
                    }

                    vaRadReportMap.put(mapKey, radReport);

                    itemId.append(contextId);
                    itemId.append(";");
                }
            }
        }

        // Step 2: Setup multi location query for both DoD and VA (DoD query is also sent to VDS, not DES for VIX)
        List<PatientLocation> locs = LocationHelper.getPatientLocations(jMeadowsQuery, true);
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();
        QueryBean qb;

        for (PatientLocation loc : locs) {
            qb = null;

            try {
                qb = VistaBeanFactory.create(jMeadowsQuery);
            }
            catch (JMeadowsException e) {
                LOGGER.error("Failed to create vista query bean");
            }

            if (qb == null) {
                continue;
            }

            if (StringUtils.equalsIgnoreCase(SYSTEM_TYPE_DOD_ADAPTER, loc.getSystemType())) {
                if (!APP_CONFIG.isVixDodRadiologyRetrieveEnabled()) {
                    continue;
                }

                // For DoD Radiology, we are querying for all images at site 200 in one call
                qb.setItemId(DOD_SITE_ID);
                configs.add(new PatientMultiLocQueryCfgBuilder().dataServiceHandler(sm_vistaHandler)
                        .dataServiceMethod("getVixStudyQueryFromSite").dataServiceMethodArgs(qb)
                        .location(loc).build());
            }
            else if (StringUtils.equalsIgnoreCase(SYSTEM_TYPE_VISTA, loc.getSystemType()) && vaRadReportMap.size() > 0) {
                // For VA Radiology, we are querying for each radiology report individually in one call
                qb.setItemId(itemId.toString());
                configs.add(new PatientMultiLocQueryCfgBuilder().dataServiceHandler(sm_vistaHandler)
                        .dataServiceMethod("getVixStudyQueryWithDetails").dataServiceMethodArgs(qb)
                        .smartBeanMappingMethod("mapStudyQueryWithDetails").location(loc).build());
            }
        }

        Object[] queryObjs = null;
        if (configs.size() > 0) {
            try {
                queryObjs = MultiLocationQuery.execute(configs, new SmartVixStudyQuery(), jMeadowsQuery);
            }
            catch (JMeadowsException e) {
                LOGGER.error("Failed to get study queries for radiologies");
            }
        }

        // Step 3: Map images to radiology reports
        if (queryObjs != null) {
            StudyQuery query;
            StudyQueryStudy queryStudy;
            StudyDetail detail;
            StudyDetailStudy detailStudy;
            RadiologyReport radReport;
            SimpleDateFormat simpleDateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
            Date studyDate;
            Date startDate = jMeadowsQuery.getStartDate().getTime();
            Date endDate = jMeadowsQuery.getEndDate().getTime();
            StringTokenizer tokenizer;
            String delim = APP_CONFIG.getVixDodRadExamIdDelim();
            String token;
            String examId;

            for (Object queryObj : queryObjs) {
                query = (StudyQuery)queryObj;

                if (query == null || query.getStudies() == null
                        || StringUtils.equalsIgnoreCase(query.getPatientIcn(), CONNECTION_UNAVAILABLE)) {
                    continue;
                }

                if (StringUtils.equalsIgnoreCase(query.getSiteNumber(), DOD_SITE_ID)) {
                    // All DoD images
                    for (Iterator<StudyQueryStudy> queryStudyItr = query.getStudies().iterator(); queryStudyItr.hasNext(); ) {
                        queryStudy = queryStudyItr.next();
                        examId = queryStudy.getExternalContextId();

                        if (queryStudy == null || StringUtils.isBlank(examId)) {
                            // Skip if query study is null or
                            // external context ID is null or empty (it means it's not a DoD radiology if null or empty)
                            continue;
                        }
                        else {
                            // Parse external context ID
                            token = "";
                            tokenizer = new StringTokenizer(examId, delim);

                            if (tokenizer != null) {
                                // Get last token
                                while (tokenizer.hasMoreTokens()) {
                                    token = tokenizer.nextToken();
                                }

                                if (StringUtils.isNotBlank(token)) {
                                    examId = token;
                                }
                            }
                        }

                        radReport = dodRadReportMap.get(examId);
                        if (radReport != null) {
                            radReport.setVixStudy(queryStudy);
                        }
                        else {
                            // Add new report not linked to an existing radiology report
                            detail = queryStudy.getStudyDetail();
                            if (detail != null) {
                                if (detail.getStudies() != null && detail.getStudies().size() > 0) {
                                    detailStudy = detail.getStudies().get(0);
                                    if (detailStudy != null) {
                                        try {
                                            studyDate = simpleDateFormat.parse(detailStudy.getStudyDate());
                                            if (studyDate.compareTo(startDate) >= 0 && studyDate.compareTo(endDate) <= 0) {
                                                radReport = new RadiologyReport();

                                                radReport.setVixStudy(queryStudy);
                                                radReport.setExamDate(Utils.toXMLGregorianCalendar(studyDate));
                                                radReport.setStudy(detailStudy.getStudyDescription());

                                                site = new Site();
                                                site.setAgency(APP_CONFIG.getAgencyDOD());
                                                site.setName("HAIMS");
                                                site.setMoniker(APP_CONFIG.getAgencyDOD());
                                                radReport.setSite(site);

                                                radiologyReports.add(radReport);
                                            }
                                        }
                                        catch (ParseException e) {
                                            LOGGER.error("Failed to parse study date");
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else {
                    // VA radiology images
                    for (Iterator<StudyQueryStudy> queryStudyItr = query.getStudies().iterator(); queryStudyItr.hasNext(); ) {
                        queryStudy = queryStudyItr.next();

                        if (queryStudy == null || !StringUtils.equalsIgnoreCase(queryStudy.getStatusCode(), STATUS_CODE_OK)) {
                            continue;
                        }

                        radReport = vaRadReportMap.get(queryStudy.getContextId());
                        if (radReport != null) {
                            radReport.setVixStudy(queryStudy);
                        }
                    }
                }
            }
        }
    }
*/
    public void mapDoDRadiologyReports(JMeadowsQuery jMeadowsQuery, List<RadiologyReport> radiologyReports) {
        // Invalid arguments passed in
        if (jMeadowsQuery == null || radiologyReports == null) {
            LOGGER.error("Could not map images: jMeadowsQuery AND/OR radiologyReports list is null");
            return;
        }

        // Step 1: Create DoD Radiology map and populate map
        Map<String, RadiologyReport> dodRadReportMap = new HashMap<String, RadiologyReport>();
        Site site;
        String contextId;
        String mapKey;
        int index;
        StringBuilder itemId = new StringBuilder();

        for (RadiologyReport radReport : radiologyReports) {
            if (radReport == null) {
                continue;
            }

            site = radReport.getSite();
            if (site == null || site.getAgency() == null) {
                continue;
            }

            if (StringUtils.equalsIgnoreCase(site.getAgency(), APP_CONFIG.getAgencyDOD())) {
                if (StringUtils.isNotBlank(radReport.getExamId())) {
                    // Mapping DoD radiology exam ID with radiology report
                    dodRadReportMap.put(radReport.getExamId(), radReport);
                }
            }
        }

        // Step 2: Setup multi location query for both DoD (DoD query is also sent to VDS, not DES for VIX)
        List<PatientLocation> locs = LocationHelper.getPatientLocations(jMeadowsQuery, true);
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();
        QueryBean qb;

        for (PatientLocation loc : locs) {
            qb = null;

            try {
                qb = VistaBeanFactory.create(jMeadowsQuery);
            } catch (JMeadowsException e) {
                LOGGER.error("Failed to create vista query bean");
            }

            if (qb == null) {
                continue;
            }

            if (StringUtils.equalsIgnoreCase(SYSTEM_TYPE_DOD_ADAPTER, loc.getSystemType())) {
                // For DoD Radiology, we are querying for all images at site 200 in one call
                qb.setItemId(DOD_SITE_ID);
                configs.add(new PatientMultiLocQueryCfgBuilder().dataServiceHandler(sm_vistaHandler)
                        .dataServiceMethod("getVixStudyQueryFromSite").dataServiceMethodArgs(qb)
                        .location(loc).build());
            }
        }

        Object[] queryObjs = null;
        if (configs.size() > 0) {
            try {
                queryObjs = MultiLocationQuery.execute(configs, new SmartVixStudyQuery(), jMeadowsQuery);
            } catch (JMeadowsException e) {
                LOGGER.error("Failed to get study queries for radiologies");
            }
        }

        // Step 3: Map images to radiology reports
        if (queryObjs != null) {
            StudyQuery query;
            StudyQueryStudy queryStudy;
            StudyDetail detail;
            StudyDetailStudy detailStudy;
            RadiologyReport radReport;
            SimpleDateFormat simpleDateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
            Date studyDate;
            Date startDate = jMeadowsQuery.getStartDate().getTime();
            Date endDate = jMeadowsQuery.getEndDate().getTime();
            StringTokenizer tokenizer;
            String delim = APP_CONFIG.getVixDodRadExamIdDelim();
            String token;
            String examId;

            for (Object queryObj : queryObjs) {
                query = (StudyQuery)queryObj;

                if (query == null || query.getStudies() == null
                        || StringUtils.equalsIgnoreCase(query.getPatientIcn(), CONNECTION_UNAVAILABLE)) {
                    continue;
                }

                if (StringUtils.equalsIgnoreCase(query.getSiteNumber(), DOD_SITE_ID)) {
                    // All DoD images
                    for (Iterator<StudyQueryStudy> queryStudyItr = query.getStudies().iterator(); queryStudyItr.hasNext(); ) {
                        queryStudy = queryStudyItr.next();

                        if (queryStudy == null || StringUtils.isBlank(queryStudy.getExternalContextId())) {
                            // Skip if query study is null or
                            // external context ID is null or empty (it means it's not a DoD radiology if null or empty)
                            continue;
                        } else {
                            // Parse external context ID
                            token = "";
                            examId = queryStudy.getExternalContextId();
                            tokenizer = new StringTokenizer(examId, delim);

                            if (tokenizer != null) {
                                // Get last token
                                while (tokenizer.hasMoreTokens()) {
                                    token = tokenizer.nextToken();
                                }

                                if (StringUtils.isNotBlank(token)) {
                                    examId = token;
                                }
                            }
                        }

                        radReport = dodRadReportMap.get(examId);
                        if (radReport != null) {
                            radReport.setVixStudy(queryStudy);
                        } else {
                            // Add new report not linked to an existing radiology report
                            detail = queryStudy.getStudyDetail();
                            if (detail != null) {
                                if (detail.getStudies() != null && detail.getStudies().size() > 0) {
                                    detailStudy = detail.getStudies().get(0);
                                    if (detailStudy != null) {
                                        try {
                                            studyDate = simpleDateFormat.parse(detailStudy.getStudyDate());
                                            if (studyDate.compareTo(startDate) >= 0 && studyDate.compareTo(endDate) <= 0) {
                                                radReport = new RadiologyReport();

                                                radReport.setVixStudy(queryStudy);
                                                radReport.setExamDate(Utils.toXMLGregorianCalendar(studyDate));
                                                radReport.setStudy(detailStudy.getStudyDescription());

                                                site = new Site();
                                                site.setAgency(APP_CONFIG.getAgencyDOD());
                                                site.setName("HAIMS");
                                                site.setMoniker(APP_CONFIG.getAgencyDOD());
                                                radReport.setSite(site);

                                                radiologyReports.add(radReport);
                                            }
                                        } catch (ParseException e) {
                                            LOGGER.error("Failed to parse study date");
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    public String getPatientDocumentVixStudyViewerUrl(JMeadowsQuery jMeadowsQuery, PatientDocument document) {
        if (jMeadowsQuery == null || document == null) {
            return null;
        }

        QueryBean qb;
        try {
            qb = VistaBeanFactory.create(jMeadowsQuery);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to create vista query bean");
            return null;
        }
        String contextId = generateVaPatientDocumentContextId(jMeadowsQuery.getPatient(), document);
        qb.setItemId(contextId);
        return getVixStudyViewerUrl(qb);
    }

    public String getEncounterDocumentVixStudyViewerUrl(JMeadowsQuery jMeadowsQuery, String documentId, String documentSiteId) {
        if (jMeadowsQuery == null || documentId == null || documentSiteId == null) {
            return null;
        }

        QueryBean qb;
        try {
            qb = VistaBeanFactory.create(jMeadowsQuery);
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to create vista query bean");
            return null;
        }
        String contextId = generateVaEncounterDocumentContextId(jMeadowsQuery.getPatient(), documentId, documentSiteId);
        qb.setItemId(contextId);
        return getVixStudyViewerUrl(qb);
    }

    public String getRadiologyVixStudyViewerUrl(JMeadowsQuery jMeadowsQuery, RadiologyReport radiologyReport) {
        if (jMeadowsQuery == null || radiologyReport == null) {
            return null;
        }

        QueryBean qb;
        try {
            qb = VistaBeanFactory.create(jMeadowsQuery);
        } catch (JMeadowsException e) {
            LOGGER.error("Failed to create vista query bean");
            return null;
        }
        String contextId = generateVaRadiologyReportContextId(jMeadowsQuery.getPatient(), radiologyReport);
        qb.setItemId(contextId);
        return getVixStudyViewerUrl(qb);
    }

    private String getVixStudyViewerUrl(QueryBean qb) {
        try {
            String viewerUrl = sm_vistaHandler.getVixStudyViewerUrl(qb);
            if (viewerUrl != null) {
                return viewerUrl;
            } else {
                LOGGER.error("Failed to get study viewer URL");
            }
        }
        catch (JMeadowsException e) {
            LOGGER.error("Failed to get study viewer URL", e);
        }
        return null;
    }
}
