/*
 * Janus 4.0 (c)
 * Copyright (c) 2012 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dao.util;

import gov.va.med.jmeadows.common.AppConfig;
import gov.va.med.jmeadows.smartbeans.patient.SmartBean;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.Callable;

public class MultiLocationQueryThread implements Callable<Object> {
    private static final Logger LOGGER = LoggerFactory.getLogger(MultiLocationQueryThread.class);

    private static final AppConfig APP_CONFIG = AppConfig.getInstance();
//    private static final String AGENCY_VA = APP_CONFIG.getAgencyVA();
//    private static final String AGENCY_DOD = APP_CONFIG.getAgencyDOD();
    private static final String IDENTIFIER_UNAVAILABLE_FLAG = APP_CONFIG.getIdentifierUnavailableFlag();
//    private static final String SITE_UNKNOWN = "Unknown";
    
    private DataQuery dataQuery;
    private MultiLocationQueryConfig queryConfig;
    private SmartBean smartBean;

    /**
     * Constructs a new DataQuery thread.
     * @param queryConfig Data Query configuration object.
     * @param smartBean SmartBean used to process data query
     */
    public MultiLocationQueryThread(MultiLocationQueryConfig queryConfig, SmartBean smartBean)
    {
        this.queryConfig = queryConfig;
        this.dataQuery = new DataQuery(queryConfig.getWebserviceHandler(),
                queryConfig.getWebserviceMethod(),
                queryConfig.getWebserviceArgs());
        this.smartBean = smartBean;
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();

        sb.append("{dataQuery: ");
        if (dataQuery != null) {
            sb.append(dataQuery.toString());
        }

        sb.append(", smartBean: ");
        if (smartBean != null) {
            sb.append(smartBean.getClass().getName());
        }

        sb.append(", dataQuery: ");
        if (dataQuery != null) {
            sb.append(dataQuery.toString());
        }
        sb.append("}");

        return sb.toString();
    }

    /**
     * Executes dataQuery and handles result post processing
     */
    public Object call() {
        Object rtc = null;
        boolean done = false;
        LOGGER.debug(this.toString());

        Object result = null;

        //do not execute query if source ien is marked as unavailable (Constants.IDENTIFIER_UNAVAILABLE)
        if (queryConfig instanceof PatientMultiLocationQueryConfig) {
//            PatientLocation loc = ((PatientMultiLocationQueryConfig)queryConfig).getLocation();


            //check if this is a bhie query (business rule: proceed with bhie query even if chcs ien is unavailable)
            Object[] args = queryConfig.getWebserviceArgs();

            if (args[0] != null && args[0] instanceof gov.va.med.bhie.QueryBean) {
                gov.va.med.bhie.QueryBean qb = (gov.va.med.bhie.QueryBean) args[0];
                String patientId = qb.getPatientId();

                //report connection unavailable bean if bhie identifier is unavailable
                if (patientId == null || patientId.contains(IDENTIFIER_UNAVAILABLE_FLAG)) {
                    result = smartBean.generateConnectionUnavailableBean(
                            queryConfig.getLocation(),
                            queryConfig.getConnectionErrorMsg());
                    rtc = result;
                    done = true;
                }
            }
        }

        if (!done) {
            result = dataQuery.execute();
    
            //if there is a problem with data query
            //report 'Connection unavailable' to client
            if (result != null && result.getClass().getName().endsWith("Exception")) {
                Exception e = (Exception) result;
    
                //log exception
                LOGGER.error(e.getMessage(), e);
    
                result = smartBean.generateConnectionUnavailableBean(
                        queryConfig.getLocation(),
                        queryConfig.getConnectionErrorMsg());
    
                rtc = result;
    
            } else { //process result
                //if non-array result is expected, set result to a single item array
                if (queryConfig.isSingleResultExpected()) {
                    Object singleResult = (Object) result;
                    Object[] resultArr = new Object[1];
                    resultArr[0] = singleResult;
                    result = resultArr;
                }
    
                //if result requires bean mapping
                if (queryConfig.resultRequiresBeanMapping()) {
                    DataQuery dq = new DataQuery(smartBean,
                            queryConfig.getSmartBeanMappingMethod(), new Object[]{result});
    
                    result = (Object[]) dq.execute();
    
                }
    
                rtc =  result;
            }
        }
        return rtc;
    }

    public DataQuery getDataQuery() {
        return dataQuery;
    }

    public void setDataQuery(DataQuery dataQuery) {
        this.dataQuery = dataQuery;
    }

    public MultiLocationQueryConfig getQueryConfig() {
        return queryConfig;
    }

    public void setQueryConfig(MultiLocationQueryConfig queryConfig) {
        this.queryConfig = queryConfig;
    }

    public SmartBean getSmartBean() {
        return smartBean;
    }

    public void setSmartBean(SmartBean smartBean) {
        this.smartBean = smartBean;
    }
}
