/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dataservicehandler;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.*;
import gov.va.med.jmeadows.dao.hdd.VLERSiteCache;
import gov.va.med.jmeadows.dao.patient.AuditDao;
import gov.va.med.jmeadows.site.Sites;
import gov.va.med.mvi.*;
import gov.va.med.mvi.ObjectFactory;
import gov.va.med.vds.webservice.*;

import java.io.Serializable;
import java.io.StringWriter;
import java.math.BigInteger;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.WebServiceException;
import javax.xml.ws.handler.Handler;

import gov.va.med.vhahon.common.Utils;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class MVIServiceHandler extends DataServiceHandler {
    private static final Logger LOGGER = LoggerFactory.getLogger(MVIServiceHandler.class);
    private static final String OID_EDIPI = "2.16.840.1.113883.3.42.10001.100001.12";

    private static int lastGoodURLindex = 0;
    private VAIdM mviService;
    private JAXBContext jaxbContext = null;

    protected static final Sites SITES = Sites.getInstance();

    public MVIServiceHandler(String url) {
        super();

        serviceURL = url;
        serviceName = "MVI_Service";
        serviceNamespace = "http://webservice.mviws.med.va.gov";
        qServiceName = new QName(serviceNamespace, serviceName);

        mviService = new VAIdM();
    }

    private VAIdMPort getMVIPort() {
        VAIdMPort port = mviService.getVAIdMPort();

        // serviceURL can be a set of URLs to try like: URL;URL;URL
        String currentURL = serviceURL;
        if (serviceURL != null && serviceURL.indexOf(';') > -1) {
            String[] urls = serviceURL.split(";");
            if (lastGoodURLindex < urls.length) {
                currentURL = urls[lastGoodURLindex];
            } else {
                currentURL = urls[0];
                lastGoodURLindex = 0;
            }
        }
        BindingProvider prov = (BindingProvider) port;
        prov.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, currentURL);
        prov.getRequestContext().put("com.sun.xml.internal.ws.connect.timeout", serviceConnectionTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.internal.ws.request.timeout", serviceRequestTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.ws.connect.timeout", serviceConnectionTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.ws.request.timeout", serviceRequestTimeoutMS);

        // SOAP message logging - Enable to show logging in output window
        if ("test".equalsIgnoreCase(APP_CONFIG.getAppEnvironment()) || "development".equalsIgnoreCase(APP_CONFIG.getAppEnvironment())) {
            prov.getBinding().getHandlerChain().add(new SOAPLoggingHandler());
            List<Handler> handlerList = prov.getBinding().getHandlerChain();
            handlerList.add(new SOAPLoggingHandler());
            prov.getBinding().setHandlerChain(handlerList);
        }

        return port;
    }

    public ResponsePatientSelect getPatientIds(JMeadowsQuery queryBean, Map<String, Site> siteMap) throws JMeadowsException {
        String errorMessage = "";
        boolean done = false;

        ResponsePatientSelect responsePatientSelect = new ResponsePatientSelect();
        List<ExternalID> externalIDs = new ArrayList<ExternalID>();

        if (jaxbContext == null) {
            try {
                jaxbContext = JAXBContext.newInstance(PRPAIN201309UV02.class);
            } catch (JAXBException e) {
                jaxbContext = null;
                incLastGoodURLindex();

                errorMessage = "VA Patient Identity Service is unavailable. VA data may not display.";

                responsePatientSelect.setSuccess(false);
                responsePatientSelect.setErrorMsg(errorMessage);

                logAndAuditException(e, errorMessage, queryBean, false);

                done = true;
            }
        }
        if (!done) {
            try {
                // Generate PRPAIN201309UV02 request message
                PRPAIN201309UV02 request = generateRequest(queryBean);

                PRPAIN201310UV02 response = getMVIPort().prpaIN201309UV02(request);

                // Parse response into patient IDs map
                List<String> patientIds = parseResponseToPatientIds(response);

                Patient patient = new Patient();

                Iterator<String> iterator = patientIds.iterator();
                while (iterator.hasNext()) {
                    // Corresponding IDs format: ID^TYPE^SOURCE^ISSUER^STATUS
                    StringTokenizer tokenizer = new StringTokenizer(iterator.next(), "^");
                    if (tokenizer.countTokens() == 5) // According to format, there should only be 5 tokens
                    {
                        String id = tokenizer.nextToken();
                        String type = tokenizer.nextToken();
                        String source = tokenizer.nextToken();
                        String issuer = tokenizer.nextToken();
                        String status = tokenizer.nextToken();

                        // Makes sure each relevant token is not null or blank and status is "A" for active
                        if (StringUtils.isNotBlank(id) && StringUtils.isNotBlank(type)
                                && StringUtils.isNotBlank(source) && StringUtils.isNotBlank(issuer)
                                && StringUtils.isNotBlank(status)) {
                            if ("NI".equalsIgnoreCase(type)) {
                                // National patient identifier
                                if ("200DOD".equalsIgnoreCase(source) && "USDOD".equalsIgnoreCase(issuer) && "A".equalsIgnoreCase(status)) {
                                    // DoD EDIPI
                                    if ("EDIPI".equalsIgnoreCase(queryBean.getStatus())) {
                                        id = queryBean.getItemId();
                                    }
                                    patient.setEDIPI(id);
                                } else if ("200M".equalsIgnoreCase(source) && "USVHA".equalsIgnoreCase(issuer) && "P".equalsIgnoreCase(status)) {
                                    // VA ICN
                                    patient.setICN(id);
                                } else {
                                    externalIDs.add(new ExternalID(id, type, source, issuer));
                                }
                            } else if ("PI".equalsIgnoreCase(type)) {
                                // Local facility patient identifier
                                if ("USVHA".equalsIgnoreCase(issuer) && "A".equalsIgnoreCase(status)) {
                                    // VA DFN (IEN)
                                    if (siteMap != null) {
                                        Site site = siteMap.get(source);
                                        if (site == null) {
                                            LOGGER.debug("Site with site code " + source + " not found in sites map");
                                        } else {
                                            patient.getVistaSites().add(site);

                                            // Add patient's VA IEN and corresponding site code to key-value pair list
                                            KeyValuePair kvp = new KeyValuePair();
                                            kvp.setKey(source);
                                            kvp.setValue(id);
                                            patient.getPatientIens().add(kvp);
                                        }
                                    }
                                } else {
                                    externalIDs.add(new ExternalID(id, type, source, issuer));
                                }
                            }
                        }
                    } else {
                        LOGGER.debug("Invalid corresponding ID string format");
                    }
                }

                // VA user restrict flag is set. Check to see if VA user is selecting a DoD-only patient
                if ("TRUE".equalsIgnoreCase(APP_CONFIG.getVaUserRestrict())
                        && APP_CONFIG.getAgencyVA().equalsIgnoreCase(queryBean.getUser().getHostSite().getAgency()) && patient.getICN() == null
                        && patient.getPatientIens().isEmpty()) {
                    responsePatientSelect.setVARestricted(true);

                    if ("TRUE".equalsIgnoreCase(APP_CONFIG.getVaUserAllowRestrictAccess())) {
                        responsePatientSelect.setAllowVAAccess(true);
                    }
                }

                externalIDs = getVLERSiteNames(externalIDs);

                responsePatientSelect.setSuccess(true);
                responsePatientSelect.setPatient(patient);
                responsePatientSelect.setExternalIDs(externalIDs);
//            } catch (JAXBException e) {
//                incLastGoodURLindex();
//                errorMessage = "VA Patient Identity Service is unavailable. VA data may not display.";
//
//                responsePatientSelect.setSuccess(false);
//                responsePatientSelect.setErrorMsg(errorMessage);
//
//                logAndAuditException(e, errorMessage, queryBean, false);
            } catch (JMeadowsException e) {
                // If VA user, they do not need to see the error message. GUI checks for empty string and skips message.
                if (APP_CONFIG.getAgencyVA().equalsIgnoreCase(queryBean.getUser().getAgency())) {
                    errorMessage = "";
                } else {
                    errorMessage = "This patient is currently not known to the VA Master Veterans Index, and therefore no VA"
                            + " data will be displayed.<br/><br/>If this patient is a veteran, please notify the help desk of"
                            + " this issue.";
                }

                responsePatientSelect.setSuccess(false);
                responsePatientSelect.setErrorMsg(errorMessage);

                logAndAuditException(e, errorMessage, queryBean, false);
            } catch (Exception e) {
                incLastGoodURLindex();
                errorMessage = "VA Patient Identity Service is unavailable. VA data may not display.";

                responsePatientSelect.setSuccess(false);
                responsePatientSelect.setErrorMsg(errorMessage);

                logAndAuditException(e, errorMessage, queryBean, true);
                // Commenting out below line will not pass fortify(?) scan, but leaving it in breaks functionality
                // to continue showing DoD patient data when there is an issue with the MVI service. Revisit. -Jimi
                //            throw new JMeadowsException("MVI Application Error: " + errorMessage);
            }

            // MVI was unsuccessful, at least keep original identifier
            if (!responsePatientSelect.isSuccess()) {

                Patient patient = new Patient();

                String type = queryBean.getStatus();
                String id = queryBean.getItemId();
                String source = queryBean.getRecordSiteCode();

                if ("EDIPI".equalsIgnoreCase(type)) {
                    patient.setEDIPI(id);
                } else if ("ICN".equalsIgnoreCase(type)) {
                    patient.setICN(id);
                } else if ("VAIEN".equalsIgnoreCase(type) && siteMap != null) {
                    Site site = siteMap.get(source);
                    if (site != null) {
                        patient.getVistaSites().add(site);

                        // Add patient's VA IEN and corresponding site code to key-value pair list
                        KeyValuePair kvp = new KeyValuePair();
                        kvp.setKey(source);
                        kvp.setValue(id);
                        patient.getPatientIens().add(kvp);
                    }
                }

                // VA user restrict flag is set. Check to see if VA user is selecting DoD-only patient
                if ("TRUE".equalsIgnoreCase(APP_CONFIG.getVaUserRestrict())
                        && APP_CONFIG.getAgencyVA().equalsIgnoreCase(queryBean.getUser().getHostSite().getAgency())
                        && patient.getICN() == null && patient.getPatientIens().isEmpty()) {
                    responsePatientSelect.setVARestricted(true);
                    responsePatientSelect.setAllowVAAccess(true);
                }

                responsePatientSelect.setErrorMsg(errorMessage);
                responsePatientSelect.setPatient(patient);
            }
        }
        return responsePatientSelect;
    }

    public ResponsePatientQuery lookupPatient(PDWSQueryBean queryBean, Map<String, Site> siteMap) {
        ResponsePatientQuery rpq = new ResponsePatientQuery();

        try {
            PRPAIN201305UV02 searchRequest = generateSearchRequest(queryBean);

            String appEnv = APP_CONFIG.getAppEnvironment();
            //print request xml
            if ("test".equalsIgnoreCase(appEnv) || "development".equalsIgnoreCase(appEnv)) {
                jaxbContext = JAXBContext.newInstance(PRPAIN201305UV02.class.getPackage().getName());
                Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
                jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
                StringWriter stringWriter = new StringWriter();
                jaxbMarshaller.marshal(new JAXBElement<PRPAIN201305UV02>(new QName("uri", "PRPAIN201305UV02"), PRPAIN201305UV02.class, searchRequest), stringWriter);
                LOGGER.debug(stringWriter.toString());
            }
            //end printout;

            PRPAIN201306UV02 response = getMVIPort().prpaIN201305UV02(searchRequest);

            //print response xml
            if ("test".equalsIgnoreCase(appEnv) || "development".equalsIgnoreCase(appEnv)) {
                jaxbContext = JAXBContext.newInstance(PRPAIN201306UV02.class.getPackage().getName());
                Marshaller jaxbMarshaller2 = jaxbContext.createMarshaller();
                jaxbMarshaller2 = jaxbContext.createMarshaller();
                jaxbMarshaller2.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
                jaxbMarshaller2.marshal(new JAXBElement<PRPAIN201306UV02>(new QName("uri", "PRPAIN201306UV02"), PRPAIN201306UV02.class, response), System.out);
            }
            //end printout;


            List<Patient> patients = parseResponseToPatientLookup(response, siteMap, queryBean.getUser().getRestrictedSiteCodesList());
            if (patients.size() > 0) {
                String msg = patients.get(0).getName();
                if (msg.contains("MVI Error:") || msg.contains("MVI Message:")) {
                    //remove prefix for dialog requested in JLV-69
                    if (msg.startsWith("MVI Message: Patient not found with search elements entered"))
                        msg = msg.substring("MVI Message: ".length());
                    rpq.setErrorMsg(msg);
                    rpq.setSuccess(false);
                    return rpq;
                }
            }
            rpq.setPatients(patients);
            rpq.setSuccess(true);

        } catch (WebServiceException e){
            LOGGER.error("Error with MVI Lookup: " + e.getMessage());
            rpq.setErrorMsg("Error with MVI Lookup:  The MVI Search capability appears to be down at this time.");
        } catch (Exception e) {
            LOGGER.error("Error with MVI Lookup: " + e.getMessage());
            rpq.setErrorMsg("There was an error with the MVI patient lookup query.");
        }
        return rpq;
    }

    private static void incLastGoodURLindex() {
        ++lastGoodURLindex;
    }


    private PRPAIN201309UV02 generateRequest(JMeadowsQuery queryBean) {
        ObjectFactory factory;
        PRPAIN201309UV02 request;
        MCCIMT000100UV01Receiver receiver;
        MCCIMT000100UV01Device device;
        MCCIMT000100UV01Sender sender;
        PRPAIN201309UV02QUQIMT021001UV01ControlActProcess controlActProcess;
        PRPAMT201307UV02QueryByParameter queryByParameter;
        PRPAMT201307UV02ParameterList parameterList;
        PRPAMT201307UV02PatientIdentifier patientIdentifier;
        SimpleDateFormat simpleDateFormat;
        II ii;
        TS ts;
        CS cs;
        CD cd;
        Date date;
        String status;
        String oid;
        String id;
        String type;
        String source;
        String issuer;

        factory = new ObjectFactory();
        request = factory.createPRPAIN201309UV02();
        date = new Date();

        //id
        simpleDateFormat = new SimpleDateFormat("yyyyMMdd", Locale.getDefault());
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.933");
        ii.setExtension("MCID_JANUS_DEV_T1_" + simpleDateFormat.format(date));
        request.setId(ii);

        //creationTime
        ts = factory.createTS();
        simpleDateFormat = new SimpleDateFormat("yyyyMMddhhmmss", Locale.getDefault());
        ts.setValue(simpleDateFormat.format(date));
        request.setCreationTime(ts);

        //interactionId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.1.6");
        ii.setExtension("PRPA_IN201309UV02");
        request.setInteractionId(ii);

        //processingCode
        cs = factory.createCS();
        if ("production".equalsIgnoreCase(APP_CONFIG.getAppEnvironment())) {
            cs.setCode("P");
        } else {
            cs.setCode("T");
        }
        request.setProcessingCode(cs);

        //processingModeCode
        cs = factory.createCS();
        cs.setCode("T");
        request.setProcessingModeCode(cs);

        //acceptAckCode
        cs = factory.createCS();
        cs.setCode("AL");
        request.setAcceptAckCode(cs);

        //receiver
        receiver = factory.createMCCIMT000100UV01Receiver();
        request.getReceiver().add(receiver);

        //receiver/typeCode
        receiver.setTypeCode(CommunicationFunctionType.RCV);

        //receiver/device
        device = factory.createMCCIMT000100UV01Device();
        receiver.setDevice(device);

        //receiver/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //receiver/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //receiver/device/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.4.349");
        ii.setExtension("200M");
        device.getId().add(ii);

        //sender
        sender = factory.createMCCIMT000100UV01Sender();
        request.setSender(sender);

        //sender/typeCode
        sender.setTypeCode(CommunicationFunctionType.SND);

        //sender/device
        device = factory.createMCCIMT000100UV01Device();
        sender.setDevice(device);

        //sender/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //sender/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //sender/device/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.4.349");
        ii.setExtension(APP_CONFIG.getMVISndDevId());
        device.getId().add(ii);

        //controlActProcess
        controlActProcess = factory.createPRPAIN201309UV02QUQIMT021001UV01ControlActProcess();
        request.setControlActProcess(controlActProcess);

        //controlActProcess/classCode
        controlActProcess.setClassCode(ActClassControlAct.CACT);

        //controlActProcess/moodCode
        controlActProcess.setMoodCode(XActMoodIntentEvent.EVN);

        //controlActProcess/code
        cd = factory.createCD();
        cd.setCode("PRPA_TE201309UV02");
        controlActProcess.setCode(cd);

        //controlActProcess/codeSystem
        cd.setCodeSystem("2.16.840.1.113883.1.6");

        //controlActProcess/queryByParameter
        queryByParameter = factory.createPRPAMT201307UV02QueryByParameter();
        controlActProcess.setQueryByParameter(factory.createPRPAIN201309UV02QUQIMT021001UV01ControlActProcessQueryByParameter(queryByParameter));

        //controlActProcess/queryByParameter/queryId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.933");
        ii.setExtension("33452");
        queryByParameter.setQueryId(ii);

        //controlActProcess/queryByParameter/statusCode
        cs = factory.createCS();
        cs.setCode("new");
        queryByParameter.setStatusCode(cs);

        //controlActProcess/queryByParameter/responsePriorityCode
        cs = factory.createCS();
        cs.setCode("I");
        queryByParameter.setResponsePriorityCode(cs);

        //controlActProcess/queryByParameter/parameterList
        parameterList = factory.createPRPAMT201307UV02ParameterList();
        queryByParameter.setParameterList(parameterList);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier
        patientIdentifier = factory.createPRPAMT201307UV02PatientIdentifier();
        parameterList.getPatientIdentifier().add(patientIdentifier);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier/value
        ii = factory.createII();
        status = queryBean.getStatus();
        if ("EDIPI".equalsIgnoreCase(status)) {
            // Get identifiers by DoD EDIPI
            oid = "2.16.840.1.113883.3.364";
            id = queryBean.getItemId();
            type = "NI";
            source = "200DOD";
            issuer = "USDOD";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type + "^" + source + "^" + issuer);

            LOGGER.debug("Getting corresponding patient identifiers via EDIPI");
        } else if ("ICN".equalsIgnoreCase(status)) {
            // Get identifiers by VA ICN
            oid = "2.16.840.1.113883.4.349";
            id = queryBean.getItemId();
            type = "NI";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type);

            LOGGER.debug("Getting corresponding patient identifiers via ICN");
        } else if ("VAIEN".equalsIgnoreCase(status)) {
            // Get identifiers by VA IEN
            oid = "2.16.840.1.113883.4.349";
            id = queryBean.getItemId();
            type = "PI";
            source = queryBean.getRecordSiteCode();
            issuer = "USVHA";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type + "^" + source + "^" + issuer);

            LOGGER.debug("Getting corresponding patient identifiers via VA IEN");
        }
//        else
//        {
//            logger.error("Could not query with identifier number type: " + status);
//        }
        patientIdentifier.getValue().add(ii);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier/semanticsText
        patientIdentifier.setSemanticsText("Patient.id");

        return request;
    }

    private PRPAIN201305UV02 generateSearchRequest(PDWSQueryBean queryBean) {
        ObjectFactory factory;
        PRPAIN201305UV02 request;
        MCCIMT000100UV01Receiver receiver;
        MCCIMT000100UV01Device device;
        MCCIMT000100UV01Sender sender;
        PRPAIN201305UV02QUQIMT021001UV01ControlActProcess controlActProcess;
        PRPAMT201306UV02QueryByParameter queryByParameter;
        PRPAMT201306UV02ParameterList parameterList;
        //PRPAMT20130UV02PatientIdentifier patientIdentifier;
        SimpleDateFormat simpleDateFormat;
        II ii;
        TS ts;
        CS vc;
        TEL tel;
        CS cs;
        CD cd;
        Date date;
        String queryID;
        String status;
        String oid;
        String id;
        String type;
        String source;
        String issuer;

        factory = new ObjectFactory();
        request = factory.createPRPAIN201305UV02();
        date = new Date();

        try {
            //id
            simpleDateFormat = new SimpleDateFormat("yyyyMMddHHmmss", Locale.getDefault());
            ii = factory.createII();
            ii.setRoot("2.16.840.1.113883.3.933");
            queryID = "MCID_JANUS_DEV_T1_" + simpleDateFormat.format(date);
            ii.setExtension(queryID);
            request.setId(ii);

            //creationTime
            ts = factory.createTS();
            simpleDateFormat = new SimpleDateFormat("yyyyMMddhhmmss", Locale.getDefault());
            ts.setValue(simpleDateFormat.format(date));
            request.setCreationTime(ts);

            //versionCode
            vc = factory.createCS();
            vc.setCode("3.0");
            request.setVersionCode(vc);

            //interactionId
            ii = factory.createII();
            ii.setRoot("2.16.840.1.113883.1.6");
            ii.setExtension("PRPA_IN201305UV02");
            request.setInteractionId(ii);

            //processingCode
            cs = factory.createCS();
            if ("production".equalsIgnoreCase(APP_CONFIG.getAppEnvironment())) {
                cs.setCode("P");
            } else {
                cs.setCode("T");
            }
            request.setProcessingCode(cs);

            //processingModeCode
            cs = factory.createCS();
            cs.setCode("T");
            request.setProcessingModeCode(cs);

            //acceptAckCode
            cs = factory.createCS();
            cs.setCode("AL");
            request.setAcceptAckCode(cs);

            //receiver
            receiver = factory.createMCCIMT000100UV01Receiver();
            request.getReceiver().add(receiver);

            //receiver/typeCode
            receiver.setTypeCode(CommunicationFunctionType.RCV);

            //receiver/device
            device = factory.createMCCIMT000100UV01Device();
            receiver.setDevice(device);

            //receiver/device/classCode
            device.setClassCode(EntityClassDevice.DEV);

            //receiver/device/determinerCode
            device.setDeterminerCode("INSTANCE");

            //receiver/device/id
            ii = factory.createII();
            ii.setRoot("1.2.840.114350.1.13.999.234");
            device.getId().add(ii);

            //receiver/device/telcom
            tel = factory.createTEL();
            tel.setValue("http://servicelocation/PDQuery");
            device.getTelecom().add(tel);

            //sender
            sender = factory.createMCCIMT000100UV01Sender();
            request.setSender(sender);

            //sender/typeCode
            sender.setTypeCode(CommunicationFunctionType.SND);

            //sender/device
            device = factory.createMCCIMT000100UV01Device();
            sender.setDevice(device);

            //sender/device/classCode
            device.setClassCode(EntityClassDevice.DEV);

            //sender/device/determinerCode
            device.setDeterminerCode("INSTANCE");

            //sender/device/id
            ii = factory.createII();
            ii.setRoot("2.16.840.1.113883.4.349");
            ii.setExtension(APP_CONFIG.getMVISndDevId());
            device.getId().add(ii);

            //controlActProcess
            controlActProcess = factory.createPRPAIN201305UV02QUQIMT021001UV01ControlActProcess();
            request.setControlActProcess(controlActProcess);

            //controlActProcess/classCode
            controlActProcess.setClassCode(ActClassControlAct.CACT);

            //controlActProcess/moodCode
            controlActProcess.setMoodCode(XActMoodIntentEvent.EVN);

            //controlActProcess/dataEnterer
            QUQIMT021001UV01DataEnterer de = factory.createQUQIMT021001UV01DataEnterer();
            controlActProcess.getDataEnterer().add(de);
            de.setContextControlCode("AP");
            de.getTypeCode().add("ENT");

            COCTMT090100UV01AssignedPerson aPerson = factory.createCOCTMT090100UV01AssignedPerson();
            de.setAssignedPerson(aPerson);

            //controlActProcess/dataEnterer/assignedPerson
            aPerson.setClassCode("ASSIGNED");


            //controlActProcess/dataEnterer/assignedPerson/id
            String personID = "";
            String userNameLast = "";
            String userNameFirst = "";
            String idRoot = "2.16.840.1.113883.4.349";

            if (queryBean.getUser() != null) {
                User user = queryBean.getUser();
                if (user.getCardId() != null && "DOD".equalsIgnoreCase(user.getAgency())) {
                    personID = user.getCardId() + "^NI^200DOD^USDOD";
                    idRoot = "2.16.840.1.113883.3.42.10001.100001.12";
                } else if("VA".equalsIgnoreCase(user.getAgency())){
                    personID = user.getUserId() + "^PN^" + user.getHostSite().getSiteCode() + "^USVHA";
                }

                String userNameFull = user.getName();
                if (!StringUtils.isEmpty(userNameFull)) {
                    userNameLast = Utils.getFirstName(userNameFull).toUpperCase(Locale.getDefault());
                    userNameFirst = Utils.getLastName(userNameFull).toUpperCase(Locale.getDefault());
                }
            }

            II aPersonId = factory.createII();
            aPersonId.setExtension(personID);
            aPersonId.setRoot(idRoot);
            aPerson.getId().add(aPersonId);

            //controlActProcess/dataEnterer/assignedPerson/
            COCTMT090100UV01Person aPersonPerson = factory.createCOCTMT090100UV01Person();
            aPersonPerson.getClassCode().add("PSN");
            aPersonPerson.setDeterminerCode("INSTANCE");

            //controlActProcess/dataEnterer/assignedPerson/name
            EN aPersonPersonName = factory.createEN();
            aPersonPersonName.getContent().add(factory.createENFamily(userNameLast));
            aPersonPersonName.getContent().add(factory.createENGiven(userNameFirst));
            aPersonPerson.getName().add(aPersonPersonName);
            aPerson.setAssignedPerson(factory.createCOCTMT090100UV01AssignedPersonAssignedPerson(aPersonPerson));

            //controlActProcess/code
            cd = factory.createCD();
            cd.setCode("PRPA_TE201305UV02");
            controlActProcess.setCode(cd);

            //controlActProcess/codeSystem
            cd.setCodeSystem("2.16.840.1.113883.1.6");

            //controlActProcess/queryByParameter
            queryByParameter = factory.createPRPAMT201306UV02QueryByParameter();
            controlActProcess.setQueryByParameter(factory.createPRPAIN201305UV02QUQIMT021001UV01ControlActProcessQueryByParameter(queryByParameter));

            //controlActProcess/queryByParameter/queryId
            ii = factory.createII();
            ii.setRoot("1.2.840.114350.1.13.28.1.18.5.999");
            ii.setExtension(queryID);
            queryByParameter.setQueryId(ii);

            //controlActProcess/queryByParameter/statusCode
            cs = factory.createCS();
            cs.setCode("new");
            queryByParameter.setStatusCode(cs);

            //controlActProcess/queryByParameter/modifyCode
            CS csModCode = factory.createCS();
            csModCode.setCode("MVI.COMP1");
            queryByParameter.setModifyCode(csModCode);

            //controlActProcess/queryByParameter/parameterList
            parameterList = factory.createPRPAMT201306UV02ParameterList();
            queryByParameter.setParameterList(parameterList);

            //the amount of results to return from a search query
            //controlActProcess/queryByParameter/initialQuantity
            INT initQty = factory.createINT();
            initQty.setValue(BigInteger.valueOf(APP_CONFIG.getMVISearchQty()));
            queryByParameter.setInitialQuantity(initQty);


            //patient

            //parameter: livingSubjectAdministrativeGender   M/F
            //controlActProcess/queryByParameter/parameterList/livingSubjectAdministrativeGender
            if (StringUtils.isNotEmpty(queryBean.getGender())) {
                PRPAMT201306UV02LivingSubjectAdministrativeGender saGender = factory.createPRPAMT201306UV02LivingSubjectAdministrativeGender();
                CE genderValue = factory.createCE();
                genderValue.setCode(queryBean.getGender());
                saGender.getValue().add(genderValue);
                saGender.setSemanticsText("Gender");
                parameterList.getLivingSubjectAdministrativeGender().add(saGender);
            }

            //parameter: livingSubjectBirthTime  YYYYMMDD
            //controlActProcess/queryByParameter/parameterList/livingSubjectBirthTime
            String dobStr = queryBean.getDob();
            if (StringUtils.isNotEmpty(dobStr)) {
                dobStr = Utils.shortDate(Utils.getCal(dobStr));
                PRPAMT201306UV02LivingSubjectBirthTime bt = factory.createPRPAMT201306UV02LivingSubjectBirthTime();
                IVLTS dob = factory.createIVLTS();
                dob.setValue(dobStr);
                bt.getValue().add(dob);
                bt.setSemanticsText("Date of Birth");
                parameterList.getLivingSubjectBirthTime().add(bt);
            }

            //parameter: livingSubjectId  (SSN)
            //controlActProcess/queryByParameter/parameterList/livingSubjectId
            if (StringUtils.isNotEmpty(queryBean.getPatientId())) {
                PRPAMT201306UV02LivingSubjectId lsId = factory.createPRPAMT201306UV02LivingSubjectId();
                II lsIdValue = factory.createII();
                lsIdValue.setRoot("2.16.840.1.113883.4.1");
                lsIdValue.setExtension(queryBean.getPatientId());
                lsId.getValue().add(lsIdValue);
                lsId.setSemanticsText("SSN");
                parameterList.getLivingSubjectId().add(lsId);
            }

            //parameter: livingSubjectName
            //controlActProcess/queryByParameter/parameterList/livingSubjectName
            PRPAMT201306UV02LivingSubjectName lsName = factory.createPRPAMT201306UV02LivingSubjectName();
            EN lsNameEN = factory.createEN();
            lsNameEN.getUse().add("L");
            lsNameEN.getContent().add(factory.createENFamily(queryBean.getLastName()));
            lsNameEN.getContent().add(factory.createENGiven(queryBean.getFirstName()));
            lsName.setSemanticsText("Legal Name");
            lsName.getValue().add(lsNameEN);
            parameterList.getLivingSubjectName().add(lsName);

            //parameter: patientTelecom
            //controlActProcess/queryByParameter/parameterList/patientTelecom
            if (StringUtils.isNotEmpty(queryBean.getTel())) {
                PRPAMT201306UV02PatientTelecom telecom = factory.createPRPAMT201306UV02PatientTelecom();
                TEL pTel = factory.createTEL();
                pTel.setValue(queryBean.getTel());
                pTel.getUse().add("HP");
                telecom.getValue().add(pTel);
                parameterList.getPatientTelecom().add(telecom);
            }

            //parameter: patientAddress
            //controlActProcess/queryByParameter/parameterList/patientAddress
            if (StringUtils.isNotEmpty(queryBean.getAddress()) || StringUtils.isNotEmpty(queryBean.getCity())
                    || StringUtils.isNotEmpty(queryBean.getState()) || StringUtils.isNotEmpty(queryBean.getZip())) {
                PRPAMT201306UV02PatientAddress patAddress = factory.createPRPAMT201306UV02PatientAddress();
                AD patAd = factory.createAD();
                patAd.getUse().add("HP");
                if (!StringUtils.isEmpty(queryBean.getAddress()))
                    patAd.getContent().add(factory.createADStreetAddressLine(queryBean.getAddress()));
                if (!StringUtils.isEmpty(queryBean.getCity()))
                    patAd.getContent().add(factory.createADCity(queryBean.getCity()));
                if (!StringUtils.isEmpty(queryBean.getState()))
                    patAd.getContent().add(factory.createADState(queryBean.getState()));
                if (!StringUtils.isEmpty(queryBean.getZip()))
                    patAd.getContent().add(factory.createADPostalCode(queryBean.getZip()));
                patAddress.getValue().add(patAd);
                parameterList.getPatientAddress().add(patAddress);
            }

            PRPAMT201306UV02OtherIDsScopingOrganization otherOrg = factory.createPRPAMT201306UV02OtherIDsScopingOrganization();
            II otherOrgValue = factory.createII();
            otherOrgValue.setExtension("VA_DOD");
            otherOrgValue.setRoot("2.16.840.1.113883.4.349");
            otherOrg.getValue().add(otherOrgValue);
            otherOrg.setSemanticsText("MVI.REGISTRATION");
            parameterList.getOtherIDsScopingOrganization().add(otherOrg);

        } catch (Exception e) {
            e.printStackTrace(System.out);
        }

        return request;
    }

    private List<String> parseResponseToPatientIds(PRPAIN201310UV02 response) throws JMeadowsException {
        Iterator<MCCIMT000300UV01Acknowledgement> acknowledgementIterator;
        MCCIMT000300UV01Acknowledgement acknowledgement;
        Iterator<MCCIMT000300UV01AcknowledgementDetail> acknowledgementDetailIterator;
        MCCIMT000300UV01AcknowledgementDetail acknowledgementDetail;
        Iterator<PRPAIN201310UV02MFMIMT700711UV01Subject1> subjectIterator;
        PRPAIN201310UV02MFMIMT700711UV01Subject1 subject;
        PRPAMT201304UV02Patient patient;
        Iterator<II> iiIterator;
        II ii;

        List<String> patientIds = new ArrayList<String>();
        String errorMessage = "No Acknowledgement detail was provided in MVI repsonse.";

        // Check for errors
        acknowledgementIterator = response.getAcknowledgement().iterator();
        while (acknowledgementIterator.hasNext()) {
            acknowledgement = acknowledgementIterator.next();
            if (!"AA".equalsIgnoreCase(acknowledgement.getTypeCode().getCode())) {
                acknowledgementDetailIterator = acknowledgement.getAcknowledgementDetail().iterator();
                while (acknowledgementDetailIterator.hasNext()) {
                    acknowledgementDetail = acknowledgementDetailIterator.next();
                    errorMessage = acknowledgementDetail.getText();
                    break;
                }
                //melcolm todo: error logs are reporting: "Correlation Does Not Exist:IQI"  filter these out from error log
                //return null?
                throw new JMeadowsException("MVI Application Error: " + errorMessage);
            }
        }

        // Parse out patient identifiers
        subjectIterator = response.getControlActProcess().getSubject().iterator();
        while (subjectIterator.hasNext()) {
            subject = subjectIterator.next();
            patient = subject.getRegistrationEvent().getSubject1().getPatient();

            iiIterator = patient.getId().iterator();
            while (iiIterator.hasNext()) {
                ii = iiIterator.next();
                patientIds.add(ii.getExtension());
            }
        }

        return patientIds;
    }

    private List<Patient> parseResponseToPatientLookup(PRPAIN201306UV02 response, Map<String, Site> siteMap, List<Object> restrictedSiteCodesList) throws JMeadowsException {
        List<Patient> patients = new ArrayList<Patient>();


        Iterator<MCCIMT000300UV01Acknowledgement> acknowledgementIterator;
        MCCIMT000300UV01Acknowledgement acknowledgement;
        Iterator<MCCIMT000300UV01AcknowledgementDetail> acknowledgementDetailIterator;
        MCCIMT000300UV01AcknowledgementDetail acknowledgementDetail;
        Iterator<PRPAIN201306UV02MFMIMT700711UV01Subject1> subjectIterator;
        PRPAIN201306UV02MFMIMT700711UV01Subject1 subject;
        PRPAMT201310UV02Patient mviPatient;
        Iterator<II> iiIterator;
        II ii;

//        List<String> patientIds = new ArrayList<String>();
        String errorMessage = "No Acknowledgement detail was provided in MVI repsonse.";
//
//        // Check for errors
        acknowledgementIterator = response.getAcknowledgement().iterator();
        while (acknowledgementIterator.hasNext()) {
            acknowledgement = acknowledgementIterator.next();
            if (!"AA".equalsIgnoreCase(acknowledgement.getTypeCode().getCode())) {
                acknowledgementDetailIterator = acknowledgement.getAcknowledgementDetail().iterator();
                errorMessage = "";
                while (acknowledgementDetailIterator.hasNext()) {
                    acknowledgementDetail = acknowledgementDetailIterator.next();
                    errorMessage += acknowledgementDetail.getText() + "; ";
                    break;
                }
                if("AR".equalsIgnoreCase(acknowledgement.getTypeCode().getCode())){
                    Patient patient = new Patient();
                    if(errorMessage.contains("DOD Error")){
                        errorMessage = "Please modify your search parameters and try again.";
                    }
                    patient.setName("MVI Error: Application Reject. There was an error attempting to process your query. " + errorMessage);
                    patients.add(patient);
                    return patients;
                } else if("AE".equalsIgnoreCase(acknowledgement.getTypeCode().getCode())){
                    Patient patient = new Patient();
                    patient.setName("MVI Error: Application Error. There was an error attempting to process your query. " + errorMessage);
                    patients.add(patient);
                    return patients;
                }
                //melcolm todo: handle this differently, so that user is informed of proper errors from MVI
                //return null?
                throw new JMeadowsException("MVI Application Error: " + errorMessage);
            }
        }

        //query response code: QE - too many results, AE - application error, NF - no records found
        String qrc = "";
        if(response.getControlActProcess() != null &&
                response.getControlActProcess().getQueryAck() != null &&
                response.getControlActProcess().getQueryAck().getQueryResponseCode() != null){
            qrc = response.getControlActProcess().getQueryAck().getQueryResponseCode().getCode();
            if("QE".equalsIgnoreCase(qrc)) {
                Patient patient = new Patient();
                patient.setName("MVI Error: Your query yields too many results. Please modify your search parameters to narrow the search.");
                patients.add(patient);
                return patients;
            } else if("AE".equalsIgnoreCase(qrc)) {
                Patient patient = new Patient();
                patient.setName("MVI Error: There was an error attempting to process your query.");
                patients.add(patient);
                return patients;
            } else if("NF".equalsIgnoreCase(qrc)) {
                Patient patient = new Patient();
                patient.setName("MVI Message: Patient not found with search elements entered, please provide additional " +
                        "patient identifying information and search again");
                patients.add(patient);
                return patients;
            }
        }

        subjectIterator = response.getControlActProcess().getSubject().iterator();
        while (subjectIterator.hasNext()) {
            Patient patient = new Patient();
            subject = subjectIterator.next();
            mviPatient = subject.getRegistrationEvent().getSubject1().getPatient();

            // Parse out patient identifiers
            List<String> patientIds = new ArrayList<String>();
            iiIterator = mviPatient.getId().iterator();
            while (iiIterator.hasNext()) {
                ii = iiIterator.next();
                if(StringUtils.isNotEmpty(ii.getRoot()) && OID_EDIPI.equals(ii.getRoot())
                        && StringUtils.isNotEmpty(ii.getExtension()) && ii.getExtension().length() == 10){
                    patient.setEDIPI(ii.getExtension());
                } else {
                    patient = parsePatientId(ii.getExtension(), siteMap, patient);
                }
            }

            // Parse patientPerson
            PRPAMT201310UV02Person patientPerson = mviPatient.getPatientPerson().getValue();
            if (patientPerson != null) {
                CE ce;
                TS ts;
                CS cs;

                // Patient gender
                ce = patientPerson.getAdministrativeGenderCode();
                if (ce != null) {
                    patient.setGender(ce.getCode());
                }

                // Patient date of birth
                ts = patientPerson.getBirthTime();
                if (ts != null) {
                    patient.setDob(ts.getValue());
                }

                // Patient name
                if (patientPerson.getName() != null) {
                    for (PN pn : patientPerson.getName()) {
                        if (pn.getUse() == null) { continue; }
                        for (String use : pn.getUse()) {
                            if ("L".equalsIgnoreCase(use)) {
                                patient.setName(parseName(pn));
                            }
                        }
                    }
                }

                // Patient SSN
                if (patientPerson.getAsOtherIDs() != null) {
                    for (PRPAMT201310UV02OtherIDs otherIDs : patientPerson.getAsOtherIDs()) {
                        if (otherIDs.getId() == null) { continue; }
                        for (II otherId : otherIDs.getId()) {
                            if ("2.16.840.1.113883.4.1".equalsIgnoreCase(otherId.getRoot())){
                                patient.setSSN(otherId.getExtension());
                            }
                        }
                    }
                }
            }

            boolean unrestrictedPatient = true;
            // if there is a restricted site list, only add patients with an id at at least one vista site in the list
            if (CollectionUtils.isNotEmpty(restrictedSiteCodesList)) {
                // since there is a restricted list, at first assume this patient is not allowed to be added to lookup
                unrestrictedPatient = false;
                for (Site vistaSite : patient.getVistaSites()) {
                    // patient's vistaSites added to patient.getVistaSites() during parsing of identifiers; check those sites against restricted list
                    if (restrictedSiteCodesList.contains(vistaSite.getSiteCode())) {
                        // this patient has an id at a vista site in the site list so set this patient back to unrestricted
                        unrestrictedPatient = true;
                        break;
                    }
                }
                if (!unrestrictedPatient && APP_CONFIG.useEnhancedDebug()) {
                    StringBuilder debugString = new StringBuilder("!!! VBAUSER SITERESTRICTED: ");
                    // Log that a patient was filtered out.
                    debugString.append("Removing from search PATIENT: ");
                    debugString.append(patient.getName());
                    debugString.append("   Restricted SITES: ");
                    for (Object restrictedSiteCode : restrictedSiteCodesList) {
                        debugString.append(restrictedSiteCode);
                        debugString.append(" ");
                    }
                    debugString.append("  Patient SITES: ");
                    for (Site site : patient.getVistaSites()) {
                        debugString.append(site.getSiteCode());
                        debugString.append(" ");
                    }
                    LOGGER.debug(debugString.toString());
                }
            }

            // add only unrestricted patients to lookup list
            if (unrestrictedPatient) {
                patients.add(patient);
            }
        }

        return patients;
    }

    private Patient parsePatientId(String idString, Map<String, Site> siteMap, Patient patient){

        // Corresponding IDs format: ID^TYPE^SOURCE^ISSUER^STATUS
        StringTokenizer tokenizer = new StringTokenizer(idString, "^");

        List<ExternalID> externalIDs = new ArrayList<ExternalID>();

        if (tokenizer.countTokens() >= 4) // According to format, there should only be 5 tokens
        {
            String id = tokenizer.nextToken();
            String type = tokenizer.nextToken();
            String source = tokenizer.nextToken();
            String issuer = tokenizer.nextToken();
            String status = "";
            if(tokenizer.hasMoreTokens()) {
               status = tokenizer.nextToken();
            }

            // Makes sure each relevant token is not null or blank and status is "A" for active
            if (StringUtils.isNotBlank(id) && StringUtils.isNotBlank(type)
                    && StringUtils.isNotBlank(source) && StringUtils.isNotBlank(issuer)) {
                if ("NI".equalsIgnoreCase(type)) {
                    // National patient identifier
                    if ("200DOD".equalsIgnoreCase(source) && "USDOD".equalsIgnoreCase(issuer)
                            && ("A".equalsIgnoreCase(status) || StringUtils.isBlank(status))) {
                        patient.setEDIPI(id);
                    } else if ("200M".equalsIgnoreCase(source) && "USVHA".equalsIgnoreCase(issuer)
                            && ("P".equalsIgnoreCase(status) || StringUtils.isBlank(status))) {
                        // VA ICN
                        patient.setICN(id);
                    } else {
                        externalIDs.add(new ExternalID(id, type, source, issuer));
                    }
                } else if ("PI".equalsIgnoreCase(type)) {
                    // Local facility patient identifier
                    if ("USVHA".equalsIgnoreCase(issuer) && ("A".equalsIgnoreCase(status) || StringUtils.isBlank(status))) {
                        // VA DFN (IEN)
                        if (siteMap != null) {
                            Site site = siteMap.get(source);
                            if (site == null) {
                                LOGGER.debug("Site with site code " + source + " not found in sites map");
                            } else {
                                patient.getVistaSites().add(site);

                                // Add patient's VA IEN and corresponding site code to key-value pair list
                                KeyValuePair kvp = new KeyValuePair();
                                kvp.setKey(source);
                                kvp.setValue(id);
                                patient.getPatientIens().add(kvp);
                            }
                        }
                    } else {
                        externalIDs.add(new ExternalID(id, type, source, issuer));
                    }
                }
            }
        } else {
            LOGGER.debug("Invalid corresponding ID string format");
        }
        return patient;
    }

    private void logAndAuditException(Throwable e, String auditMsg, JMeadowsQuery qb, boolean isAudit) {
        AuditDao auditDao = new AuditDao();
//        StringWriter sw = new StringWriter();
//        PrintWriter pw = new PrintWriter(sw);

        // Log error
        if (e != null) {
//            e.printStackTrace(pw);
//            logger.error(sw.toString());
            LOGGER.error(e.getMessage(), e);
        }

        // Audit error
        if (isAudit) {
            String requestingApp = qb.getRequestingApp();
            String siteCode = null;
            String userIen = null;
            String userNPI = null;
            String userName = null;
            String patId;
            String siteMoniker = qb.getSiteMoniker();
            String siteAgency = qb.getSiteAgency();
            String recordID = qb.getItemId();
            String complexTransaction = qb.getComplexTransaction();
            User user = qb.getUser();

            if (user != null) {
                siteCode = user.getHostSite().getSiteCode();
                userIen = user.getUserIen();

                if (StringUtils.isNotBlank(user.getProviderIen())) {
                    userIen += "^" + user.getProviderIen();
                }

                userNPI = user.getNPI();
                userName = user.getName();
            }

            patId = qb.getItemId();
            if (StringUtils.isNotBlank(qb.getRecordSiteCode())) {
                patId += "^" + qb.getRecordSiteCode();
            }

            try {
                auditDao.setAudit(siteCode, siteMoniker, siteAgency, recordID, userIen, userNPI, userName, patId, auditMsg, requestingApp, null, null, complexTransaction);
            } catch (JMeadowsException exception) {
//                exception.printStackTrace(pw);
//                logger.error(sw.toString());
                if (e != null) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private List<ExternalID> getVLERSiteNames(List<ExternalID> ids) {
        for (ExternalID id : ids) {
            if (id.getSource() != null) {
                Code code = VLERSiteCache.getCode("", id.getSource(), "");
                if (code != null && code.getDisplay() != null) {
                    id.setSourceName(code.getDisplay());
                }
            }
        }
        return ids;
    }

    private String parseName(EN en) {
        StringBuilder familyName = new StringBuilder();
        StringBuilder givenName = new StringBuilder();
        StringBuilder prefixName = new StringBuilder();
        StringBuilder suffixName = new StringBuilder();
        List<Serializable> serializableList1;
        List<Serializable> serializableList2;
        String rtc;
        Object object;

        serializableList1 = en.getContent();
        if (serializableList1 == null) {
            rtc = null;
        } else {

            for (Serializable serializable1 : serializableList1) {
                if (!(serializable1 instanceof JAXBElement)) { continue; }

                object = ((JAXBElement) serializable1).getValue();
                if (object == null) { continue; }

                if ("family".equals(( (JAXBElement) serializable1).getName().getLocalPart()))  {
                    if(familyName.length()>0){
                        familyName.append(" ");
                    }
                   familyName.append(object);
                } else if ("given".equals(( (JAXBElement) serializable1).getName().getLocalPart()))  {
                    if(givenName.length()>0){
                        givenName.append(" ");
                    }
                    givenName.append(object);
                } else if ("prefix".equals(( (JAXBElement) serializable1).getName().getLocalPart())) {
                    if(prefixName.length()>0){
                        prefixName.append(" ");
                    }
                    prefixName.append(object);
                } else if ("suffix".equals(( (JAXBElement) serializable1).getName().getLocalPart())) {
                    if(suffixName.length()>0){
                        suffixName.append(" ");
                    }
                    suffixName.append(object);
                }
            }

            // add prefix before given name
            if (prefixName.length() > 0) {
                givenName = prefixName.append(givenName);
            }

            // add suffix after given name
            if (suffixName.length() > 0) {
                givenName = givenName.append(suffixName);
            }

            rtc = familyName.toString().trim() + "," + givenName.toString().trim();
        }
        return rtc;
    }

    private String parsePhoneNumber(String telecom) {
        StringBuilder phoneNumber = new StringBuilder();
        StringTokenizer tokenizer = new StringTokenizer(telecom, " ");
        String token;
        char code;

        while (tokenizer.hasMoreTokens()) {
            token = tokenizer.nextToken();
            code = token.toLowerCase(Locale.getDefault()).charAt(0);
            // Do not accept tokens that begin with (p)riority, (y)es text, (n)o text
            if ((code != 'p') && (code != 'y') && (code != 'n')) {
                phoneNumber.append(token);
                phoneNumber.append(" ");
            }
        }

        return phoneNumber.toString().trim();
    }
}
