/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.jmeadows.smartbeans.patient;

import gov.va.med.bhie.*;
import gov.va.med.bhie.SiteStatus;
import gov.va.med.jmeadows.common.AppConfig;
import gov.va.med.jmeadows.dao.hdd.VLERSiteCache;
import gov.va.med.jmeadows.dao.util.Location;
import gov.va.med.jmeadows.site.Sites;
import gov.va.med.vds.webservice.*;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.XMLGregorianCalendar;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public abstract class SmartBean implements Comparable {


    protected static final AppConfig APP_CONFIG = AppConfig.getInstance();

    //system types
    protected static final String SYSTEM_TYPE_SHARE = APP_CONFIG.getConstSystemTypeShare();
    protected static final String SYSTEM_TYPE_CHCS = APP_CONFIG.getSystemTypeCHCS();
    protected static final String SYSTEM_TYPE_DOD_ADAPTER = APP_CONFIG.getSystemTypeDODAdapter();
    protected static final String SYSTEM_TYPE_VISTA = APP_CONFIG.getSystemTypeVISTA();

    //protocols
    protected static final String SOURCE_PROTOCOL_CACHE = APP_CONFIG.getSourceProtocolCache();
    protected static final String SOURCE_PROTOCOL_DOD_ADAPTER = APP_CONFIG.getSourceProtocolDODAdapter();
    protected static final String SOURCE_PROTOCOL_SHARE = APP_CONFIG.getSourceProtocolShare();

    protected static final String ERROR_MSG_CONNECTION_UNAVAILABLE = APP_CONFIG.getErrorMsgConnectionUnavailable();
    protected static final String BHIE_ERROR_SITE_CODE = APP_CONFIG.getBHIEErrorSiteCode();
    protected static final String BHIE_SOURCE_STATUS_REPORT_FLAG = APP_CONFIG.getDODAdapterSourceStatusReportFlag();
    protected static final String AGENCY_DOD = APP_CONFIG.getAgencyDOD();
    protected static final String AGENCY_VA = APP_CONFIG.getAgencyVA();

    //meant for temporary use during transition to enterprise
    protected static final String LOCAL_CHCS_SITECODE = APP_CONFIG.getLocalCHCSSiteCode();

    protected static final Location BHIE_STATUS_REPORT_LOCATION = new Location();

    static {
        BHIE_STATUS_REPORT_LOCATION.setSystemType(BHIE_SOURCE_STATUS_REPORT_FLAG);
    }

    protected static final Sites SITES = Sites.getInstance();


    protected static final String BHIE_SITE_STATUS_UP = "UP";
    protected static final String BHIE_SITE_STATUS_DOWN = "DOWN";
    protected static final String BHIE_SITE_STATUS_NO_RESPONSE = "NO_RESPONSE";
    protected static final String BHIE_SITE_STATUS_TIMED_OUT = "TIMED_OUT";
    protected static final String BHIE_SITE_STATUS_UNKNOWN = "UNKNOWN";

    //for workarounds to filter consult notes
    protected static final String LOINC_CODE_NOTES = "11536-0";
    protected static final Set<String> NOTE_TYPE_CODE_CONSULT = Collections.unmodifiableSet(new HashSet<String>(Arrays.asList("258", "15149159")));

    private static final Logger LOGGER = LoggerFactory.getLogger(SmartBean.class);
    
    protected int m_sortBy;
    protected boolean m_bAscending;
    
    public int compareString(String string1, String string2) {
        int rtc;
        if (string1 == null && string2 == null) {
            rtc = 0;
        } else  if (string1 == null) {
            rtc = -1;
        } else if (string2 == null) {
            rtc = 1;
        } else {
            rtc = string1.compareToIgnoreCase(string2);
        }
        return rtc;
    }

    public int compareDate(Date date1, Date date2) {
        int rtc;
        if (date1 == date2) {
            rtc = 0;
        } else if (date1 == null) {
            rtc = -1;
        } else if (date2 == null) {
            rtc = 1;
        } else {
            rtc = date1.compareTo(date2);
        }
        return rtc;
    }

    public int compareDate(Calendar cal, Calendar cal2) {
        int rtc;
        if (cal == cal2) {
            rtc = 0;
        } else if (cal == null) {
            rtc = -1;
        } else if (cal2 == null) {
            rtc = 1;
        } else {
            rtc = cal.compareTo(cal2);
        }
        return rtc;
    }

    public int compareDate(XMLGregorianCalendar date1, XMLGregorianCalendar date2) {
        int rtc;
        if (date1 == date2) {
            rtc = 0;
        } else if (date1 == null) {
            rtc = -1;
        } else if (date2 == null) {
            rtc = 1;
        } else {
            rtc = date1.compare(date2);
        }
        return rtc;
    }

    /**
     * Compares String date in the format MMM dd, yyyy, or MMM, yyyy, or yyyy
     *
     * @param date1
     * @param date2
     * @return
     */
    public int compareDate(String date1, String date2) {
        return compareDate(dateStringToCal(date1), dateStringToCal(date2));
    }

    private Calendar dateStringToCal(String date) {
        Calendar rtc = null;
        String formatString = null;

        if (date != null) {
            if (date.length() == 12) {
                formatString = "MMM dd, yyyy";
            } else if (date.length() == 8) {
                formatString = "MMM, yyyy";
            } else if (date.length() == 4) {
                formatString = "yyyy";
            }
            if (formatString != null) {
                try {
                    SimpleDateFormat sdf = new SimpleDateFormat(formatString);
                    rtc = Calendar.getInstance();
                    rtc.setTime(sdf.parse(date));
                } catch (ParseException e) {
                    rtc = null;
                }
            }
        }

        return rtc;
    }

    public String setAscending(String sort) {
        String rtc;
        if (sort.length() < 1) {
            rtc = sort;
        } else {
            rtc = sort;
            m_bAscending = true;
            if (sort.indexOf("-") == 0) {
                m_bAscending = false;
                rtc = sort.substring(1, sort.length());
            }
        }
        return rtc;
    }

    private static Site getBHIESite(String siteCode, String siteName, String siteMoniker) {
        Site site = new Site();
        String moniker = siteMoniker;
        String name = siteName;
        site.setSiteCode(siteCode);

        if (name != null) {
            name = name.toUpperCase();
        }

        if (StringUtils.isBlank(moniker)) {
            moniker = name;
        } else if (StringUtils.isBlank(name)) {
            name = moniker;
        }

        site.setName(name);
        site.setMoniker(moniker);

        site.setAgency(bhieCustodianIdToSiteAgency(siteCode));

        return site;
    }

    protected static void addBHIESiteDataToVistaBean(String siteCode, String siteName, String siteMoniker, String cdrEventId, DataBean bean) {
        bean.setSite(getBHIESite(siteCode, siteName, siteMoniker));
        bean.setSourceProtocol(SOURCE_PROTOCOL_DOD_ADAPTER);

        if (StringUtils.isNotBlank(cdrEventId)) {
            bean.setCdrEventId(cdrEventId);
        }
    }

    protected static void addBHIESiteDataToVistaBean(String siteCode, String siteName, String siteMoniker, String cdrEventId, PatientDemographics bean) {
        bean.setSite(getBHIESite(siteCode, siteName, siteMoniker));
        bean.setSourceProtocol(SOURCE_PROTOCOL_DOD_ADAPTER);

        if (StringUtils.isNotBlank(cdrEventId)) {
            bean.setCdrEventId(cdrEventId);
        }
    }

    protected static String bhieCustodianIdToSiteAgency(String siteCode) {
        String rtc;
        if (APP_CONFIG.getBHIECustodianOidVA().equalsIgnoreCase(siteCode)) {
            rtc = APP_CONFIG.getAgencyVA();
        } else {
            rtc = AGENCY_DOD;
        }
        return rtc;
    }

    protected static String generateBhieStatusReport(String clinicalDomain, List<SiteStatus> siteStatusList, List<gov.va.med.bhie.Error> errorList) {
        String report = generateBhieStatusReport(siteStatusList, errorList);

        report = "<div class='domainStatusWrap'><div class='statusWrap vlerStatus'><div class='title'>" + clinicalDomain + "</div>" + report  + "</div>";

        return report;
    }

    protected static String generateBhieStatusReportV4(String clinicalDomain, List<SiteStatusV4> siteStatusList, List<String> errorList) {
        String report = generateBhieStatusReportV4(siteStatusList, errorList, false);

        report = "<div class='domainStatusWrap'><div class='title'>" + clinicalDomain + "</div>" + report + "</div>";

        return report;
    }

    protected static String generateBhieStatusReportV4(List<SiteStatusV4> siteStatusList, List<String> errorList, boolean isVler) {
        StringBuilder report = new StringBuilder();

        Map<String, String> siteStatusMap = new LinkedHashMap<String, String>();

        if (isVler) {
            report.append("<div class='statusWrap vlerStatus'><div class='title'>Community Partners Source Status List</div>");
        } else {
            report.append("<div class='statusWrap bdaStatus'><div class='title'>Source Status List</div>");
        }

        if (siteStatusList == null) {
            report.append("Source status list is unavailable.");
        } else {
            for (SiteStatusV4 siteStatus : siteStatusList) {
                if (siteStatus == null || siteStatus.getPartner() == null) {
                    continue;
                }
                if (siteStatus.getPartner().getPartnerName() != null && siteStatus.getPartnerStatus() != null) {
                    siteStatusMap.put(siteStatus.getPartner().getPartnerName(), siteStatus.getPartnerStatus());
                }
            }

            report.append("<table><thead><th scope='col' class='table-header'>Source Name</th><th scope='col' class='table-header'>Source Status</th></thead>");

            Set<String> keys = siteStatusMap.keySet();

            String[] keyArray = keys.toArray(new String[keys.size()]);
            Arrays.sort(keyArray);

            report.append("<tbody>");

            if (keyArray.length < 1) {
                report.append("<tr class='even'><td>Source status list is unavailable</td></tr>");
            } else {
                int cnt = 0;
                for (String siteName : keyArray) {
                    String statusStr = "";
                    String siteNameStr = "";

                    if (siteName != null) {
                        if (siteStatusMap.get(siteName) != null) {
                            statusStr = siteStatusMap.get(siteName).toUpperCase();
                        }
                        siteNameStr = siteName.toUpperCase();
                    }
                    String trClass = "odd";

                    if (cnt % 2 == 0) {
                        trClass = "even";
                    }

                    report.append("<tr class='" + trClass + "'>");
                    report.append("<td>");
                    report.append(siteNameStr);
                    report.append("</td>");
                    report.append("<td>");
                    report.append(statusStr);
                    report.append("</td>");
                    report.append("</tr>");

                    cnt++;
                }
            }

            report.append("</tbody>");
            report.append("</table>");
        }

        report.append("<div class='title statusErrors'>Errors</div>");

        if (errorList == null) {
            report.append("Error report is unavailable.");
        } else {
            if (errorList.size() < 1) {
                report.append("No errors to report.");
            } else {
                report.append("<table><thead><th scope='col' colspan='2' class='table-header'>Error Code</th></thead>");

                int cnt = 0;
                for (String error : errorList) {
                    String trClass = "odd";

                    if (cnt % 2 == 0) {
                        trClass = "even";
                    }

                    report.append("<tr class='" + trClass + "'>");
                    report.append("<td align='center'><div style='margin:10px;background-color:red;width:15px;'></div></td>");
                    report.append("<td>");
                    report.append(error);
                    report.append("</td>");
                    report.append("</tr>");

                    cnt++;
                }

                report.append("</tbody>");
                report.append("</table>");
            }
        }
        report.append("</div>");
        return report.toString();
    }

    protected static String generateBhieStatusReport(List<SiteStatus> siteStatusList, List<gov.va.med.bhie.Error> errorList) {
        StringBuilder report = new StringBuilder();

        Map<String, String> siteStatusMap = new LinkedHashMap<String, String>();

        report.append("<div class='statusWrap bdaStatus'><div class='title'>Source Status List</div>");

        if (siteStatusList == null) {
            report.append("Source status list is unavailable.");
        } else {
            StringBuilder siteStatusEntries = new StringBuilder();
            for (SiteStatus siteStatus : siteStatusList) {
//                String status = siteStatus.getSiteStatus();
//
//                siteStatusEntries.setLength(0);
//                if (siteStatusMap.containsKey(siteStatus.getSiteName())) {
//                    siteStatusEntries.append(siteStatusMap.get(siteStatus.getSiteName()));
//                    siteStatusEntries.append(", ");
//                    siteStatusEntries.append(status);
//                } else {
//                    siteStatusEntries.append(status);
//                }
//
                if (siteStatus != null && siteStatus.getSiteName() != null && siteStatusEntries.toString() != null) {
                    siteStatusMap.put(siteStatus.getSiteName(), siteStatusEntries.toString());
                }
            }

            report.append("<table><thead><th scope='col' class='table-header'>Source Name</th><th scope='col' class='table-header'>Source Status</th></thead>");

            Set<String> keys = siteStatusMap.keySet();

            String[] keyArray = keys.toArray(new String[keys.size()]);
            Arrays.sort(keyArray);

            report.append("<tbody>");

            if (keyArray.length < 1) {
                report.append("<tr class='even'><td>Source status list is unavailable</td></tr>");
            } else {
                int cnt = 0;
                for (String siteName : keyArray) {
                    String statusStr = "";
                    String siteNameStr = "";

                    if (siteName != null) {
                        if (siteStatusMap.get(siteName) != null) {
                            statusStr = siteStatusMap.get(siteName).toUpperCase();
                        }
                        siteNameStr = siteName.toUpperCase();
                    }
                    String trClass = "odd";

                    if (cnt % 2 == 0) {
                        trClass = "even";
                    }

                    report.append("<tr class='" + trClass + "'>");
                    report.append("<td>");
                    report.append(siteNameStr);
                    report.append("</td>");
                    report.append("<td>");
                    report.append(statusStr);
                    report.append("</td>");
                    report.append("</tr>");

                    cnt++;
                }
            }

            report.append("</tbody>");
            report.append("</table>");
        }

        report.append("<div class='title statusErrors'>Errors</div>");

        if (errorList == null) {
            report.append("Error report is unavailable.");
        } else {
            if (errorList.size() < 1) {
                report.append("No errors to report.");
            } else {
                report.append("<table><thead><th scope='col' colspan='2' class='table-header'>Error Code</th></thead>");

                int cnt = 0;
                for (gov.va.med.bhie.Error error : errorList) {
                    String trClass = "odd";

                    if (cnt % 2 == 0) {
                        trClass = "even";
                    }

                    report.append("<tr class='" + trClass + "'>");
                    report.append("<td align='center'><div style='margin:10px;background-color:red;width:15px;'></div></td>");
                    report.append("<td>");
                    report.append(error.getErrorCode());
                    report.append("</td>");
                    report.append("</tr>");

                    cnt++;
                }

                report.append("</tbody>");
                report.append("</table>");
            }
        }
        report.append("</div>");
        return report.toString();
    }

//    protected static String generateBhieVLERStatusReport(List<SiteStatusV4> siteStatusList, List<String> errorList) {
//        StringBuilder report = new StringBuilder();
//
//        Map<String, String> siteStatusMap = new LinkedHashMap<String, String>();
//
//        report.append("<div class='title'>Source Status List</div>");
//
//        if (siteStatusList == null) {
//            report.append("Source status list is unavailable.");
//        } else {
//            StringBuilder siteStatusEntries = new StringBuilder();
//            for (SiteStatusV4 siteStatus : siteStatusList) {
//                String status = siteStatus.getPartnerStatus();
//
//                siteStatusEntries.setLength(0);
//                if (siteStatusMap.containsKey(siteStatus.getPartner().getPartnerName())) {
//                    siteStatusEntries.append(siteStatusMap.get(siteStatus.getPartner().getPartnerName()));
//                    siteStatusEntries.append(", ");
//                    siteStatusEntries.append(status);
//                } else {
//                    siteStatusEntries.append(status);
//                }
//
//                siteStatusMap.put(siteStatus.getPartner().getPartnerName(), siteStatusEntries.toString());
//            }
//
//            report.append("<table><thead><th scope='col' class='table-header'>Source Name</th><th scope='col' class='table-header'>Source Status</th></thead>");
//
//            Set<String> keys = siteStatusMap.keySet();
//
//            String[] keyArray = keys.toArray(new String[0]);
//            Arrays.sort(keyArray);
//
//            report.append("<tbody>");
//
//            if (keyArray.length < 1) {
//                report.append("<tr class='even'><td>Source status list is unavailable</td></tr>");
//            } else {
//                int cnt = 0;
//                for (String siteName : keyArray) {
//                    String statusStr = siteStatusMap.get(siteName);
//
//                    String trClass = "odd";
//
//                    if (cnt % 2 == 0) trClass = "even";
//
//                    report.append("<tr class='" + trClass + "'>");
//                    report.append("<td>");
//                    report.append(siteName.toUpperCase());
//                    report.append("</td>");
//                    report.append("<td>");
//                    report.append(statusStr.toUpperCase());
//                    report.append("</td>");
//                    report.append("</tr>");
//
//                    cnt++;
//                }
//            }
//
//            report.append("</tbody>");
//            report.append("</table>");
//        }
//
//        report.append("<br/><br/><div class='title'>Errors</div>");
//
//        if (errorList == null) {
//            report.append("Error report is unavailable.");
//        } else {
//            if (errorList.size() < 1) {
//                report.append("No errors to report.");
//            } else {
//                report.append("<table><thead><th scope='col' colspan='2' class='table-header'>Error Code</th></thead>");
//
//                int cnt = 0;
//                for (String error : errorList) {
//                    String trClass = "odd";
//
//                    if (cnt % 2 == 0) trClass = "even";
//
//                    report.append("<tr class='" + trClass + "'>");
//                    report.append("<td align='center'><div style='margin:10px;background-color:red;width:15px;'></div></td>");
//                    report.append("<td>");
//                    report.append(error);
//                    report.append("</td>");
//                    report.append("</tr>");
//
//                    cnt++;
//                }
//
//                report.append("</tbody>");
//                report.append("</table>");
//            }
//        }
//
//        return report.toString();
//    }



        /**
         * Abstract sorting method
         *
         * @param array1
         * @param sort
         * @return
         */
    public abstract Object[] sort(Object[] array1, String sort);

    /**
     * Abstract method combineData: method combines data from
     * from different sources, all smart beans must implement this method.
     *
     * @param array1
     * @param array2
     * @param sort
     * @return
     */
    public abstract Object[] combineData(Object[] array1, Object[] array2, String sort);

    /**
     * Generate a single entry object array that contains connection unavailable databean.
     *
     * @return a single entry object array that contains connection unavailable databean..
     */
    public abstract Object[] generateConnectionUnavailableBean(Location location, String errorMsg);

    protected Site getConnectionUnavailableSite(Location location) {
        Site site = new Site();

        if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(location.getSystemType())
                || SYSTEM_TYPE_CHCS.equalsIgnoreCase(location.getSystemType())
                || SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(location.getSystemType())
                || SYSTEM_TYPE_SHARE.equalsIgnoreCase(location.getSystemType())) {
            site.setSiteCode(AGENCY_DOD);
            site.setName(AGENCY_DOD);
            site.setMoniker(AGENCY_DOD);
            site.setAgency(AGENCY_DOD);
        } else if (BHIE_SOURCE_STATUS_REPORT_FLAG.equalsIgnoreCase(location.getSystemType())) {
            site.setSiteCode(AGENCY_DOD);
            site.setMoniker(BHIE_SOURCE_STATUS_REPORT_FLAG);
            site.setName(BHIE_SOURCE_STATUS_REPORT_FLAG);
            site.setAgency(AGENCY_DOD);
        } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(location.getSystemType())) {
            site.setSiteCode(AGENCY_VA);
            site.setName(AGENCY_VA);
            site.setMoniker(AGENCY_VA);
            site.setAgency(AGENCY_VA);
        } else { //this case should never be reached!!
            final String unknown = "Unknown";

            site.setSiteCode(unknown);
            site.setName(unknown);
            site.setMoniker(unknown);
            site.setAgency(unknown);
        }

        return site;
    }

    protected static void addCHCSSiteDataToVistaBean(DataBean vistaBean) {
        vistaBean.setSite(SITES.getSite(LOCAL_CHCS_SITECODE));
        vistaBean.setSourceProtocol(SOURCE_PROTOCOL_CACHE);
    }

    protected static void addCHCSSiteDataToVistaBean(PatientDemographics vistaBean) {
        vistaBean.setSite(SITES.getSite(LOCAL_CHCS_SITECODE));
        vistaBean.setSourceProtocol(SOURCE_PROTOCOL_CACHE);
    }

    protected static void addSHARESiteDataToVistaBean(DataBean vistaBean) {
        vistaBean.setSite(SITES.getSite(LOCAL_CHCS_SITECODE));
        vistaBean.setSourceProtocol(SOURCE_PROTOCOL_CACHE);
    }

    /**
     * GregorianCalendar representation of date format string.
     * @param dateStr Date format string
     * @return GregorianCalendar representation of date format string.
     */
    protected static GregorianCalendar bhieToCal(String dateStr) {

        Date date = null;
        String format = "";
        GregorianCalendar retCal = null;
        SimpleDateFormat df = null;
        if (dateStr != null && dateStr.length() != 0) {
            try {

                if (dateStr.indexOf("/") > 0) {
                    format = "MM/dd/yy";
                } else if (dateStr.indexOf(" ") == 2) {
                    format = "dd MMM yyyy";
                } else if (dateStr.indexOf("T") > 0 && dateStr.indexOf("Z") > 0 && dateStr.length() == 17) {
                    format = "yyyy-MM-dd'T'HH:mm'Z'";
                } else if (dateStr.indexOf("T") > 0 && dateStr.indexOf("Z") > 0) {
                    format = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
                } else if (dateStr.indexOf("T") > 0 && dateStr.indexOf(":") > 0) {
                    dateStr = dateStr.replace("-00:00", "");
                    format = "yyyy-MM-dd'T'HH:mm:ss";
                } else if (dateStr.indexOf("-") > 0 && dateStr.indexOf(":") > 0) {
                    format = "yyyy-MM-dd HH:mm:ss";
                } else if (dateStr.indexOf("-") > 0 && dateStr.indexOf("@") > 0) {
                    format = "dd-MMM-yyyy@HHmm";
                } else if (dateStr.indexOf(",") > 0 && dateStr.indexOf("@") > 0) {
                    format = "MMM dd, yyyy@HH:mm";
                } else if (dateStr.indexOf("@") < 0 && dateStr.indexOf("-") == 4) {
                    format = "yyyy-MM-dd";
                } else if (dateStr.indexOf("@") > 0 && dateStr.indexOf("-") < 0) {
                    format = "ddMMMyy@HHmm";
                } else if (dateStr.length() == 8 && dateStr.indexOf("-") < 0) {
                    format = "yyyyMMdd";
                } else if (dateStr.length() == 12) {
                    format = "yyyyMMddHHmm";
                } else if (dateStr.length() == 14) {
                    format = "yyyyMMddHHmmss";
                } else if (dateStr.indexOf(".") > 12 && dateStr.length() > 14 && dateStr.length() < 19) {
                    //assumeUTC = false;
                    format = "yyyyMMddHHmmss.SSS";
                } else if (dateStr.indexOf(".") > 12 && dateStr.length() > 18) {
                    //assumeUTC = false;
                    format = "yyyyMMddHHmmss.SSSZ";
                } else if (dateStr.length() == 7) {
                    format = "ddMMMyy";
                } else if (dateStr.length() == 19 && (dateStr.contains("AM") || dateStr.contains("PM"))) {
                    format = "MMM dd yyyy hh:mma";
                } else {
                    return null;
                }
//                if (isUTC) {
//                    df = new SimpleDateFormat(format);
//                    df.setTimeZone(TimeZone.getTimeZone("UTC"));
//                } else {
                df = new SimpleDateFormat(format, Locale.getDefault());
//                }

                date = df.parse(dateStr);

                retCal = new GregorianCalendar();

                retCal.setTime(date);
                retCal.setTimeZone(df.getTimeZone());

            } catch (ParseException e) {
                LOGGER.error(e.getMessage());
            }
        }

        return retCal;

    }

//    /**
//     * GregorianCalendar representation of date format string.
//     * @param indateStr Date format string
//     * @param isUTC Assume date format string timezone is UTC if timezone not specified.
//     * @return GregorianCalendar representation of date format string.
//     */
//    protected GregorianCalendar bhieToCal(String indateStr, boolean isUTC) {
//        String dateStr = indateStr;
//        Date date = null;
//        String format = "";
//        GregorianCalendar retCal = null;
//        SimpleDateFormat df = null;
//        if (dateStr != null && dateStr.length() != 0) {
//            try {
//
//                if (dateStr.indexOf("/") > 0) {
//                    format = "MM/dd/yy";
//                } else if (dateStr.indexOf(" ") == 2) {
//                    format = "dd MMM yyyy";
//                } else if (dateStr.indexOf("T") > 0 && dateStr.indexOf("Z") > 0) {
//                    format = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
//                } else if (dateStr.indexOf("T") > 0 && dateStr.indexOf(":") > 0) {
//                    dateStr = dateStr.replace("-00:00", "");
//                    format = "yyyy-MM-dd'T'HH:mm:ss";
//                } else if (dateStr.indexOf("-") > 0 && dateStr.indexOf(":") > 0) {
//                    format = "yyyy-MM-dd HH:mm:ss";
//                } else if (dateStr.indexOf("-") > 0 && dateStr.indexOf("@") > 0) {
//                    format = "dd-MMM-yyyy@HHmm";
//                } else if (dateStr.indexOf(",") > 0 && dateStr.indexOf("@") > 0) {
//                    format = "MMM dd, yyyy@HH:mm";
//                } else if (dateStr.indexOf("@") < 0 && dateStr.indexOf("-") == 4) {
//                    format = "yyyy-MM-dd";
//                } else if (dateStr.indexOf("@") > 0 && dateStr.indexOf("-") < 0) {
//                    format = "ddMMMyy@HHmm";
//                } else if (dateStr.length() == 8 && dateStr.indexOf("-") < 0) {
//                    format = "yyyyMMdd";
//                } else if (dateStr.length() == 12) {
//                    format = "yyyyMMddHHmm";
//                } else if (dateStr.length() == 14) {
//                    format = "yyyyMMddHHmmss";
//                } else if (dateStr.indexOf(".") > 12 && dateStr.length() > 14 && dateStr.length() < 19) {
//                    //assumeUTC = false;
//                    format = "yyyyMMddHHmmss.SSS";
//                } else if (dateStr.indexOf(".") > 12 && dateStr.length() > 18) {
//                    //assumeUTC = false;
//                    format = "yyyyMMddHHmmss.SSSZ";
//                } else if (dateStr.length() == 7) {
//                    format = "ddMMMyy";
//                } else if (dateStr.length() == 19 && (dateStr.contains("AM") || dateStr.contains("PM"))) {
//                    format = "MMM dd yyyy hh:mma";
//                } else {
//                    return null;
//                }
////                if (isUTC) {
////                    df = new SimpleDateFormat(format);
////                    df.setTimeZone(TimeZone.getTimeZone("UTC"));
////                } else {
//                    df = new SimpleDateFormat(format, Locale.getDefault());
////                }
//
//                date = df.parse(dateStr);
//
//                retCal = new GregorianCalendar();
//
//                retCal.setTime(date);
//                retCal.setTimeZone(df.getTimeZone());
//
//            } catch (ParseException e) {
//                LOGGER.error(e.getMessage());
//            }
//        }
//
//        return retCal;
//    }

    /**
     *  XMLGregorianCalendar representation of date format string.
     * @param dateStr Date format string
     * @return XMLGregorianCalendar representation of date format string.
     */
    protected XMLGregorianCalendar bhieToXMLCal(String dateStr) {
        XMLGregorianCalendar rtc;
        GregorianCalendar cal = bhieToCal(dateStr);
        //GregorianCalendar oldDate = new GregorianCalendar(1900, Calendar.JANUARY, 1);

        if (cal == null) {
            rtc = null;
        } else if ("18991230050000".equals(dateStr)) {        //BDA invalid/unknown date
            rtc = null;
        } else {
            try {
                rtc = javax.xml.datatype.DatatypeFactory.newInstance().newXMLGregorianCalendar(cal);
            } catch (DatatypeConfigurationException e) {
                LOGGER.error(e.getMessage());
                rtc = null;
            }
        }
        return rtc;
    }



    /**
     *  XMLGregorianCalendar representation of date format string.
     * @param dateRange Date format string
     * @return XMLGregorianCalendar representation of date format string.
     */
    public static XMLGregorianCalendar bhieToXMLCalV4(DateRangeV4 dateRange) {
        XMLGregorianCalendar rtc;
        String dateStr = "";
        if (dateRange != null) {
            if (dateRange.getStart() != null && dateRange.getStart().length() > 0) {
                dateStr = dateRange.getStart();
            } else if (dateRange.getEnd() != null && dateRange.getEnd().length() > 0) {
                dateStr = dateRange.getEnd();
            }
        }
        GregorianCalendar cal = bhieToCal(dateStr);
        //GregorianCalendar oldDate = new GregorianCalendar(1900, Calendar.JANUARY, 1);

        if (cal == null) {
            rtc = null;
        } else if ("18991230050000".equals(dateStr) || (dateStr.contains("1899-12-30"))) {        //BDA invalid/unknown date
            rtc =  null;
        } else {
            try {
                rtc =  javax.xml.datatype.DatatypeFactory.newInstance().newXMLGregorianCalendar(cal);
            } catch (DatatypeConfigurationException e) {
                LOGGER.error(e.getMessage());

                rtc = null;
            }
        }
        return rtc;
    }

//    protected Provider mapProvider(gov.va.med.bhie.Provider bhieProvider){
//        Provider provider = new Provider();
//        if (bhieProvider != null)
//            provider.setName(bhieProvider.getName());
////        provider.setPrimary(bhieProvider.isPrimary());
////        provider.setCode(bhieProvider.getCode());
////        provider.setTaxonomyCode(bhieProvider.getTaxonomyCode());
////        provider.setProviderType(bhieProvider.getTaxonomyDisplay());
//        if (bhieProvider != null && bhieProvider.getId() != null && !bhieProvider.getId().isEmpty()) {
//            provider.setCode(bhieProvider.getId().get(0).getId());
//            if (bhieProvider.getTaxonomy() != null && !bhieProvider.getTaxonomy().isEmpty()) {
//                provider.setTaxonomyCode(bhieProvider.getTaxonomy().get(0).getCode());
//                provider.setProviderType(bhieProvider.getTaxonomy().get(0).getDisplay());
//            }
//        }
//
//        return provider;
//    }

    protected Provider mapProvider(gov.va.med.bhie.ProviderV4 bhieProvider) {
        Provider provider = new Provider();
        if (bhieProvider != null) {
            provider.setName(bhieProvider.getName());

            if (bhieProvider.getAddress() != null) {
                AddressV4 bhieAddress = bhieProvider.getAddress();
                provider.setCity(bhieAddress.getCity());
                provider.setStateProvince(bhieAddress.getState());
                provider.setPostalCode(bhieAddress.getPostalCode());
                if (bhieAddress.getStreet() != null && bhieAddress.getStreet().size() > 0) {
                    List<String> streetList = bhieAddress.getStreet();
                    int size;
                    if (streetList.size() > 3) {
                        size = 3;
                    } else {
                        size = streetList.size();
                    }

                    switch (size) {
                        case 3:
                            provider.setStreet3(streetList.get(2)); // fall through
                        case 2:
                            provider.setStreet2(streetList.get(1)); // fall through
                        case 1:
                            provider.setStreet1(streetList.get(0)); // fall through
                        default:
                            break;
                    }

                }

            }
            if (bhieProvider.getTelecoms() != null && bhieProvider.getTelecoms().size() > 0) {
                StringBuilder phoneSb = new StringBuilder();
                if (bhieProvider.getTelecoms().get(0).getUseType() != null
                        && StringUtils.isNotEmpty(bhieProvider.getTelecoms().get(0).getUseType())) {
                    phoneSb.append(bhieProvider.getTelecoms().get(0).getUseType());
                    phoneSb.append(" ");
                }
                if (bhieProvider.getTelecoms().get(0).getUriPrefix() != null
                        && bhieProvider.getTelecoms().get(0).getUriPrefix().length() > 0) {
                    phoneSb.append(bhieProvider.getTelecoms().get(0).getUriPrefix());
                    phoneSb.append(" ");
                }
                if (bhieProvider.getTelecoms().get(0).getValue() != null
                        && bhieProvider.getTelecoms().get(0).getValue().length() > 0) {
                    phoneSb.append(bhieProvider.getTelecoms().get(0).getValue());
                }

                provider.setPhone(phoneSb.toString());
            }
        }
        if (bhieProvider != null
                && bhieProvider.getIds() != null
                && !bhieProvider.getIds().isEmpty()
                && bhieProvider.getIds().get(0) != null) {
            provider.setCode(bhieProvider.getIds().get(0).getId());
            if (bhieProvider.getTaxonomy() != null
                    && !bhieProvider.getTaxonomy().isEmpty()
                    && bhieProvider.getTaxonomy().get(0) != null) {
                provider.setTaxonomyCode(bhieProvider.getTaxonomy().get(0).getCode());
                provider.setProviderType(bhieProvider.getTaxonomy().get(0).getDisplay());
            }
        }

        return provider;
    }

    protected SpdcBeanV4 getFirstSpdcV4(List<SpdcBeanV4> list) {
        SpdcBeanV4 rtc = new SpdcBeanV4();
        if (list != null && list.iterator().hasNext()) {
            rtc = list.iterator().next();
        }
        return rtc;
    }

    protected SpdcBeanV4 getFirstSpdcV4FromWrapper(List<SpdcWrapperV4> list) {
        SpdcWrapperV4 spdcWrapper = null;
        SpdcBeanV4 rtc = new SpdcBeanV4();
        if (list != null && list.iterator().hasNext()) {
            spdcWrapper = list.iterator().next();

            if(spdcWrapper != null && spdcWrapper.getSpdcList() != null && spdcWrapper.getSpdcList().iterator().hasNext()){
                rtc = spdcWrapper.getSpdcList().iterator().next();
            }
        }
        return rtc;
    }

    protected SpdcBeanV4 getPrimarySpdcV4(List<SpdcBeanV4> list) {
        SpdcBeanV4 rtc = null;
        for (SpdcBeanV4 spdcBean : list) {
            if (spdcBean != null && spdcBean.isPrimary()) {
                rtc = spdcBean;
                break;
            }

        }
        if (rtc == null) {
            rtc = getFirstSpdcV4(list);
        }
        return rtc;
    }

    protected Code[] getCodes(List<SpdcBeanV4> list) {
        Code[] rtc = null;
        if (list != null && !list.isEmpty()) {
            rtc = new Code[list.size()];
            Iterator<SpdcBeanV4> liter = list.iterator();
            int cnt = 0;
            while (liter.hasNext()) {
                SpdcBeanV4 bean = liter.next();
                rtc[cnt] = new Code();
                rtc[cnt].setCode(bean.getCode());
                rtc[cnt].setDisplay(bean.getDisplay());
                if (bean.isPrimary()) {
                    rtc[cnt].setRanking("primary");
                }
                rtc[cnt].setSystem(bean.getSystem());
            }
        }
        return rtc;
    }

    protected Code[] getCodesV4(List<SpdcBeanV4> list) {
        Code[] rtc = null;
        if (list != null && !list.isEmpty()) {
            rtc = new Code[list.size()];
            Iterator<SpdcBeanV4> liter = list.iterator();
            int cnt = 0;
            while (liter.hasNext()) {
                SpdcBeanV4 bean = liter.next();
                rtc[cnt] = new Code();
                rtc[cnt].setCode(bean.getCode());
                rtc[cnt].setDisplay(bean.getDisplay());
                if (bean.isPrimary()) {
                    rtc[cnt].setRanking("primary");
                }
                rtc[cnt].setSystem(bean.getSystem());
            }
        }
        return rtc;
    }

    protected static Set<String> getNarrativeIdsByLoinc(List<NarrativeV4> narratives, String loinc) {
        Set<String> narrativeIds = new HashSet<String>();
        for (NarrativeV4 narrative : narratives) {
            if (narrative.getNarrativeIdentifier().contains(loinc)) {
                narrativeIds.add(narrative.getNarrativeIdentifier());
            }
        }

        return narrativeIds;
    }

    public List<NarrativeV4> mapDASNarratives(List<Narrative> narrativeList){
        List<NarrativeV4> retVal = new ArrayList<NarrativeV4>();
        for(Narrative narrative : narrativeList){
            NarrativeV4 narrativeV4 = new NarrativeV4();
            narrativeV4.setNarrativeAssigningAuthorityIdentifier(narrative.getOrganization());
            narrativeV4.setNarrativeHomeCommunityIdentifier(narrative.getSourceSystemID());
            narrativeV4.setNarrativeIdentifier(narrative.getNarrativeId());
            narrativeV4.setNarrativeText(narrative.getNarrativeText());
            retVal.add(narrativeV4);
        }

        return retVal;
    }

    public List<SiteStatusV4> mapDASStatusList(List<gov.va.med.vds.webservice.SiteStatus> siteStatusList){
        List<SiteStatusV4> retVal = new ArrayList<SiteStatusV4>();
        for(gov.va.med.vds.webservice.SiteStatus siteStatus : siteStatusList){
            SiteStatusV4 siteStatusV4 = new SiteStatusV4();
            PartnerV4 partnerV4 = new PartnerV4();
            String name = VLERSiteCache.getVLERSiteName(siteStatus.getSiteName());
            if(StringUtils.isNotEmpty(name)) {
                partnerV4.setPartnerName(name);
            } else {
                partnerV4.setPartnerName(siteStatus.getSiteName());
            }
            partnerV4.setPartnerIdentifier(siteStatus.getSiteName());
            siteStatusV4.setPartner(partnerV4);
            siteStatusV4.setPartnerStatus(siteStatus.getStatus());
            retVal.add(siteStatusV4);
        }

        return retVal;
    }

}
