package gov.va.med.common;


import gov.va.med.guibeans.GUISite;
import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.codec.net.URLCodec;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.LinkedHashMap;
import java.util.Map;

@Component
public class TokenUtils {
    @Autowired
    private AESCryptoUtil cryptoUtil;

    private static final Logger logger = Logger.getLogger(TokenUtils.class.getName());

    public byte[] generateSecurityKey() {
        return cryptoUtil.generateSecurityKey();
    }

    /**
     * Generates a patient token that is derived from input data. \
     * <p/>
     * A patient token is essential an AES encrypted string of patient parameters.
     *
     * @param encryptSecurityKey encryption key (required)
     * @param idType             patient identifier type - EDIPI, IEN, etc. (required)
     * @param patientId          patient identifier (required)
     * @param patSiteCode        site code associated with patient identifier (optional)
     * @return An patient token.
     */
    public String generatePatientToken(byte[] encryptSecurityKey, String idType, String patientId, String patSiteCode) {
        //do not generate patient token if idType or patientId is null or empty
        if (StringUtils.isBlank(idType) || StringUtils.isBlank(patientId)) return null;

        if (patSiteCode == null) patSiteCode = "";

        Map<String, String> patTokenMap = new LinkedHashMap<String, String>();

        patTokenMap.put(HttpKeys.PATIENT_ID, patientId);
        patTokenMap.put(HttpKeys.ID_TYPE, idType);
        patTokenMap.put(HttpKeys.PATIENT_SITE_CODE, patSiteCode);
        patTokenMap.put(HttpKeys.TIMESTAMP, Long.toString(System.currentTimeMillis()));

        StringBuilder paramString = new StringBuilder();

        try {
            for (Map.Entry<String, String> entry : patTokenMap.entrySet()) {
                if (paramString.length() > 0) {
                    paramString.append("&");
                }

                String key = entry.getKey();
                String val = entry.getValue();

                if (val == null) {
                    val = "";
                }

                paramString.append(key);
                paramString.append("=");
                paramString.append(URLEncoder.encode(val, "UTF-8"));
            }
        } catch (UnsupportedEncodingException e) {
            logger.error(e);
        }

        return cryptoUtil.generateEncryptedQuery(encryptSecurityKey, paramString.toString());
    }

    public String generateTMDSPatientToken(byte[] encryptSecurityKey,
        String userEDIPI ,
        String userName ,
        String userNCID ,
        String patientEDIPI ,
        String resourceUserID ,
        String resourcePassword ,
        String facilityNCID
        ) {
        
        if (facilityNCID == null) facilityNCID = "";

        Map<String, String> patTokenMap = new LinkedHashMap<String, String>();

        patTokenMap.put("patientEDIPI", patientEDIPI);
        patTokenMap.put("facilityNCID", facilityNCID);

        patTokenMap.put("userEDIPI", userEDIPI);
        patTokenMap.put("userName", userName);
        patTokenMap.put("userNCID", userNCID);
        patTokenMap.put("patientEDIPI", patientEDIPI);
        patTokenMap.put("resourceUserID", resourceUserID);
        patTokenMap.put("resourcePassword", resourcePassword);
        patTokenMap.put("facilityNCID", facilityNCID);

        patTokenMap.put(HttpKeys.TIMESTAMP, Long.toString(System.currentTimeMillis()));

        StringBuilder paramString = new StringBuilder();

        try {
            for (Map.Entry<String, String> entry : patTokenMap.entrySet()) {
                if (paramString.length() > 0) {
                    paramString.append("&");
                }

                String key = entry.getKey();
                String val = entry.getValue();

                if (val == null) {
                    val = "";
                }

                paramString.append(key);
                paramString.append("=");
                paramString.append(URLEncoder.encode(val, "UTF-8"));
            }
        } catch (UnsupportedEncodingException e) {
            logger.error(e);
        }

        return cryptoUtil.generateEncryptedQuery(encryptSecurityKey, paramString.toString());
    }

    public String generateRecordDetailsToken(byte[] encryptSecurityKey,
                                             String patToken, String recordControllerName,
                                             Map<String, Object> recordQuery, GUISite recordSite,
                                             String recordSourceProtocol, boolean isSessionCached, int cacheCounter) {
        if (recordQuery == null) {
            throw new RuntimeException("record details token requires a record query map");
        }

        try {
            recordQuery.put(HttpKeys.RECORD_CONTROLLER_NAME, recordControllerName);
            recordQuery.put(HttpKeys.RECORD_SITE_CODE, recordSite.getSiteCode());
            recordQuery.put(HttpKeys.RECORD_SOURCE_PROTOCOL, recordSourceProtocol);
            recordQuery.put(HttpKeys.SESSION_CACHED_RECORD, Boolean.toString(isSessionCached));
            recordQuery.put(HttpKeys.TIMESTAMP, Long.toString(System.currentTimeMillis()));

            if (isSessionCached) {
                if (org.apache.commons.lang.StringUtils.isNotBlank(patToken)) {
                    MessageDigest messageDigest = MessageDigest.getInstance("SHA-256");
                    messageDigest.update(patToken.getBytes("UTF-8"));

                    String patTokenHash = new BigInteger(1, messageDigest.digest()).toString(16);
                    patTokenHash = StringUtils.leftPad(patTokenHash, 40, '0');

                    //create unique record cache key
                    recordQuery.put(HttpKeys.RECORD_CACHE_KEY, patTokenHash + "^" + cacheCounter);
                }
            }

            StringBuilder recordQueryStr = new StringBuilder();

            for (Map.Entry<String, Object> entry : recordQuery.entrySet()) {
                if (recordQueryStr.length() > 0) {
                    recordQueryStr.append("&");
                }

                String key = entry.getKey();
                Object val = entry.getValue();

                if (val == null) {
                    val = "";
                }
                recordQueryStr.append(key);
                recordQueryStr.append("=");
                recordQueryStr.append(URLEncoder.encode("" + val, "UTF-8"));
            }

            return cryptoUtil.generateEncryptedQuery(encryptSecurityKey, recordQueryStr.toString());
        } catch (UnsupportedEncodingException e) {
            logger.error(e);
        } catch (NoSuchAlgorithmException e) {
            logger.error(e);
        }

        return null;
    }

    public String decryptTokenToString(byte[] securityKey, String encryptedQuery) {
        String params;

        if (cryptoUtil.isEncryptedTokenDebug()) {
            try {
                URLCodec urlCodec = new URLCodec();
                params = urlCodec.decode(encryptedQuery);
            } catch (DecoderException ex) {
                logger.error(ex.getMessage(), ex);
                params = "";
            }
        } else {
            byte[] decryptBytes = cryptoUtil.decrypt(securityKey, Base64.decodeBase64(encryptedQuery));

            try {
                params = new String(decryptBytes, "UTF-8");
            } catch (UnsupportedEncodingException ex) {
                logger.error(ex.getMessage(), ex);
                params = "";
            }
        }

        return params;
    }

    public Map decryptTokenToMap(byte[] securityKey, String encryptedQuery) {

        Map<String, String> map = generateURLParameterMap(decryptTokenToString(securityKey, encryptedQuery));

        //decode URL parameters
        for (Map.Entry<String, String> entry : map.entrySet()) {
            String key = entry.getKey();
            String val = entry.getValue();

            if (val != null) {
                try {
                    String decodedVal = URLDecoder.decode(val, "UTF-8");
                    map.put(key, decodedVal);
                } catch (IllegalArgumentException e) {
                    logger.error(e);
                } catch (UnsupportedEncodingException e) {
                    logger.error(e);
                }
            }
        }

        return map;
    }

    private Map<String, String> generateURLParameterMap(String params) {
        Map<String, String> paramMap = new LinkedHashMap<String, String>();

        String[] paramVals = params.split("&");
        for (String param : paramVals) {
            String[] vals = param.split("=");
            if (vals.length > 1) {
                String key = vals[0];
                String val = vals[1];
                paramMap.put(key, val);
            }
        }

        return paramMap;
    }
}
