package gov.va.med.vds.common;

import gov.va.med.foundations.utilities.AESCryptoUtil;
import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.EncoderException;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.codec.net.URLCodec;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;

import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Created by melcolm on 1/5/17.
 */
public class TokenUtils {

    private static final Logger logger = Logger.getLogger(TokenUtils.class.getName());

    private static final AESCryptoUtil cryptoUtil = new AESCryptoUtil();

    public static AESCryptoUtil getCryptoUtil() {
        return cryptoUtil;
    }

    private static final TokenUtils INSTANCE = new TokenUtils();

    public static TokenUtils getInstance() {
        return INSTANCE;
    }

    private TokenUtils() {
        super();
    }


    public byte[] generateSecurityKey() {
        return cryptoUtil.generateSecurityKey();
    }

    /**
     * Generates a patient token that is derived from input data. \
     * <p/>
     * A patient token is essential an AES encrypted string of patient parameters.
     *
     * @param encryptSecurityKey encryption key (required)
     * @param patientICN         patient identifier (required)
     * @param userID             user identifier (required)
     * @param domain             domain of the query (required)
     * @return An patient token.
     */
    public String generateCallbackToken(byte[] encryptSecurityKey, String patientICN,
                                        String userID,
                                        String domain, String forceRefresh) {
        //do not generate patient token if idType or patientId is null or empty
        if (StringUtils.isBlank(patientICN) || StringUtils.isBlank(userID)) return null;

        Map<String, String> callbackTokenMap = new LinkedHashMap<String, String>();

        callbackTokenMap.put("patientICN", patientICN);
        callbackTokenMap.put("userID", userID);
        callbackTokenMap.put("domain", domain);
        callbackTokenMap.put("forceRefresh", forceRefresh);
        callbackTokenMap.put("timeStamp", Long.toString(System.currentTimeMillis()));

        StringBuilder paramString = new StringBuilder();

        try {
            for (Map.Entry<String, String> entry : callbackTokenMap.entrySet()) {
                if (paramString.length() > 0) {
                    paramString.append("&");
                }

                String key = entry.getKey();
                String val = entry.getValue();

                if (val == null) {
                    val = "";
                }

                paramString.append(key);
                paramString.append("=");
                paramString.append(URLEncoder.encode(val, "UTF-8"));
            }
        } catch (UnsupportedEncodingException e) {
            logger.error(e);
        }

        return cryptoUtil.generateEncryptedQuery(encryptSecurityKey, paramString.toString());
    }

//    public String generateEncryptedString(byte[] encryptSecurityKey, String queryParams) {
////        if (isEncryptedTokenDebug()) {
////            try {
////                URLCodec urlCodec = new URLCodec();
////                return urlCodec.encode(queryParams);
////            } catch (EncoderException ex) {
////                logger.error(ex.getMessage(), ex);
////                return null;
////            }
////        } else {
//            byte[] encryptedBytes = cryptoUtil.encrypt(encryptSecurityKey, queryParams);
//            return Base64.encodeBase64URLSafeString(encryptedBytes);
////        }
//
//    }

    public String decryptTokenToString(byte[] securityKey, String encryptedQuery) {
        String params;

        byte[] decryptBytes = cryptoUtil.decrypt(securityKey, Base64.decodeBase64(encryptedQuery));

        try {
            params = new String(decryptBytes, "UTF-8");
        } catch (UnsupportedEncodingException ex) {
            logger.error(ex.getMessage(), ex);
            params = "";
        }
        return params;
    }

    public Map decryptTokenToMap(byte[] securityKey, String encryptedQuery) {

        Map<String, String> map = generateURLParameterMap(decryptTokenToString(securityKey, encryptedQuery));

        //decode URL parameters
        for (Map.Entry<String, String> entry : map.entrySet()) {
            String key = entry.getKey();
            String val = entry.getValue();

            if (val != null) {
                try {
                    String decodedVal = URLDecoder.decode(val, "UTF-8");
                    map.put(key, decodedVal);
                } catch (IllegalArgumentException e) {
                    logger.error(e);
                } catch (UnsupportedEncodingException e) {
                    logger.error(e);
                }
            }
        }

        return map;
    }

    private Map<String, String> generateURLParameterMap(String params) {
        Map<String, String> paramMap = new LinkedHashMap<String, String>();

        String[] paramVals = params.split("&");
        for (String param : paramVals) {
            String[] vals = param.split("=");
            if (vals.length > 1) {
                String key = vals[0];
                String val = vals[1];
                paramMap.put(key, val);
            }
        }

        return paramMap;
    }
}